#ifndef GSMP_ENHWAVE_VIEWS
#define GSMP_ENHWAVE_VIEWS

// GSML stuff:

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "WavePlot.hh"

#include "View.hh"

// gtk-- gui stuff:
#include <gtk--/main.h>

#include <gtk--/window.h>
#include <gtk--/adjustment.h>
#include <gtk--/box.h>
#include <gtk--/drawingarea.h>
#include <gtk--/menu.h>
#include <gtk--/menubar.h>
#include <gtk--/ruler.h>
#include <gtk--/scrollbar.h>
#include <gtk--/style.h>

#include <vector>
#include <deque>

#define SCROLL_CACHE_FACTOR 20
#define SMOTH_SCROLL_FACTOR 4
#define AUTO_SCOLL_ACCEL 6

using namespace GSML;

class ViewOverlay;
class ViewOverlayManager;

struct Area
{
  Area ()
    : x (), y (), w (), h ()
  { }
  
  int x, y, w, h;
};

typedef int Priority;

/*  0: wave-form
 *  5: grid
 * 10: selection
 * 20: temp selection
 * 30: marker
 * 50: parameter-line
 *     ...
 */

class SensitiveArea
{
public:
  
  SensitiveArea (ViewOverlay* i_view_overlay);
  
  Area area;
  Priority priority;
  
  ViewOverlay* view_overlay;
};

class ViewOverlayManager
{
public:
  
  virtual void RegisterSensitiveArea (const SensitiveArea& area) = 0;
  virtual void UnRegisterSensitiveArea (const SensitiveArea& area) = 0;
  
  virtual void DrawBackground (const Area& area, Priority priority) = 0;
  virtual void FinishedDrawing (const Area& area, Priority priority) = 0;
  
  Signal1 <void, GdkEventButton*> SignalButtonPress;
  Signal1 <void, GdkEventButton*> SignalButtonRelease;
  Signal2 <void, GdkEventMotion*, sample_index> SignalMotionNotification;
};

class ViewOverlay
{
public:
  
  Area area;
  int priority;
  
  ViewOverlay (ViewOverlayManager* i_manager, Gdk_Pixmap* i_pixmap);
  virtual ~ViewOverlay () = 0;
  
  virtual void Regenerate (const Area& n_size, sample_index real_start, sample_index real_end) = 0;
  virtual void Draw (const Area& area) = 0;
  
  virtual void EnterNotivication (GdkEventMotion* event, sample_index real_pos) = 0;
  virtual void LeaveNotification () = 0;
  virtual void MotionNotification (GdkEventMotion* event, GdkModifierType state, sample_index real_pos) = 0;
  
  virtual void ButtonPress (GdkEventButton* event, sample_index real_pos) = 0;
  virtual void ButtonRelease (GdkEventButton* event, sample_index real_pos) = 0;

protected:
  ViewOverlayManager* m_manager;
  Gdk_Pixmap* m_pixmap;
};

class WaveFormViewOverlay : ViewOverlay
{
public:
  
  WaveFormViewOverlay (WavePlotTracer<MinMaxAvgWavePlot>* i_waveplot, ViewOverlayManager* i_manager, Gdk_Pixmap* i_pixmap);
  ~WaveFormViewOverlay ();
  
  void Regenerate (const Area& n_size, sample_index real_start, sample_index real_end);
  void Draw (const Area& area);
  
  void EnterNotivication (GdkEventMotion* event);
  void LeaveNotification ();
  void MotionNotification (GdkEventMotion* event, GdkModifierType state, sample_index);
  
  void ButtonPress (GdkEventButton* event, sample_index real_pos);
  void ButtonRelease (GdkEventButton* event, sample_index real_pos);

protected:
  WavePlotTracer<MinMaxAvgWavePlot>* m_waveplot;
  Gdk_GC wave_gc;  // for the wave
  
  int waveform_middle;
  int waveform_vscale;
  int waveform_dBfs;
};

class MarkerViewOverlay : ViewOverlay
{
public:
  
  MarkerViewOverlay (ViewOverlayManager* i_manager, Gdk_Pixmap* i_pixmap);
  ~MarkerViewOverlay ();
  
  void Regenerate (const Area& n_size, sample_index real_start, sample_index real_end);
  void Draw (const Area& area);
  
  void EnterNotivication (GdkEventMotion* event);
  void LeaveNotification ();
  void MotionNotification (GdkEventMotion* event, GdkModifierType state, sample_index);
  
  void ButtonPress (GdkEventButton* event, sample_index real_pos);
  void ButtonRelease (GdkEventButton* event, sample_index real_pos);

protected:
};

class EnhNormalView; // forward

// sub class for a single track
class EnhTrackView : public Gtk::DrawingArea, public ViewOverlayManager
{
public:
  EnhTrackView (WavePlotTracer<MinMaxAvgWavePlot>* i_plot, const WavePlotConfig* i_config,
		EnhNormalView* i_parent, bool i_master = false); // master will draw additional graphic stuff (marker heads ...)
  ~EnhTrackView ();
  
  void SetStartDelta (int n_delta);
  
  void SetTempSelection (sample_index n_start, sample_index n_end);
  void RemoveTempSelection ();
  
  void SetRefresh (); // only temp. debug hack
  
  int Width ();
  int Height ();
  
  void SlotGUIRefresh ();
  
  void Debug () const;
  
  // ViewOverlayManager
  void UnRegisterSensitiveArea (const SensitiveArea& area);
  void RegisterSensitiveArea (const SensitiveArea& area);
  void DrawBackground (const Area& area, Priority priority);
  void FinishedDrawing (const Area& area, Priority priority);
  
private:
  
  void add_refresh (int n_start, int n_end);
  void add_refresh_from_real (sample_index start, sample_index length);
  
  void add_refresh_all ();
  
  void connect_selection ();
  void connect_markers ();
  
  // Signal Slots:
  void SlotWavePlotChanged (MinMaxAvgWavePlot* n_plot);
  void SlotWavePlotUpdated (sample_index start, sample_index length);
  
  SigC::Connection connection_selection_selected;
  void SlotSelectionSelected (sample_index, sample_index);
  SigC::Connection connection_selection_unselected;
  void SlotSelectionUnselected (sample_index, sample_index);
  SigC::Connection connection_selection_active_changed;
  void SlotSelectionActiveChanged (bool n_state);
  
  SigC::Connection connection_marker_added;
  void SlotMarkerAdded (const string& n_name, sample_index n_pos);
  SigC::Connection connection_marker_deleted;
  void SlotMarkerDeleted (const string& n_name, sample_index n_pos);
  SigC::Connection connection_marker_changed;
  void SlotMarkerChanged (const string& n_name, sample_index o_pos, sample_index n_pos);
  
  // gtk-- event-impl overwrites ...
  
  virtual gint configure_event_impl (GdkEventConfigure* event);
  virtual gint expose_event_impl (GdkEventExpose* event);
  virtual gint button_press_event_impl (GdkEventButton* event);
  virtual gint button_release_event_impl (GdkEventButton* event);
  virtual gint motion_notify_event_impl (GdkEventMotion* event);
  
  // Gtk-- stuff
  Gdk_Window win;
  Gdk_Visual visual;
  
  Gdk_Pixmap m_pixmap; // the cached image data
  Gdk_GC back_gc;  // the normal background
  
  // internal states
  int m_width; // current width and height of our window
  int m_height;
  
  int m_start_delta; // the start in plot we should dispaly
  int m_internal_delta; // our start in plot of cache pixmap
  
  int m_cache_length; // the length of our cache pixmap
  
  sample_index m_temp_sel_start;
  sample_index m_temp_sel_end;
  bool m_use_temp_sel;
  
  SigC::Threads::Mutex refresh_lock;
  bool refresh_needed;
  bool refresh_all;
  sample_index refresh_start;
  sample_index refresh_end;
  
  // Overlay(tm)-Architecture
  deque <ViewOverlay*> m_view_overlays;
  deque <SensitiveArea> m_sensitive_areas;
  
  // GSML Stuff
  WavePlotTracer<MinMaxAvgWavePlot>* m_waveplot;
  const WavePlotConfig* m_config;
  EnhNormalView* m_parent;
  bool m_master;
  
  MinMaxAvgWavePlot* m_current_plot;
  TrackLayer::Selection* m_current_sel;
  DataLayer::Markers* m_current_markers;
};
     
class EnhNormalView : public Gtk::VBox, public View
{
public:
  EnhNormalView (WavePlotManager* m_manager, Signal0<void>* i_timeout);
  EnhNormalView (const EnhNormalView& src);
  ~EnhNormalView ();
  
  void SetStart (sample_index n_start);
  sample_index Start ();
  
  void SetZoom (sample_index n_zoom_factor);
  void SetZoomAll ();
  int Zoom ();
  
  sample_index Width ();
  sample_index Length ();
  
  void SetTrackVisible (track_index track_no, bool status);
  bool IsTrackVisible (track_index track_no);
  
  View* Clone ();
  
protected:
  friend class EnhTrackView;
  
  /********************
   * internal helpers *
   ********************/
  void initialize ();
  
  void set_zoom (int n_sps);
  void set_start (sample_index n_start, bool adjustment_update = true, bool force_plot_reconf = false);
  void reconfig_plots (sample_index n_start, sample_index n_sps);
  void set_views_start_delta ();
  void update_adjustment ();
  
  void start_selection (sample_index n_current);
  void set_selection (sample_index n_current);
  void end_normal_selection ();
  void end_state_selection (bool plus_minus);
  void end_no_selection_but_marker ();
  
  void debug_internals ();

  /***********************************
   * "signal" handlers / connections *
   ***********************************/
  // some child callbacks
  void SlotChildConfigured (int n_width);
  void SlotButtenPress (GdkEventButton* event);
  void SlotButtenRelease (GdkEventButton* event);
  void SlotMotionNotify (GdkEventMotion* event, GdkModifierType state);
  
  // Data
  void SlotLayerChanged (DataLayer* n_layer);
  
  void SlotAdjustmentChanged ();
  void SlotAdjustmentValueChanged ();
  
  void SlotGUIRefresh (); // clock signal for delayed GUI refresh 
  
  // gtk-- callbacks (hacks)
  void slot_context_zoom_out ();
  void slot_context_zoom_in ();
  void slot_context_zoom_one2one ();
  void slot_context_zoom_all ();
  void slot_context_zoom_sel ();
  
  void slot_context_select_all ();
  void slot_context_select_clear ();
  void slot_context_select_invert ();
  
  void slot_context_debug_redraw ();
  void slot_context_debug_reload ();
  void slot_context_debug_dump_internal ();
  
  /*****************
   * my properties *
   *****************/
  
  // internal (important) states
  WavePlotConfig m_config;
  
  // all values we need
  sample_index m_start;        // our display start
  sample_index m_layer_length; // whole DataLayer length
  sample_index m_plot_start;   // plot's start
  sample_index m_plot_length;  // plot's length
  sample_index m_cache_length; // real length of plot's
  int m_view_width;            // view's width
  sample_index m_view_length;  // view's real length
  int m_sps;                   // out zoom factor
  
  int m_view_start_delta;      // current view start delta in the plot
  bool m_auto_zoom_all;        // sould be auto-track on the whole track
  
  // selection states
  bool selecting;
  bool selected;
  sample_index selection_start;
  sample_index selection_current;
  
  // autoscrolling states
  bool m_autoscroll;
  int m_autoscroll_delta;
  
  // GSML configs
  class view_stuff
  {
  public:
    EnhTrackView* view;
    WavePlotTracer<MinMaxAvgWavePlot>* plot;
  };
  
  deque<view_stuff> m_views;
  
  WavePlotManager* m_manager;
  Signal0<void>* m_timeout;
  
  // gtk-- gui
  Gtk::VBox view_box;
  
  Gtk::HRuler m_hruler;
  Gtk::HScrollbar m_scrollbar;
  Gtk::Adjustment m_adjustment;
  
  Gtk::Menu menu_popup;
};

#endif // GSML_ENHWAVE_VIEWS
