/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   grouping of DataLayers to the "undo-stack" Data
 */

#ifndef GSML_DATA
#define GSML_DATA

#include "include.hh"
#include "typedefs.hh"

#include "Basics.hh"
#include "SampleBuffer.hh"
#include "DataFile.hh"
#include "DataFileIndex.hh"
#include "DataLayer.hh"

#include <list>
#include <deque>

#include <algorithm>

namespace GSML
{

  // #define DEBUG_GSML_DATA // for debuging some "couts" ...

  using namespace std;
  using namespace SigC;
  
  class Data
    {
    public:
      class Except : public BasicExcept {};
      class ExceptLogicalError : public Except {};
      class ExceptUnsupported : public Except {};
      class ExceptIteratorInvaild : public Except {};
      
      /* coding convention:
       *   - this info should be approximately updated 5 times per second! ;-)
       * other Info:
       *   - the InfoUpdatedSignal must be called manually (or should we add and SetInfo like the one
       *     in the Markers class??
       *   - the top layer's properties can be accessed via the Text() and Progress() methods
       */
      
      class ProgressInfo
      {
      public:
	class Except : public BasicExcept {};
	class ExceptFixated : public Except {};
	class ExceptEmpty : public Except {};
	
	class Info
	{
	public:
	  Info (ProgressInfo* init_parent);
	  Info (ProgressInfo* init_parent, const Info& init_info);
	  
	  void SetText (const string& new_text);
	  void SetProgress (double new_progress);
	  
	  const string& Text () const;
	  const double Progress () const;
	  
	  Info& operator= (const Info& src);
	  
	protected:
	  
	  friend class ProgressInfo;
	  
	  ProgressInfo* parent;
	  
	  string text; // short info-string whats going on
	  double progress; // progress in percent
	};
	
	typedef deque<Info*>::iterator iterator;
	typedef deque<Info*>::const_iterator const_iterator;
	
	ProgressInfo ();
	~ProgressInfo ();
	
	// for very basic access
	const string& Text ();
	double Progress ();
	
	Info* PushInfo (const string& text);
	void PopInfo ();
	
	void Clear ();
	
	int Count () const;
	
	const_iterator begin ();
	const_iterator end ();
	
	void Debug ();
	
	Signal1<void, const Info*> SignalInfoUpdated;
	Signal1<void, const Info*> SignalInfoPushed;
	Signal0<void> SignalInfoPopped;
	
      private:
	ProgressInfo& operator= (const ProgressInfo& src)
	{
	}
	
	deque <Info*> infos;
      };
      
      typedef deque<DataLayer*>::iterator iterator;
      typedef deque<DataLayer*>::const_iterator const_iterator;
      
      
      Data (DataFileIndex* init_fileindex,  unsigned int init_data_layer,
	    PCMDatatype* init_sampletype, sample_rate init_samplerate = 44100);
      ~Data ();
      
      // Wraped stuff for easier access:
      sample_index Length () const;
      track_index Tracks () const;
      
      DataLayer* CurrentLayerPtr ();
      int Layers ();
      
      iterator CurrentLayer ();
      iterator BeginLayer ();
      iterator EndLayer ();
      
      void SetCurrentLayer (iterator new_layer);
      
      // creates and return a new data_layar which will be connected to
      // the wave_plots for the dynamic graphic preview!!
      DataLayer* PrepareNewLayer ();
      void FixateNewLayer ();
      void CancelNewLayer ();
      
      // config methods
      DataFileIndex* GetDataFileIndex ();
      PCMDatatype* TempSampletype ();
      sample_rate TempSamplerate ();
      
      void SetTempSampletype (PCMDatatype* new_type);
      void SetTempSamplerate (sample_rate new_rate);
      
      void SetMaxLayers (unsigned int new_max_layers);
      unsigned int MaxLayers ();
      bool IsPrepared ();
      
      bool IsModified ();
      void SetModified (bool new_state);
      
      Data& operator= (const Data& src);
      
      void Debug ();
      
      ProgressInfo progress_info;
      
      Signal1<void, DataLayer*> SignalCurrentLayerChanged; // also emited afet a layer-fixate!
      Signal1<void, DataLayer*> SignalNewLayerPrepared;
      Signal0<void> SignalNewLayerFixated;
      Signal0<void> SignalNewLayerCanceled;
      
      /* TODO: rework for most (all?) such signals ...
       * 
       * Signal1<void, const Selection&> SignalSelectionUpdated;
       * Signal1<void, const ProgressInfo&> SignalProgressUpdated;
       */
      
    private:
      void delete_if_too_many_layers ();
      
      DataFileIndex* my_fileindex;
      unsigned int max_layers;
      
      PCMDatatype* my_temp_sampletype;
      sample_rate my_temp_samplerate;
      
      // real data / states
      
      bool modified;
      
      iterator current_layer;
      DataLayer* prepared_layer;
      
      deque<DataLayer*>* data;
    };

} // end namespace GSML

#endif //GSML_DATA
