/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Base class for data-file read/write interface and RAW-data
 *   implementation.
 */

#ifndef GSML_DATA_FILE
#define GSML_DATA_FILE

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "SampleBuffer.hh"

namespace GSML
{
  
  // #define DEBUG_GSML_DATA_FILE
  
  /* a (very simply) Cache to improve the performance of interleaved/compressed files
   *
   * although elements can be definded (via _bound_elements) do belong together
   * the indices are specified in real position (e.g. bound == tracks -> indices
   * in frames ... (like in the OggVorbis DataFile implementation) )
   */
  
  template <class T_data, class T_index>
  class SimpleCache
  {
  public:
    SimpleCache (int i_bound_elements = 1)
    {
      m_bound_elements = i_bound_elements;
      
      data = 0;
      vaild_start = vaild_length = 0;
      cache_start = cache_length = 0;
      cache_size = 0;
      
    }
    
    ~SimpleCache ()
    {
      if (data != 0)
	delete [] data;
    }
    
    void SetBoundElements (int n_bound_elements)
    {
      m_bound_elements = n_bound_elements;
    }
    
    // moves the cache data and adjusts vaild_start /_end
    void SetCachedRange (T_index n_start, T_index n_length)
    {
      // resize cache ??
      if (cache_size < n_length)
	{
	  //cout << "Cache: Cache too small. (" << n_length << " is " << cache_size << ")"<< endl;
	  if (data != 0)
	    delete [] data;
	  data = new T_data [n_length * m_bound_elements];
	  
	  vaild_start = vaild_length = 0;
	  cache_start = cache_length = 0;
	  cache_size = n_length;
	}
  
      //cout << ">>  Chache: " << cache_start << " + " << cache_length
      //	   << ", vaild: " << vaild_start << " + " << vaild_length << endl;
      
      // try to use cache
      T_index t_cache_end = cache_start + cache_length;
      T_index t_end = n_start + n_length;
      
      T_index t_use_start = limit_min_max (n_start, cache_start, t_cache_end);
      T_index t_use_end = limit_max (t_end, t_cache_end);
      T_index t_use_length = t_use_end - t_use_start;
      
      T_index t_new_start = limit_min (cache_start - n_start, 0); 
      
      if (t_use_length > 0) // s.th. to use
	{
	  //cout << ">> DataFile: Something usefull in cache." << endl;
	  
	  T_index t_start_in_cache = t_use_start - cache_start;
	  
	  //cout << ">>  used from cache: " << t_use_start << " + " << t_use_length
	  //     << ", t_new_start: " << t_new_start << ", t_start_in_cache: " << t_start_in_cache << endl;
	  
	  // move data
	  if (t_new_start < t_start_in_cache) {
	    for (int i = 0; i < t_use_length * m_bound_elements; ++i)
	      data [t_new_start * m_bound_elements + i]
		= data [t_start_in_cache * m_bound_elements + i];
	  }
	  else if (t_new_start > t_start_in_cache) {
	    for (int i = t_use_length * m_bound_elements - 1; i >= 0; --i)
	      data [t_new_start * m_bound_elements + i]
		= data [t_start_in_cache * m_bound_elements + i];
	  }
	  //else
	  //  cout << "t_use_length == t_start_in_cache -> nothing to do ..." << endl;
	  
	  T_index t_vaild_end = vaild_start + vaild_length;
	  vaild_start = limit_min (vaild_start + t_new_start, 0);
	  vaild_length = limit_min_max (t_vaild_end - t_new_start, 0, t_use_length);
	}
      else // n.th. to be used
	{
	  vaild_start = 0;
	  vaild_length = 0;
	}
      
      cache_start = n_start;
      cache_length = n_length;
      
      //cout << "   @End:  Chache: " << cache_start << " + " << cache_length
      //	   << ", vaild: " << vaild_start << " + " << vaild_length << endl;
    }
    
    T_data* data;
    
    T_index cache_start;  // start of cache
    T_index cache_length; // length of cache
    
    int vaild_start, vaild_length;
    
  protected:
    int m_bound_elements; // number of objects (elements) that belongs together
    int cache_size;       // current cache size (never shrinked ...)
  };
  
  // This DataFiles will only be used for read access during editing work
  // and (ONLY the normal RAW classes) for temporary file write ...
  // for saving data special export filter will be used
  // when writeable no read is allowed - and is is only possible to write once ...
  
  class DataFile
    {
    public:
      class Except : public BasicExcept
	{
	public:
	  string filename;
	  DataFile* data_file;
	};
      
      class ExceptNotWriteable : public Except {};
      class ExceptNotFixated : public Except {};
      class ExceptFileNotReadable : public Except {};
      class ExceptFileNotWriteable : public Except {};
      class ExceptBindingError : public Except {};
      class ExceptLogicalError : public Except {};
      class ExceptBufferError : public Except {};
      class ExceptUnsupported : public Except {};
      
      DataFile (const string& i_filename, bool i_writeable=false, bool i_autodelete=false);
      virtual ~DataFile ();
      
      virtual void ReadData (sample_index start,
			     sample_index count,
			     track_index track,
			     SampleBuffer* destination) = 0;
  
      // No start! Because all has to be written linear!!
      // (normally) track has to be "0" !!!
      virtual void WriteData (sample_index count,
			      track_index track,
			      SampleBuffer* source,
			      sample_index buffer_index) = 0;

      // this will fixate the writen data!
      virtual void FixateWrite () = 0;
      
      // config methods
      bool IsAutoDelete ();
      bool IsWriteable ();
      
      string& Filename ();
      virtual void SetFilename (string& new_filename);
      
      int Bindings ();  //retunrs bindings
      void Bind (); // inc bindings counter
      void Unbind (); // dec bindings counter (exeption when zero ??)
  
    protected:
      string m_filename;
      bool m_writeable; // only for tmp files true!!
      bool m_autodelete; // only for original files false!!
      
      int m_bindings;
    };

  // first short impl. for RAW file access ... - needed fot tmp files ...
  
  class DataFileRAW : public DataFile
    {
    public:
      DataFileRAW (const string& i_filename, PCMDatatype* i_type, track_index i_tracks = 1,
		   sample_rate i_rate=44100, int i_leading_trash = 0, bool i_writeable=false,
		   bool i_autodelete=false);
      
      virtual ~DataFileRAW ();
  
      virtual void ReadData (sample_index start,
			     sample_index count,
			     track_index track, 
			     SampleBuffer* destination);
      
      virtual void WriteData (sample_index count,
			      track_index track, 
			      SampleBuffer* source,
			      sample_index buffer_index);
      
      virtual void FixateWrite ();
      
      virtual void SetFilename (string& new_filename);
      
      PCMDatatype* SampleTyp ();
      track_index TrackCount ();
      sample_rate SampleRate ();
      
    protected:
      
      off_t read_from_file (off_t start, off_t length);
      
      SigC::Threads::Mutex m_lock;
      
      // states
      int file_descriptor; // if in use
      
      SimpleCache <uint8, off_t> m_cache;
      
      // config
      PCMDatatype* m_sample_type;
      track_index m_tracks;
      sample_rate m_rate;
      int m_leading_trash;
  };

} // end namespace GSML
#endif //GSML_DATA_FILE
