/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Grouping of several TrackLayer to one "undo-step", multi-track
 *   data storage.
 */

#ifndef GSML_DATA_LAYER
#define GSML_DATA_LAYER

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "SampleBuffer.hh"
#include "DataFile.hh"
#include "DataFileIndex.hh"
#include "TrackLayer.hh"

#include <deque>
#include <list>

// #define DEBUG_GSML_DATA_LAYER // for debuging some "couts" ...

namespace GSML
{
  using namespace std;
  using namespace SigC;

  class DataLayer
    {
    public:
      class Except : public BasicExcept {};
      class ExceptFixated : public Except {};
      class ExceptNotFixated : public Except {};
      class ExceptLogicalError : public Except {};
      class ExceptUnsupported : public Except {};
      
      class Markers
	{
	public:
	  
	  class Except : public BasicExcept {};
	  class ExceptFixated : public Except {};
	  class ExceptUndeleteable : public Except {};
	  
	  class Marker
	  {
	  public:
	    
	    Marker ();	    
	    Marker (const Marker& source);
	    
	    Marker& operator= (const Marker& source);
	    
	    string name;
	    sample_index position;
	    bool undeleteable;
	  };
	  
	  typedef list<Marker>::iterator iterator;
	  typedef list<Marker>::const_iterator const_iterator;
	  
	  Markers (DataLayer* i_layer);
	  Markers (const Markers& source);
	  
	  // Attention: the layer must be adjusted after copying the object!!
	  Markers& operator= (const Markers& source);
	  void SetLayer (DataLayer* n_layer);
	  
	  bool AddMarker (const string& name, sample_index pos, bool undeleteable = false);
	  bool AddMarker (Marker& new_marker);
	  
	  bool DeleteMarker (const string& name);	  
	  
	  bool GetMarker (const string& name, sample_index* pos);
	  bool SetMarker (const string& name, sample_index pos);
	  
	  const_iterator begin ();
	  const_iterator end ();
	  
	  void Debug ();
	  
	  Signal2<void, const string&, sample_index> SignalAdded;
	  Signal2<void, const string&, sample_index> SignalDeleted;
	  Signal3<void, const string&, sample_index, sample_index> SignalChanged;
	  
	protected:
	  DataLayer* m_layer;
	  list <Marker> markers;
	};
      
      typedef deque<TrackLayer*>::iterator iterator;
      typedef deque<TrackLayer*>::const_iterator const_iterator;
      
      DataLayer (const string& init_comment, DataFileIndex* init_index,
		 PCMDatatype* init_temp_type, sample_rate init_rate=44100);
      DataLayer (const DataLayer& source_layer);
      ~DataLayer ();
      
      iterator Begin ();
      iterator End ();
      track_index Tracks () const;
      
      TrackLayer* CreateTrack (track_index pos);
      iterator CreateTrack (const iterator& pos);
      void CreateNEmptyTracks (track_index count);
      
      void EraseTrack (track_index pos);
      iterator EraseTrack (const iterator& pos);
      void EraseAllTracks ();
      
      const string& Comment () const;
      void Comment (const string& new_comment);
      
      sample_index Length () const; // returns track length

      // we fix (or lock) writen layer - because they shouldn't be modified later - anyway ...
      void Fixate ();
      void UnFixate ();
      bool IsFixated ();

      // create and return a new data_layer with the copy of the given region ...
      DataLayer* CreateDataLayer (sample_index start,
				  sample_index length) const;
  
      // inserts the given data_layer at the start position
      void InsertDataLayer (sample_index start,
			    const DataLayer& source_layer);

      void RemoveRegion (sample_index start,
			 sample_index length);
      
      DataFileIndex* GetDataFileIndex ();
      PCMDatatype* TempSampletype ();
      sample_rate TempSamplerate ();
      
      void SetTempSampletype (PCMDatatype* new_type);
      void SetTempSamplerate (sample_rate new_rate);
      
      // creates an new copy of this
      DataLayer& operator= (const DataLayer& source_layer);
  
      // for direct track acces with [] ...
      TrackLayer* operator[] (track_index track) const;

      void Debug () const;
      
      Signal1<void, track_index> SignalTrackCreated;
      Signal1<void, track_index> SignalTrackErased;
     
      Markers markers;
      
    private:
      friend class Markers;
      
      string my_comment; // for undo history or so ...
      
      DataFileIndex* my_fileindex;
      PCMDatatype* my_temp_sampletype;
      sample_rate my_temp_samplerate;
      
      bool fixated; // when fixated (and laying around in the undo_history) ... ??
      deque<TrackLayer*>* data;
  
    };

} // end namespae GSML

#endif //DATA_LAYER
