/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   ALSA Mixer Node implementation
 */

#ifndef GSML_IO_ALSA_H
#define GSML_IO_ALSA_H

#include "generated/config.hh"

#ifdef GSML_WITH_ALSA

#include "IO.hh"

#include "ALSAHelper.hh" // our ALSA wrappings

namespace GSML {
  
  // format converter helpers
  class ReadHandler
  {
  public:
    ReadHandler (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    virtual ~ReadHandler () {};
      
    virtual int TryRead () = 0;
    virtual bool TryMix (Mixer* mixer) = 0;
    
  protected:
    ALSAHelper::ALSAPCM* m_card;
    track_index m_channels;
    sample_index m_period_size;
  };
    
  class WriteHandler
  {
  public:
    WriteHandler (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    virtual ~WriteHandler () {};
      
    virtual bool TryMix (Mixer* mixer) = 0;
    virtual int TryWrite () = 0;
    
  protected:
    ALSAHelper::ALSAPCM* m_card;
    track_index m_channels;
    sample_index m_period_size;
  };
    
  class WriteHandler16Bit : public WriteHandler
  {
  public:
    WriteHandler16Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    ~WriteHandler16Bit ();
      
    bool TryMix (Mixer* mixer);
    int TryWrite ();
      
  protected:
    SampleBufferDefault* buffer;
    pcm_16bit::value_type* interleaved;
      
    bool pl_vaild;
  };
    
  class ReadHandler16Bit : public ReadHandler
  {
  public:
    ReadHandler16Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    ~ReadHandler16Bit ();
      
    int TryRead ();
    bool TryMix (Mixer* mixer);
      
  protected:
    SampleBufferDefault* buffer;
    pcm_16bit::value_type* interleaved;
      
    bool ca_vaild;
  };
  
  class WriteHandler32Bit : public WriteHandler
  {
  public:
    WriteHandler32Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    ~WriteHandler32Bit ();
      
    bool TryMix (Mixer* mixer);
    int TryWrite ();
      
  protected:
    SampleBufferDefault* buffer;
    int32* interleaved;
      
    bool pl_vaild;
  };
    
  class ReadHandler32Bit : public ReadHandler
  {
  public:
    ReadHandler32Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size);
    ~ReadHandler32Bit ();
      
    int TryRead ();
    bool TryMix (Mixer* mixer);
      
  protected:
    SampleBufferDefault* buffer;
    int32* interleaved;
      
    bool ca_vaild;
  };
  
  // the ALSA implementation which uses our Job to perform the sound card IO
  class ALSAMixerNode : public HWMixerNode, public GSMX_utils::Threads::Job
  {
  public:
    ALSAMixerNode ();
    virtual ~ALSAMixerNode ();
    
    // HWMixerNode interface
    virtual bool Init ();
    
    virtual const vector<string>& Interfaces ();
    virtual const string& Interface ();
    virtual bool SetInterface (const string& n_interface);
    
    // MixerNode interface
    
    virtual track_index PlaybackChannels ();
    virtual track_index CaptureChannels ();
    
    virtual int PlaybackPeriodSize ();
    virtual int CapturePeriodSize ();
    
    virtual int PlaybackPeriods ();
    virtual int CapturePeriods ();
    
    virtual int PlaybackXRuns ();
    virtual int CaptureXRuns ();
    
    virtual void SetPlayback (bool state);
    virtual void SetCapture (bool state);
    
    virtual bool IsPlayback ();
    virtual bool IsCapture ();
    
    virtual void StartIO (Mixer& source);
    virtual void StopIO ();
    
    virtual bool IsIORunning ();
    
    virtual sample_index GetCurrentPos ();
    
    virtual void Debug ();
    
  protected:
    
    bool reget_card_params ();
    
    void dump_alsa_params_info (ALSAHelper::ALSAPCM* card);
    bool internal_config_params (ALSAHelper::ALSAPCM* card, bool is_pl, bool overwrite_params);
    void handle_xrun (ALSAHelper::ALSAPCM* card);
    
    // the second thread ...
    void Loop ();
    
    // states
    Mixer* m_mixer;
    
    ALSAHelper::ALSAPCM* playback_card;
    ALSAHelper::ALSAPCM* capture_card;
    
    ReadHandler* read_handler;
    WriteHandler* write_handler;
    
    track_index playback_channels;
    track_index capture_channels; 
    
    bool initialized;
    bool io_running;
    bool cap_running; // only indicated whether cp was started via card->Start ()
    
    int pl_period_size;
    int ca_period_size;
    
    int pl_periods;
    int ca_periods;
    
    bool stop_requested;
    bool m_is_first_loop;
    bool m_worked_in_last_loop;
    bool m_using_realtime_sched;
    
    /* frame counting */
    sample_index current_playback_period;
    sample_index current_capture_period;
    
    int pl_xruns;
    int ca_xruns;
    
    // config
    string cname;
    
    bool do_playback;
    bool do_capture;
  };
  
} // end namespace GSMP

#endif // GSML_WITH_ALSA
#endif // GSMP_IO_ALSA_H
