/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Different IO architecture implementations.
 */

#ifndef GSML_IO_IMPL_H
#define GSML_IO_IMPL_H

#include "Data.hh" // GSML stuff
#include "Threads.hh" // Thread / Job support
#include "IO.hh"

#include <vector>
#include <deque>

namespace GSML {
  
  // A little helper to scatter IO
  class ScatterHelperSource
  {
  protected:
    
    void BuildTablesHelper (track_index pl_tracks, track_index ca_tracks, bool n_state);
    void ClearTablesHelper ();
    
    bool DoRead (MixerSource* source, track_index track, sample_index pos, sample_index length,
		 SampleBufferDefault* dest, bool is_master);
    void InvaildRead (track_index track, bool is_master);
    
    bool DoWrite (MixerSource* source, track_index track, sample_index pos, sample_index length,
		  SampleBufferDefault* src, sample_index buffer_index, bool is_master);
    void InvaildWrite (track_index track, bool is_mixer);
    
    // states
    vector <bool> io_read_pos_vaild;
    vector <bool> io_write_pos_vaild;
    
    bool master_read_pos_vaild;
    bool master_write_pos_vaild;
  };

  // A GSML::Data implementation
  class DataSource : public MixerSource
  {
  public:
    
    DataSource (GSML::DataLayer* pl_layer, GSML::DataLayer* ca_layer);
    ~DataSource ();
    
    // custom interface
    
    void SetDataLayer (GSML::DataLayer* pl_layer, GSML::DataLayer* ca_layer);
    
    // common source interface
    
    track_index PlaybackTracks () const;
    track_index CaptureTracks () const;
    
    sample_index MinStart () const;
    sample_index MaxEnd () const;
    bool IsInputable () const;
    
  private:
    
    void StartO_impl ();
    void StartIO_impl ();
    void StopIO_impl ();
    
    void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master);
    void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master);
    bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
		    bool is_master);
    bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
		     sample_index buffer_index, bool is_master);
    
    void Debug_impl () const;
    
    Data::ProgressInfo::Info* m_pinfo;
    
    // internal states
    
    sample_index m_playback_end_pos;

    // config
    GSML::DataLayer* m_pl_layer;
    GSML::DataLayer* m_ca_layer;
    
    vector <bool> io_write_prepared;
  };
  
  class SelectionFilterSource : public MixerSource, ScatterHelperSource
  {
  public:
    SelectionFilterSource (MixerSource* i_src, DataLayer* i_layer);
    ~SelectionFilterSource ();
    
    // custom interface
    
    void SetDataLayer (GSML::DataLayer* n_data);
    
    void SetSelectionForPlayback (bool n_state);
    bool IsSelectionForPlayback () const;
    
    void SetSelectionForCapture (bool n_state);
    bool IsSelectionForCapture () const;
    
    // common source interface
    
    track_index PlaybackTracks () const;
    track_index CaptureTracks () const;
    
    sample_index MinStart () const;
    sample_index MaxEnd () const;
    bool IsInputable () const;
    
  protected:
    
    void StartO_impl ();
    void StartIO_impl ();
    void StopIO_impl ();
    
    void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master);
    void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master);
    bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
		    bool is_master);
    bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
		     sample_index buffer_index, bool is_master);
    
    void Debug_impl () const;
    
    track_index build_selection_data ();

    class track_data
    {
    public:
      TrackLayer::Selection* selection;
      TrackLayer::Selection::const_iterator pl_segment;
      TrackLayer::Selection::const_iterator ca_segment;
    };
    
    vector <track_data> m_data;
    
    // internal states
    vector <bool> io_read_pos_vaild;
    vector <bool> io_write_pos_vaild;
    
    bool master_read_pos_vaild;
    bool master_write_pos_vaild;
    
    sample_index m_start_pos;
    sample_index m_end_pos;
    
    // config
    MixerSource* m_source;
    DataLayer* m_layer;
    
    bool m_for_playback;
    bool m_for_capture;
  };
  
  class MarkerFilterSource : public MixerSource, ScatterHelperSource
  {
  public:
    
    MarkerFilterSource (MixerSource* i_src, DataLayer* i_layer);
    ~MarkerFilterSource ();
    
    // custom interface
    
    void SetDataLayer (DataLayer* n_layer);
    
    void SetPlaybackStart (const string& start_mark);
    void SetPlaybackEnd (const string& start_mark);
    void SetCaptureStart (const string& start_mark);
    void SetCaptureEnd (const string& start_mark);
    
    const string& SetPlaybackStart ();
    const string& SetPlaybackEnd ();
    const string& SetCaptureStart ();
    const string& SetCaptureEnd ();
    
    // common source interface
    
    track_index PlaybackTracks () const;
    track_index CaptureTracks () const;
    
    sample_index MinStart () const;
    sample_index MaxEnd () const;
    bool IsInputable () const;
    
  protected:
    
    void StartO_impl ();
    void StartIO_impl ();
    void StopIO_impl ();
    
    void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master);
    void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master);
    bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
		    bool is_master);
    bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
		     sample_index buffer_index, bool is_master);
    
    void Debug_impl () const;
    
    void reget_positions ();
    void limit_start_end ();
    
    void SlotMarkerDeleted (const string& marker, sample_index old_pos);
    void SlotMarkerChanged (const string& marker, sample_index old_pos, sample_index new_pos);
    
    // internal states
    
    Connection connection_marker_deleted;
    Connection connection_marker_changed;
    
    string m_playback_start_mark;
    string m_playback_end_mark;
    string m_capture_start_mark;
    string m_capture_end_mark;
    
    sample_index m_playback_start_pos;
    sample_index m_playback_end_pos;
    sample_index m_capture_start_pos;
    sample_index m_capture_end_pos;
    
    sample_index m_start_pos;
    sample_index m_end_pos;
    
    // config
    MixerSource* m_source;
    DataLayer* m_layer;
  };
  
  class LoopFilterSource : public MixerSource
  {
  public:
    
    LoopFilterSource (MixerSource* i_src);
    ~LoopFilterSource ();
    
    // custom interface
    
    void SetLooping (bool n_state);
    bool Looping ();
    
    void SetStopAtEnd (bool n_state);
    bool StopAtEnd ();
    
    // common source interface
    
    track_index PlaybackTracks () const;
    track_index CaptureTracks () const;
    
    sample_index MinStart () const;
    sample_index MaxEnd () const;
    bool IsInputable () const;
    
  protected:
    
    void StartO_impl ();
    void StartIO_impl ();
    void StopIO_impl ();
    
    void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master);
    void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master);
    bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
		    bool is_master);
    bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
		     sample_index buffer_index, bool is_master);
    
    void Debug_impl () const;
    
    void SlotMinStartChanged (sample_index n_min_start);
    void SlotMaxEndChanged (sample_index n_max_end);
    
    MixerSource* m_source;
    bool m_looping;
    bool m_stop_at_end;
    
    // internal states
    sample_index m_start_pos;
    sample_index m_end_pos;
  };
  
  class MultiplexerFilterSource : public MixerSource, ScatterHelperSource
  {
  public:
    
    MultiplexerFilterSource (MixerSource* pri_src, MixerSource* sec_src);
    ~MultiplexerFilterSource ();
    
    // custom interface
    
    void SetPriSource (MixerSource* pri_src);
    void SetSecSource (MixerSource* sec_src);
    
    void SetWay (bool use_sec); // switch the way: false (default): primary; true: secondary
    bool Way ();
    
    // common source interface
    
    sample_index MinStart () const;
    sample_index MaxEnd () const;
    bool IsInputable () const;
    
  protected:
    
    void StartO_impl ();
    void StartIO_impl ();
    void StopIO_impl ();
    
    void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master);
    void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master);
    bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
		    bool is_master);
    bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
		     sample_index buffer_index, bool is_master);
    
    void Debug_impl () const;
    
    void SlotMinStartChanged (sample_index n_min_start);
    void SlotMaxEndChanged (sample_index n_max_end);
    
    // internal states
    SigC::Threads::Mutex m_lock;
    
    sample_index m_start_pos;
    sample_index m_end_pos;
    
    // config
    MixerSource* m_pri_source;
    MixerSource* m_sec_source;
    
    MixerSource* m_cur_source;
    
    bool m_way;
  };
  
  // The implementation we are normally using
  class UniversalMixerParams : public MixerParams
  {
  public:
    
    UniversalMixerParams ();
    ~UniversalMixerParams ();
    
    // custom interface
    
    void SetTrackValue (track_index track, pcm_default::value_type n_value);
    void SetChannelValue (track_index channel, pcm_default::value_type n_value);
    void SetMatrixValue (track_index track, track_index channel, pcm_default::value_type n_value);
    
    pcm_default::value_type TrackValue (track_index track);
    pcm_default::value_type ChannelValue (track_index channel);
    pcm_default::value_type MatrixValue (track_index track, track_index channel);
    
    Signal2<void, track_index, pcm_default::value_type> SignalTrackValueChanged;
    Signal2<void, track_index, pcm_default::value_type> SignalChannelValueChanged;
    Signal3<void, track_index, track_index, pcm_default::value_type> SignalMatrixValueChanged;
    
    // generic interface
    void SetTracks (track_index n_tracks);
    track_index Tracks ();
    
    void SetChannels (track_index n_channels);
    track_index Channels ();
    
    void SetTrackMute (track_index track, bool n_state);
    bool IsTrackMuted (track_index track);
    
    void SetChannelMute (track_index channel, bool n_state);
    bool IsChannelMuted (track_index channel);
    
    pcm_default::value_type Value (track_index track, track_index channel);
    
    void StartIO ();
    void StopIO ();
    bool IsIORunning ();
    
    void Debug ();
    
    UniversalMixerParams& operator= (const UniversalMixerParams& src);
    
  private:
    
    inline pcm_default::value_type effective_value (track_index track, track_index channel);
    
    bool io_running;
    
    track_index m_tracks;
    track_index m_channels;
    
    vector <pcm_default::value_type> m_track_values;
    vector <pcm_default::value_type> m_channel_values;
    vector <vector <pcm_default::value_type> > m_matrix_values;
    vector <bool> m_track_mutes;
    vector <bool> m_channel_mutes;
  };
  
  // A quick debug implementation ...
  class DebugMixerNode : public MixerNode, public GSMX_utils::Threads::Job
  {
  public:
    
    DebugMixerNode (track_index init_pl_ch, track_index init_cp_ch, int init_period_size, int init_delay);
    ~DebugMixerNode ();
    
    // custom interface
    
    void SetBenchmarking (bool state);
    void SetDelay (int new_delay);
    
    // common interface
    
    bool Init ();
    
    track_index PlaybackChannels ();
    track_index CaptureChannels ();
    
    int PlaybackPeriodSize ();
    int CapturePeriodSize ();
    
    int PlaybackPeriods ();
    int CapturePeriods ();
    
    int PlaybackXRuns ();
    int CaptureXRuns ();
    
    void SetPlayback (bool state);
    void SetCapture (bool state);
    
    bool IsPlayback ();
    bool IsCapture ();
    
    void StartIO (Mixer& source);
    void StopIO ();
    
    bool IsIORunning ();
    
    sample_index GetCurrentPos ();
    
    void Debug ();
    
  protected:
    
    // the second thread ...
    void Loop ();
    
    /* states */
    bool initialized;
    bool io_running;
    
    bool do_playback;
    bool do_capture;
    
    bool stop_requested;
    
    /* frame counting */
    sample_index current_playback_period;
    sample_index current_capture_period;
    
    vector <sample_index> io_playback_periods;
    vector <sample_index> io_capture_periods;
    
    Mixer* m_mixer;
    
    /* config */
    track_index user_capture_channels;
    track_index user_playback_channels;
    
    bool benchmarking;
    timeval bench_start;
    
    int user_period_size;
    int user_delay;
  };
  
} // namespace GSML

#endif //GSML_IO_IMPL_H
