/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   One track data-pattern container.
 */

#ifndef GSML_TRACK_LAYER
#define GSML_TRACK_LAYER

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "SampleBuffer.hh"
#include "DataFile.hh"
#include "DataFileIndex.hh"

#include <deque>
#include <list>

// #define DEBUG_GSML_TRACK_LAYER // for debugging some "couts" ...

namespace GSML
{

  using namespace std;
  using namespace SigC;

  class TrackLayer
    {
    public:
      
      class Except : public BasicExcept {};
      class ExceptNotReadable : public Except {};
      class ExceptNotWriteable : public Except {};
      class ExceptFixated : public Except {};
      class ExceptNotFixated : public Except {};
      class ExceptUnsupported : public Except {};
      class ExceptLogicalError : public Except {};
      
      // only used internal (or for import plug-ins ...)
      // it keeps track of the binding flags of data files ...
      class Entry
	{
	public:
	  // defauts to virtual zero data (see below), register a new file - or use other Constructor
	  Entry ();
	  Entry (DataFile* init_file);
	  Entry (const Entry& source_entry);
	  ~Entry ();
    
	  Entry& operator=  (const Entry& source_entry);
	  
	  void RegisterFile (DataFile* new_file); // registers a new file (and unbinds!!!! the old)

	  DataFile* File (); // returns 0 if virutal zero data
	  const DataFile* File () const;
	  
	  void Debug () const;

	  sample_index length;
	  sample_index start_in_file;
	  track_index track_in_file;
    
	private:
	  DataFile* myfile;
	  // if this is true myfile is a vaild file; if false this is only a holder for virtual empty (zero) data ...
	  bool real_data; 
	};

      class Selection
      {
      public:
	
	class Except : public BasicExcept {};
	class ExceptFixated : public Except {};
	
	typedef IntervalList<sample_index>::iterator iterator;
	typedef IntervalList<sample_index>::const_iterator const_iterator;
	
	Selection (TrackLayer* i_track);
	Selection (const Selection& source);
	
	// some Helper methods
	// returns start of the 1st and end of the last selection segments and a bool if there was a segment ...
	bool MinStart (sample_index* value);
	bool MaxEnd (sample_index* value);
	sample_index SummedLength (); // length of all segments
	
	void Select (sample_index start, sample_index length);
	void Unselect (sample_index start, sample_index length);
	void Clear ();
	void All ();
	void Invert ();
	
	// active flag - indicating if effects should use this selection
	void SetActive (bool state);
	bool IsActive ();
	
	const_iterator begin ();
	const_iterator end ();
	
	Selection& operator= (const Selection& source);
	
	void Debug ();
	
	Signal2<void, sample_index, sample_index> SignalSelected;
	Signal2<void, sample_index, sample_index> SignalUnselected;
	
	Signal1<void, bool> SignalActiveChanged;
	
      protected:
	TrackLayer* m_track;
	
	IntervalList <sample_index> m_intervals;
	
	bool active;
      };
      
      typedef deque<Entry*>::iterator iterator;
      typedef deque<Entry*>::const_iterator const_iterator;
      
      TrackLayer (DataFileIndex* init_index, PCMDatatype* init_temp_type,
		  sample_rate init_temp_rate=44100);
      TrackLayer (const TrackLayer& source_track);
      ~TrackLayer ();
      
      const string& Name () const;
      void Name (const string& n_name);
      
      // for random read access
      bool ReadData (sample_index start, sample_index length, SampleBuffer* destination);
  
      // for !linear! write access (only when write is "prepared" ...)
      void WriteData (sample_index length, SBHandle source, sample_index buffer_index);
      
      sample_index Length () const;

      void PrepareOverWrite (sample_index start);
      void PrepareInsertWrite (sample_index start);
      void FixateWrite ();
      void CancelWrite ();

      void Fixate ();
      void UnFixate ();

      // this two sould only used to load and save files
      // here you have DIRECT access to the internal entries ... - keep care what ya doing!!
      const_iterator Begin () const;
      const_iterator End () const;
      
      const_iterator InsertRawEntry (const_iterator i, const Entry& n_entry);
      void InsertRawEntry (sample_index index, const Entry& n_entry);
      
      // create and return a new data_layer with the copy of the given region ...
      TrackLayer* CreateTrackLayer (sample_index start,
				    sample_index length) const;

      void InsertTrackLayer (sample_index start, const TrackLayer& source_track);
      void RemoveRegion (sample_index start, sample_index length);
      
      DataFileIndex* GetDataFileIndex ();
      PCMDatatype* TempSampletype ();
      sample_rate TempSamplerate ();
      
      void SetTempSampletype (PCMDatatype* new_type);
      void SetTempSamplerate (sample_rate new_rate);
      
      TrackLayer& operator= (const TrackLayer& source_track);
  
      // signals
      // use SBHandle to be able to have multiple threads reading the data cleanly ...
      Signal1<void, const string&> SignalNameChanged;
      
      Signal3<void, sample_index, SBHandle, sample_index> SignalWroteData;
      Signal2<void, sample_index, sample_index> SignalRegionRemoved;
      Signal2<void, sample_index, sample_index> SignalRegionChanged;
  
      Signal1<void, sample_index> SignalPreparedOverWrite;
      Signal1<void, sample_index> SignalPreparedInsertWrite;
      Signal0<void> SignalFixatedWrite;
      Signal0<void> SignalCanceledWrite;
  
      void Debug () const;
      
      Selection selection;
      
    protected:
      friend class Selection;
      
      // internal helper methods ...
      
      sample_index length () const;
      
      void prepare_write (sample_index start);
      
      // returns iter to the inserted (new) entry
      iterator insert_entry (sample_index pos, const Entry& n_entry);
      void remove_region (sample_index start, sample_index length);
      
      iterator  get_entry4index (sample_index index) const;
      sample_index get_start4entry (const iterator& entry) const;
      // returns iter to 2nd (new) entry
      iterator split_entry (const iterator& split_entry, sample_index index_in_entry);

      deque<Entry*>* m_data;
      
      // config data
      DataFileIndex* m_fileindex;
      PCMDatatype* m_temp_sampletype;
      sample_rate m_temp_samplerate;
      string m_name;
      
      // internal states
      bool fixated; // is the layer fixated an though locked for write access
      bool write_locked; // is in write state
      bool overwrite; // the data is overwritten during write
      
      sample_index write_start; // pos, write started
      sample_index write_length; // written samples during write
      DataFile* write_file; // data file for write
  
      // BasicLock busy; // lock for thread safty (I hope ...)
    };

} // end namespae GSML

#endif //GSML_TRACK_LAYER
