/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   WavePlot architecture classes, interfaces and MinMaxAvg implementaion.
 *   This stuff stream and calculate the data for graphic views in the
 *   background.
 */

#ifndef GSML_WAVE_PLOT
#define GSML_WAVE_PLOT

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "Threads.hh"

#include "Data.hh"

#include <typeinfo>

#include <deque>
#include <list>
#include <vector>

namespace GSML
{
  // TODO: let the Manager have one WavePlot per Track for global caching
  
  // #define DEBUG_GSML_WAVE_PLOT
  // #define DEBUG_GSML_WAVE_PLOT_MANAGER
  // #define DEBUG_GSML_WAVE_PLOT_MIN_MAX
  
  class WavePlotConfig
  {
  public:
    WavePlotConfig ();
    
    void SetStart (sample_index n_start);
    void SetPlotLength (sample_index n_plot_length);
    void SetSamplesPerSample (sample_index n_sps);
    
    void SetViewStart (sample_index n_view_start);
    void SetViewLength (sample_index n_view_length);
    
    sample_index Start () const;
    sample_index Length () const;
    sample_index PlotLength () const;
    sample_index SamplesPerSample () const;
    
    sample_index ViewStart () const;
    sample_index ViewLength () const;
    
    bool operator == (const WavePlotConfig& src) const;
    
    void Debug () const;
    
  protected:
    sample_index m_start;
    sample_index m_plot_length;
    sample_index m_sps;
    
    sample_index m_view_start;
    sample_index m_view_length;
  };
  
  class WavePlotProgress
  {
  public:
    WavePlotProgress ();
    
    sample_index Length () const;
    virtual void SetLength (sample_index n_length);
    
    // per index requests - should be inlined
    bool IsUseable (sample_index pos) const {
      if (pos > m_length) {
	BasicExcept e;
	e.info = "WavePlotProgress::IsUseable: wrong size";
      }
      return m_progress [pos] == VAILD  || m_progress [pos] == INTERPOLATED;
    }

    bool IsVaild (sample_index pos) const {
      if (pos > m_length) {
	BasicExcept e;
	e.info = "WavePlotProgress::IsUsable: wrong size";
      }
      return m_progress [pos] == VAILD;
    }
    
    bool IsInterpolated (sample_index pos) const {
      if (pos > m_length) {
	BasicExcept e;
	e.info = "WavePlotProgress::IsUseable: wrong size";
      }
      return m_progress [pos] == INTERPOLATED;
    }
    
    // range requests
    bool IsVaild (sample_index pos, sample_index length) const;
    
    // complete requests
    bool IsAllVaild () const;
    sample_index GetInvaildStart () const;
    sample_index GetInvaildEnd () const; // end of first segment
    
    // actions
    void Invaild (sample_index start, sample_index length);
    void InvaildAll ();
    
    void Vaild (sample_index start, sample_index length);
    void VaildAll ();
    
    void Interpolate (sample_index start, sample_index length);
    
    // complex actions
    virtual void Move (sample_index start, sample_index length, sample_index new_start);
    virtual void Zero (sample_index start, sample_index length);
    
    virtual void Debug (bool verbose = false) const;
    
  protected:
    void set_length_impl (sample_index n_length);
    void move_impl (sample_index start, sample_index length, sample_index new_start);
    void zero_impl (sample_index start, sample_index length);
    void debug_impl (bool verbose) const;
    
    typedef enum {INVAILD, INTERPOLATED, VAILD} progress_state;

    vector<progress_state> m_progress;
    
    sample_index m_length;
  };
  
  class WavePlot; // forward
  
  class WavePlotManager : public SigC::Object, public GSMX_utils::Threads::Job
  {
  public:
    class Except : public BasicExcept {};
    
    WavePlotManager (Data* i_data);
    ~WavePlotManager ();
    
    Data* GetData ();
    DataLayer* CurrentLayerPtr ();
    
    void SetCacheLayers (unsigned int n_max_layers);
    unsigned int CacheLayers ();
    
    DataLayer* RegisterWavePlot (WavePlot* new_plot);
    bool UnRegisterWavePlot (WavePlot* old_plot);
    
    // hacked halt interface
    void WaitRequest ();
    void ContinueRequest ();
    
    // TODO: some try to recycle whatever hack / hook
    
    void Debug (bool verbose=false);
    
  private:
    const WavePlotManager& operator= (const WavePlotManager& src);
    
    // internal data-types
    struct PlotStuff
    {
    public:
      WavePlot* plot;
      string type; // don't ask ... - the c++ type_info can't be copied #argh# !!
    };
    
    struct LayerStuff
    {
    public:
      DataLayer* layer;
      deque <PlotStuff> plots;
    };
    
    typedef deque <LayerStuff>::iterator iterator_l;
    typedef deque <PlotStuff>::iterator  iterator_p;
    
    // Slots from Data
    void SlotCurrentLayerChanged (DataLayer* n_layer);
    void SlotNewLayerPrepared (DataLayer* n_layer);
    void SlotNewLayerFixated ();
    void SlotNewLayerCanceled ();
    
    // Slots from Plots
    void SlotConfigChanged (const WavePlotConfig& n_config, bool real_change);
    
    void CloneLayer (LayerStuff* n_stuff, DataLayer* n_layer, bool is_prepared);
    void DeleteIfTooManyLayer ();
    
    void InternalWait ();
    void InternalContinue ();
    
    // 2nd thread stuff
    int CreateListsForStage (int stage);
    void Loop ();
    
    struct LoopPlotStuff
    {
      LoopPlotStuff ()
      {
	plot = 0;
	start = end = 0;
	prepared = false;
      };
      
      WavePlot* plot;
      sample_index start;
      sample_index end;
      bool prepared;
    };
    
    // start / end is out-dated replaces wiht loop-IntervalList - maybe a per track list ??
    struct LoopTrackStuff
    {
      LoopTrackStuff ()
      {
	//start = end = 0;
      };
      //sample_index start;
      //sample_index end;
      
      list <LoopPlotStuff> plots;
    };
    
    bool stop;
    bool m_first_loop;
    int m_loop_stage; // 0: visible ranges, 1: all 
    
    track_index loop_tracks;
    vector <LoopTrackStuff> loop_plots;
    IntervalList <sample_index> loop_intervals;
    typedef list <LoopPlotStuff>::iterator l_plots_it;
    
    SigC::Threads::Mutex loop_plots_lock;
    
    // END 2nd stuff
    
    // config:
    Data* m_data;
    int max_layers;
    
    sample_index m_stream_chunk_size; // chunk during Loop::streaming
    
    // states
    bool prepared;
    
    int halt_level; // how often got HaltRequest called
    
    LayerStuff prepared_layer;
    iterator_l current_layer;
    deque <LayerStuff> layers;
  };
  
  /* The common interface of WavePlots
   * Short infos: The managers Register can't be called in the ctor (virutal obj is not full constructed ...)!
   *              It is done in the Config Method. It is the current and get the layer.
   */
  class WavePlot : public SigC::Object
  {
  public:
    class Except : public BasicExcept {};
    class ExceptLogicalError : public Except {};

    WavePlot (WavePlotManager* i_manager, int i_track_no, WavePlotProgress* i_progress);
    virtual ~WavePlot ();
    
    void Config (const WavePlotConfig& n_config);
    const WavePlotConfig& GetConfig ()  const;
    
    const WavePlotProgress& GetProgress () const;
    
    // used for cloning by Manager - only way to change the layer
    // the inheritenced class MUST use the helper below to copy our basic properties!
    virtual WavePlot* Clone (DataLayer* n_layer, bool is_prepared) = 0;
    void CloneHelper (DataLayer* n_layer, WavePlot* n_plot, bool is_prepared);
    
    void UnsetCurrent (WavePlot* n_current);
    void SetCurrent ();
    
    DataLayer* GetDataLayer ();
    TrackLayer* GetTrackLayer ();
    
    inline track_index GetTrack ();
    
    virtual void Debug () const;
    
    Signal1<void, WavePlot*> SignalUnsetCurrent;
    Signal2<void, sample_index, sample_index> SignalProgressUpdated;
    
    Signal2<void, const WavePlotConfig&, bool> SignalConfigChanged;
    
  protected:
    
    const WavePlot& operator= (const WavePlot& src);
    
    friend class WavePlotManager; // for data streaming ...
    
    /*****************************************
     * Stream write from the WavePlotManager *
     *****************************************/
    
    void PrepareStreamWrite (sample_index start);
    void StreamWrite (sample_index length, SampleBufferDefault* source, sample_index buffer_index);
    void FixateStreamWrite ();
    
    /****************************************
     * "signal" handlers / connectionws ... *
     ****************************************/
    
    // Data_Layer
    void SlotTrackCreated (track_index track);
    void SlotTrackErased (track_index track);
    
    // Track_Layer
    void SlotPreparedOverWrite (sample_index start);
    Connection connection_prepared_overwrite;
    void SlotPreparedInsertWrite (sample_index start);
    Connection connection_prepared_insertwrite;
    void SlotFixatedWrite ();
    Connection connection_fixated_write;
    void SlotCanceledWrite ();
    Connection connection_canceled_write;
    
    void SlotWroteData (sample_index length, SBHandle source, sample_index buffer_index);
    Connection connection_wrote_data;
    
    void SlotRegionRemoved (sample_index start, sample_index length);
    Connection connection_region_removed;
    void SlotRegionChanged (sample_index start, sample_index length);
    Connection connection_region_changed;
    
    /************************************
     * some other internal helper stuff *
     ************************************/
    
    void internal_prepare_write (sample_index start, bool do_buffer=true);
    void internal_write (sample_index length, SampleBuffer* source, sample_index buffer_index);
    void internal_fixate (bool was_insert=false);
    void internal_cancel ();
    
    void connect_layer ();
    void connect_track ();
    void disconnect_track ();
    
    /******************************
     * for the inheritanced class *
     ******************************/
    
    virtual void prepare_buffer_impl () = 0;
    virtual void fixate_buffer_impl () = 0;
    virtual void cancel_buffer_impl () = 0;
    
    virtual void process_data_impl (sample_index length, const SampleBuffer* buffer,
				    sample_index buffer_index, sample_index progress_index) = 0;
    
    // states:
    TrackLayer* connected_track;
    
    bool registered;
    bool current;
    
    bool external_stream; // if external stream is processed (WP-Manager)
    SigC::Threads::Mutex external_stream_lock;
    
    bool connected;  // if connected to a prepared track
    bool listening; // if connected to a TrackLayer during FX write
    bool passing; // if passing the stream to the inheritenced class
    bool insert_write; // if in overwrite mode
    
    bool stream_buffered; // is the write stream buffered?? (true in FX writes)
    sample_index stream_start; // write - data - stream values
    sample_index stream_first_alligned;
    sample_index stream_length;
    
    sample_index progress_signal_pos; // last emmited signal's pos
    
    sample_index m_progress_pos;
    
    SampleBufferDefault* m_between_write_buffer;
    sample_index m_in_write_buffer;
    
    // config:
    WavePlotManager* m_manager;
    
    DataLayer* m_layer;
    WavePlotConfig m_config;
    WavePlotProgress* m_progress;
    int m_track_no;
  };
  
  class MinMaxAvgWavePlot : public WavePlot
    {
    public:
      
      class Progress : public WavePlotProgress
      {
      public:
	
	virtual void SetLength (sample_index n_length);
	
	virtual void Move (sample_index start, sample_index length, sample_index new_start);
	virtual void Zero (sample_index start, sample_index length);
	virtual void Debug (bool verbose = false) const;
	
	vector<pcm_default::value_type> min_data;
	vector<pcm_default::value_type> max_data;
	vector<pcm_default::value_type> neg_avg_data;
	vector<pcm_default::value_type> pos_avg_data;
      
      protected:
	void set_length_impl (sample_index n_length);
	void move_impl (sample_index start, sample_index length, sample_index new_start);
	void zero_impl (sample_index start, sample_index length);
	void debug_impl (bool verbose) const;
      };
      
      MinMaxAvgWavePlot (WavePlotManager* i_manager, int i_track);
      ~MinMaxAvgWavePlot ();
      
      const MinMaxAvgWavePlot::Progress& GetProgress () const;
      
      WavePlot* Clone (DataLayer* n_layer, bool is_prepred);
      
      void Debug ();
      
    protected:
      
      void prepare_buffer_impl ();
      void fixate_buffer_impl ();
      void cancel_buffer_impl ();
      
      void process_data_impl (sample_index length, const SampleBuffer* buffer,
			      sample_index buffer_index, sample_index progress_index);
      
      bool buffering;
      
      Progress m_progress;
      
      // temp data when the Plot recieves write_data (cause this can be canceled)
      vector<pcm_default::value_type> min_temp_data;
      vector<pcm_default::value_type> max_temp_data;
      vector<pcm_default::value_type> neg_avg_temp_data;
      vector<pcm_default::value_type> pos_avg_temp_data;
  
      // points to the internal data which has to be used during the next operations
      vector<pcm_default::value_type>* min_working_data;
      vector<pcm_default::value_type>* max_working_data;
      vector<pcm_default::value_type>* neg_avg_working_data;
      vector<pcm_default::value_type>* pos_avg_working_data;
    };

  /*
    class FFTWavePlot : public WavePlot
    {
    };
    
  */
  
  /*
   * This is an WavePlot _auto-trace_ template. It _follows_ the changeing WavePlots
   * and tracks the current one ...
   */

  template <class PT>
  class WavePlotTracer : public SigC::Object
  {
  public:
    
    WavePlotTracer (GSML::WavePlotManager* i_manager, int i_track_no);
    ~WavePlotTracer ();
    
    
    // Wrapped WavePlot methods
    void Config (const GSML::WavePlotConfig& n_config);
    
    const GSML::WavePlotProgress& GetProgress () const;
    
    void Debug () const;
    
    // additional stuff
    PT* Plot ();
    
  protected:
    SigC::Connection connection_progress_updated;
  public:
    // wrapped signals
    SigC::Signal2<void, GSML::sample_index, GSML::sample_index> SignalProgressUpdated;
    SigC::Signal1<void, PT*> SignalWavePlotChanged;
    
  protected:
    
    SigC::Connection connection_unset_current;
    void SlotUnsetCurrent (GSML::WavePlot* new_current_plot);
    
    PT* current_plot;
    
    GSML::WavePlotManager* m_manager;
    GSML::track_index m_track_no;
  };
  
  template <class PT>
  WavePlotTracer<PT>::WavePlotTracer (GSML::WavePlotManager* i_manager, int i_track_no)
  {
    m_manager = i_manager;
    m_track_no = i_track_no;
    
    current_plot = new PT (m_manager, m_track_no);
    connection_unset_current
      = current_plot->SignalUnsetCurrent.connect (slot (this, &WavePlotTracer<PT>::SlotUnsetCurrent) );
    connection_progress_updated
      = current_plot->SignalProgressUpdated.connect(SignalProgressUpdated.slot() );
  }
  
  template <class PT>
  WavePlotTracer<PT>::~WavePlotTracer ()
  {
    // (no plots have to be deleted - the manager does this!!)
    if (!m_manager->UnRegisterWavePlot (current_plot) )
      cout << "~WavePlotTracer: UnRegister of current WP failed!! !?!" << endl;
  }
  
  
  template <class PT>  
  void WavePlotTracer<PT>::Config (const GSML::WavePlotConfig& new_config)
  {
    current_plot->Config (new_config);
  }
  
  template <class PT>
  const GSML::WavePlotProgress& WavePlotTracer<PT>::GetProgress () const
  {
    return current_plot->GetProgress ();
  }
  
  template <class PT>
  void WavePlotTracer<PT>::Debug () const
  {
    current_plot->Debug ();
  }
  
  template <class PT>
  PT* WavePlotTracer<PT>::Plot ()
  {
    return current_plot;
  }
  
  template <class PT>
  void WavePlotTracer<PT>::SlotUnsetCurrent (GSML::WavePlot* new_current_plot)
  {
    // cout << "disconnecting" << endl;
    connection_unset_current.disconnect ();
    connection_progress_updated.disconnect ();
    
    // TODO: find the BUG with this dynamic_cast ...
    
    // cout << "dyn_cast" << endl;
    current_plot = dynamic_cast<PT*> (new_current_plot);
    // cout << "addr: " << current_plot << endl;
    
    // cout << "BUG: using normal cast - because dyn_cast is buggy here ..." << endl;
    
    current_plot = (PT*) new_current_plot;
    
    // cout << "current_plot: " << current_plot << " (got: " << new_current_plot << ")"<<  endl;
    // new_current_plot->Debug ();
    // current_plot->Debug ();
    
    if (!current_plot)
      cout << "BUG: WavePlotTracer: Can't dynamic_cast the new current plot ??? ..." << endl;
    else
      {
	// cout << "reconnecting" << endl;
	connection_unset_current
	  = current_plot->SignalUnsetCurrent.connect (slot (this, &WavePlotTracer<PT>::SlotUnsetCurrent) );
	connection_progress_updated
	  = current_plot->SignalProgressUpdated.connect(SignalProgressUpdated.slot() );
	
	SignalWavePlotChanged.emit (current_plot);
      }
  }
  
} // end namespace GSML

#endif // GSML_WAVE_PLOT
