/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Our first type-convertion try.
 *   Very old-style C / C++. Will be replaced with a new dynamic-type system
 *   in a future release ... .
 */

#ifndef GSML_TYPEDEFS
#define GSML_TYPEDEFS

#include "include.hh"

#include <stddef.h>
#include <list>

namespace GSML
{

#define DIR_CREATE_MODE 0755
#define FILE_CREATE_MODE 0644
  
  // some nicer names for basic types
  
  typedef signed char int8;
  typedef signed short int int16;
  typedef signed int int32;
  typedef signed long long int64;
  
  typedef unsigned char uint8;
  typedef unsigned short int uint16;
  typedef unsigned int uint32;
  typedef unsigned long long uint64;
  
  typedef int64 sample_index;
  typedef uint32 sample_rate;
  typedef uint32 track_index;
  typedef uint32 sample_rate;
  
  class PCMDatatype
  {
  public:
    virtual int Size () = 0;
  };
  
  template <class RTZ>
  class PCMDatatypeTemplate : public PCMDatatype
  {
  public:
    
    typedef struct 
    {
      RTZ data;
      inline void operator () (RTZ x) {data = x;};
      inline operator RTZ () {return data;}
    } value_type;
    
    virtual value_type Min () = 0;
    virtual value_type Max () = 0;
    
    virtual value_type dBfs () = 0; // dB full scale
    
    int Size ()
     {
       return sizeof (RTZ);
     } 
  };
  
  /*
   * 8 bit implementation
   */
  
  class pcm_8bit : public PCMDatatypeTemplate <int8>
  {
  public:

    value_type Min ()
    {
      return (value_type) {0x80};
    }

    value_type Max ()
    {
      return (value_type) {0x7f};
    }
  
    value_type dBfs ()
    {
      return (value_type) {0x7f};
    }

  };
  
  /*
   * 16 bit implementation
   */
  
  class pcm_16bit : public PCMDatatypeTemplate <int16>
  {
  public:

    value_type Min ()
    {
      return (value_type) {0x8000};
    }

    value_type Max ()
    {
      return (value_type) {0x7fff};
    }
  
    value_type dBfs ()
    {
      return (value_type) {0x7fff};
    }
  };

  /*
   * 24+8 bit fixt point implementation
   */
  
  class pcm_24_8bit : public PCMDatatypeTemplate <int32>
  {
  public:

    value_type Min ()
    {
      return (value_type) {0x80000000};
    }

    value_type Max ()
    {
      return (value_type) {0x7fffffff};
    }
  
    value_type dBfs ()
    {
      return (value_type) {0x00800000};
    }
    
    static int16 Convert2int16 (value_type src)
    {
      if (src > 0x007fffff)
	return 0x7fff;
      if (src < 0x00800000 * (-1) )
	return 0x8000;
      return (int16) (src >> 8);
    };
    
    static int8 Convert2int8 (value_type src)
    {
      if (src > 0x007fffff)
	return 0x7f;
      if (src < 0x00800000 * (-1) )
	return 0x80;
      return (int8) (src >> 16);
    };
  };

  
  inline pcm_24_8bit::value_type operator * (const pcm_24_8bit::value_type& factor1, const pcm_24_8bit::value_type& factor2)
  {
    return (pcm_24_8bit::value_type)  
      { (int32) ( ((int64) factor1.data * (int64) factor2.data) >> 23) }; // Illuminati were here ;-)
  }
  
  inline pcm_24_8bit::value_type operator + (const pcm_24_8bit::value_type& summand1, const pcm_24_8bit::value_type& summand2)
  {
    return (pcm_24_8bit::value_type) 
      { summand1.data + summand2.data };
  }
  
  inline pcm_24_8bit::value_type operator - (const pcm_24_8bit::value_type& minu, const pcm_24_8bit::value_type& sub)
  {
    return (pcm_24_8bit::value_type) 
      { minu.data - sub.data };
  }
  
  typedef pcm_24_8bit pcm_default; // sampletype used for ALL computations in GSMx
  
  // two Region classes to be able to have common computation classes
  template <class T>
  class StartEndInterval
  {
  public:
    T start;
    T end; // half-open: the end doen't belong to the interval
    // -> end - start == length
  };
  
  template <class T>
  class StartLengthInterval
  {
  public:
    T start;
    T length;
  };
  
  template <class T>
  class IntervalList
  {
  public:
    typedef StartEndInterval<T> Interval;
    
    typedef list<Interval>::iterator iterator;
    typedef list<Interval>::const_iterator const_iterator;
    
    void Add (T start, T length);
    void Remove (T start, T length);
    void Clear ();
    
    void All (T all_end);
    void Invert (T invert_end);
    
    const_iterator begin () const;
    const_iterator end () const;
    
    // some helpers
    bool MinStart (T* value) const;
    bool MaxEnd (T* value) const;
    T SummedLength () const; // length of all intervals
    
    IntervalList& operator= (const IntervalList& source);
    
    SigC::Signal2<void, T, T> SignalAdded;
    SigC::Signal2<void, T, T> SignalRemoved;
    
  protected:
    
    list <Interval> intervals;
    
  };
  
  template <class T>
  void IntervalList<T>::Add (T start, T length) {
    iterator i;
    bool end = false;
    
    Interval s;
    s.start = start;
    s.end = start + length;
    
    // remove all segments in the new selection
    i = intervals . begin ();
    while (!end && i != intervals.end () )
      {
	// in this segment ??
	if (s.start <= i -> end && s.end >= i -> start)
	  {
	    s.start = limit_max (s.start, i -> start);
	    s.end = limit_min (s.end, i -> end);
	    
	    i = intervals.erase (i);
	  }
	else if  (s.start < i -> start)
	  end = true;
	else
	  ++ i;
      }
    intervals.insert (i, s);
    SignalAdded (start, length);
  }
  
  template <class T>
  void IntervalList<T>::Remove (T start, T length) {
    iterator i;
    sample_index t_end = start + length;
    
    bool end = false;
	  
    // shorten and remove all segments in the new selection
    i = intervals.begin ();
    while (!end && i != intervals.end () )
      {
	if (start <= i -> start && t_end >= i -> end)
	  {
	    i = intervals.erase (i);
	  }
	else
	  {
	    if (start >= i -> start && t_end <= i -> end)
	      {
		Interval s;
		s.start = i -> start;
		s.end = start;
		i -> start = t_end;
		i = intervals.insert (i, s);
	      }
	    else if (start >= i -> start && start <= i -> end)
	      i->end = start;
	    else if (t_end >= i -> start && t_end <= i -> end)
	      {
		i -> start = t_end;
		end = true; // sorted -> must be the last affect
	      }
	  
	    ++ i;
	  }
      }
    SignalRemoved (start, length);
  }
  
  template <class T>
  void IntervalList<T>::Clear () {
    iterator i;
    
    i = intervals.begin ();
    while (i != intervals.end () ) {
      T t_start = i -> start;
      T t_length = i -> end - i -> start;
      
      i = intervals.erase (i);
      SignalRemoved (t_start, t_length);
    }
  }
    
  template <class T>
  void IntervalList<T>::All (T all_end) {
    Interval s;
    s.start = 0;
    s.end = all_end;
  
    intervals.clear ();
    intervals.push_back (s);
    SignalAdded (0, all_end);
  }
  
  template <class T>
  void IntervalList<T>::Invert (T invert_end) {
    iterator i;
    T cache_pos = 0;
    T end_pos = invert_end;
    
    // remove first if start at 0
    i = intervals.begin ();
    if (i != intervals.end () && i -> start == 0)
      {
	cache_pos = i -> end;
	i = intervals.erase (i);
	SignalRemoved (0, cache_pos);
      }
  
    while (i != intervals.end () )
      {
	T t_pos = i -> end;
      
	T sel_start = cache_pos;
	T sel_length = i -> start - cache_pos;
	T unsel_start = i -> start;
	T unsel_length = i -> end - i -> start;
      
	i -> end = i -> start;
	i -> start = cache_pos;
	cache_pos = t_pos;
	
	SignalAdded (sel_start, sel_length);
	SignalRemoved (unsel_start, unsel_length);
	
	++ i;
      }
  
    if (cache_pos < end_pos)
      {
	Interval s;
	s.start = cache_pos;
	s.end = end_pos;
	intervals.push_back (s);
	
	SignalAdded (s.start, s.end - s.start );
      }
  }
  
  template <class T>
  IntervalList<T>::const_iterator IntervalList<T>::begin () const {
    return intervals.begin ();
  }
  
  template <class T>
  IntervalList<T>::const_iterator IntervalList<T>::end () const {
    return intervals.end ();
  }
  template <class T>
  IntervalList<T>& IntervalList<T>::operator= (const IntervalList<T>& source)
  {
    if (this != &source) {
      Clear ();
      for (const_iterator i = source . begin (); i != source . end (); ++ i)
	Add (i -> start, i -> end - i -> start);
    }
  }
  
  template <class T>
  bool IntervalList<T>::MinStart (T* value) const {
    if (intervals . begin () == intervals . end () )
      return false;
    
    *value = intervals . front () . start;
    return true;
  }
  
  template <class T>
  bool IntervalList<T>::MaxEnd (T* value) const {
    if (intervals . begin () == intervals . end () )
      return false;
    
    *value = intervals . back () . end;
    return true;
  }
  
  template <class T>
  T IntervalList<T>::SummedLength () const {
    T t_length = 0;
    for (const_iterator i = intervals . begin (); i != intervals . end (); ++ i)
      t_length += i -> end - i -> start;
    return t_length;
  }
  
} // end namespace GSML

#endif //GSML_TYPEDEFS
