/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   ALSA C++ wrapper
 */

#include "ALSAHelper.hh"

#ifdef GSML_WITH_ALSA

using namespace ALSAHelper;

/***************
 * ALSAControl *
 ***************/

ALSAControl::ALSAControl ()
{
  opened = false;
}

ALSAControl::~ALSAControl ()
{
  if (opened)
    {
      cout << "ERRORR: BUG: WARNING!!!!~ALSAControl: card is opend in dtor!?!?!?! ..." << endl;
      cout << "             try to close ..." << endl;
      Close ();
    }
}

string ALSAControl::GetName (int card)
{
  char* t_c_name;
  string t_name;
  
  if ( (err = snd_card_get_name (card, &t_c_name) ) < 0)
    {
      cout << "  Cannot open Control device!" << endl;
      return "";
    }
  
  t_name = t_c_name;
  free (t_c_name);
  return t_name;
}

bool ALSAControl::Open (string name, int mode)
{
  cname = name;
  
  if ( (err = snd_ctl_open (&handle, (char*) cname.c_str(), mode) ) < 0)
    {
      cout << "  Cannot open Control device!" << endl;
      return false;
    }
  
  
  opened = true;
  
  return true;
}

bool ALSAControl::Close ()
{
  if (!opened)
    return false;
  
  if ( (err = snd_ctl_close (handle) ) < 0)
    {
      cout << "Cannot close Control device!" << endl;
      return false;
    }
  opened = false;
  return true;
}

/************
 * ALSAPCM *
 ************/

ALSAPCM::ALSAPCM ()
{
  
  if ( (err = snd_pcm_hw_params_malloc (&hw_params) ) < 0)
    {
      ExceptOutOfMemory e;
      e.info = "could no allocate hw params structure";
      throw e;
    }
  
  if ( (err = snd_pcm_sw_params_malloc (&sw_params) ) < 0)
    {
      ExceptOutOfMemory e;
      e.info = "could no allocate sw params structure";
      throw e;
    }
  
  if ( (err = snd_pcm_status_malloc (&status) ) < 0)
    {
      ExceptOutOfMemory e;
      e.info = "could no allocate status structure";
      throw e;
    }
  opened = false;
}

ALSAPCM::~ALSAPCM ()
{
  if (opened)
    {
      cout << "ERRORR: BUG: WARNING!!!!~ALSAPCM: card is opend in dtor!?!?!?! ..." << endl;
      cout << "             try to close ..." << endl;
      Close ();
    }
  
  snd_pcm_status_free (status);
  snd_pcm_sw_params_free (sw_params);
  snd_pcm_hw_params_free (hw_params);
}

bool ALSAPCM::Open (const string& name, snd_pcm_stream_t stream, int mode)
{
  cname = name;
  
  if ( (err = snd_pcm_open (&handle, (char*) cname.c_str(),
			   stream, mode) ) < 0)
    {
      cout << "  Cannot open PCM device!" << endl;
      return false;
    }
  
  if ( (err = snd_pcm_hw_params_any (handle, hw_params) ) < 0)
    {
      cout << "Cannot get hw_params!" << endl;
      return false;
    } 
  
  opened = true;
  
  return true;
}

bool ALSAPCM::Close ()
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_close (handle) ) < 0)
    {
      cout << "Cannot close PCM device!" << endl;
      return false;
    }
  opened = false;
  return true;
       
}

snd_pcm_type_t  ALSAPCM::Type ()
{
  return snd_pcm_type (handle);
}

bool ALSAPCM::InstallParams ()
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params (handle, hw_params) ) < 0)
    {
      cout << "Cannot install hw_params!" << endl;
      return false;
    }
  
  // (re)get sw params
  if ((err = snd_pcm_sw_params_current (handle, sw_params) ) < 0)
    {
      cout << "Cannot get sw_params!" << endl;
      return false;
    }
  return true;
}

bool ALSAPCM::SetAccess (snd_pcm_access_t access_mode)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_access (handle, hw_params,
					   access_mode) ) < 0) {
    cout << "  cannot set access mode!" << endl;
    return false;
  }
  return true;
}

int ALSAPCM::GetAccess ()
{
  return snd_pcm_hw_params_get_access (hw_params);
}

bool ALSAPCM::SetFormat (snd_pcm_format_t format)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_format (handle, hw_params, format) ) < 0)
    {
      cout << "Hardware sample format is not supported!" << endl;
      return false;
    }
  return true;
}

int ALSAPCM::GetFormat ()
{
  return snd_pcm_hw_params_get_format (hw_params);
}

bool ALSAPCM::SetRate (unsigned int rate)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_rate (handle, hw_params, rate,  0) ) < 0)
    {
    cout << "Cannot set sample/period rate to " << rate << "!" << endl;
    return false;
    }
  return true;
}

bool ALSAPCM::SetRateNear (unsigned int rate)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_rate_near (handle, hw_params, rate,  0) ) < 0)
    {
    cout << "Cannot set sample/period rate near " << rate << "!" << endl;
    return false;
    }
  return true;
}

int ALSAPCM::GetRateMin ()
{
  return snd_pcm_hw_params_get_rate_min (hw_params, 0);
}

int ALSAPCM::GetRateMax ()
{
  return snd_pcm_hw_params_get_rate_max (hw_params, 0);
}

int ALSAPCM::GetRate ()
{
  return snd_pcm_hw_params_get_rate (hw_params, 0);
}

bool ALSAPCM::SetChannels (unsigned int channels)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_channels (handle, hw_params,
					     channels) ) < 0)
    {
      cout << "Cannot set channels to " << channels << "!" << endl;
      return false;
    }
  return true;
}

int ALSAPCM::GetChannels ()
{
  return snd_pcm_hw_params_get_channels (hw_params);
}

int ALSAPCM::GetChannelsMin ()
{
  return snd_pcm_hw_params_get_channels_min (hw_params);
}

int ALSAPCM::GetChannelsMax ()
{
  return snd_pcm_hw_params_get_channels_max (hw_params);
}

bool ALSAPCM::SetBufferTimeNear (unsigned int time)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_buffer_time_near (handle, hw_params,
						    time, 0) ) < 0) {
    cout << "Cannot set buffer time near to " << time << "!" << endl;
    return false;
  }
  return true;
}

int ALSAPCM::GetBufferTime ()
{
  return snd_pcm_hw_params_get_buffer_time (hw_params, 0);
}

bool ALSAPCM::SetBufferSize (snd_pcm_uframes_t size)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_buffer_size (handle, hw_params,
						size) ) < 0) {
    cout << "Cannot set buffer size to " << size << "!" << endl;
    return false;
  }
  return true;
}

int ALSAPCM::GetBufferSize ()
{
  return snd_pcm_hw_params_get_buffer_size (hw_params);
}

bool ALSAPCM::SetBufferSizeNear (snd_pcm_uframes_t size)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_buffer_size_near (handle, hw_params,
						size) ) < 0) {
    cout << "Cannot set buffer size near to " << size << "!" << endl;
    return false;
  }
  return true;
}

bool ALSAPCM::SetPeriodTimeNear (unsigned int time)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_hw_params_set_period_time_near (handle, hw_params,
						    time, 0) ) < 0) {
    cout << "Cannot set periods time near to " << time << "!" << endl;
    return false;
  }
  return true;
}

int ALSAPCM::GetPeriodTime ()
{
  return snd_pcm_hw_params_get_period_time (hw_params, 0);
}

bool ALSAPCM::SetPeriodSizeNear (snd_pcm_uframes_t size)
{
  if (!opened)
    return false;
  
  if ((err =  snd_pcm_hw_params_set_period_size_near (handle, hw_params, size, 0) ) < 0) {
    cout << "Cannot set periods size near to " << size << "!" << endl;
    return false;
  }
  return true;
}

int ALSAPCM::GetPeriodSize ()
{
  return snd_pcm_hw_params_get_period_size (hw_params, 0);
}

bool ALSAPCM::SetPeriodSize (snd_pcm_uframes_t size)
{
  if (!opened)
    return false;
  
  if ((err =  snd_pcm_hw_params_set_period_size (handle, hw_params, size, 0) ) < 0) {
    cout << "Cannot set periods size to " << size << "!" << endl;
    return false;
  }
  return true;
}

bool ALSAPCM::SetPeriods (unsigned int periods)
{
  return snd_pcm_hw_params_set_periods (handle, hw_params, periods, 0);
}

bool ALSAPCM::SetPeriodsNear (unsigned int periods)
{
  return snd_pcm_hw_params_set_periods_near (handle, hw_params, periods, 0);
}

int ALSAPCM::GetPeriodsMin ()
{
  return snd_pcm_hw_params_get_periods_min (hw_params, 0);
}

int ALSAPCM::GetPeriodsMax ()
{
  return snd_pcm_hw_params_get_periods_max (hw_params, 0);
}

int ALSAPCM::GetPeriods ()
{
  return snd_pcm_hw_params_get_periods (hw_params, 0);
}

bool ALSAPCM::InstallSwParams ()
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_sw_params (handle, sw_params)) < 0)
    {
      cout << "Cannot set software parameters!" << endl;
      return false;
    }
  return true;
}

bool ALSAPCM::SetSwStartThreshold (snd_pcm_uframes_t count)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_sw_params_set_start_threshold (handle, sw_params,
						    count) ) < 0) {
    cout << "Cannot set periods time near to " << count << "!" << endl;
    return false;
  }
  return true;
}

snd_pcm_uframes_t ALSAPCM::GetSwStartThreshold ()
{
  return snd_pcm_sw_params_get_start_threshold (sw_params);
}

bool ALSAPCM::SetSwStopThreshold (snd_pcm_uframes_t count)
{
  if (!opened)
    return false;
  
  if ((err = snd_pcm_sw_params_set_stop_threshold (handle, sw_params,
						   count) )  < 0) {
    cout << "Cannot set periods time near to " << count << "!" << endl;
    return false;
  }
  return true;
}

snd_pcm_uframes_t ALSAPCM::GetSwStopThreshold ()
{
  return snd_pcm_sw_params_get_stop_threshold (sw_params);
}

bool ALSAPCM::Prepare ()
{
  if ((err = snd_pcm_prepare (handle)) < 0)
    {
      cout << "Cannot prepare channel!" << endl;
      return false;
    }
  return true;
}

bool ALSAPCM::Drain ()
{
  if ((err = snd_pcm_drain (handle)) < 0)
    {
      cout << "Cannot drain channel!" << endl;
      return false;
    }
  return true;
}

bool ALSAPCM::Start ()
{
  if ((err = snd_pcm_start (handle)) < 0)
    {
      cout << "Cannot start channel!" << endl;
      return false;
    }
  return true;
}

snd_pcm_sframes_t ALSAPCM::WriteI (const void* buffer, snd_pcm_uframes_t size)
{
  return snd_pcm_writei (handle, buffer, size);
}

snd_pcm_sframes_t ALSAPCM::ReadI (void* buffer, snd_pcm_uframes_t size)
{
  return snd_pcm_readi (handle, buffer, size);
}

snd_pcm_sframes_t ALSAPCM::WriteN (snd_pcm_t* pcm, void** bufs, snd_pcm_uframes_t size)
{
  return snd_pcm_writen (handle, bufs, size);
}

snd_pcm_sframes_t ALSAPCM::ReadN (snd_pcm_t* pcm, void** bufs, snd_pcm_uframes_t size)
{
  return snd_pcm_readn (handle, bufs, size);
}


bool ALSAPCM::UpdateStatus ()
{
  return snd_pcm_status (handle, status);
}

snd_pcm_state_t ALSAPCM::GetStatusState ()
{
  return snd_pcm_status_get_state (status);
}

void ALSAPCM::GetTriggerTStamp (snd_timestamp_t* ptr)
{
  return snd_pcm_status_get_trigger_tstamp (status, ptr);
}

void ALSAPCM::GetTStamp (snd_timestamp_t* ptr)
{
  return snd_pcm_status_get_tstamp (status, ptr);
}

snd_pcm_sframes_t ALSAPCM::GetStatusDelay ()
{
  return snd_pcm_status_get_delay (status);
}

snd_pcm_uframes_t ALSAPCM::GetStatusAvail ()
{
  return snd_pcm_status_get_avail (status);
}

snd_pcm_uframes_t ALSAPCM::GetStatusAvailMax ()
{
  return snd_pcm_status_get_avail_max (status);
}

#endif // GSML_WITH_ALSA
