/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   grouping of DataLayers to the "undo-stack" Data
 */

#include "Data.hh"

#include <algorithm>
#include <functional>

using namespace std;
using namespace GSML;

/****************************
 * Data::ProgressInfo::Info *
 ****************************/

Data::ProgressInfo::Info::Info (ProgressInfo* init_parent)
    : text (), progress (0)
{
  parent = init_parent;
}

Data::ProgressInfo::Info::Info (ProgressInfo* init_parent, const Data::ProgressInfo::Info& init_info)
  : text (init_info.text), progress (init_info.progress)
{
  parent = init_parent;
}

void Data::ProgressInfo::Info::SetText (const string& new_text)
{
  text = new_text;
  parent->SignalInfoUpdated.emit (this);
}

void Data::ProgressInfo::Info::SetProgress (double new_progress)
{
  progress = new_progress;
  parent->SignalInfoUpdated.emit (this);
}
	  
const string& Data::ProgressInfo::Info::Text () const
{
  return text;
}

const double Data::ProgressInfo::Info::Progress () const
{
  return progress;
}
	  
Data::ProgressInfo::Info& Data::ProgressInfo::Info::operator= (const Data::ProgressInfo::Info& src)
{
  text = src.text;
  progress = src.progress;
  parent = src.parent;
}

/**********************
 * Data::ProgressInfo *
 *********************/

Data::ProgressInfo::ProgressInfo ()
{
}
	
Data::ProgressInfo::~ProgressInfo ()
{
  if (infos.size () != 0)
    cout << "~ProgressInfo: WARNING infos deque not empty?!?" << endl;
}

// for very basic access
const string& Data::ProgressInfo::Text ()
{
  return infos.back () -> Text ();
}
	
double Data::ProgressInfo::Progress ()
{
  return infos.back () -> Progress ();
}
	
Data::ProgressInfo::Info* Data::ProgressInfo::PushInfo (const string& text)
{
  Info* i = new Info (this);
  i->text = text;
	  
  infos.push_back (i);
	  
  SignalInfoPushed.emit (i);
  return i;
}
	
void Data::ProgressInfo::PopInfo ()
{
  if (infos.size () > 0)
    {
      Info* i = infos.back ();
      
      infos.pop_back ();
      SignalInfoPopped.emit ();
      delete i;
    }
  else
    {
      ExceptEmpty e;
      e.info = "can't pop info - empty";
      throw e;
    }
}

void Data::ProgressInfo::Clear ()
{
  infos.clear ();
}

int Data::ProgressInfo::Count () const
{
  return infos.size ();
}

Data::ProgressInfo::const_iterator Data::ProgressInfo::begin ()
{
  return infos.begin ();
}
	
Data::ProgressInfo::const_iterator Data::ProgressInfo::end ()
{
  return infos.end ();
}

void  Data::ProgressInfo::Debug ()
{
  cout << "Data::ProgressInfo::Debug:" << endl;
  
  cout << "  info objects: " << infos.size () << endl;
  
  for (iterator i = infos.begin (); i != infos.end (); i++)
    {
      cout << "  " << i - infos.begin () << ": " << (*i)->text << endl;
    }
}

/********
 * Data *
 ********/

Data::Data (DataFileIndex* init_fileindex,  unsigned int init_max_layers,
	    PCMDatatype* init_sampletype, sample_rate init_samplerate = 44100)
{
  
  my_fileindex = init_fileindex;
  max_layers = init_max_layers;
  my_temp_sampletype = init_sampletype;
  my_temp_samplerate = init_samplerate;
  
  data = new deque<DataLayer*>;
  
  // create initial data_layer
  DataLayer* new_layer = new DataLayer ("Ctor Layer", my_fileindex,
					my_temp_sampletype, my_temp_samplerate);

  data->push_back (new_layer);

  // init some stuff
  current_layer = data->begin ();
  prepared_layer = 0;
  modified = true; // cause a new data instance ...
}

Data::~Data ()
{
  if (prepared_layer != 0)
    {
      cout << "BUG: ~Data: Layer currently prepared ?!?" << endl;
      delete prepared_layer;
    }
  
  // destroy all tracks ...
  del_helper<DataLayer*> p;
  for_each (data->begin (), data->end (), p);
  
  delete data;
}

track_index Data::Tracks () const
{
  return (*current_layer) -> Tracks ();
}

sample_index Data::Length () const
{
  return (*current_layer) -> Length ();
}

DataLayer* Data::CurrentLayerPtr ()
{
  return *current_layer;
}

int Data::Layers ()
{
  return data->size ();
}

Data::iterator Data::CurrentLayer ()
{
  return current_layer;
}

Data::iterator Data::BeginLayer ()
{
  return data->begin ();
}

Data::iterator Data::EndLayer ()
{
  return data->end ();
}
      
void Data::SetCurrentLayer (iterator new_layer)
{
  if (prepared_layer != 0) // already prepared ??
    {
      ExceptLogicalError e;
      e.info = "A new layer is already prepared ... ";
      throw e;
    }
  
  if (new_layer == data->end () ) // vaild
    {
      ExceptIteratorInvaild e;
      e.info = "The iterator points to a not existing layer";
      throw e;
    }
  
  (*current_layer)->Fixate ();
  current_layer = new_layer;
  (*current_layer)->UnFixate ();
  
  SignalCurrentLayerChanged.emit (*current_layer);
}
 
// creates and return a new data_layar which will be connected to
// the wave_plots for the dynamic graphic preview!!
DataLayer* Data::PrepareNewLayer ()
{
#ifdef DEBUG_DATA
  cout << "DATA::PrepareNewLayer:" << endl;
#endif
  
  if (prepared_layer != 0) // already prepared ??
    {
      ExceptLogicalError e;
      e.info = "A new layer is already prepared ... ";
      throw e;
    }
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: an effect has left an Info object!!!" << endl;
      progress_info.Debug ();
      progress_info.Clear ();
    }

  prepared_layer = new DataLayer (**current_layer);
  prepared_layer->Comment("not set, yet");
  
  SignalNewLayerPrepared.emit (prepared_layer);
  
  return prepared_layer;
}

// fixate the "in work" data layer
void Data::FixateNewLayer ()
{
#ifdef DEBUG_GSML_DATA
  cout << "DATA::FixateNewLayer:" << endl;
#endif
  if (prepared_layer == 0)
    {
      ExceptLogicalError e;
      e.info = "No new layer prepared ... ";
      throw e;
    }
  
  (*current_layer)->Fixate ();
  
  // remove all undo layer between current and end
  if (current_layer != data->end () )
    current_layer++;
  data->erase (current_layer, data->end () );
  
  // add layer
  data->push_back (prepared_layer);
  
  prepared_layer = 0;
  current_layer = -- data->end ();
  modified = true;
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: a function has left an Info object!!! (I fix it ...)" << endl;
      progress_info.Clear ();
    }
  
  // check and remove more than max layer
  delete_if_too_many_layers ();
  
  // emit Signals
  SignalNewLayerFixated.emit ();
  SignalCurrentLayerChanged.emit (*current_layer);
}

// destroy the "in work" data_layer if the effect fails in some way ...
void Data::CancelNewLayer ()
{
#ifdef DEBUG_GSML_DATA
  cout << "DATA::CancelNewLayer:" << endl;
#endif
  if (prepared_layer == 0)
    {
      Data::Except e;
      e.info = "No new layer prepared ... ";
      throw e;
    }
  
  delete prepared_layer;
  prepared_layer = NULL;
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: an effect has left an Info object!!!" << endl;
      progress_info.Clear ();
    }
  
  // emit Signal
  SignalNewLayerCanceled.emit ();
}

// configure with new options

DataFileIndex* Data::GetDataFileIndex ()
{
  return my_fileindex;
}

PCMDatatype* Data::TempSampletype ()
{
  return my_temp_sampletype;
}

sample_rate Data::TempSamplerate ()
{
  return my_temp_samplerate;
}

void Data::SetTempSampletype (PCMDatatype* new_type)
{
  my_temp_sampletype = new_type;
}

void Data::SetTempSamplerate (sample_rate new_rate)
{
  my_temp_samplerate = new_rate;
}

void Data::SetMaxLayers (unsigned int new_max_layers)
{
  max_layers = new_max_layers;
  delete_if_too_many_layers ();
}

unsigned int Data::MaxLayers ()
{
  return max_layers;
}

bool Data::IsPrepared ()
{
  return prepared_layer != 0;
}

bool Data::IsModified ()
{
  return modified;
}

void Data::SetModified (bool new_state)
{
  modified = new_state;
}

void Data::Debug ()
{
  cout << "Data::Debug:" << endl;
  
  cout << "  Layer: " << data->size () << endl
       << "  Current_DataLayer: " << current_layer - data->begin ()<< endl
       << "  Prepared_Layer: " << prepared_layer << endl
       << "  Modified: " << modified << endl;
  
  cout << "  config:" << endl
       << "    max_layers: " << max_layers << endl
       << "    file_index: " << my_fileindex << endl
       << "    temp_sampletype: " << my_temp_sampletype << endl
       << "    temp_samplerate: " << my_temp_samplerate << endl;
  
  for (iterator i = data->begin (); i != data->end (); i++)
    {
      cout << "  Layer: " << i - data->begin () << " -> Debug:" << endl;
      (*i)->Debug ();
    }
    
  cout << "End ..." << endl;
}

// better than it has be ...
void Data::delete_if_too_many_layers ()
{
  cout << "Data::delete_if_too_many_layers" << endl;
  if ( data->size () > max_layers)
    {
      int current_delta;
      int delete_count;
      
#ifdef DEBUG_GSML_DATA
      cout << "more than max_layer (max: " << may_layers
	   << ", now: " << data->size () << ")"<< endl;
#endif
      current_delta = current_layer - data->begin ();
      delete_count = data->size () - max_layers;
      
      if (delete_count > current_delta)
	{
	  cout << "delete_if_too_many_layers: can't delete max ..." << endl;
	  delete_count = current_delta;
	}
      
      data->erase (data->begin (), data->begin () + delete_count);
      
      // adjust the current data layer ...
      current_layer = data->begin () + current_delta - delete_count;
    }
}
