/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Base class for data-file read/write interface and RAW-data
 *   implementation.
 */

#include "DataFile.hh"

using namespace GSML;

// abstact class - interface definition

DataFile::DataFile (const string& i_filename,
		    bool i_writeable=false,
		    bool i_autodelete=false)
{
  
  m_bindings = 0; // init binding counter...
  
  m_filename = i_filename;
  m_writeable = i_writeable;
  m_autodelete = i_autodelete;
}

DataFile::~DataFile (void)
{
  if (m_bindings > 0)
    {
      cout << endl;
      cout << "DataFile::~DataFile: bindings: " << m_bindings << endl;
      cout << "WW AA RR NN II NN GG !! !! !! - File in use!!!!" << endl;
    }
}

bool DataFile::IsAutoDelete ()
{
  return m_autodelete;
}

bool DataFile::IsWriteable ()
{
  return m_writeable;
}

string& DataFile::Filename ()
{
  return m_filename;
}

void DataFile::SetFilename (string &new_filename)
{
  m_filename = new_filename;
}

int DataFile::Bindings (void)  // retunr bindings
{
  return m_bindings;
}

void DataFile::Bind (void) // increase binding counter
{
  ++ m_bindings;
#ifdef DEBUG_GSML_DATA_FILE
  cout << "DataFile::Bind: (now: " << m_bindings << ")" << endl;
#endif
}

void DataFile::Unbind (void)
{
  if (m_bindings > 0)
    {
      --m_bindings;
#ifdef DEBUG_GSML_DATA_FILE
      cout << "DataFile::Unbind: (now: " << m_bindings  << ")" << endl;
#endif
    }
  else
    {
      ExceptBindingError e;
      e.filename = m_filename;
      e.info = "Tried to unbind unbindet file!! - m_bindings == 0!!!";
      throw e;
    }
}

// RAW implementation

DataFileRAW::DataFileRAW (const string& i_filename, PCMDatatype* i_type,
			  track_index i_tracks, sample_rate i_rate=44100,
			  int i_leading_trash = 0, bool i_writeable=false, bool i_autodelete=false)
  : DataFile (i_filename, i_writeable, i_autodelete)
{
  m_sample_type = i_type;
  m_tracks = i_tracks;
  m_rate = i_rate;
  m_leading_trash = i_leading_trash;
  
  // 1st check params
  if (m_tracks < 1)
    {
      DataFile::DataFile::ExceptLogicalError e;
      e.filename = m_filename;
      e.info = "Critiical! config invaild!";
      throw e;
    }
  
  // check writeable specific
  if (m_writeable && m_tracks > 1)
    {
      DataFile::DataFile::ExceptUnsupported e;
      e.filename = m_filename;
      e.info = "Critiical! config invaild! (more than one track for write)";
      throw e;
    }
  
  // do some init stuff
  file_descriptor = -1; // no open file!!
}

DataFileRAW::~DataFileRAW ()
{
  if (file_descriptor > 0)
    close (file_descriptor);
}


void DataFileRAW::ReadData (sample_index start,
			    sample_index count,
			    track_index track,
			    SampleBuffer* destination)
{
  // is a unfixated write in progress ??
  if (m_writeable)
    {
      DataFile::ExceptNotFixated e;
      e.filename = m_filename;
      e.info = "Tried to read in a writeable file! ...";
      throw e;
    }
  if (track >= m_tracks)
    {
      DataFile::DataFile::ExceptLogicalError e;
      e.filename = m_filename;
      e.info = "Critical: not enough tracks (was: ";
      e.info += track;
      e.info += ", have: ";
      e.info += m_tracks;
      e.info += ")";
      throw e;
    } 
  
  SigC::Threads::MLock lock (m_lock);
    
  // already open ??
  if (file_descriptor <= 0)
    {
      file_descriptor = open (m_filename . c_str (), O_RDONLY);
      // an error should not happen ...
      if (file_descriptor < 0)
	{
	  DataFile::DataFile::ExceptFileNotReadable e;
	  e.filename = m_filename;
	  e.info = "Critiical! Can't open file for read ... return:";
	  e.info += file_descriptor;
	  throw e;
	}
    }
  // if we are here the file is open!
  
  // 1st calculate some stuff
  int sample_size; // size for one sample
  int sample_size_in_file; // size of one sample (all! tracks!) in this file
  sample_index start_in_file = 0; // position where this read starts
  
  sample_size = m_sample_type->Size ();
  sample_size_in_file = sample_size * m_tracks;
  start_in_file = m_leading_trash + sample_size_in_file * start;
  
  // get buffer (cache or file ...)
  off_t data_length = read_from_file (start_in_file, sample_size_in_file * count) / sample_size_in_file;
  	  
  // if we have only one track we can read the entire buffer at once
  if (m_tracks == 1)
    {
      try {
	// try the types we know of ...
	
	if (pcm_24_8bit* t = dynamic_cast<pcm_24_8bit*> (m_sample_type) )
	  destination->WriteData (data_length, (pcm_24_8bit::value_type*) m_cache.data);
	
	else if (pcm_16bit* t = dynamic_cast<pcm_16bit*> (m_sample_type) )
	  destination->WriteData (data_length, (pcm_16bit::value_type*) m_cache.data);
	
	else if (pcm_8bit* t = dynamic_cast<pcm_8bit*> (m_sample_type) )
	  destination->WriteData (data_length, (pcm_8bit::value_type*) m_cache.data);
	
	else
	  {
	    DataFile::DataFile::ExceptUnsupported e;
	    e.filename = m_filename;
	    e.info = "BUG! pcm_sampletyp unsupported!! ...:";
	    throw e;
	  }
      }
      catch (SampleBuffer::Except e)
	{
	  DataFile::DataFile::ExceptBufferError e;
	  e.filename = m_filename;
	  e.info = "Critiical! Got buffer_exeption! - buffer too small!";
	  throw e;
	}
    } // end if
  // if more than one track we have to de-interleave the data!!! :-( ...
  else
    {
      try {
	if (pcm_24_8bit* t = dynamic_cast<pcm_24_8bit*> (m_sample_type) )
	  {
	    pcm_24_8bit::value_type t_data [data_length];
	    pcm_24_8bit::value_type* m_data = (pcm_24_8bit::value_type*) m_cache.data;
	      
	      for (sample_index i = 0; i < data_length; ++i)
		t_data [i] ( m_data [(i * m_tracks) + track] );
	      
	      destination->WriteData (data_length, t_data);
	    }
	  else if (pcm_16bit* t = dynamic_cast<pcm_16bit*> (m_sample_type) )
	    {
	      pcm_16bit::value_type t_data [data_length];
	      pcm_16bit::value_type* m_data = (pcm_16bit::value_type*) m_cache.data;
	      
	      for (sample_index i = 0; i < data_length; ++i)
		t_data [i] = m_data [(i * m_tracks) + track];
	      
	      destination->WriteData (data_length, t_data);
	    }
	  else if (pcm_8bit* t = dynamic_cast<pcm_8bit*> (m_sample_type) )
	    {
	      pcm_8bit::value_type t_data [data_length];
	      pcm_8bit::value_type* m_data = (pcm_8bit::value_type*) m_cache.data;
	      
	      for (sample_index i = 0; i < data_length; ++i)
		t_data [i] = m_data [(i * m_tracks) + track];
	      
	      destination->WriteData (data_length, t_data);
	    }
	  else
	    {
	      DataFile::DataFile::ExceptUnsupported e;
	      e.filename = m_filename;
	      e.info = "Critiical! pcm_sampletyp unsupported!! ...:";
	      throw e;
	    }
	}
      catch (SampleBuffer::Except e)
	{
	  DataFile::DataFile::ExceptBufferError e;
	  e.filename = m_filename;
	  e.info = "Critiical! Got buffer_exeption! - buffer too small!";
	  throw e;
	}
    } // end else -> multi track (interleave)
  
  // if some sample are left (file too short) fill the buffer ...
  if (data_length < count)
    {
      cout << "WARNING: The file is too short!! - filling with zeros!! .." << endl;
      
      try
	{
	  destination->Zero (count - data_length);
	}
      catch (SampleBuffer::Except e)
	{
	  DataFile::DataFile::ExceptBufferError e;
	  e.filename = m_filename;
	  e.info = "Critiical! Got buffer_exeption! - buffer too small!";
	  throw e;
	}
    } // end if not all data
}

void DataFileRAW::WriteData (sample_index count,
			     track_index track,
			     SampleBuffer* source,
			     sample_index buffer_index)
{
  // am I writeable ??
  if (!m_writeable)
    {
      DataFile::ExceptNotWriteable e;
      e.filename = m_filename;
      e.info = "Tried to write in a readonly (or already fixated) file! ...";
      throw e;
    }
  // unsupported multi-track file?? (interleaved)
  if (m_tracks != 1 || track != 0)
    {
      DataFile::ExceptUnsupported e;
      e.filename = m_filename;
      e.info = "More than one track per file currently unsupported! ...";
      e.info += " track_count: ";
      e.info += itos ( (int)m_tracks);
      e.info += " track: ";
      e.info += itos ( (int)track);
      throw e;
    }
  
  SigC::Threads::MLock lock (m_lock);
  
  // already open ??
  if (file_descriptor <= 0)
    {
      file_descriptor = open ( m_filename . c_str (), O_WRONLY | O_CREAT | O_TRUNC, FILE_CREATE_MODE);
      // an error should not happen ...
      if (file_descriptor <= 0)
	{
	  // busy.UnLock ();
	      
	  DataFile::DataFile::ExceptFileNotWriteable e;
	  e.filename = m_filename;
	  e.info = "Critiical! Can't open file for write ... return:";
	  e.info += itos (file_descriptor);
	  throw e;
	}
    }
  
  // if we are here the file is open!
  // 1st calculate some stuff
  int sample_size_in_file; // size of one sample (all! tracks!) in this file
  sample_index wrote_to_file; // the samples wrote into the file
  unsigned char* data; // buffer for write
	  
  sample_size_in_file = m_sample_type->Size () * m_tracks;
	  
  data = new unsigned char [sample_size_in_file * count];
      
  try
    {
      // read data from sample_buffer
      if (pcm_24_8bit* t = dynamic_cast<pcm_24_8bit*> (m_sample_type) )
	{
	  source->ReadData (buffer_index, count, (pcm_24_8bit::value_type*) data);
	}
      else if (pcm_16bit* t = dynamic_cast<pcm_16bit*> (m_sample_type) )
	{
	  source->ReadData (buffer_index, count, (pcm_16bit::value_type*) data);
	}
      else if (pcm_8bit* t = dynamic_cast<pcm_8bit*> (m_sample_type) )
	{
	  source->ReadData (buffer_index, count, (pcm_8bit::value_type*) data);
	}
      else
	{
	  delete [] data;
	  
	  DataFile::DataFile::ExceptUnsupported e;
	  e.filename = m_filename;
	  e.info = "Critiical! pcm_sampletyp unsupported!! ...:";
	  throw e;
	}
    }
  catch (SampleBuffer::Except e)
    {
      delete [] data;
	  
      DataFile::DataFile::ExceptBufferError e;
      e.filename = m_filename;
      e.info = "Critiical! Got buffer_exeption! - buffer too small!";
      throw e;
    }
      
  // write to file
  wrote_to_file = write (file_descriptor, data, sample_size_in_file * count);

  // check if all was written ... (disc full ??)
  if (wrote_to_file < sample_size_in_file * count)
    {
      delete [] data;
      
      DataFile::ExceptFileNotWriteable e;
      
      e.filename = m_filename;
      e.info = "Critical! Can't write to file ... return:";
      e.info += file_descriptor;
      e.info += " disc full ?? ...";
      throw e;
    }
  
  delete [] data;
}


void DataFileRAW::FixateWrite (void)
{
  if (!m_writeable)
    {
      DataFile::ExceptNotWriteable e;
      e.filename = m_filename;
      e.info = "Tried to fixate a readonly (or already fixated) file! ...";
      throw e;
    }
  else
    {
      m_writeable = false;
      
      if (file_descriptor > 0)
	close (file_descriptor);
      
      file_descriptor = -1; // not vaild anymore ...
    }
}

void DataFileRAW::SetFilename (string& new_filename)
{
  // only modifyable when no oepn file and writeabel ...
  if (file_descriptor < 0 && m_writeable)
    m_filename = new_filename;
  else
    {
      DataFile::ExceptLogicalError e;
      e.filename = m_filename;
      e.info = "Tryed to config the filename of an open or !writeable file! ...";
      throw e;
    }
    
}


PCMDatatype* DataFileRAW::SampleTyp ()
{
  return m_sample_type;
}

track_index DataFileRAW::TrackCount ()
{
  return m_tracks;
}

sample_rate DataFileRAW::SampleRate ()
{
  return m_rate;
}

off_t DataFileRAW::read_from_file (off_t start, off_t length)
{
  //cout << "rff: " << start << " + " << length << endl;
  m_cache.SetCachedRange (start, length);
  
  if (m_cache.vaild_start > 0) // at the front
    {
      off_t real_start = lseek (file_descriptor, (off_t) start, SEEK_SET);
      off_t read_count = read (file_descriptor, m_cache.data, m_cache.vaild_start);
    }
  
  if (m_cache.vaild_start + m_cache.vaild_length < length) // at the end
    {
      sample_index t_vaild_end = m_cache.vaild_start + m_cache.vaild_length;
      sample_index t_length = length - t_vaild_end;
      
      off_t real_start = lseek (file_descriptor, (off_t) start + t_vaild_end, SEEK_SET);
      off_t read_count = read (file_descriptor, m_cache.data + t_vaild_end, t_length);
      
    }
  
  // update, because we have read all, now.
  m_cache.vaild_start = 0;
  m_cache.vaild_length = length;

  if (false) // check cache consistency
    {
      uint8 validate_buffer [length];
      off_t real_start = lseek (file_descriptor, (off_t) start, SEEK_SET);
      off_t read_count = read (file_descriptor, validate_buffer, length);
  
      bool difference = false;
      int i;
      for (i = 0; i < length && !difference; ++i)
	if (m_cache.data [i] != validate_buffer [i] )
	  difference = true;
  
      if (difference) {
	cerr << ">>>>>>>>>>>>>>>> D I F F E R E N C E ! ! ! ! ! ("
	     << i << "/" << length << ") <<<<<<<<<<<<<<<<" << endl;
	StopInDebugger ();
      }
    }
  
  //cout << "<< rff" << endl;
  return length;
}
