/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Index to keep track of used files, temp-file creation and removal.
 */

#include "DataFileIndex.hh"

using namespace std;
using namespace GSML;

DataFileIndex::DataFileIndex (const DataFileIndex::Config& i_config)
{
  m_config = i_config;
  m_configured = false;
  m_next_temp_file_number = 0;
}


DataFileIndex::~DataFileIndex ()
{
  // perform a some clearing
  DeleteUnusedIndexes ();
}

bool DataFileIndex::Configure (bool correct_minor_problems)
{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
  cout << "DataFileIndex::Configure:" << endl;
#endif
  
  if (m_config . working_dir . length () == 0)
    {
#ifdef DEBUG_GSML_DATA_FILE_INDEX
      cout << "working directory string is empty! - using \"./\"" << endl;
#endif
      if (correct_minor_problems)
	m_config . working_dir = "./";
      else
	return false;
    }
  
  if (access (m_config . working_dir . c_str (), R_OK | W_OK | X_OK) != 0)
    {
      cout << "--ERROR: DataFileIndex: working directory ("
	   << m_config . working_dir << ") does not exist or is not accessible ...!!" << endl;
      if (correct_minor_problems)
	{
	  if (mkdir (m_config . working_dir.c_str(), DIR_CREATE_MODE) != 0);
	}
      else
	{
	  return false;
	}
    }
  m_configured = true;
  return true;
}

void DataFileIndex::RegisterFile (DataFile* n_file)
{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
  cout << "DataFileIndex::RegisterFile:" << endl;
#endif
  if (m_configured)
    {
      AddDataFile (n_file);
    }
  else
    {
      ExceptLogicalError e;
      e.info = "config not checked ...";
      throw e;
    }
}

void DataFileIndex::RegisterTempFile (DataFile* n_file)
{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
  cout << "DataFileIndex::RegisterTempFile:" << endl;
#endif
    
  if (!m_configured)
    {
      ExceptLogicalError e;
      e.info = "config not checked ...";
      throw e;
    }
  
  string t_fname;
  
  m_next_temp_file_number = FindUniqueName (t_fname,
					    m_config . working_dir + m_config . filename_base,
					    m_config . filename_ext, m_next_temp_file_number);
  
  int fd;
  // create file ..
  if ( (fd = open ( t_fname . c_str (), O_WRONLY | O_CREAT | O_TRUNC, FILE_CREATE_MODE) ) < 0)
    {
      ExceptLogicalError e;
      e.info = "Can't create tmp_file ... ???";
      throw e;
    }
  if ( close (fd) < 0)
    {
      ExceptLogicalError e;
      e.info = "Can't close tmp_file ... ???";
      throw e;
    }
	  
  // config with filename ...
  n_file -> SetFilename (t_fname);
  
  AddDataFile (n_file);
	  
  ++m_next_temp_file_number; // already assigned above; use the when called again next ...
}

void DataFileIndex::DeleteUnusedIndexes ()
{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
  cout << "DataFileIndex::DeleteUnusedIndexes:" << endl;
#endif
  
  // test all files
  iterator i = m_file_index . begin ();
  
  while (i != m_file_index . end () )
    {
#ifdef DEBUG_GSML_DATA_FILE_INDEX
      cout << "testing " << i - m_file_index . begin () << ": " << endl;
#endif
      
      // file in use (bindings)
      if ( i -> datafile -> Bindings () == 0)
	{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
	  cout << "OK! - unused!  (bindings == 0) - \""
	       << i -> datafile -> Filename () << "\"" << endl;
#endif
	  // should we auto-deleted the file?? (only temp files ...)
	  if ( i -> datafile -> IsAutoDelete () )
	    {
#ifdef DEBUG_GSML_DATA_FILE_INDEX
	      cout << "  is autodelete -> unlink ..." << endl;
		   
#endif
	      if (unlink ( i -> datafile -> Filename () . c_str () ) != 0)
		{
		  ExceptDeleteError e;
		  e.info = "Cant delete tmp_file ... ???";
		  throw e;
		}
	      delete i -> datafile;
	      i = m_file_index . erase (i);
	    }
	  else
	    ++ i;
	}
      else
	{
#ifdef DEBUG_GSML_DATA_FILE_INDEX
	  cout << "used! (bindings " << i -> datafile -> Bindings () << ") - \"" 
	       << i -> datafile -> Filename () << "\"" << endl;
#endif
	  ++ i;
	}
    }
#ifdef DEBUG_GSML_DATA_FILE_INDEX
  cout << " <<- END - DataFileIndex::DeleteUnusedIndexes:" << endl;
#endif
}

bool DataFileIndex::IsInIndex (string fname)
{
  struct stat t_stat;
  if (stat (fname . c_str (), &t_stat) > 0)
    {
      cout << "Unable to stat the given file in AddDataFile ... ???" << endl;
      return false;
    }
  
  for (iterator i = m_file_index . begin (); i != m_file_index . end (); ++ i)
    {
      if (fname == i -> datafile -> Filename () )
	{
	  cout << "filename match" << endl;
	  return true;
	}
      else if (i -> device == t_stat.st_dev && i -> inode == t_stat.st_ino)
	{
	  cout << "device / inode match" << endl;
	  return true;
	}
      else
	cout << "no match in this iteration ..." << endl;
    }
  
  return false;
}

DataFileIndex::const_iterator DataFileIndex::Begin ()
{
  return m_file_index . begin ();
}

DataFileIndex::const_iterator DataFileIndex::End ()
{
  return m_file_index . end ();
}

void DataFileIndex::Debug ()
{
  cout << "DataFileIndex::Debug:" << endl << endl;
  
  iterator i = m_file_index . begin ();
  int count = 0;
  while (i != m_file_index . end () )
    {
      cout << count
	   << ": filename: " << i -> datafile -> Filename ()
	   << ", writeable: " <<  i -> datafile -> IsWriteable ()
	   << ", autodelete: " << i -> datafile -> IsAutoDelete ()
	   << ", bindings: " << i -> datafile -> Bindings ()
	   << endl;
      ++ i;
      ++ count;
    }
  cout << "count: " << m_file_index . size () << endl;
  cout << "End ..." << endl;
}

bool DataFileIndex::AddDataFile (DataFile* n_file)
{
  for (iterator i = m_file_index . begin (); i != m_file_index . end (); ++ i)
    if (i -> datafile == n_file) {
      cout << "WARNING: obj already in the file index ... ???!?" << endl;
      return false;
    }
    
  struct stat t_stat;
  if (stat (n_file -> Filename () . c_str (), &t_stat) > 0)
    {
      ExceptLogicalError e;
      e.info = "Unable to stat the given file in AddDataFile ... ???";
      throw e;
    }
  
  file_stuff t_stuff;
  t_stuff . datafile = n_file;
  t_stuff . device = t_stat.st_dev;
  t_stuff . inode = t_stat.st_ino;
  
  m_file_index . push_back (t_stuff);
  return true;
}
