/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   IO architecture. Base classes and Mixer implementation.
 */

#include "IO.hh"

#include <algorithm>
#include <functional>

using namespace GSML;

/*******************
 * The MixerSource *
 *******************/

MixerSource::MixerSource ()
{
  io_running = false;
  io_input_running = false;
  
  m_playback_tracks = 0;
  m_capture_tracks = 0;
}

MixerSource::~MixerSource ()
{
  cout << "MixerSource::~MixerSource: DESTRUCTED !!" << endl;
}

track_index MixerSource::PlaybackTracks () const
{
  return m_playback_tracks;
}

track_index MixerSource::CaptureTracks () const
{
  return m_capture_tracks;
}

void MixerSource::StartO ()
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  StartO_impl ();
  
  io_running = true;
  io_input_running = false;
}

void MixerSource::StartIO ()
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  StartIO_impl ();
  
  io_running = true;
  io_input_running = true;
}

void MixerSource::StopIO ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is already running!";
      throw e;
    }
  
  StopIO_impl ();
  
  ClearTables ();
  
  io_running = false;
  io_input_running = false;
}

bool MixerSource::IsIORunning () const
{
  return io_running;
}

sample_index MixerSource::CurrentReadPos () const
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is already running!";
      throw e;
    }
  
  return master_read_position;
}

sample_index MixerSource::CurrentWritePos () const
{
  if (!io_input_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is already running!";
      throw e;
    }
  return master_write_position;
}

void MixerSource::SetReadPos (track_index track, sample_index n_pos, bool is_master = false)
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is already running!";
      throw e;
    }
  if (track > m_playback_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "The specified read track is invaild!";
      throw e;
    }
  SetReadPos_impl (track, n_pos, is_master);
}

void MixerSource::SetWritePos (track_index track, sample_index n_pos, bool is_master = false)
{
  if (!io_input_running)
    {
      ExceptNotInputable e;
      e.info = "No Input IO is running!";
      throw e;
    }
  if (track > m_playback_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "The specified read track is invaild!";
      throw e;
    }
  SetWritePos_impl (track, n_pos, is_master);
}

bool MixerSource::Read (track_index track, sample_index length, SampleBufferDefault* dest,
			bool is_master = false)
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is already running!";
      throw e;
    }
  if (track > m_playback_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "The specified io-track is not available!";
      throw e;
    }
  return Read_impl (track, length, dest, is_master);
}

bool MixerSource::Write (track_index track, sample_index length, SampleBufferDefault* src,
			 sample_index buffer_index, bool is_master = false)
{
  if (!io_input_running)
    {
      ExceptNotInputable e;
      e.info = "No input IO is running!";
      throw e;
    }
  if (track > m_capture_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "The specified io-track is not available!";
      throw e;
    }
  return Write_impl (track, length, src, buffer_index, is_master);
}

void MixerSource::Debug () const
{
  cout << "MixerSource::Debug:" << endl;
  
  cout << "  States: io_running: " << io_running
       << ", input_running: " << io_input_running << endl;
  
  cout << "  Tracks: playback: " << m_playback_tracks
       << ", capture: " << m_capture_tracks << endl;
  
  if (io_running)
    {
      cout << "  Read positions:" << endl;
      for (track_index i = 0; i < m_playback_tracks; ++i)
	cout << io_read_positions [i] << ", ";
      cout << "master: " << master_read_position;
      cout << endl;
      
      if (io_input_running)
	{
	  cout << "  Write positions:" << endl;
	  for (track_index i = 0; i < m_capture_tracks; ++i)
	    cout << io_write_positions [i] << ", ";
	  cout << "master: " << master_write_position;
	  cout << endl;
	}
    }
  Debug_impl ();
}

void MixerSource::BuildTables (track_index n_start)
{
  ClearTables ();
  
  for (track_index i = 0; i < m_playback_tracks; ++i)
    io_read_positions.push_back (n_start);
  master_read_position = n_start;
  
  for (track_index i = 0; i < m_capture_tracks; ++i)
    io_write_positions.push_back (n_start);
  master_write_position = n_start;
}

void MixerSource::ClearTables ()
{
  io_read_positions.clear ();
  io_write_positions.clear ();
}

/*******************
 * The HWMixerNode *
 *******************/

const HWMixerNode::Params& HWMixerNode::GetParams ()
{
  return m_params;
}

sample_rate HWMixerNode::SetRate (sample_rate n_rate)
{
  m_params.cur_rate = limit_min_max (n_rate, m_params.min_rate, m_params.max_rate);
  return m_params.cur_rate;
}

int HWMixerNode::SetDepth (int n_depth)
{
  m_params.cur_depth = limit_min_max (n_depth, m_params.min_depth, m_params.max_depth);
  return m_params.cur_depth;
}

int HWMixerNode::SetPeriodTime (int n_period_time)
{
  m_params.cur_period_time = limit_min_max (n_period_time, m_params.min_period_time, m_params.max_period_time);
  return m_params.cur_period_time;
}

int HWMixerNode::SetPeriods (int n_periods)
{
  m_params.cur_periods = limit_min_max (n_periods, m_params.min_periods, m_params.max_periods);
  return m_params.cur_periods;
}

/*************
 * The Mixer *
 *************/

Mixer::Mixer ()
{
  m_node = 0;
  
  io_playback = true;
  io_capture = false;
  
  playback_zero_buffer = 0;
  
  m_io_config.cache_size = 10;
  m_io_config.cache_min = 2;
  m_io_config.cache_grow = 5;
  m_io_config.write_burst = 5;
  
  m_first_loop = false;
  
  io_running = false;
  
  Ready (false);
  usleep (250);
}

Mixer::~Mixer ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~Mixer: io is running in dtor!?!?!?! ..." << endl;
  // TODO ...
}

void Mixer::SetPlayback (bool state)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  io_playback = state;
}

void Mixer::SetCapture (bool state)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  io_capture = state;
}

bool Mixer::IsPlayback ()
{
  return io_playback;
}

bool Mixer::IsCapture ()
{
  return io_capture;
}

track_index Mixer::PlaybackChannels ()
{
  if (io_running)
    return vaild_playback_channels;
  else
    if (m_node != 0)
      return m_node->PlaybackChannels();
    else
      return 0;
}

track_index Mixer::CaptureChannels ()
{
  if (io_running)
    return vaild_capture_channels;
  else
    if (m_node != 0)
      return m_node->CaptureChannels();
    else
      return 0;
}

void Mixer::StartIO (MixerSource* source, MixerParams* playback_params, MixerParams* capture_params)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  if (!io_playback && !io_capture)
     {
      ExceptNoIOSelected e;
      e.info = "Neither playback nor capture selected!";
      throw e;
    }
  if (m_node == 0)
     {
      ExceptNoNode e;
      e.info = "No node registered!";
      throw e;
    }
  
  m_source = source;
  m_playback_params = playback_params;
  m_capture_params = capture_params;
  
  if (m_source->IsIORunning () )
    {
      ExceptSourceBusy e;
      e.info = "The source-class is already running!";
      throw e;
    }
  if (m_playback_params->IsIORunning () )
    {
      ExceptSourceBusy e;
      e.info = "The playback param-class is already running!";
      throw e;
    }
  if (m_capture_params->IsIORunning () )
    {
      ExceptSourceBusy e;
      e.info = "The capture param-class is already running!";
      throw e;
    }
  
  m_playback_params->StartIO ();
  m_capture_params->StartIO ();
  
  try {
    if (io_capture)
      m_source->StartIO ();
    else
      m_source->StartO ();
  }
  catch (Mixer::Except e)
    {
      cout << "BUG (normally): Exception during MixerSource::Start(I)O!" << endl;
      cout << "  info: " << e.info << endl;
      
      Except m_e;
      m_e.info = "The source class exception (" + e.info +")";
      throw m_e;
    }
  
  build_values ();
  
  connection_playback_value_changed =
    m_playback_params->SignalValueChanged.connect (slot (this, &Mixer::SlotPlaybackValueChanged) );
  connection_capture_value_changed =
    m_capture_params->SignalValueChanged.connect (slot (this, &Mixer::SlotCaptureValueChanged) );

  connection_playback_track_mute_changed =
    m_playback_params->SignalTrackMuteChanged.connect (slot (this, &Mixer::SlotPlaybackTrackMuteChanged) );
  connection_capture_track_mute_changed =
    m_capture_params->SignalTrackMuteChanged.connect (slot (this, &Mixer::SlotCaptureTrackMuteChanged) );
  
  connection_playback_channel_mute_changed =
    m_playback_params->SignalChannelMuteChanged.connect (slot (this, &Mixer::SlotPlaybackChannelMuteChanged) );
  connection_capture_channel_mute_changed =
    m_capture_params->SignalChannelMuteChanged.connect (slot (this, &Mixer::SlotCaptureChannelMuteChanged) );
  
  // m_source->Debug ();
  
  m_using_realtime_sched = false;
  io_running = true;
  m_first_loop = true;
  
  // the node is stated (after pre-mixing) in the Loop
  cout << "Mixer StartIO finished. Continueing" << endl;
  Continue ();
}

void Mixer::StopIO ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "IO not running!";
      throw e;
    }
  
  // only notify the node - the rest is done in IOFinishedNotification
  m_node->StopIO ();
}

void Mixer::IOFinishedNotification ()
{
#ifdef DEBUG_GSML_IO
  cout << "Mixer::IOFinishedNotification:" << endl;
#endif
  
  if (!io_running)
    {
      cout << "BUG: IOFinishedNotification called when no IO is running .. ?!?" << endl;
    }
  else
    {
      // stop the 2nd thread nicely
      AsyncWait ();
      loop_waiting_for_work.unlock ();
      SyncWait ();
      
      m_source->StopIO ();
      m_playback_params->StopIO ();
      m_capture_params->StopIO ();
      
      connection_playback_value_changed.disconnect ();
      connection_capture_value_changed.disconnect ();
      
      connection_playback_track_mute_changed.disconnect ();
      connection_capture_track_mute_changed.disconnect ();
      
      connection_playback_channel_mute_changed.disconnect ();
      connection_capture_channel_mute_changed.disconnect ();
      
      clear_values ();
      
      io_running = false;
    }
}
bool Mixer::IsIORunning ()
{
  return io_running;
}

bool Mixer::Read (track_index channel, SampleBufferDefault* destination)
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "IO not running!";
      throw e;
    }
  
  bool data = true;
  // cout << "Mixer::Read: " << channel << endl;
  
  if (channel >= vaild_playback_channels)
    {
      destination->WriteData (pl_period_size, playback_zero_buffer->data);
    }
  else
    {
      while (playback_channel_queues [channel].size () == 0)
	{
	  cout << "BUG: NO DATA - should never get here (SYSTEM TOO SLOW ?? ;-) ..." << endl
	       << "     channel: " << channel << endl;
	  cout << "Waiting ..." << endl;
	  
	  ++m_io_status.pl_data_queue_outs;
	  
	  io_waiting_for_data.lock();
	  io_waiting_for_data.lock();
	}
      
      // write data from cache
      SampleBufferDefault* t_buffer = playback_channel_queues [channel].front ().buffer;
      data = playback_channel_queues [channel].front ().data;
      destination->WriteData (pl_period_size, t_buffer->data);
      
      // remove data from cache
      SigC::Threads::MLock lock (playback_channel_queues_lock);
      del_helper2 (playback_channel_queues [channel].front ().buffer);
      playback_channel_queues [channel].pop ();
      
      ++io_read_period [channel];
      
      // pop to next track vu-meter data when at next period
      if (channel == vaild_playback_channels - 1)
	{
	  for (int i = 0; i < vaild_playback_tracks; ++i)
	    {
	      if (playback_track_queues [i].size () > 0)
		playback_track_queues [i].pop ();
	      else
		cout << "BUG: pl-track-vu-meter-queue is empty ... ?!? (Queue corrupted ...)" << endl;
	    }
	  loop_waiting_for_work.unlock ();
	}
      
    }
  
  //cout << "Mixer::Read returning: " << (bool) (current_read_period <= last_full_read_period) << endl;
  return data;
}

bool Mixer::Write (track_index channel, SampleBufferDefault* source)
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "IO not running!";
      throw e;
    }
  //cout << "Mixer::Write: " << channel << endl;
  
  if (channel >= vaild_capture_channels)
    {
      // throw away ...
    }
  else
    {
      // calc vu-meter
      vu_meter_from_buffer (source, capture_channel_vu_meter [channel] );
      
      SampleBufferDefault* t_buffer = new SampleBufferDefault (ca_period_size);
      t_buffer->WriteData (ca_period_size, source->data);
      
      SigC::Threads::MLock lock (capture_channel_queues_lock);
      capture_channel_queues [channel].push (t_buffer);
      
      ++io_write_period [channel];
      if (channel == vaild_capture_channels - 1)
	loop_waiting_for_work.unlock ();
    }
  
  return true;
}

void Mixer::SetNode (MixerNode* new_node)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  m_node = new_node;
}

MixerNode* Mixer::GetNode ()
{
  return m_node;
}

const Mixer::io_status& Mixer::GetIOStatus ()
{
  m_io_status.pl_data_queue = playback_channel_queues [0].size ();
  m_io_status.pl_data_queue_max = m_running_io_config.cache_size;
  m_io_status.ca_data_queue = capture_channel_queues [0].size ();
  m_io_status.pl_xruns = m_node->PlaybackXRuns ();
  m_io_status.ca_xruns = m_node->CaptureXRuns ();
  
  return m_io_status;
}

sample_index Mixer::CurrentReadPos ()
{
  return io_read_period [0] * pl_period_size;
}

sample_index Mixer::CurrentWritePos ()
{
  return io_write_period [0] * ca_period_size;
}

int Mixer::PlaybackPeriodSize ()
{
  return pl_period_size;
}

int Mixer::CapturePeriodSize ()
{
  return ca_period_size;
}

Mixer::vu_meter_data Mixer::PlaybackChannelVU (sample_index channel)
{
  if (channel < vaild_playback_channels && playback_channel_queues [channel].size () > 0)
    return playback_channel_queues [channel].front ().vu_meter;
  
  return (vu_meter_data) { {0}, {0} };
}
    
Mixer::vu_meter_data Mixer::CaptureChannelVU (sample_index channel)
{
  if (channel < vaild_capture_channels)
    return capture_channel_vu_meter [channel];
  
  return (vu_meter_data) { {0}, {0} };
}

Mixer::vu_meter_data Mixer::PlaybackTrackVU (sample_index track)
{
  if (track < vaild_playback_tracks && playback_track_queues [track].size () > 0)
    return playback_track_queues [track].front ();
  
  return (vu_meter_data) { {0}, {0} };
}
    
Mixer::vu_meter_data Mixer::CaptureTrackVU (sample_index track)
{
  if (track < vaild_capture_tracks)
    return capture_track_vu_meter [track];
  
  return (vu_meter_data) { {0}, {0} };
}

void Mixer::Debug ()
{
  cout << "Mixer::Debug:" << endl;
  
  cout << "  io_running: " << io_running
       << ", io_playback: " << io_playback
       << ", io_capture: " << io_capture
       << ", Node: " << m_node
       << endl;

  if (io_running)
    {
      cout << endl;
      cout << "  Channels: playback: " << playback_channels
	   << ", capture: " << capture_channels << endl;
      
      cout << "  Vaild channels: playback: " << vaild_playback_channels
	   << ", capture: " << vaild_capture_channels << endl;
      
      cout << "  Vaild tracks: playback: " << vaild_playback_tracks
	   << ", capture: " << vaild_capture_tracks << endl;
      
      cout << "  period size: pl: " << pl_period_size
	   << ", ca: " << ca_period_size << endl;
      
      cout << "  detailed IO period count:" << endl
	   << "  Read: " << endl
	   << "    ";
      for (track_index ch = 0; ch < playback_channels; ++ch)
	cout << "[" << ch << "]: " << io_read_period [ch] << ", ";
      
      cout << endl
	   << "  Write: " << endl
	   << "    ";
      for (track_index ch = 0; ch < capture_channels; ++ch)
	cout << "[" << ch << "]: " << io_write_period [ch] << ", ";
      
      cout << endl;
    }
}

void Mixer::Loop ()
{
  bool worked = false;
  
  loop_waiting_for_work.unlock ();
  loop_waiting_for_work.lock ();
  
  if (!io_running)
    {
      cout << "BUG: Mixer::Loop and NO io running ?!?!?!?! ... Sleeping ..." << endl;
      Wait ();
      return;
    }
  
  if (io_playback)
    {
      int t_cache_size = m_running_io_config.cache_size;
      
      if (m_first_loop) // fill a bit more at IO start ...
	t_cache_size += m_node->PlaybackPeriods ();
      
      bool t_resized = false;
      while (playback_channel_queues [0].size () <= t_cache_size)
	{
	  // cout << "Mixer::Loop: mix next playback period ... ";
	  
	  // enlarge queue ??
	  if (!m_first_loop && !t_resized
	      && playback_channel_queues [0].size () < m_running_io_config.cache_min)
	    {
	      cout << "QUEUE TOO SMALL -->> RESIZE ..." << endl;
	      m_running_io_config.cache_size += m_running_io_config.cache_grow;
	      
	      t_cache_size = m_running_io_config.cache_size;
	      t_resized = true;
	    }
	  
	  mix_next_period ();
	  
	  io_waiting_for_data.unlock(); // data available ...
	  
	  worked = true;
	}
    }
  
  if (io_capture && !worked)
    {
      int t_write_periods = capture_channel_queues [0].size ();
      
      for (track_index tr = 1; tr < vaild_capture_channels; ++tr)
	{
	  t_write_periods = limit_max (t_write_periods, capture_channel_queues [tr].size () );
	}
      
      t_write_periods = limit_max (t_write_periods, m_running_io_config.write_burst);
      
      while (t_write_periods > 0)
	{
	  //   cout << "Mixer::Loop: write next playback period ..." << endl;
	  write_period ();
	  --t_write_periods;
	}
    }
  
  if (m_first_loop)
    {
      m_using_realtime_sched = ThreadHelper::EnableRealtimeScheduling ();
      if (!m_using_realtime_sched)
	cout << "Warning: Mixer: Can't enable realtime scheduling ..." << endl;
      
      cout << "Mixer::Loop: Starting Node's IO ..." << endl;
      m_node->SetPlayback (io_playback);
      m_node->SetCapture (io_capture);
      
      m_node->StartIO (*this);
      
      worked = true; // although it must have been set during read-buffering above ...
      m_first_loop = false;
    }
  
  if (!worked)
    {
      loop_waiting_for_work.lock ();
    }
  // cout << "END.." << endl;
}

void Mixer::clear_values ()
{
  playback_values.clear ();
  capture_values.clear ();
  
  playback_track_mutes.clear ();
  capture_track_mutes.clear ();
  
  playback_channel_mutes.clear ();
  capture_channel_mutes.clear ();
  
  io_read_period.clear ();
  io_write_period.clear ();
  
  // io_read_pos_vaild.clear ();
  io_write_pos_vaild.clear ();
  
  del_helper<SampleBufferDefault*> dh;
  
  // delete the SampleBuffer (cache)
  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
    {
      while (playback_channel_queues [ch].size () > 0)
	{
	  del_helper2 (playback_channel_queues [ch].front ().buffer);
	  playback_channel_queues [ch].pop ();
	}
    }
  playback_channel_queues.clear ();
  
  for (track_index ch = 0; ch < vaild_capture_channels; ++ch)
    {
      while (capture_channel_queues [ch].size () > 0)
	{
	  del_helper2 (capture_channel_queues [ch].front () );
	  capture_channel_queues [ch].pop ();
	}
    }
  capture_channel_queues.clear ();
  capture_channel_vu_meter.clear ();
  
  playback_track_queues.clear ();
  capture_track_vu_meter.clear ();
  
  // clear m_io_status
  m_io_status.pl_data_queue = 0;
  m_io_status.pl_data_queue_outs = 0;
  
  m_io_status.ca_data_queue = 0;
  
  m_io_status.pl_xruns = 0;
  m_io_status.ca_xruns = 0;
  
  dh (playback_zero_buffer);
}

void Mixer::build_values ()
{
  cout << "Mixer::build_values:" << endl;
  playback_channels = m_node->PlaybackChannels ();
  capture_channels = m_node->CaptureChannels ();
  
  vaild_playback_channels = limit_max (playback_channels, m_playback_params->Channels () );
  vaild_capture_channels = limit_max (capture_channels, m_capture_params->Channels () );
  
  track_index playback_tracks = m_source->PlaybackTracks ();
  track_index capture_tracks = m_source->CaptureTracks ();
  
  vaild_playback_tracks = limit_max (playback_tracks, m_playback_params->Tracks () );
  vaild_capture_tracks = limit_max (capture_tracks, m_capture_params->Tracks () );
  
  cout << "Debug: using:" << endl
       << "  Node:" << endl
       << "    playback_channels: " << playback_channels << endl
       << "    capture_channels: " << capture_channels << endl
       << "  After comparing with params:" << endl
       << "    vaild_playback_channels: " << vaild_playback_channels << endl
       << "    vaild_capture_channels: " << vaild_capture_channels << endl
       << "  After comparing with source: " << endl
       << "    vaild_playback_tracks: " << vaild_playback_tracks << endl
       << "    vaild_capture_tracks: " << vaild_capture_tracks << endl;
  
  // build my pl-track value arrays
  for (track_index tr = 0; tr < vaild_playback_tracks; ++tr)
    {
      playback_values.push_back ();
      for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
	{
	  cout << "adding track: " << tr << ", channel: " << ch << endl;
	  playback_values.back () . push_back (m_playback_params->Value (tr, ch) );
	}
      playback_track_mutes.push_back (m_playback_params->IsTrackMuted (tr) );
    }
  
  // build my ca-track value arrays
  for (track_index tr = 0; tr < vaild_capture_tracks; ++tr)
    {
      capture_values.push_back ();
      for (track_index ch = 0; ch < vaild_capture_channels; ++ch)
	{
	  cout << "adding track: " << tr << ", channel: " << ch << endl;
	  capture_values.back () . push_back ( m_capture_params->Value (tr, ch) );
	}
      capture_track_mutes.push_back (m_capture_params->IsTrackMuted (tr) );
      io_write_pos_vaild.push_back (true);
    }
  
  pl_period_size = m_node->PlaybackPeriodSize ();
  ca_period_size = m_node->CapturePeriodSize ();
  
  // build my pl-channel value arrays
  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
    {
      io_read_period.push_back (0);
      playback_channel_mutes.push_back (m_playback_params->IsChannelMuted (ch) );
    }
  
  // build my ca-channel value arrays
  for (track_index ch = 0; ch < vaild_capture_channels; ++ch)
    {
      io_write_period.push_back (0);
      capture_channel_mutes.push_back (m_capture_params->IsChannelMuted (ch) );
    }
  
  // create the SampleBuffer (cache) / cu-meter stuff
  playback_channel_queues.resize (vaild_playback_channels);
  capture_channel_queues.resize (vaild_capture_channels);
  capture_channel_vu_meter.resize (vaild_capture_channels);
  
  playback_track_queues.resize (vaild_playback_tracks);
  capture_track_vu_meter.resize (vaild_capture_tracks);
  
  playback_zero_buffer = new SampleBufferDefault (pl_period_size);
  playback_zero_buffer->Zero (pl_period_size);
  
  // copy io_config
  m_running_io_config = m_io_config;
  cout << "Mixer::build_values: returning ..." << endl;
}

void Mixer::vu_meter_from_buffer (SampleBufferDefault* buffer, vu_meter_data& data)
{
  sample_index t_length = buffer->UsedLength ();
  
  // calculate vu-meter
  pcm_default::value_type t_min = buffer->data [0];
  pcm_default::value_type t_max = buffer->data [0];
  for (int i = 0; i < t_length; ++i)
    {
      t_min.data = limit_max (buffer->data [i].data, t_min.data);
      t_max.data = limit_min (buffer->data [i].data, t_max.data);
    }
  data.min = t_min;
  data.max = t_max;
}

void Mixer::mix_next_period ()
{
#ifdef DEBUG_GSML_IO
  cout << "Mixer::mix_next_period:" << endl;
#endif
  
  vector <channel_period> t_period (vaild_playback_channels);
  
  SampleBufferDefault temp_buffer (pl_period_size);
  sample_index all_mixed = true;
  
  // bild zero period
  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
    {
      t_period [ch].buffer = new SampleBufferDefault (pl_period_size);
      t_period [ch].buffer->Zero (pl_period_size);
    }
  
  // mix
  for (track_index tr = 0; tr < vaild_playback_tracks; ++tr)
    {
      temp_buffer.Clear ();
      
      if (!m_source->Read (tr, pl_period_size, &temp_buffer) )
	all_mixed = false; // update if not all was read
	  
      // calc track vu-meter
      vu_meter_data t_data;
      vu_meter_from_buffer (&temp_buffer, t_data);
      playback_track_queues [tr].push (t_data);
      
      if (!playback_track_mutes [tr] ) // cpu time saving when muted ...
	{
	  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
	    {
	      if (!playback_channel_mutes [ch] ) {
		pcm_default::value_type value = playback_values [tr] [ch];
		// cpu-time saving - unmuted?
		if (value.data != 0) {
		  for (sample_index i = 0; i < pl_period_size; ++i)
		    {
		      pcm_default::value_type d = temp_buffer.data [i] * value;
		      t_period [ch].buffer->data [i] (t_period [ch].buffer->data [i] + d);
		    }
		}
	      } // end if !channel muted
	    } // end for all channels
	} // end if !track muted
    } // end for all tracks
  
  // calc vu-meter and do post(tm) things
  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
    {
      vu_meter_from_buffer (t_period [ch].buffer, t_period [ch].vu_meter);
      t_period [ch].data = all_mixed;
    }
  
  // read master
  m_source->Read (0, pl_period_size, 0, true); // for master pos

  // push to queue
  SigC::Threads::MLock lock (playback_channel_queues_lock);
  for (track_index ch = 0; ch < vaild_playback_channels; ++ch)
    {
      playback_channel_queues [ch].push (t_period [ch] );
    }
  
  //cout << "Mixer::mix_next_period returning: " << all_mixed << endl;
}

void Mixer::write_period ()
{
#ifdef DEBUG_GSML_IO
  cout << "Mixer::write_period:" << endl;
#endif
  
  SampleBufferDefault temp_buffer (ca_period_size);
  bool all_written = true;
  
  // mix
  
  for (track_index tr = 0; tr < vaild_capture_tracks; ++tr)
    {
      // zero temp buffer
      temp_buffer.Clear ();
      temp_buffer.Zero (ca_period_size);
      
      if (!capture_track_mutes [tr] ) // cpu time saving
	{
	  for (track_index ch = 0; ch < vaild_capture_channels; ++ch)
	    {
	      if (!capture_channel_mutes [ch] ) {
		SampleBufferDefault* ch_buffer = capture_channel_queues [ch].front ();
		
		pcm_default::value_type value = capture_values [tr] [ch];
		      // cpu-time saving - unmuted?
		      if (value.data != 0)
			for (sample_index i = 0; i < ca_period_size; ++i)
			  {
			    pcm_default::value_type d = ch_buffer->data [i] * value;
				  temp_buffer.data [i] (temp_buffer.data [i] + d);
			  }
	      } // end if !channel muted
	    } // end for all channels
	  if (!io_write_pos_vaild [tr] )
	    {
	      cout << "write_period: io_pos invaild (must have been muted) -> setting ..." << endl;
 	      m_source->SetWritePos (tr, m_source->CurrentWritePos () );
	      io_write_pos_vaild [tr] = true;
	    }
	  if (!m_source->Write (tr, ca_period_size, &temp_buffer, 0) )
	    all_written = false;
	}
      else // end if !muted
	io_write_pos_vaild [tr] = false;
      
      // post mixing calc track vu-meter
      vu_meter_from_buffer (&temp_buffer, capture_track_vu_meter [tr] );
    }
    
  SigC::Threads::MLock lock (capture_channel_queues_lock);
  // post write - pop from write queue
  for (track_index ch = 0; ch < vaild_capture_channels; ++ch)
    {
      del_helper2 (capture_channel_queues [ch].front () );
      capture_channel_queues [ch].pop ();
    }
  
  m_source->Write (0, ca_period_size, 0, 0, true); // for master pos
  
  //return all_written;
}

void Mixer::SlotPlaybackValueChanged (track_index track, track_index channel, pcm_default::value_type value)
{
#ifdef DEBUG_GSML_IO
  cout << "Mixer::SlotPlaybackValueChanged:" << endl;
#endif
  if (!io_running)
    cout << "BUG: Mixer::SlotPlaybackValueChanged connected while no IO is runngin?!?" << endl;
  else
    {
      if (track <= vaild_playback_tracks && channel <= vaild_playback_channels)
	playback_values [track] [channel] = value;
      else
	cout << " track/channel not cached (invaild): track: " << track 
	     << ", channel: " << channel << endl;
    }
}

void Mixer::SlotCaptureValueChanged (track_index track, track_index channel, pcm_default::value_type value)
{
#ifdef DEBUG_GSML_IO
  cout << "Mixer::SlotCaptureValueChanged:" << endl;
#endif
  if (!io_running)
    cout << "BUG: Mixer::SlotCaptureValueChanged connected while no IO is runngin?!?" << endl;
  else
    {
      if (track <= vaild_capture_tracks && channel <= vaild_capture_channels)
	capture_values [track] [channel] = value;
      else
	cout << " track/channel not cached (invaild): track: " << track 
	     << ", channel: " << channel << endl;
    }
}

void Mixer::SlotPlaybackTrackMuteChanged (track_index track, bool n_state)
{
  if (!io_running)
    cout << "BUG: Mixer::SlotPlaybackTrackMuteChanged connected while no IO is running?!?" << endl;
  else
    {
      if (track <= vaild_playback_tracks)
	playback_track_mutes [track] = n_state;
      else
	cout << " track not cached (invaild): track: " << track << endl;
    }
}

void Mixer::SlotCaptureTrackMuteChanged (track_index track, bool n_state)
{
  if (!io_running)
    cout << "BUG: Mixer::SlotCaptureTrackMuteChanged connected while no IO is running?!?" << endl;
  else
    {
      if (track <= vaild_capture_tracks)
	capture_track_mutes [track] = n_state;
      else
	cout << " track not cached (invaild): track: " << track << endl;
    }
}


void Mixer::SlotPlaybackChannelMuteChanged (track_index channel, bool n_state)
{
  if (!io_running)
    cout << "BUG: Mixer::SlotPlaybackTrackMuteChanged connected while no IO is running?!?" << endl;
  else
    {
      if (channel <= vaild_playback_channels)
	playback_channel_mutes [channel] = n_state;
      else
	cout << " channel not cached (invaild): channel: " << channel << endl;
    }
}

void Mixer::SlotCaptureChannelMuteChanged (track_index channel, bool n_state)
{
  if (!io_running)
    cout << "BUG: Mixer::SlotCaptureChannelMuteChanged connected while no IO is running?!?" << endl;
  else
    {
      if (channel <= vaild_capture_channels)
	capture_channel_mutes [channel] = n_state;
      else
	cout << " channel not cached (invaild): channel: " << channel << endl;
    }
}
