/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   ALSA Mixer Node implementation
 */

#include "IO_ALSA.hh"

#ifdef GSML_WITH_ALSA

#include <sstream>

using namespace GSML;
using namespace ALSAHelper;

/********************************
 * The ALSA Node implementation *
 ********************************/

ALSAMixerNode::ALSAMixerNode ()
{
  cname = "hw:0";
  
  initialized = false;
  io_running = false;
  
  do_playback = true;
  do_capture = false;
  
  pl_period_size = 0;
  ca_period_size = 0;
  
  playback_card = new ALSAPCM;
  capture_card = new ALSAPCM;
  
  read_handler = 0;
  write_handler = 0;
  
  Ready (false);
}

ALSAMixerNode::~ALSAMixerNode ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~ALSAMixerNode: io is running in dtor!?!?!?! ..." << endl;
  
  delete playback_card;
  delete capture_card;
  
  // TODO ...
}

bool ALSAMixerNode::Init ()
{
  int err;
  
  if (initialized)
    {
      ExceptInitialized e;
      e.info = "Already Initialized!";
      throw e;
    }
  
  // build / hack list of possible device names ...
  ALSAHelper::ALSAControl t_alsa;
  
  for (int i = 0; i < 8; ++i)
    {
      string card_name = t_alsa.GetName (i);
      if (card_name != "")
	{
	  stringstream interface;
	  interface << "hw:" << i;
	  m_interfaces.push_back (interface.str () );
	}
    }
  
  if (!reget_card_params () )
    return false;
  
  initialized = true;
  return true;
}

const vector<string>& ALSAMixerNode::Interfaces ()
{
  return m_interfaces;
}

const string& ALSAMixerNode::Interface ()
{
  return cname;
}

bool ALSAMixerNode::SetInterface (const string& n_interface)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  string old_name = cname;
  cname = n_interface;
  
  pl_period_size = 0;
  ca_period_size = 0;
  
  if (!reget_card_params () ) {
    cname = old_name;
    return false;
  }
  return true;
}

track_index ALSAMixerNode::PlaybackChannels ()
{
  return playback_channels;
}

track_index ALSAMixerNode::CaptureChannels ()
{
  return capture_channels;
}

int ALSAMixerNode::PlaybackPeriodSize ()
{
  return pl_period_size;
}

int ALSAMixerNode::CapturePeriodSize ()
{
  return ca_period_size;
}

int ALSAMixerNode::PlaybackPeriods ()
{
  return pl_periods;
}

int ALSAMixerNode::CapturePeriods ()
{
  return ca_periods;
}

int ALSAMixerNode::PlaybackXRuns ()
{
  return pl_xruns;
}

int ALSAMixerNode::CaptureXRuns ()
{
  return ca_xruns;
}

void ALSAMixerNode::SetPlayback (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  do_playback = state;
}

void ALSAMixerNode::SetCapture (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  do_capture = state;
}

bool ALSAMixerNode::IsPlayback ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  return do_playback;
}

bool ALSAMixerNode::IsCapture ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  return do_capture;
}

void ALSAMixerNode::StartIO (Mixer& source)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  m_mixer = &source;
  
  if (do_playback)
    {
      if (!playback_card->Open (cname, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK) )
	throw ExceptHWError ("ALSA - can't open pl");
      if (!internal_config_params (playback_card, true, false) )
	throw ExceptHWError ("ALSA - can't config params pl");
      
      // no prepare needed (ALSA 0.9-beta6)
      
      // create write handler
      switch (playback_card->GetFormat () )
	{
	case SND_PCM_FORMAT_S16_LE:
	  write_handler = new WriteHandler16Bit (playback_card, playback_channels, pl_period_size);
	  break;
	case SND_PCM_FORMAT_S32_LE:
	  write_handler = new WriteHandler32Bit (playback_card, playback_channels, pl_period_size);
	  break;
	default:
	  cout << "No write Handler for format." << endl;
	}
    }
  
  if (do_capture)
    {
      if (!capture_card->Open (cname, SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK) )
	throw ExceptHWError ("ALSA - can't open ca");
      if (!internal_config_params (capture_card, false, false) )
	throw ExceptHWError ("ALSA - can't config params ca");
      cap_running = false;
      
      // create read handler
      switch (capture_card->GetFormat () )
	{
	case SND_PCM_FORMAT_S16_LE:
	  read_handler = new ReadHandler16Bit (capture_card, capture_channels, ca_period_size);
	  break;
	case SND_PCM_FORMAT_S32_LE:
	  read_handler = new ReadHandler32Bit (capture_card, capture_channels, ca_period_size);
	  break;
	default:
	  cout << "No read Handler for format." << endl;
	}
    }
  
  // prepare states
  stop_requested = false;
  
  current_playback_period = 0;
  current_capture_period = 0;
  
  pl_xruns = 0;
  ca_xruns = 0;
  
  m_is_first_loop = true;
  m_worked_in_last_loop = true;
  m_using_realtime_sched = false;
  
  io_running = true;
  
#ifdef DEBUG_GSML_IO_ALSA
  cout << "  StartIO: calling _Continue_" << endl;
#endif
  Continue (); // unlock Loop ...
}
    
void ALSAMixerNode::StopIO ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO is currently running ...";
      throw e;
    }
  
  stop_requested = true;
}

bool ALSAMixerNode::IsIORunning ()
{
  return io_running;
}

sample_index ALSAMixerNode::GetCurrentPos ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO currently running ...";
      throw e;
    }
  
  if (do_playback)
    return current_playback_period * pl_period_size;
  if (do_capture)
    return current_capture_period * ca_period_size;
  
  cout << "Warning: Neither playback nor capture is running!" << endl;
  return 0;
}

/*
  SOME ALSA DOKU (do not delete it is rare!!! ...)
  
  Only to know the area you want to fill:
  
  while (!done) {
  poll(...)
  nframes = snd_pcm_avail_update(pcm);
  snd_pcm_mmap_begin(pcm, &areas, &offset, &nframes);
  process(..., areas, offset, nframes); // Contiguous frames
  snd_pcm_mmap_commit(pcm, offset, nframes);
  // Here you can choose to cycle or complete the residual nframes
  }
  
  "offset" is the frame offset of the first frame you have to fill and not
  always it's 0 on the first access (as you seems to expect, considered
  you didn't call old snd_pcm_mmap_offset)
  
  snd_pcm_mmap_begin(pcm, &areas, &offset, &frames);
  >snd_pcm_channel_area_t *a = &areas[channel];
  >address = a->addr + (a->first + a->step * offset) / 8;
*/

void ALSAMixerNode::Debug ()
{
  cout << "ALSAMixerNode::Debug:" << endl;
  
  cout << "  Period size: pl: " << pl_period_size
       << ", ca: " << ca_period_size << endl;
  
  cout << "  Channels: playback: " << capture_channels
       << ", capture: " << playback_channels << endl;
  
  cout << "  Initialized: " << initialized
       << ", io_running: " << io_running
       << ", do_playback: " << do_playback
       << ", do_capture: " << do_capture << endl;
  
  if (io_running)
    {
      cout << "  stop requested: " << stop_requested << endl;
      
      cout << "  Current period: playback: " << current_playback_period
	   << ", capture: " << current_capture_period << endl;
      
      cout << endl;
    }
}

bool ALSAMixerNode::reget_card_params ()
{
  if (!capture_card->Open (cname, SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK) )
    return false;
  if (!internal_config_params (capture_card, false, true) )
    return false;
  if (!capture_card->Close () )
    return false;
  
  if (!playback_card->Open (cname, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK) )
    return false;
  if (!internal_config_params (playback_card, true, false) )
    return false;
  if (!playback_card->Close () )
    return false;
  return true;
}

void ALSAMixerNode::dump_alsa_params_info (ALSAHelper::ALSAPCM* card)
{
  cout << "format: " << card->GetFormat () << endl;
  
  cout << "min_ch: " << card->GetChannelsMin () << endl;
  cout << "max_ch: " << card->GetChannelsMax () << endl;

  cout << "min_rate: " << card->GetRateMin () << endl;
  cout << "max_rate: " << card->GetRateMax () << endl;
  
  cout << "period_size: " << card->GetPeriodSize () << endl;
  cout << "buffer_size: " << card->GetBufferSize () << endl;
  
  cout << "periods_min: " << card->GetPeriodsMin () << endl;
  cout << "periods_max: " << card->GetPeriodsMax () << endl;
  
  cout << "periods: " << card->GetPeriods () << endl;
}

bool ALSAMixerNode::internal_config_params (ALSAPCM* card, bool is_pl, bool overwrite_params)
{
  dump_alsa_params_info (card);
  
  // Use real channels if hw:? and use some hardcoded value (plug layer returns bogus values ...)
  if (card->Type () == SND_PCM_TYPE_HW) {
    if (is_pl)
      playback_channels = card->GetChannelsMax ();
    else
      capture_channels = card->GetChannelsMax ();
  }
  else {
    if (is_pl)
      playback_channels = 2;
    else
      capture_channels = 2;
  }
  
  if (!card->SetAccess (SND_PCM_ACCESS_RW_INTERLEAVED) )
    return false;
  
  // try the formats we support ...
  if (card->SetFormat (SND_PCM_FORMAT_S16_LE) ) {
  }
  else if (card->SetFormat (SND_PCM_FORMAT_S32_LE) ) {
  }
  else
    return false;
  
  if (is_pl) {
    if (!card->SetChannels (playback_channels) )
      return false;
  }
  else {
    if (!card->SetChannels (capture_channels) )
      return false;
  }
  
  if (overwrite_params) {
    m_params.min_rate = card->GetRateMin ();
    m_params.max_rate = card->GetRateMax ();
    m_params.cur_rate = 44100;
  }
  
  if (!card->SetRateNear (m_params.cur_rate) )
    return false;
  
  if (overwrite_params) {
    m_params.min_periods = card->GetPeriodsMin ();
    m_params.max_periods = card->GetPeriodsMax ();
    m_params.cur_periods = limit_min_max (8, m_params.min_periods, m_params.max_periods);
  }
  
  cout << "try to configure " << m_params.cur_periods << " periods ..." << endl;
  if (!card->SetPeriodsNear (m_params.cur_periods) )
    {
      cout << "SetPeriodsNear faild ..." << endl;
      return false;
    }
  
  if (overwrite_params) {
    m_params.min_period_time = 1000;
    m_params.max_period_time = 200000;
    m_params.cur_period_time = 800000 / m_params.cur_periods;
  }
  
  int t_period_size;
  if (is_pl)
    t_period_size = pl_period_size;
  else
    t_period_size = ca_period_size;
  
  if (t_period_size == 0)
    {
      cout << "t_period_size == 0 -> using time near ..." << endl;
      
      if (!card->SetPeriodTimeNear (m_params.cur_period_time) )
	return false;
    }
  else
    {
      cout << "period_size == " << t_period_size << " -> using size ..." << endl;
       if (!card->SetPeriodSize (t_period_size) )
	return false;
    }
  
  if (!card->InstallParams () )
    return false;
  
  dump_alsa_params_info (card);
  
  if (!card->SetSwStartThreshold (card->GetPeriods () / 2) )
    return false;
  
  if (!card->InstallSwParams () )
    return false;
  
  cout << "start_th: " << card->GetSwStartThreshold () << endl;
  cout << "stop_th: " << card->GetSwStopThreshold () << endl;
  
  if (is_pl) {
    pl_period_size = card->GetPeriodSize ();
    cout << "Set pl_period_size to: " << pl_period_size << endl;
    
    pl_periods = card->GetPeriods ();
    cout << "Set pl_periods to: " << pl_periods << endl;
  }
  else
  {
    ca_period_size = card->GetPeriodSize ();
    cout << "Set ca_period_size to: " << ca_period_size << endl;
    
    ca_periods = card->GetPeriods ();
    cout << "Set ca_periods to: " << ca_periods << endl;
  }
  
  return true;
}

void ALSAMixerNode::handle_xrun (ALSAHelper::ALSAPCM* card)
{
  cout << "ALSAMixerNode::handle_xrun:" << endl;
  
  card->UpdateStatus ();
  
  if (card->GetStatusState () == SND_PCM_STATE_XRUN)
    {
      struct timeval now, diff, tstamp;
      gettimeofday(&now, 0);
      
      card->GetTriggerTStamp (&tstamp);
      timersub(&now, &tstamp, &diff);
      cout << "  !!!-->> XRUNx !!! (at least " << diff.tv_sec * 1000 + diff.tv_usec / 1000.0 << " ms long)"
	   << endl;
      
      if (!card->Prepare () )
	cout << "  BUR: can't reprepare during XRun!! " << endl;
    }
}

// the second thread ...
void ALSAMixerNode::Loop ()
{
  bool data = true;
  
  if (m_is_first_loop) // do first loop things (real-time scheduling ...)
    {
      m_using_realtime_sched = ThreadHelper::EnableRealtimeScheduling ();
      if (!m_using_realtime_sched)
	cout << "Warning: ALSANode: Can't enable realtime scheduling ..." << endl;
      m_is_first_loop = false;
    }
  
#ifdef DEBUG_GSML_IO_ALSA
  cout << "ALSAMixerNode::Loop:" << endl;
#endif
  try {
  if (!stop_requested)
    {
      // start capture ??
      if (do_capture && !cap_running)
	{
	  cout << "capture is not running ... -> starting now." << endl;
	  if (!capture_card->Start () )
	    cout << "can't start capture!" << endl;
	  else
	    cap_running = true;
	}
      
      // sleep when not worked in last loop
      if (!m_worked_in_last_loop)
	{
	  // get best delay from ALSA
	  int pl_avail, ca_avail, xx_needed;
	  
	  if (do_playback) {
	    playback_card->UpdateStatus ();
	    pl_avail = playback_card->GetStatusAvail ();
	  }
	  if (do_capture) {
	    capture_card->UpdateStatus ();
	    ca_avail = capture_card->GetStatusAvail ();
	  }
	  
	  if (!do_capture)
	    xx_needed = pl_period_size - pl_avail;
	  else if (!do_playback)
	    xx_needed = ca_period_size - ca_avail;
	  else
	    xx_needed = min (pl_period_size - pl_avail, ca_period_size - ca_avail);
	  
	  //	  cout << "pl_avail: " << pl_avail << ", ca_avail: " << ca_avail
	  //           << ", xx_needed: " << xx_needed << endl;
	  
	  xx_needed = limit_min (xx_needed, 0);
	  
	  // TODO: hw-samplerate usage ...
	  int xx_delay = (1000000 / 44100) * xx_needed;
	  
	  // cout << "xx_delay: " << xx_delay << endl;
	  
	  // make shoure other tasks get time ...
	  if (m_using_realtime_sched)
	    ThreadHelper::USleep (xx_delay);
	  else
	    usleep (xx_delay);
	}
      
      m_worked_in_last_loop = false;
      
      if (do_playback)
	{
	  
	  data = write_handler->TryMix (m_mixer);
	  
	  int written = write_handler->TryWrite ();
	  
	  if (written > 0) {
	    ++current_playback_period;
	    m_worked_in_last_loop = true;
	  }
	  else {
	    if (written == -EAGAIN) {
	    }
	    else if (written == -EPIPE) { // xrun
	      cout << "P: XRUN ..." << endl;
	      handle_xrun (playback_card);
	      ++pl_xruns;
	    }
	    else
	      cout << "P: WriteI faild: " << snd_strerror (written) << " " << written << endl;
	  }
	}
      if (do_capture)
	{
	  int read = read_handler->TryRead ();
	  
	  if (read > 0) {
	    ++current_capture_period;
	    m_worked_in_last_loop = true;
	  }
	  else {
	    if (read == -EAGAIN) {
	    }
	    else if (read == -EPIPE) { // xrun
	      cout << "C: XRUN ..." << endl;
	      handle_xrun (capture_card);
	      ++ca_xruns;
	    }
	    else
	      cout << "C: Read faild: " << snd_strerror (read) << " " << read << endl;
	  }
	  bool temp = read_handler->TryMix (m_mixer);
	  data = temp && data;
	}
    }
  
  // no else - because it might be changed in the io loop-part ...
  if (!data || stop_requested)
    {
      //#ifdef DEBUG_GSML_IO_ALSA
      cout << "  Loop: END OF IO" << endl;
      //#endif
      
      /* clear some states and data */
      
      cout << "Closind cards ..." << endl;
      if (do_playback)
	playback_card->Close ();
      if (do_capture)
	capture_card->Close ();
      
      del_helper2 (read_handler);
      del_helper2 (write_handler);
      
      io_running = false;
      m_mixer->IOFinishedNotification ();
      
      Wait ();
    }
  }
  catch (GSML::BasicExcept e)
    {
      cout << endl 
	   << "!! ALSAMixerNode::Loop caught exception: !!" << endl
	   << e.info << endl;
      throw;
    }
}

/*******************
 * IO Hander Bases *
 *******************/

ReadHandler::ReadHandler (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size)
{
  m_card = i_card;
  m_channels = i_channels;
  m_period_size = i_period_size;
}

WriteHandler::WriteHandler (ALSAHelper::ALSAPCM* i_card, track_index i_channels, sample_index i_period_size)
{
  m_card = i_card;
  m_channels = i_channels;
  m_period_size = i_period_size;
}

/**********************
 * 16 Bit IO Handlers *
 **********************/

ReadHandler16Bit::ReadHandler16Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels,
				    sample_index i_period_size)
  : ReadHandler (i_card, i_channels, i_period_size)
{
  buffer = new SampleBufferDefault (m_period_size);
  interleaved = new pcm_16bit::value_type [m_period_size * m_channels];
  ca_vaild = false;
}

ReadHandler16Bit::~ReadHandler16Bit ()
{
  delete buffer;
  delete [] interleaved;
}

int ReadHandler16Bit::TryRead ()
{
  int read = m_card->ReadI (interleaved, m_period_size);
  if (read > 0)
    ca_vaild = true;
  return read;
}

bool ReadHandler16Bit::TryMix (Mixer* mixer)
{
  if (ca_vaild) {
    bool data = true;
    
    buffer->SetUsedLength (m_period_size);
    for (track_index ch = 0; ch < m_channels; ++ch)
      {
	for (int i = 0; i < m_period_size; ++i)
	  buffer->data [i] ( interleaved [i * m_channels + ch] << 8 );
	
	bool temp = mixer->Write (ch, buffer);
	data = data && temp;
      }
    ca_vaild = false;
    return data;
  }
  else
    return true;
}
  
WriteHandler16Bit::WriteHandler16Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels,
				      sample_index i_period_size)
  : WriteHandler (i_card, i_channels, i_period_size)
{
  buffer = new SampleBufferDefault (m_period_size);
  interleaved = new pcm_16bit::value_type [m_period_size * m_channels];
  
  pl_vaild = false;
}

WriteHandler16Bit::~WriteHandler16Bit ()
{
  delete buffer;
  delete interleaved;
}

bool WriteHandler16Bit::TryMix (Mixer* mixer)
{
  if (!pl_vaild)
    {
      bool data = true;
      // building interleaved buffer ...
      for (track_index ch = 0; ch < m_channels; ++ch)
	{
	  buffer->Clear ();
	  bool temp = mixer->Read (ch, buffer);
	  data = data && temp;
	  
	  for (int i = 0; i < m_period_size; ++i)
	    interleaved [i * m_channels + ch]
	      ( pcm_24_8bit::Convert2int16 (buffer->data [i] ) );
	}
      pl_vaild = true;
      return data;
    }
  else
    return true;
}
      
int WriteHandler16Bit::TryWrite ()
{
  int write = m_card->WriteI (interleaved, m_period_size);
  if (write > 0)
    pl_vaild = false;
  return write;
}

/**********************
 * 32 Bit IO Handlers *
 **********************/

ReadHandler32Bit::ReadHandler32Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels,
				    sample_index i_period_size)
  : ReadHandler (i_card, i_channels, i_period_size)
{
  buffer = new SampleBufferDefault (m_period_size);
  interleaved = new int32 [m_period_size * m_channels];
  ca_vaild = false;
}

ReadHandler32Bit::~ReadHandler32Bit ()
{
  delete buffer;
  delete [] interleaved;
}

int ReadHandler32Bit::TryRead ()
{
  int read = m_card->ReadI (interleaved, m_period_size);
  if (read > 0)
    ca_vaild = true;
  return read;
}

bool ReadHandler32Bit::TryMix (Mixer* mixer)
{
  if (ca_vaild) {
    bool data = true;
    
    buffer->SetUsedLength (m_period_size);
    for (track_index ch = 0; ch < m_channels; ++ch)
      {
	for (int i = 0; i < m_period_size; ++i)
	  buffer->data [i] ( interleaved [i * m_channels + ch] >> 8 );
	
	bool temp = mixer->Write (ch, buffer);
	data = data && temp;
      }
    ca_vaild = false;
    return data;
  }
  else
    return true;
}
  
WriteHandler32Bit::WriteHandler32Bit (ALSAHelper::ALSAPCM* i_card, track_index i_channels,
				      sample_index i_period_size)
  : WriteHandler (i_card, i_channels, i_period_size)
{
  buffer = new SampleBufferDefault (m_period_size);
  interleaved = new int32 [m_period_size * m_channels];
  
  pl_vaild = false;
}

WriteHandler32Bit::~WriteHandler32Bit ()
{
  delete buffer;
  delete [] interleaved;
}

bool WriteHandler32Bit::TryMix (Mixer* mixer)
{
  if (!pl_vaild)
    {
      int32 t_conv, t_result;
      
      bool data = true;
      // building interleaved buffer ...
      for (track_index ch = 0; ch < m_channels; ++ch)
	{
	  buffer->Clear ();
	  bool temp = mixer->Read (ch, buffer);
	  data = data && temp;
	  
	  for (int i = 0; i < m_period_size; ++i)
	    {
	      t_conv = buffer->data [i];
	      
	      if (t_conv > 0x007fffff)
		t_result = 0x7fffffff;
	      else if (t_conv < 0x00800000 * (-1) )
		t_result = 0x80000000;
	      else t_result = int32 (t_conv << 8);
      
	      interleaved [i * m_channels + ch] = t_result;
	    }
	}
      pl_vaild = true;
      return data;
    }
  else
    return true;
}
      
int WriteHandler32Bit::TryWrite ()
{
  int write = m_card->WriteI (interleaved, m_period_size);
  if (write > 0)
    pl_vaild = false;
  return write;
}

#endif GSML_WITH_ALSA
