/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Different IO architecture implementations.
 */

#include "IO_impl.hh"

#include <algorithm>
#include <functional>

#include <sstream>

using namespace GSML;

/***********************
 * ScatterHelperSource *
 ***********************/

void ScatterHelperSource::BuildTablesHelper (track_index pl_tracks, track_index ca_tracks, bool n_state)
{
  ClearTablesHelper ();
  
  for (track_index i = 0; i < pl_tracks; ++i)
    io_read_pos_vaild.push_back (n_state);
  master_read_pos_vaild = n_state;

  for (track_index i = 0; i < ca_tracks; ++i)
    io_write_pos_vaild.push_back (n_state);
  master_write_pos_vaild = n_state;
}

void ScatterHelperSource::ClearTablesHelper ()
{
  io_read_pos_vaild.clear ();
  io_write_pos_vaild.clear ();
}

bool ScatterHelperSource::DoRead (MixerSource* source, track_index track, sample_index pos,
				  sample_index length,  SampleBufferDefault* dest, bool is_master)
{
  if (is_master && !master_read_pos_vaild) {
    source->SetReadPos (track, pos, is_master);
    master_read_pos_vaild = true;
  }
  else if (!is_master && !io_read_pos_vaild [track] ) {
    source->SetReadPos (track, pos, is_master);
    io_read_pos_vaild [track] = true;
  }
  return source->Read (track, length, dest, is_master);
}

void ScatterHelperSource::InvaildRead (track_index track, bool is_master)
{
  if (is_master)
    master_read_pos_vaild = false;
  else
    io_read_pos_vaild [track] = false;
}

bool ScatterHelperSource::DoWrite (MixerSource* source, track_index track, sample_index pos,
				   sample_index length, SampleBufferDefault* src,
				   sample_index buffer_index, bool is_master)
{
  if (is_master && !master_write_pos_vaild) {
    source->SetWritePos (track, pos, is_master);
    master_write_pos_vaild = true;
  }
  else if (!is_master && !io_write_pos_vaild [track] ) {
    source->SetWritePos (track, pos, is_master);
    io_write_pos_vaild [track] = true;
  }
  
  return source->Write (track, length, src, buffer_index, is_master);
}

void ScatterHelperSource::InvaildWrite (track_index track, bool is_master)
{
  if (is_master)
    master_write_pos_vaild = false;
  else
    io_write_pos_vaild [track] = false;
}

/******************
 * The DataSource *
 ******************/

DataSource::DataSource (GSML::DataLayer* pl_layer, GSML::DataLayer* ca_layer)
{
  m_pl_layer = pl_layer;
  m_ca_layer = ca_layer;
  
  m_pinfo = 0;
}

DataSource::~DataSource ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~DataSource io is running in dtor!?!?!?! ..." << endl;
}

void DataSource::SetDataLayer (GSML::DataLayer* pl_layer, GSML::DataLayer* ca_layer)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_pl_layer = pl_layer;
  m_ca_layer = ca_layer;
  
  if (m_pl_layer != 0)
    m_playback_tracks = m_pl_layer->Tracks ();
  else
    m_playback_tracks = 0;
  
  if (m_ca_layer != 0)
    m_capture_tracks = m_ca_layer->Tracks ();
  else
    m_capture_tracks = 0;
}

track_index DataSource::PlaybackTracks () const
{
  return m_playback_tracks;
}

track_index DataSource::CaptureTracks () const
{
  return m_capture_tracks;
}

bool DataSource::IsInputable () const
{
  return m_ca_layer != 0;
}

sample_index DataSource::MinStart () const
{
  return 0;
}

sample_index DataSource::MaxEnd () const
{
  return m_pl_layer->Length ();
}

void DataSource::StartO_impl ()
{
  m_playback_tracks = m_pl_layer->Tracks ();
  
  if (m_ca_layer != 0)
    m_capture_tracks = m_ca_layer->Tracks ();
  else
    m_capture_tracks = 0;
  
  m_playback_end_pos = m_pl_layer->Length ();
  
  BuildTables (0);
}

void DataSource::StartIO_impl ()
{
  if (m_ca_layer == 0)
    {
      ExceptNotInputable e;
      e.info = "Not capture layer specified (pointer nil)";
      throw e;
    }
  
  if (m_ca_layer->IsFixated () )
    {
      ExceptNotInputable e;
      e.info = "Not capture layer is fixated. No write would be possible.";
      throw e;
    }
  
  m_playback_tracks = m_pl_layer->Tracks ();
  m_capture_tracks = m_ca_layer->Tracks ();
  
  m_playback_end_pos = m_pl_layer->Length ();

  BuildTables (0);
  
  // additinal tables
  io_write_prepared.clear ();
  for (track_index i = 0; i < m_capture_tracks; ++i)
    io_write_prepared.push_back (false);
}

void DataSource::StopIO_impl ()
{
  if (io_input_running)
    {
      for (track_index i = 0; i < m_capture_tracks; ++i)
	try {
	  if (io_write_prepared [i] )
	    (*m_ca_layer) [i]->FixateWrite ();
	}
	catch (TrackLayer::Except e)
	  {
	    cout << "Got a TrackLayer exception during fixate ..." << endl;
	    cout << "  " << e.info << endl;
	  }
    }
  io_write_prepared.clear ();
}

void DataSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  if (is_master)
    master_read_position = n_pos;
  else
    io_read_positions [track] = n_pos;
}

void DataSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  if (is_master)
    master_write_position = n_pos;
  else
    {
      // fixate current write
      if (io_write_prepared [track])
	(*m_ca_layer) [track]->FixateWrite ();
      
      io_write_positions [track] = n_pos;
      io_write_prepared [track] = false;
    }
}

bool DataSource::Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
			    bool is_master)
{
  sample_index cur_pos = CurrentReadPosition (track, is_master);
  bool data = true;
  sample_index t_length;
  
  if (cur_pos < m_playback_end_pos)
    {
      if (!is_master)
	{
	  sample_index t_length = limit_max (m_playback_end_pos - cur_pos, length);
	  
	  if (t_length > 0) {
	    try {
	      sample_index t_used_length = dest->UsedLength ();
	      if (! (*m_pl_layer) [track] -> ReadData (cur_pos, t_length, dest) ) {
		// not all got read -> zero
		sample_index t_read_length = dest->UsedLength () - t_used_length;
		//cout << "zeroing: " << t_length - t_read_length << endl;
		dest->Zero (t_length - t_read_length);
	      }
	    }
	    catch (BasicExcept e)
	      {
		cout << "DataSource::Read:" << endl;
		cout << "EE XX EE PP TT II OO NN !! ?? ::" << endl;
		cout << e.info << endl;
	      }
	  }
	  
	  if (t_length < length) { // playback end reached ?
	    dest->Zero (length - t_length);
	  }
	  
	}
    }
  else {
    if (!is_master)
      dest->Zero (length);
    data = false;
  }
  
  CurrentReadPosition (track, is_master) += length;
  
  //cout << "DateMixerSource::Read returning: " << data << endl;
  return data;
}

bool DataSource::Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
			     sample_index buffer_index, bool is_master)
{
  sample_index cur_pos = CurrentWritePosition (track, is_master);
  cout << "DataSource::Write_impl: track: " << track << ", is_master: " << is_master
       << ", cur_pos: " << cur_pos << endl;
    
  if (!is_master)
    {
      if (!io_write_prepared [track])
	{
	  (*m_ca_layer) [track] -> PrepareOverWrite (cur_pos);
	  io_write_prepared [track] = true;
	}
      
      cout << "writeing" << endl;
      
      // build a SBHandle that has "reference counting" ...
      SBHandle data = manage (new SampleBufferDefault (length) ); // , true for destroy couts
      //cout << "SBuffer @: " << data << endl;
      try {
	data->WriteData (length, src->data + buffer_index);
      }
      catch (BasicExcept e)
	{
	  cout << "DataSource::WriteData in SampleBuffer:" << endl;
	  cout << "EE XX EE PP TT II OO NN !! ?? ::" << endl;
	  cout << e.info << endl;
	}
      
      try {
	(*m_ca_layer) [track] -> WriteData (length, data, 0);
      }
      catch (BasicExcept e)
	{
	  cout << "DataSource::Write:" << endl;
	  cout << "EE XX EE PP TT II OO NN !! ?? ::" << endl;
	  cout << e.info << endl;
	}
    }
  
  CurrentWritePosition (track, is_master) += length;
  
  return true;
}

void DataSource::Debug_impl () const
{
  cout << "DataSource::Debug_impl:" << endl;
}

/*************************
 * SelectionFilterSource *
 *************************/

SelectionFilterSource::SelectionFilterSource (MixerSource* i_src, DataLayer* i_layer)
{
  m_source = i_src;
  m_layer = i_layer;
  
  m_for_playback = false;
  m_for_capture = false;
}

SelectionFilterSource::~SelectionFilterSource ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~SelectionFilterSource io is running in dtor!?!?!?! ..." << endl;
}

void SelectionFilterSource::SetDataLayer (DataLayer* n_layer)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_layer = n_layer;
}

void SelectionFilterSource::SetSelectionForPlayback (bool n_state)
{
  if (io_running)
    {
      for (track_index i = 0; i < m_playback_tracks; ++i)
	InvaildRead (i, false);
      InvaildRead (0, true);
    }
  
  m_for_playback = n_state;
}

bool SelectionFilterSource::IsSelectionForPlayback () const
{
  return m_for_playback;
}

void SelectionFilterSource::SetSelectionForCapture (bool n_state)
{
  if (io_running) {
    for (track_index i = 0; i < m_capture_tracks; ++i)
      InvaildWrite (i, false);
    InvaildWrite (0, true);
  }
  m_for_capture = n_state;
}

bool SelectionFilterSource::IsSelectionForCapture () const
{
  return m_for_capture;
}

track_index SelectionFilterSource::PlaybackTracks () const
{
  return m_source->PlaybackTracks ();
}

track_index SelectionFilterSource::CaptureTracks () const
{
  return m_source->CaptureTracks ();
}

sample_index SelectionFilterSource::MinStart () const
{
  if (!io_running) {
    cout << "WARNING: SelectionFilterSource::MinStart called and no IO is running!" << endl;
    return 0;
  }
  
  return m_start_pos;
}

sample_index SelectionFilterSource::MaxEnd () const
{
  if (!io_running) {
    cout << "WARNING: SelectionFilterSource::MinStart called and no IO is running!" << endl;
    return 0;
  }
   
  return m_end_pos;
}

bool SelectionFilterSource::IsInputable () const
{
  return m_source->IsInputable ();
}

void SelectionFilterSource::StartO_impl ()
{
  m_source->StartO ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->CaptureTracks ();
  
  track_index t_tracks = build_selection_data ();
  m_playback_tracks = limit_max (m_playback_tracks, t_tracks);
  m_capture_tracks = limit_max (m_capture_tracks, t_tracks);
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void SelectionFilterSource::StartIO_impl ()
{
  m_source->StartIO ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->PlaybackTracks ();
  
  track_index t_tracks = build_selection_data ();
  m_playback_tracks = limit_max (m_playback_tracks, t_tracks);
  m_capture_tracks = limit_max (m_capture_tracks, t_tracks);
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void SelectionFilterSource::StopIO_impl ()
{
  m_source->StopIO ();
  
  ClearTables ();
  ClearTablesHelper ();
  
  for (vector <track_data>::iterator i = m_data.begin (); i != m_data.end (); ++i)
    del_helper2 (i->selection);
  m_data.resize (0);
}

void SelectionFilterSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  // m_source->SetReadPos (track, n_pos, is_master);
  
  InvaildRead (track, is_master);
  CurrentReadPosition (track, is_master) = n_pos;
  
  if (!is_master)
    m_data [track].pl_segment = m_data  [track].selection->begin ();
}

void SelectionFilterSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  // m_source->SetWritePos (track, n_pos, is_master);
  
  InvaildWrite (track, is_master);
  CurrentWritePosition (track, is_master) = n_pos;
  
  if (!is_master)
    m_data [track].ca_segment = m_data  [track].selection->begin ();
  
}

bool SelectionFilterSource::Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
				       bool is_master)
{
  sample_index proc_length = 0;
  bool data = true;
  
  sample_index cur_pos = CurrentReadPosition (track, is_master);
  
  if (m_for_playback) {
    cout << "selection USED for playback (please wait ;-)" << endl;
    
    while (proc_length < length && data)
      {
	if (is_master)
	  {
	    cout << "is_master" << endl;
	    if (cur_pos < m_end_pos)
	      {
		sample_index t_length = limit_max (m_end_pos - cur_pos, length - proc_length);
		
		cout << "reading: " << t_length << " for master" << endl;
		
		data = DoRead (m_source, track, cur_pos, t_length, dest, is_master);
		proc_length += t_length;
	      }
	    else{
	      cout << "end for master" << endl;
	      InvaildRead (track, is_master);
	      data = false;
	    }
	  }
	else
	  {
	    track_data& t_data = m_data [track];
	    
	    // iterate to right segment
	    while (t_data.pl_segment != t_data.selection->end () && t_data.pl_segment->end < cur_pos)
	      {
		cout << "not at right segment ..." << endl;
		++t_data.pl_segment;
	      }
	    
	    // end reached ??
	    if (t_data.pl_segment == t_data.selection->end () )
	      {
		cout << "end reached ..." << endl;
		if (cur_pos < m_end_pos)
		  {
		    sample_index t_length = limit_max (m_end_pos - cur_pos, length - proc_length);
		    cout << "zeroing: " << t_length << endl;
		    dest->Zero (t_length);
		    
		    proc_length += t_length;
		    cur_pos += t_length;
		  }
		if (proc_length < length)
		  {
		    cout << "behind pl-region data now == false" << endl;
		    data = false;
		  }
		
		InvaildRead (track, is_master);
		
	      } // end if end
	    else
	      {
		cout << "ok segment found" << endl;
		if (cur_pos < t_data.pl_segment->start)
		  {
		    cout << "before sel-segment: cur_pos: " << cur_pos
			 << ", sel-seg-start: " << t_data.pl_segment->start << endl;
		    sample_index t_length = limit_max (t_data.pl_segment->start - cur_pos,
						       length - proc_length);
		    cout << "zeroing: " << t_length << endl;
		    dest->Zero (t_length);
		    
		    InvaildRead (track, is_master);
		    
		    proc_length += t_length;
		    cur_pos += t_length;
		  }
		else if (cur_pos >= t_data.pl_segment->start)
		  {
		    cout << "ok. in sel-segment: cur_pos: " << cur_pos << ", length: " << length
			 << ", proc_length: " << proc_length
			 << ", sel-seg-end: " << t_data.pl_segment->end << endl;
		    
		    sample_index t_length = limit_max (t_data.pl_segment->end - cur_pos,
						       length - proc_length);
		    if (t_length > 0)
		      {
			cout << "reading: " << t_length << endl;
			DoRead (m_source, track, cur_pos, t_length, dest, false);
			
			proc_length += t_length;
			cur_pos += t_length;
		      }
		    else
		      {
			cout << "length !> 0" << endl;
			++t_data.pl_segment;
		      }
		    
		  }
	      }
	  }
      }
  }
  else {
    // cout << "selection NOT used for playback passing through ..." << endl;
    
    data = DoRead (m_source, track, cur_pos, length, dest, is_master);
    proc_length = length;
  }
  
  // zero left space
  if (!is_master && proc_length < length)
    dest->Zero (length - proc_length);
  
  CurrentReadPosition (track, is_master) += length;
  
  return data;
}

bool SelectionFilterSource::Write_impl (track_index track, sample_index length,
					SampleBufferDefault* src, sample_index buffer_index,
					bool is_master)
{
  sample_index proc_length = 0;
  
  bool data = true;
  
  sample_index cur_pos = CurrentWritePosition (track, is_master);
  
  if (m_for_capture) {
    cout << "selection USED for capture (please wait ;-)" << endl;
    
    while (proc_length < length && data)
      {
	if (is_master)
	  {
	    cout << "is_master" << endl;
	    if (cur_pos < m_end_pos)
	      {
		sample_index t_length = limit_max (m_end_pos - cur_pos, length - proc_length);
		
		cout << "writeing: " << t_length << " for master" << endl;
		data = DoWrite (m_source, track, cur_pos, t_length, src, buffer_index + proc_length, is_master);
		proc_length += t_length;
	      }
	    else{
	      cout << "end for master" << endl;
	      InvaildWrite (track, is_master);
	      data = false;
	    }
	  }
	else
	  {
	    track_data& t_data = m_data [track];
	    cout << "serch to right segment" << endl;
	    // iterate to right segment
	    while (t_data.ca_segment != t_data.selection->end () && t_data.ca_segment->end < cur_pos)
	      {
		cout << "not at right segment ..." << endl;
		++t_data.ca_segment;
	      }
	    
	    // end reached ??
	    if (t_data.ca_segment == t_data.selection->end () )
	      {
		cout << "end reached ..." << endl;
		if (cur_pos < m_end_pos)
		  {
		    sample_index t_length = limit_max (m_end_pos - cur_pos, length - proc_length);
		    
		    proc_length += t_length;
		    cur_pos += t_length;
		  }
		if (proc_length < length)
		  {
		    cout << "behind ca-region data now == false" << endl;
		    data = false;
		  }
		InvaildWrite (track, is_master);
	      } // end if end
	    else
	      {
		cout << "ok segment found" << endl;
		if (cur_pos < t_data.ca_segment->start)
		  {
		    cout << "before sel-segment: cur_pos: " << cur_pos
			 << ", sel-seg-start: " << t_data.ca_segment->start << endl;
		    sample_index t_length = limit_max (t_data.ca_segment->start - cur_pos,
						       length - proc_length);
		    
		    InvaildWrite (track, is_master);
		    
		    proc_length += t_length;
		    cur_pos += t_length;
		  }
		else if (cur_pos >= t_data.ca_segment->start)
		  {
		    cout << "ok. in sel-segment: cur_pos: " << cur_pos << ", length: " << length
			 << ", proc_length: " << proc_length
			 << ", sel-seg-end: " << t_data.ca_segment->end << endl;
		    
		    sample_index t_length = limit_max (t_data.ca_segment->end - cur_pos,
						       length - proc_length);
		    if (t_length > 0)
		      {
			cout << "writeing: " << t_length << endl;
			
			DoWrite (m_source, track, cur_pos, t_length, src, buffer_index + proc_length, is_master);
			
			proc_length += t_length;
			cur_pos += t_length;
		      }
		    else
		      {
			cout << "length !> 0" << endl;
			++t_data.ca_segment;
		      }
		  }
	      }
	  }
      }
  }
  else {
    cout << "selection NOT used for capture passing through ..." << endl;
    
    data = DoWrite (m_source, track, cur_pos, length, src, buffer_index, is_master);
    proc_length = length;
  }
  
  CurrentWritePosition (track, is_master) += length;
  
  return data;
}

void SelectionFilterSource::Debug_impl () const
{
  cout << "SelectionFilterSource::Debug_impl:" << endl;
  m_source->Debug();
}

track_index SelectionFilterSource::build_selection_data ()
{
  cout << "SelectionFilterSource::build_selection_data:" << endl;
  
  track_index t_tracks = m_layer->Tracks ();
  
  m_start_pos = m_source->MinStart ();
  m_end_pos = m_source->MaxEnd ();
  
  bool t_first_segment = true;
  
  for (track_index tr = 0; tr < t_tracks; ++tr)
    {
      // create / copy selection
      m_data.push_back ();
      
      m_data.back ().selection = new TrackLayer::Selection ( (*m_layer)[tr]->selection );
      track_data& t_data = m_data.back ();
      t_data.pl_segment = t_data.selection->begin ();
      t_data.ca_segment = t_data.selection->begin ();
      
      // update start / end position (when sel. used for pl or ca)
      if (m_for_playback || m_for_capture)
	{
	  sample_index t_start, t_end;
	  
	  if (t_data.selection->MinStart (&t_start) && t_data.selection->MaxEnd (&t_end) )
	    {
	      cout << "  got start (" << t_start << ") / end (" << t_end << ") from selection" << endl;
	      if (t_first_segment)
		{
		  m_start_pos = t_start;
		  m_end_pos = t_end;
		  
		  t_first_segment = false;
		}
	      else
		{
		  m_start_pos = limit_max (t_start, m_start_pos);
		  m_end_pos = limit_min (t_end, m_end_pos);
		}
	    }
	}
      else
      cout << "  selection has no segments ... - no start / end available ..." << endl;
    }
  
  cout << "m_start_pos: " << m_start_pos << ", m_end_pos: " << m_end_pos << endl;
  
  return t_tracks;
}

/**********************
 * MarkerFilterSource *
 **********************/

MarkerFilterSource::MarkerFilterSource (MixerSource* i_src, DataLayer* i_layer)
{
  m_source = i_src;
  m_layer = i_layer;
  
  m_playback_start_pos = 0;
  m_playback_end_pos = m_source->MaxEnd ();
  m_capture_start_pos = 0;
  m_capture_end_pos = m_source->MaxEnd ();

  connection_marker_deleted =
    m_layer->markers.SignalDeleted.connect (slot (this, &MarkerFilterSource::SlotMarkerDeleted) );
  connection_marker_changed =
    m_layer->markers.SignalChanged.connect (slot (this, &MarkerFilterSource::SlotMarkerChanged) );
}

MarkerFilterSource::~MarkerFilterSource ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~MarkerFilterSource io is running in dtor!?!?!?! ..." << endl;
}

void MarkerFilterSource::SetDataLayer (DataLayer* n_layer)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_layer = n_layer;
    
  connection_marker_deleted.disconnect ();
  connection_marker_changed.disconnect ();
  
  connection_marker_deleted =
    m_layer->markers.SignalDeleted.connect (slot (this, &MarkerFilterSource::SlotMarkerDeleted) );
  connection_marker_changed =
    m_layer->markers.SignalChanged.connect (slot (this, &MarkerFilterSource::SlotMarkerChanged) );
  
  reget_positions ();
}

void MarkerFilterSource::SetPlaybackStart (const string& start_mark)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_playback_start_mark = start_mark;
}

void MarkerFilterSource::SetPlaybackEnd (const string& end_mark)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_playback_end_mark = end_mark;
}

void MarkerFilterSource::SetCaptureStart (const string& start_mark)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  
  m_capture_start_mark = start_mark;
}

void MarkerFilterSource::SetCaptureEnd (const string& end_mark)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  
  m_capture_end_mark = end_mark;
}

track_index MarkerFilterSource::PlaybackTracks () const
{
  return m_source->PlaybackTracks ();
}

track_index MarkerFilterSource::CaptureTracks () const
{
  return m_source->CaptureTracks ();
}

sample_index MarkerFilterSource::MinStart () const
{
  return m_start_pos;
}

sample_index MarkerFilterSource::MaxEnd () const
{
  return m_end_pos;
}

bool MarkerFilterSource::IsInputable () const
{
  return m_source->IsInputable ();
}

void MarkerFilterSource::SlotMarkerDeleted (const string& marker, sample_index old_pos)
{
  if (marker == m_playback_start_mark)
    {
      m_playback_start_mark = "";
      m_playback_start_pos = 0;
    }
  if (marker == m_capture_start_mark)
    {
      m_capture_start_mark = "";
      m_capture_start_pos = 0;
    }
  
  if (marker == m_playback_end_mark)
    {
      m_playback_end_mark = "";
      m_playback_end_pos = m_source->MaxEnd ();
    }
  if (marker == m_capture_end_mark)
    {
      m_capture_end_mark = "";
      m_capture_end_pos = m_source->MaxEnd ();
    }
}

void MarkerFilterSource::SlotMarkerChanged (const string& marker, sample_index old_pos, sample_index new_pos)
{
  if (marker == m_playback_start_mark)
    {
      m_playback_start_pos = new_pos;
    }
  if (marker == m_playback_end_mark)
    {
      m_playback_end_pos = new_pos;
    }
  
  if (marker == m_capture_start_mark)
    {
      m_capture_start_pos = new_pos;
    }
  if (marker == m_capture_end_mark)
    {
      m_capture_end_pos = new_pos;
    }
  limit_start_end ();
}

void MarkerFilterSource::StartO_impl ()
{
  m_source->StartO ();
  
  reget_positions ();
  limit_start_end ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->CaptureTracks ();
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void MarkerFilterSource::StartIO_impl ()
{
  m_source->StartIO ();
  
  reget_positions ();
  limit_start_end ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->CaptureTracks ();
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void MarkerFilterSource::StopIO_impl ()
{
  m_source->StopIO ();
}

void MarkerFilterSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  // m_source->SetReadPos (track, n_pos, is_master);
  
  InvaildRead (track, is_master);
  CurrentReadPosition (track, is_master) = n_pos;
}

void MarkerFilterSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  // m_source->SetWritePos (track, n_pos, is_master);
  
  InvaildWrite (track, is_master);
  CurrentWritePosition (track, is_master) = n_pos;
}

bool MarkerFilterSource::Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
				    bool is_master)
{
  sample_index proc_length = 0;
  sample_index t_length;
  
  sample_index cur_pos = CurrentReadPosition (track, is_master);
  
  bool data = true;
  
  // before pl-start ??
  if (cur_pos < m_playback_start_pos)
    {
      t_length = limit_max (m_playback_start_pos - cur_pos, length);
      
      if (!is_master)
	dest->Zero (t_length);
      
      InvaildRead (track, is_master);
      
      cur_pos += t_length;
      proc_length += t_length;
    }
  
  //  cout << "  used_length: " << dest->UsedLength () << endl;
  
  // in pl-range ??
  if (proc_length < length)
    {
      t_length = length - proc_length;
      
      // limit end (if marker)
      if (m_playback_end_mark != "")
	t_length = limit_max (t_length, m_playback_end_pos - cur_pos);
      
      if (t_length >= 0)
	{
	  try {
	    if (! DoRead (m_source, track, cur_pos, t_length, dest, is_master) )
	      data = false;
	  }
	  catch (BasicExcept e)
	    {
	      cout << "EE XX EE PP TT II OO NN !! !! ::" << endl;
	      cout << e.info << endl;
	    }
	}
      else
	{
	  if (!is_master)
	    dest->Zero (length - proc_length);
	  data = false;
	}
    }
  else
    {
      if (!is_master)
	dest->Zero (length - proc_length);
      
      data = false;
    }
  
  CurrentReadPosition (track, is_master) += length;
  
  if (!data)
    InvaildRead (track, is_master);
  
  //cout << "DateMixerSource::Read returning: " << data << endl;
  return data;
}

bool MarkerFilterSource::Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
				     sample_index buffer_index, bool is_master)
{
  sample_index cur_pos = CurrentWritePosition (track, is_master);
  
  bool data = false;
  
  // before capture range
  if (cur_pos + length <= m_capture_start_pos)
    {
      InvaildWrite (track, is_master);
    }
  else
    {
      // calc skip
      sample_index skip = 0;
      if (cur_pos < m_capture_start_pos)
	skip = m_capture_start_pos - cur_pos;
      
      sample_index t_length = length - skip;
      // limit end (if marker)
      if (m_capture_end_mark != "")
	t_length = limit_max (m_capture_end_pos - cur_pos - skip, t_length);
      
      if (t_length > 0)
	{
	  // cout << "  WRITE: skip: " << skip << ", length: " << t_length << endl;
	  DoWrite (m_source, track, cur_pos + skip, length, src, skip, is_master);
	  data = true;
	}
      
      
      if (t_length <= 0 || skip + t_length < length) // after capture-range
	{
	  InvaildWrite (track, is_master);
	  data = false;
	}
    }
  
  if (!data)
    InvaildWrite (track, is_master);
  
  CurrentWritePosition (track, is_master) += length;
  
  return data;
}

void MarkerFilterSource::reget_positions ()
{
  if (m_playback_start_mark != "")
    m_layer->markers.GetMarker (m_playback_start_mark, &m_playback_start_pos);
  else
    m_playback_start_pos = 0;
  
  if (m_capture_start_mark != "")
    m_layer->markers.GetMarker (m_capture_start_mark, &m_capture_start_pos);
  else
    m_capture_start_pos = 0;
  
  if (m_playback_end_mark != "")
    m_layer->markers.GetMarker (m_playback_end_mark, &m_playback_end_pos);
  else
    m_playback_end_pos = m_source->MaxEnd ();
  
  if (m_capture_end_mark != "")
    m_layer->markers.GetMarker (m_capture_end_mark, &m_capture_end_pos);
  else
    m_capture_end_pos = m_source->MaxEnd ();
}

void MarkerFilterSource::limit_start_end ()
{
  if (io_input_running)
    {
      m_start_pos = min (m_playback_start_pos, m_capture_end_pos);
      m_end_pos = max (m_capture_start_pos, m_capture_end_pos);
    }
  else
    {
      m_start_pos = m_playback_start_pos;
      m_end_pos = m_playback_end_pos;
    }
  SignalMinStartChanged (m_start_pos);
  SignalMaxEndChanged (m_end_pos);
}

void MarkerFilterSource::Debug_impl () const
{
  cout << "MarkerFilterSource::Debug_impl:" << endl;

  cout << "  Playback Marker: start: " << m_playback_start_mark
       << ", end: " << m_playback_end_mark << endl;
  
  cout << "  Capture Marker: start: " << m_capture_start_mark
       << ", end: " << m_capture_end_mark << endl;
  
  cout << "  Playback Pos: start: " << m_playback_start_pos
       << ", end: " << m_playback_end_pos << endl;
  
  cout << "  Capture Pos: start: " << m_capture_start_pos
       << ", end: " << m_capture_end_pos << endl;
  
  cout << "  Effective: start: " << m_start_pos << ", end: " << m_end_pos << endl;
  
  cout << "  SOURCE:" << endl;
  m_source->Debug ();
}

/************************
 * The LoopFilterSource *
 ************************/

LoopFilterSource::LoopFilterSource (MixerSource* i_src)
{
  m_source = i_src;
  m_looping = true;
  m_stop_at_end = false;
  
  m_source->SignalMinStartChanged.connect (slot (this, &LoopFilterSource::SlotMinStartChanged) );
  m_source->SignalMaxEndChanged.connect (slot (this, &LoopFilterSource::SlotMaxEndChanged) );
  
  m_start_pos = m_source->MinStart ();
  m_end_pos = m_source->MaxEnd ();
}

LoopFilterSource::~LoopFilterSource ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~LoopFilterSource io is running in dtor!?!?!?! ..." << endl;
}

void LoopFilterSource::SetLooping (bool n_state)
{
  m_looping = n_state;
}

bool LoopFilterSource::Looping ()
{
  return m_looping;
}

void LoopFilterSource::SetStopAtEnd (bool n_state)
{
  m_stop_at_end = n_state;
}

bool LoopFilterSource::StopAtEnd ()
{
  return m_stop_at_end;
}

track_index LoopFilterSource::PlaybackTracks () const
{
  return m_source->PlaybackTracks ();
}

track_index LoopFilterSource::CaptureTracks () const
{
  return m_source->CaptureTracks ();
}
    
sample_index LoopFilterSource::MinStart () const
{
  return m_start_pos;
}

sample_index LoopFilterSource::MaxEnd () const
{
  return m_end_pos;
}

bool LoopFilterSource::IsInputable () const
{
  return m_source->IsInputable ();
}

void LoopFilterSource::StartO_impl ()
{
  m_source->StartO ();
  
  m_start_pos = m_source->MinStart ();
  m_end_pos = m_source->MaxEnd ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->CaptureTracks ();
  
  BuildTables (m_start_pos);
  
  for (track_index i = 0; i < m_playback_tracks; ++i)
    m_source->SetReadPos (i, m_start_pos, false);
  m_source->SetReadPos (0, m_start_pos, true);
  
  Debug ();
}

void LoopFilterSource::StartIO_impl ()
{
  m_source->StartIO ();
  
  m_start_pos = m_source->MinStart ();
  m_end_pos = m_source->MaxEnd ();
  
  m_playback_tracks = m_source->PlaybackTracks ();
  m_capture_tracks = m_source->PlaybackTracks ();
  
  BuildTables (m_start_pos);
  
  for (track_index i = 0; i < m_playback_tracks; ++i)
    m_source->SetReadPos (i, m_start_pos, false);
  m_source->SetReadPos (0, m_start_pos, true);
  
  for (track_index i = 0; i < m_capture_tracks; ++i)
    m_source->SetWritePos (i, m_start_pos, false);
  m_source->SetWritePos (0, m_start_pos, true);
}

void LoopFilterSource::StopIO_impl ()
{
  m_source->StopIO ();
}
    
void LoopFilterSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  m_source->SetReadPos (track, n_pos, is_master);
  CurrentReadPosition (track, is_master) = n_pos;
}

void LoopFilterSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  m_source->SetWritePos (track, n_pos, is_master);
  CurrentWritePosition (track, is_master) = n_pos;
}

bool LoopFilterSource::Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
				  bool is_master)
{
  sample_index proc_length = 0;
  sample_index t_length;
  
  bool data = true;
  
  sample_index cur_pos = CurrentReadPosition (track, is_master);
  
  while (proc_length < length && data)
    {
      if (cur_pos < m_end_pos || !m_stop_at_end)
	{
	  if (m_stop_at_end)
	    t_length = limit_max (length - proc_length, m_end_pos - cur_pos);
	  else
	    t_length = length;
	  
	  if (t_length <= 0)
	    {
	      if (!is_master)
		dest->Zero (length - proc_length);
	      data = false;
	    }
	  else
	    {
	      try {
		m_source -> Read (track, t_length, dest, is_master);
	      }
	      catch (BasicExcept e)
		{
		  cout << "EE XX EE PP TT II OO NN !! !! ::" << endl;
		  cout << e.info << endl;
		}
	    }
	  proc_length += t_length;
	  cur_pos += t_length;
	}
      if (cur_pos >= m_end_pos && m_stop_at_end) // wrap to start
	{
	  if (m_looping)
	    {
	      cur_pos = m_start_pos;
	      m_source->SetReadPos (track, cur_pos, is_master);
	    }
	  else
	    data = false;
	}
    }
  
  if (!is_master && proc_length < length)
    dest->Zero (length - proc_length);
  
  CurrentReadPosition (track, is_master) = cur_pos;
  
  return data;
}

bool LoopFilterSource::Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
				   sample_index buffer_index, bool is_master)
{
  sample_index proc_length = 0;
  sample_index t_length;
  
  bool data = true;

  sample_index cur_pos = CurrentWritePosition (track, is_master);
  
  while (proc_length < length && data)
    {
      if (cur_pos < m_end_pos || !m_stop_at_end)
	{
	  if (m_stop_at_end)
	    t_length = limit_max (length - proc_length, m_end_pos - cur_pos);
	  else
	    t_length = length;
	  
	  if (t_length <= 0)
	    {
	      data = false;
	    }
	  else
	    {
	      try {
		m_source -> Write (track, t_length, src, proc_length, is_master);
	      }
	      catch (BasicExcept e)
		{
		  cout << "EE XX EE PP TT II OO NN !! !! ::" << endl;
		  cout << e.info << endl;
		}
	    }
	  proc_length += t_length;
	  cur_pos += t_length;
	}
      
      if (cur_pos >= m_end_pos && m_stop_at_end) // wrap to start
	{
	  if (m_looping)
	    {
	      cur_pos = m_start_pos;
	      m_source->SetWritePos (track, cur_pos, is_master);
	    }
	  else
	    data = false;
	}
    }
  
  CurrentWritePosition (track, is_master) = cur_pos;
  
  return data;
}

void LoopFilterSource::Debug_impl () const
{
  cout << "LoopFilterSource::Debug_impl:" << endl;
  cout << "  looping: " << m_looping << endl;
  cout << "  start_pos: " << m_start_pos << "; end_pos: " << m_end_pos << endl;
  cout << "  SOURCE:" << endl;
  m_source->Debug ();
}

void LoopFilterSource::SlotMinStartChanged (sample_index n_min_start)
{
  m_start_pos = n_min_start;
}

void LoopFilterSource::SlotMaxEndChanged (sample_index n_max_end)
{
  m_end_pos = n_max_end;
}

/***************************
 * MultiplexerFilterSource *
 ***************************/

MultiplexerFilterSource::MultiplexerFilterSource (MixerSource* pri_src, MixerSource* sec_src)
{
  m_pri_source = pri_src;
  m_sec_source = sec_src;
  
  m_cur_source = pri_src;
  
  m_start_pos = m_cur_source->MinStart ();
  m_end_pos = m_cur_source->MaxEnd ();
  
  m_way = false;
}

MultiplexerFilterSource::~MultiplexerFilterSource ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~MultiplexerFilterSource io is running in dtor!?!?!?! ..." << endl;
}

void MultiplexerFilterSource::SetPriSource (MixerSource* pri_src)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  m_pri_source = pri_src;
}

void MultiplexerFilterSource::SetSecSource (MixerSource* sec_src)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO is already running!";
      throw e;
    }
  m_sec_source = sec_src;
}

void MultiplexerFilterSource::SetWay (bool use_sec)
{
  if (m_way != use_sec)
    {
      SigC::Threads::MLock t_lock (m_lock);
      
      if (io_running) {
	for (track_index tr = 0; tr < m_playback_tracks; ++tr)
	  InvaildRead (tr, false);
	InvaildRead (0, true);
	
	if (io_input_running) {
	  for (track_index tr = 0; tr < m_capture_tracks; ++tr)
	    InvaildWrite (tr, false);
	  InvaildWrite (0, true);
	}
      }
      m_way = use_sec;
      
      if (!m_way)
	m_cur_source = m_pri_source;
      else
	m_cur_source = m_sec_source;
      
      m_start_pos = m_cur_source->MinStart ();
      m_end_pos = m_cur_source->MaxEnd ();
      
      SignalMinStartChanged (m_start_pos);
      SignalMaxEndChanged (m_end_pos);
    }
}

bool MultiplexerFilterSource::Way ()
{
  return m_way;
}

sample_index MultiplexerFilterSource::MinStart () const
{
  return m_start_pos;
}

sample_index MultiplexerFilterSource::MaxEnd () const
{
  return m_end_pos;
}

bool MultiplexerFilterSource::IsInputable () const
{
  if (!m_way)
    return m_pri_source->IsInputable ();
  else
    return m_sec_source->IsInputable ();
}

void MultiplexerFilterSource::StartO_impl ()
{
  m_pri_source->StartO ();
  m_sec_source->StartO ();
  
  if (!m_way)
    m_cur_source = m_pri_source;
  else
    m_cur_source = m_sec_source;
  
  m_playback_tracks = limit_max (m_pri_source->PlaybackTracks (), m_sec_source->PlaybackTracks () );
  m_capture_tracks = m_pri_source->CaptureTracks ();
  
  m_start_pos = m_cur_source->MinStart ();
  m_end_pos = m_cur_source->MaxEnd ();
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void MultiplexerFilterSource::StartIO_impl ()
{
  m_pri_source->StartIO ();
  m_sec_source->StartO ();
  
  if (!m_way)
    m_cur_source = m_pri_source;
  else
    m_cur_source = m_sec_source;
  
  m_playback_tracks = limit_max (m_pri_source->PlaybackTracks (), m_sec_source->PlaybackTracks () );
  m_capture_tracks = m_pri_source->CaptureTracks ();
  
  m_start_pos = m_cur_source->MinStart ();
  m_end_pos = m_cur_source->MaxEnd ();
  
  BuildTables (m_start_pos);
  BuildTablesHelper (m_playback_tracks, m_capture_tracks, false);
}

void MultiplexerFilterSource::StopIO_impl ()
{
  m_pri_source->StopIO ();
  m_sec_source->StopIO ();
}

void MultiplexerFilterSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  SigC::Threads::MLock t_lock (m_lock);
  
  InvaildRead (track, is_master);
  CurrentReadPosition (track, is_master) = n_pos;
}

void MultiplexerFilterSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  SigC::Threads::MLock t_lock (m_lock);
  
  InvaildWrite (track, is_master);
  CurrentWritePosition (track, is_master) = n_pos;
}

bool MultiplexerFilterSource::Read_impl (track_index track, sample_index length,
					 SampleBufferDefault* dest, bool is_master)
{
  SigC::Threads::MLock t_lock (m_lock);
  
  sample_index cur_pos = CurrentReadPosition (track, is_master);
  bool data = DoRead (m_cur_source, track, cur_pos, length, dest, is_master);
  CurrentReadPosition (track, is_master) += length;
  return data;
}

bool MultiplexerFilterSource::Write_impl (track_index track, sample_index length,
					  SampleBufferDefault* src, sample_index buffer_index,
					  bool is_master)
{
  SigC::Threads::MLock t_lock (m_lock);
  
  sample_index cur_pos = CurrentWritePosition (track, is_master);
  bool data = DoWrite (m_pri_source, track, cur_pos, length, src, buffer_index, is_master);
  CurrentReadPosition (track, is_master) += length;
  return data;
}

void MultiplexerFilterSource::Debug_impl () const
{
  cout << "MultiplexerFilterSource::Debug_impl:" << endl;
  cout << "  m_way: " << m_way << endl;
}

/****************************
 * The UniversalMixerParams *
 ****************************/

UniversalMixerParams::UniversalMixerParams ()
{
  m_tracks = 0;
  m_channels = 0;
  
  io_running = false;
}

UniversalMixerParams::~UniversalMixerParams ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~UniversalMixerParams: io is running in dtor!?!?!?! ..." << endl;
  // TODO ...
}

void UniversalMixerParams::SetTrackValue (track_index track, pcm_default::value_type n_value)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  
  m_track_values [track] = n_value;
  
  SignalTrackValueChanged (track, n_value);
  for (track_index ch = 0; ch < m_channels; ++ch)
    SignalValueChanged (track, ch, effective_value (track, ch) );
}

void UniversalMixerParams::SetChannelValue (track_index channel, pcm_default::value_type n_value)
{
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  
  m_channel_values [channel] = n_value;
  
  SignalChannelValueChanged (channel, n_value);
  
  for (track_index tr = 0; tr < m_tracks; ++tr)
    SignalValueChanged (tr, channel, effective_value (tr, channel) );
}

void UniversalMixerParams::SetMatrixValue (track_index track, track_index channel,
					   pcm_default::value_type n_value)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  
  m_matrix_values [track] [channel] = n_value;
  
  SignalMatrixValueChanged (track, channel, n_value);
  SignalValueChanged (track, channel, effective_value (track, channel) );
}

pcm_default::value_type UniversalMixerParams::TrackValue (track_index track)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  
  return m_track_values [track];
}

pcm_default::value_type UniversalMixerParams::ChannelValue (track_index channel)
{
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  
  return m_channel_values [channel];
}

pcm_default::value_type UniversalMixerParams::MatrixValue (track_index track, track_index channel)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  
  return m_matrix_values [track] [channel];
}

void UniversalMixerParams::SetTracks (track_index n_tracks)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  
  pcm_default t;
  
  // remove last if too many
  while (m_track_values.size () > n_tracks)
    {
      m_track_values.pop_back ();
      m_matrix_values.pop_back ();
      m_track_mutes.pop_back ();
    }
  // add zeros if too less
  while (m_track_values.size () < n_tracks)
    {
      m_track_values.push_back ( (pcm_default::value_type) {t.dBfs () } );
      m_matrix_values.push_back ();
      for (track_index i = 0; i < m_channels; ++i)
	m_matrix_values.back () . push_back ( (pcm_default::value_type) {0} );
      m_track_mutes.push_back (false);
    }
  
  m_tracks = n_tracks;
  SignalTracksChanged (n_tracks);
}

track_index UniversalMixerParams::Tracks ()
{
  return m_tracks;
}


void UniversalMixerParams::SetChannels (track_index n_channels)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO running - not allowed!";
      throw e;
    }
  
  pcm_default t;
  
  // resize all track matrixs ...
  for (vector <vector<pcm_default::value_type> >::iterator i = m_matrix_values.begin ();
       i != m_matrix_values.end (); ++i)
    {
      while ( i->size () > n_channels)
	i->pop_back ();
      // add zeros if too less
      while ( i->size () < n_channels)
	i->push_back ( (pcm_default::value_type) {0} );
    }
			    
  while (m_channel_values.size () > n_channels)
    {
      m_channel_values.pop_back ();
      m_channel_mutes.pop_back ();
    }
  while (m_channel_values.size () < n_channels)
    {
      m_channel_values.push_back ( (pcm_default::value_type) {t.dBfs () } );
					  m_channel_mutes.push_back (false);
    }	    
  m_channels = n_channels;
  SignalChannelsChanged (n_channels);
}

track_index UniversalMixerParams::Channels ()
{
  return m_channels;
}

void UniversalMixerParams::SetTrackMute (track_index track, bool n_state)
{
  if (track > m_tracks)
  {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  m_track_mutes [track] = n_state;
  SignalTrackMuteChanged (track, n_state);
}

bool UniversalMixerParams::IsTrackMuted (track_index track)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  return m_track_mutes [track];
}

void UniversalMixerParams::SetChannelMute (track_index channel, bool n_state)
{
  if (channel > m_channels)
  {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  m_channel_mutes [channel] = n_state;
  SignalChannelMuteChanged (channel, n_state);
}

bool UniversalMixerParams::IsChannelMuted (track_index channel)
{
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  return m_channel_mutes [channel];
}

pcm_default::value_type UniversalMixerParams::Value (track_index track, track_index channel)
{
  if (track > m_tracks)
    {
      ExceptTrackInvaild e;
      e.info = "Track invaild!";
      throw e;
    }
  if (channel > m_channels)
    {
      ExceptChannelInvaild e;
      e.info = "Channel invaild!";
      throw e;
    }
  
  return effective_value (track, channel);
}

void UniversalMixerParams::StartIO ()
{
  io_running = true;
}

void UniversalMixerParams::StopIO ()
{
  io_running = false;
}

bool UniversalMixerParams::IsIORunning ()
{
  return io_running;
}

void UniversalMixerParams::Debug ()
{
  cout << "UniversalMixerParams::Debug:" << endl;
  
  cout << "  Config: Tracks: " << m_tracks << ", channels: " << m_channels << endl;
  
  // internal states
  cout << "  States: io_running: " << io_running
       << endl;
  
  cout << "  Values:" << endl;
  for (int tr = 0; tr < m_tracks; ++tr)
    {
      cout << "  track [" << tr << "]: ";
      
      cout << " track: " << m_track_values [tr] << "; ";

      for (track_index ch = 0; ch < m_channels; ++ch)
	cout << m_matrix_values [tr] [ch] . data << ", ";
      cout << endl;
    }
  cout << "  Channel values:" << endl;
  for (int ch = 0; ch < m_tracks; ++ch)
    cout << "  ch " << ch << ": " << m_channel_values [ch] << endl;
}

pcm_default::value_type UniversalMixerParams::effective_value (track_index track, track_index channel)
{
  return m_track_values [track] * m_matrix_values [track] [channel] * m_channel_values [channel];
}

UniversalMixerParams& UniversalMixerParams::operator= (const UniversalMixerParams& src)
{
  // io_running is explicitely not copyed.
  // (although an io shouldn't be on on copy, anyway ... (exceptio??) )
    
  m_tracks = src.m_tracks;
  m_channels = src.m_channels;
    
  m_track_values = src.m_track_values;
  m_channel_values = src.m_channel_values;
  m_matrix_values = src.m_matrix_values;
  m_track_mutes = src.m_track_mutes;
  m_channel_mutes = src.m_channel_mutes;
}

/*********************************
 * The Debug Node implementation *
 *********************************/

DebugMixerNode::DebugMixerNode (track_index init_pl_ch, track_index init_cp_ch,
				int init_period_size, int init_delay)
{
  user_playback_channels = init_pl_ch;
  user_capture_channels = init_cp_ch;

  user_period_size = init_period_size;
  user_delay = init_delay;
  
  do_playback = true;
  do_capture = false;
  
  initialized = false;
  io_running = false;
  
  benchmarking = false;
  
  Ready (false); // tell the threadlib that we are ready - but don't exec the loop ...
}

DebugMixerNode::~DebugMixerNode ()
{
  if (io_running)
    cout << "ERRORR: BUG: WARNING!!!!~DebugMixerNode: io is running in dtor!?!?!?! ..." << endl;
  // TODO ...
}

void DebugMixerNode::SetBenchmarking (bool state)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  benchmarking = state;
}

void DebugMixerNode::SetDelay (int new_delay)
{
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  user_delay = new_delay;
}

bool DebugMixerNode::Init ()
{
  initialized = true;
  
  return true;
}

track_index DebugMixerNode::PlaybackChannels ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  
  return user_playback_channels;
}

track_index DebugMixerNode::CaptureChannels ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  
  return user_capture_channels;
}

int DebugMixerNode::PlaybackPeriodSize ()
{
  return user_period_size;
}

int DebugMixerNode::CapturePeriodSize ()
{
  return user_period_size;
}

int DebugMixerNode::PlaybackPeriods ()
{
  return 2;
}

int DebugMixerNode::CapturePeriods ()
{
  return 2;
}

int DebugMixerNode::PlaybackXRuns ()
{
  return 0;
}

int DebugMixerNode::CaptureXRuns ()
{
  return 0;
}

void DebugMixerNode::SetPlayback (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  do_playback = state;
}

void DebugMixerNode::SetCapture (bool state)
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  do_capture = state;
}

bool DebugMixerNode::IsPlayback ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  return do_playback;
}

bool DebugMixerNode::IsCapture ()
{
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  return do_capture;
}

void DebugMixerNode::StartIO (Mixer& source)
{
#ifdef DEBUG_GSML_IO
  cout << "DebugMixerNode::StartIO:" << endl;
#endif
  if (!initialized)
    {
      ExceptNotInitialized e;
      e.info = "Not Initialized!";
      throw e;
    }
  if (io_running)
    {
      ExceptIORunning e;
      e.info = "IO currently running ...";
      throw e;
    }
  
  m_mixer = &source;
  
  // prepare states
  stop_requested = false;
  
  current_playback_period = 0;
  current_capture_period = 0;

  io_playback_periods.clear ();
  for (track_index i = 0; i < user_playback_channels; i++)
    io_playback_periods.push_back (0);
  
  io_capture_periods.clear ();
  for (track_index i = 0; i < user_capture_channels; i++)
    io_capture_periods.push_back (0);
  
  io_running = true;
  
  if (benchmarking)
    gettimeofday (&bench_start, NULL);
  
#ifdef DEBUG_GSML_IO
  cout << "  StartIO: calling _Continue_" << endl;
#endif
  Continue (); // unlock Loop ...
}

void DebugMixerNode::StopIO ()
{
#ifdef DEBUG_GSML_IO
  cout << "DebugMixerNode::StopIO:" << endl;
#endif
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO currently running ...";
      throw e;
    }

  stop_requested = true;
}

bool DebugMixerNode::IsIORunning ()
{
  return io_running;
}

sample_index DebugMixerNode::GetCurrentPos ()
{
  if (!io_running)
    {
      ExceptNoIORunning e;
      e.info = "No IO currently running ...";
      throw e;
    }
  
  if (do_playback)
    return current_playback_period * user_period_size;
  if (do_capture)
    return current_capture_period * user_period_size;
  
  cout << "Warning: Neiter playback nor capture is running!" << endl;
  return 0;
}

void DebugMixerNode::Loop ()
{
  SampleBufferDefault temp (user_period_size);
  bool no_data = false;
  
#ifdef DEBUG_GSML_IO
  cout << "DebugMixerNode::Loop:" << endl;
#endif
  
  if (!stop_requested)
    {
      if (do_playback)
	{
#ifdef DEBUG_GSML_IO
	  cout << "  Loop: do_playback ..." << endl;
#endif
	  for (track_index ch = 0; ch < user_playback_channels; ch++)
	    {
	      try {
		if (!m_mixer->Read (ch, &temp) )
		  no_data = true;
	      }
	      catch (BasicExcept e)
		{
		  cout << "EE XX EE PP TT II OO NN !! !! ::" << endl;
		  cout << e.info << endl;
		}
	      cout << "1st DATA SAMPLE: " << temp.data [0] << endl;
	      
	      io_playback_periods [ch] ++;
	    }
	  current_playback_period++;
	}
      
      if (do_capture)
	{
#ifdef DEBUG_GSML_IO
	  cout << "  Loop: do_capture ..." << endl;
#endif
	  temp.Zero (user_period_size);
	  for (track_index ch = 0; ch < user_capture_channels; ch++)
	    {
	      try {
		if (!m_mixer->Write (ch, &temp) )
		  no_data = true;
	      }
	       catch (BasicExcept e)
		 {
		   cout << "EE XX EE PP TT II OO NN !! !! ::" << endl;
		   cout << e.info << endl;
		 }
	      io_capture_periods [ch] ++;
	    }
	  current_capture_period++;
	}
      if (!no_data && user_delay > 0)
	usleep (user_delay);
    }
  
  // no else - because it might be changed in the io loop-part ...
  if (no_data || stop_requested)
    {
#ifdef DEBUG_GSML_IO
      cout << "  Loop: END OF IO" << endl;
#endif
      io_running = false;
      m_mixer->IOFinishedNotification ();
      
      if (benchmarking)
	{
	  timeval cur_time;
	  
	  long long ctime;
	      
	  gettimeofday (&cur_time, NULL);
	  
	  ctime = (cur_time.tv_sec * 1000000) + cur_time.tv_usec;
	  ctime = ctime - (bench_start.tv_sec * 1000000) - bench_start.tv_usec;
	  
	  sample_index periods;
	  if (do_playback)
	    periods = current_playback_period;
	  else
	    periods = current_capture_period;
	  
	  cout << endl << "  ** BENCHMARK **" << endl
	       << "    Processing " << periods << " periods, took: " << ctime << " usecs!" << endl;
	}
      
      Wait ();
    }
}

void DebugMixerNode::Debug ()
{
  cout << "DebugMixerNode::Debug:" << endl;
  
  cout << "  period_size: " << user_period_size
       << ", delay: " << user_delay << endl;
  
  cout << "  Channels: playback: " << user_capture_channels
       << ", capture: " << user_playback_channels << endl;
  
  cout << "  Initialized: " << initialized
       << ", io_running: " << io_running
       << ", do_playback: " << do_playback
       << ", do_capture: " << do_capture << endl;
  
  if (io_running)
    {
      cout << "  stop requested: " << stop_requested << endl;
      
      cout << "  Current period: playback: " << current_playback_period
	   << ", capture: " << current_capture_period << endl;
      
      cout << "  detailed IO period count:" << endl
	   << "  Playback: " << endl
	   << "    ";
      
      for (track_index ch = 0; ch < user_playback_channels; ch++)
	cout << "[" << ch << "]: " << io_playback_periods [ch] << ", ";
      cout << endl;
	
      cout << "  Capture: " << endl
	   << "    ";
      for (track_index ch = 0; ch < user_capture_channels; ch++)
	cout << "[" << ch << "]: " << io_capture_periods [ch] << ", ";
      cout << endl;
    }
}
