/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   One track data-pattern container.
 */

#include "TrackLayer.hh"
#include <algorithm>

using namespace std;
using namespace GSML;

/*********************
 * TrackLayer::Entry *
 *********************/

TrackLayer::Entry::Entry ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry::Entry ()" << endl;
#endif
  
  length = 0;
  start_in_file = 0;
  track_in_file = 0;
  
  real_data = false;
  myfile = 0;
}

TrackLayer::Entry::Entry (DataFile* init_file)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry::Entry (DataFile *init_file)" << endl;
#endif
  
  length = 0;
  start_in_file = 0;
  track_in_file = 0;
  
  real_data = true;
  myfile = init_file;
  
  myfile->Bind (); // it is used by me, too ...
}

TrackLayer::Entry::Entry (const Entry& source_entry)
{
  real_data = false; // init for copy const. ...
  this->operator= (source_entry);
}

TrackLayer::Entry& TrackLayer::Entry::operator= (const TrackLayer::Entry& source_entry)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry& TrackLayer::Entry::operator= (const TrackLayer::Entry& source_entry)" << endl;
#endif
  if (this != &source_entry)
    {
      if (real_data)
      	myfile->Unbind (); // not use be me anymore ...
      
      length = source_entry.length;
      start_in_file = source_entry.start_in_file;
      track_in_file = source_entry.track_in_file;
      myfile = source_entry.myfile;
      real_data = source_entry.real_data;
      
      if (real_data)
      	myfile->Bind (); // it is used by me, too ...
    }
  return *this;
}

TrackLayer::Entry::~Entry ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry::~Entry: " << endl;
#endif
  
  if (real_data)
    myfile->Unbind ();
}

void TrackLayer::Entry::RegisterFile (DataFile* new_file)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry::Register_File:" << endl;
#endif
  // any file already "registered" and binded??
  if (real_data)
    myfile->Unbind ();
  else
    real_data = true; // in a few instructins we'll have real_data
  
  myfile = new_file;
  myfile->Bind ();
}

GSML::DataFile* TrackLayer::Entry::File ()
{
  if (real_data)
    return myfile;
  else
    return 0;
}

const DataFile* TrackLayer::Entry::File () const
{
  if (real_data)
    return myfile;
  else
    return 0;
}

void TrackLayer::Entry::Debug () const
{
  cout << "Data_Layer::debug:" << endl;
  
  cout << "  Length: " << length << "\t S_In_File: " << start_in_file
       << "\t t_in_file: " << track_in_file << "\t Real_Data: " << real_data
       << "\t MyFile: " << myfile << endl;
  
}

/*************************
 * TrackLayer::Selection *
 *************************/

TrackLayer::Selection::Selection (TrackLayer* i_track)
{
  m_track = i_track;
  active = true;
  
  m_intervals . SignalAdded . connect ( SignalSelected . slot () );
  m_intervals . SignalRemoved . connect ( SignalUnselected . slot () );
}
	
TrackLayer::Selection::Selection (const Selection& source)
{
  operator= (source);
}

bool TrackLayer::Selection::MinStart (sample_index* value)
{
  return m_intervals . MinStart (value);
}

bool TrackLayer::Selection::MaxEnd (sample_index* value)
{
  return m_intervals . MaxEnd (value);
}

sample_index TrackLayer::Selection::SummedLength ()
{
  return m_intervals . SummedLength ();
}

void TrackLayer::Selection::Select (sample_index start, sample_index length)
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  m_intervals . Add (start, length);
}
	
void TrackLayer::Selection::Unselect (sample_index start, sample_index length)
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  m_intervals . Remove (start, length);
}
	
void TrackLayer::Selection::Clear ()
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  m_intervals . Clear ();
}
	
void TrackLayer::Selection::All ()
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  m_intervals . All (m_track -> Length () );
}
	
void TrackLayer::Selection::Invert ()
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  m_intervals . Invert (m_track -> Length () );
}
	
// active flag - indicating if effects should use this selection
void TrackLayer::Selection::SetActive (bool state)
{
  if (m_track -> fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data track is fixated";
      throw e;
    }
  
  active = state;
  SignalActiveChanged (active);
}

bool TrackLayer::Selection::IsActive ()
{
  return active;
}
	
TrackLayer::Selection::const_iterator TrackLayer::Selection::begin ()
{
  return m_intervals . begin ();
}
	
TrackLayer::Selection::const_iterator TrackLayer::Selection::end ()
{
  return m_intervals . end ();
}
	
TrackLayer::Selection& TrackLayer::Selection::operator= (const Selection& source)
{
  if (this != &source)
    {
      m_intervals = source . m_intervals;
      active = source . active;
    }
}
	
void TrackLayer::Selection::Debug ()
{
  cout << "TrackLayer::Selection::Debug:" << endl;
  cout << "  is active: " << active << endl;
  for (const_iterator i = m_intervals . begin (); i != m_intervals . end (); ++ i)
    cout << "  start: " << i -> start << ", end: " << i -> end << endl;
}

/**************
 * TrackLayer *
 **************/

TrackLayer::TrackLayer (DataFileIndex* init_index, PCMDatatype* init_temp_type,
			sample_rate init_temp_rate=44100)
  : selection (this)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::TrackLayer ()" << endl;
#endif

  m_fileindex = init_index;
  m_temp_sampletype = init_temp_type;
  m_temp_samplerate = init_temp_rate;
  
  m_data = new deque<Entry*>;

  write_locked = false;
  fixated = false;
}

TrackLayer::TrackLayer (const TrackLayer& source_track)
  : selection (this)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::TrackLayer (const TrackLayer& source_track)" << endl;
#endif

  m_data = new deque<Entry*>;

  // for the overloaded = operator !!! - we MUST have vaild inits !!!
  write_locked = false;
  fixated = false;
  
  this->operator= (source_track);
}

TrackLayer::~TrackLayer ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::~TrackLayer: " << endl;
#endif
  
  // destroy all entrys ...
  del_helper<Entry*> p;
  for_each (m_data->begin (), m_data->end (), p);
  
  delete m_data;
}

const string& TrackLayer::Name () const
{
  return m_name;
}

void TrackLayer::Name (const string& n_name)
{
  m_name = n_name;
  SignalNameChanged (m_name);
}

bool TrackLayer::ReadData (sample_index start,
			   sample_index length,
			   SampleBuffer* destination)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::ReadData: start: " << start << ", length: " << length << endl;
#endif
  
  // is a write prepared ???
  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  // busy.DoLock ();
      
  // 1st get pos and sample for start ...
  iterator entry = get_entry4index (start);
  sample_index entry_index = get_start4entry (entry);
  
  sample_index read_length = 0;
      
  // is the start in an entry ??
  if (entry_index <= start && entry != m_data->end() )
    {
      sample_index start_in_entry;
      sample_index length_in_entry;
      
      start_in_entry = start - entry_index;
      length_in_entry = (*entry)->length - start_in_entry;
      // limit size to real wanted size ...
      if (length_in_entry > length)
	length_in_entry = length;
	  
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "start IN an entry: entry: " << "itor"
	   << ", start_in_entry: " << start_in_entry 
	   << ", length_in_entry: " << length_in_entry << endl;
#endif

      // real data ??
      if ( (*entry)->File () != 0 )
	{
	  DataFile* t_file;
	      
	  t_file = (*entry)->File ();
	  
	  try
	    {
	      t_file->ReadData ( (*entry)->start_in_file + start_in_entry,
				 length_in_entry,
				 (*entry)->track_in_file,
				 destination);
	    }
	  catch (DataFile::Except e)
	    {
	      // busy.UnLock ();
	      
	      TrackLayer::ExceptNotReadable e;
	      e.info = "Exception during Read_Data ...";
	      throw e;
	    }
	} // end real data
      else // virtual (zero) data ...
	destination->Zero (length_in_entry);
      
      // we continue with the next entry ...
      ++entry;
      read_length = length_in_entry;
    }
  
  // as long as we don't hit the end of data ...
  while (read_length < length && entry != m_data->end () )
    {
      sample_index length_in_entry;
      
      // limit to entry length
      length_in_entry = limit_max (length - read_length, (*entry)->length);
      
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "read: entry: " << "itor"
	   << ", length_in_entry: " << length_in_entry << endl;
#endif
      
      // real data ??
      if ( (*entry)->File () != 0)
	{
	  DataFile* t_file;
	  
	  t_file = (*entry)->File ();
	  t_file->ReadData ( (*entry)->start_in_file,
			     length_in_entry,
			     (*entry)->track_in_file,
			     destination);
	} // end real data
      else // virtual (zero) data ...
	destination->Zero (length_in_entry);
      
      read_length += length_in_entry;
      ++entry;
    }
  if (read_length < length)
    {
      // cout << "--WARNING: not enough data to read  ..." << endl;
    }
  if (read_length > length)
    {
      cout << "BUG: !!!!!!!!!! READ LENGTH BIGGER ?????????????" << endl;
      cout << "     is: " << read_length << ", should be: " << length << endl;
    }
  
  
  // busy.UnLock ();
  return read_length == length;
}

void TrackLayer::WriteData (sample_index length, SBHandle source, sample_index buffer_index)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::WriteData: " << endl;
#endif
  // is a write prepared ???
  if (!write_locked)
    {
      TrackLayer::ExceptNotWriteable e;
      e.info = "Not prepared for write ... ";
      throw e;
    }
  
  // write to DataFile
  write_file->WriteData (length, 0, source, buffer_index);
  write_length += length;
  
  // emit Signal
  SignalWroteData (length, source, buffer_index);
}

// returns track length
GSML::sample_index TrackLayer::Length () const
{
  // is a write prepared ???
  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  return length ();
}

void TrackLayer::PrepareOverWrite (sample_index start)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::PrepareOverWrite: " << endl;
#endif
  
  prepare_write (start);
  overwrite = true;
  
  // emit Signal
  SignalPreparedOverWrite (start);
}

void TrackLayer::PrepareInsertWrite (sample_index start)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::PrepareInsertWrite: " << endl;
#endif
  
  prepare_write (start);
  overwrite = false;
  
  // emit Signal
  SignalPreparedInsertWrite (start);
}

void TrackLayer::FixateWrite ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::FixateWrite: " << endl;
#endif
  // is a write prepared ???
  if (!write_locked)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Not prepared for write ... ";
      throw e;
    }
  
  // if something was written into the Data_file insert an entry
  if (write_length > 0)
    {
      if (overwrite)
	remove_region (write_start, write_length);
      
      // fixate data_file
      write_file->FixateWrite ();
      
      // build entry
      Entry new_entry;
      
      new_entry.start_in_file = 0;
      new_entry.track_in_file = 0;
      new_entry.length = write_length;
      new_entry.RegisterFile (write_file);
      
      insert_entry (write_start, new_entry);
    }
  else
    {
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "special case!! - no data written into the file ... ???" << endl;
#endif
    }  
  
  // our special case bind ...
  // WE ARE NOT ALLOWED TO DELETE THE DATA_FILE!!! - IT IS IN THE FILE_INDEX !!!
  // so only unbind ... - now the file_index can delete it itselves
  write_file->Unbind ();
  
  write_file = 0;
  write_locked = false;
  
  // emit Signal
  SignalFixatedWrite ();
}

void TrackLayer::CancelWrite ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::CancelWrite: " << endl;
#endif
  // is a write prepared ???
  if (!write_locked)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Not prepared for write ... ";
      throw e;
    }
  
  // WE ARE NOT ALLOWED TO DELETE THE DATA_FILE!!! - IT IS IN THE FILE_INDEX !!!
  // so only unbind ... - now the file_index can delete it itselves
  write_file->Unbind ();
  write_file = 0;
  
  write_locked = false;
  
  // emit Signal
  SignalCanceledWrite ();
}

void TrackLayer::Fixate ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Fixate: " << endl;
#endif
  if (fixated)
    {
      TrackLayer::ExceptFixated e;
      e.info = "Already fixated ... ";
      throw e;
    }
  if (write_locked)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Currently prepared for write ... ";
      throw e;
    }
  
  fixated = true;
}

void TrackLayer::UnFixate ()
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::UnFixate: " << endl;
#endif
  if (!fixated)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Not fixated ... ";
      throw e;
    }
  
  if (write_locked)
    cout << "BUG: TrackLayer::UnFixate: write prepared during UnFicate ?!?" << endl;
  
  fixated = false;
}

TrackLayer::const_iterator TrackLayer::Begin () const
{
  return m_data->begin ();
}

TrackLayer::const_iterator TrackLayer::End () const
{
  return m_data->end ();
}

TrackLayer::const_iterator TrackLayer::InsertRawEntry (const_iterator i, const Entry& n_entry)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::InsertRawEntry (const_iterator):" << endl;
#endif
  
  iterator t_i = m_data->begin ();
  advance (t_i, distance ( (const_iterator)m_data->begin (), i) );
  Entry* t_entry = new Entry (n_entry);
  
  iterator t_return = m_data->insert (t_i, t_entry);
  
  // emit Signal
  sample_index t_start = get_start4entry (t_return);
  SignalRegionChanged (t_start, length () - t_start);
  
  return t_return;
}

void TrackLayer::InsertRawEntry (sample_index pos, const Entry& n_entry)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::InsertRawEntry (sample_index):" << endl;
#endif
  if (fixated)
    {
      TrackLayer::ExceptFixated e;
      e.info = "Already fixated ... ";
      throw e;
    }
  
  if (write_locked)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  insert_entry (pos, n_entry);
  
  // emit Signal
  SignalRegionChanged (pos, length () - pos);
}

TrackLayer* TrackLayer::CreateTrackLayer (sample_index start,
					  sample_index length) const
{
  TrackLayer* temp_track;
  sample_index temp_length;

#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::CreateTrackLayer: " << endl;
#endif

  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  // init temp_track by copying it
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "  copy the whole layer" << endl;
#endif

  temp_track = new TrackLayer (*this);

#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "  cut the ends" << endl;
#endif
  // cut the ends ;-)
  temp_track->remove_region (0, start);
  temp_length = temp_track->length  ();
  
  cout << "TEMP_LEMNGTH: " << temp_length << endl;
  temp_track->remove_region (length, temp_length - length);
  
  return temp_track;
}

void TrackLayer::InsertTrackLayer (sample_index start, const TrackLayer& source_track)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::InsertTrackLayer: " << endl;
#endif

  if (fixated)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  
  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  iterator src_entry = source_track.m_data->begin ();
  
  // something to copy ??
  if (src_entry != source_track.m_data->end () )
    {
      // let the internal method do the spliting tricks ...
      iterator cur_entry = insert_entry (start, **src_entry);
      
      ++cur_entry; // the next
      ++src_entry;
      
      while (src_entry != source_track.m_data->end () )
	{
	  Entry* t_entry = new Entry (**src_entry);
	  cur_entry = m_data->insert (cur_entry, t_entry);
	  ++cur_entry;
	  ++src_entry;
	}
      
      // emit Signal
      SignalRegionChanged (start, length () - start);
    }
  else
    {
      cout << "DEBUG: nothing to do ... - empty source!" << endl;
    }
}

void TrackLayer::RemoveRegion (sample_index start, sample_index length)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Remove_Region start_index: " << start 
       << ", length: " << length << endl;
#endif
  
  if (fixated)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  
  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  remove_region (start, length);
  // emit signal
  SignalRegionRemoved (start, length);
}

void TrackLayer::remove_region (sample_index start, sample_index length)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::remove_region: " << start << " + " << length << endl;
#endif
  iterator cur_entry = get_entry4index (start);
  
  // start in current vaild data ??
  if (cur_entry != m_data->end () )
    {
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "OK - in vaild data" << endl;
#endif
      sample_index removed_length = 0;
      
      sample_index entry_index = get_start4entry (cur_entry);
      // is the start in the entry ??
      if (entry_index < start)
	{
#ifdef DEBUG_DATA_LAYER
	  cout << "start IN an entry. I'll split it ..." << endl;
#endif
	  cur_entry = split_entry (cur_entry, start -  entry_index);
	  entry_index = start;
	}
      
      // as long as we don't hit the end of data ...
      while (removed_length < length && cur_entry != m_data->end () )
	{
	  // is the end IN the current entry ??
	  if ( (*cur_entry)->length > length - removed_length)
	    {
#ifdef DEBUG_GSML_TRACK_LAYER
	      cout << "the end is IN this entry. I'll split it ..." << endl;
#endif
	      cur_entry = split_entry (cur_entry, length - removed_length);
	      --cur_entry;
	      
	    }
	  // ok - we can delete it in all cases now ...
	  
#ifdef DEBUG_GSML_TRACK_LAYER
	  cout << "removing: " << cur_entry - m_data->begin () << ", length: " << (*cur_entry)->length << endl;
#endif
	  
	  removed_length += (*cur_entry)->length;
	  delete (*cur_entry);
	  cur_entry = m_data->erase (cur_entry);
	}
      if (removed_length < length)
	cout << "--WARNING: not enough data to delete  ..." << endl;
    }
  else
    cout << "--WARNING: start for Remove_Region behind vaild data !!! ..." << endl;
}

DataFileIndex* TrackLayer::GetDataFileIndex ()
{
  return m_fileindex;
}

PCMDatatype* TrackLayer::TempSampletype ()
{
  return m_temp_sampletype;
}

sample_rate TrackLayer::TempSamplerate ()
{
  return m_temp_samplerate;
}

void TrackLayer::SetTempSampletype (PCMDatatype* new_type)
{
  if (fixated)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  m_temp_sampletype = new_type;
}

void TrackLayer::SetTempSamplerate (sample_rate new_rate)
{
  if (fixated)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  m_temp_samplerate = new_rate;
}

TrackLayer& TrackLayer::operator=  (const TrackLayer& source_track)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::Entry& TrackLayer::Entry::operator=  (const Entry&)" << endl;
#endif

  if (fixated)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  
  if (write_locked)
    {
      TrackLayer::ExceptNotFixated e;
      e.info = "Prepared for write ... ";
      throw e;
    }
  
  if (this != &source_track)
    {
      m_fileindex = source_track.m_fileindex;
      m_temp_sampletype = source_track.m_temp_sampletype;
      m_temp_samplerate = source_track.m_temp_samplerate;
      m_name = source_track.m_name;
      
      // clean all
      // if statement for the call from copy-constructor ...
	      
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "cleaning old dynamic array:" << endl;
#endif
      // destroy all entrys ...
      del_helper<Entry*> p;
      for_each (m_data->begin (), m_data->end (), p);
      m_data->erase (m_data->begin (), m_data->end () );

#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "copy all entries:" << endl;
#endif
      const_iterator src_entry;
      for (src_entry = source_track.m_data->begin(); src_entry < source_track.m_data->end (); src_entry++)
	{
	  Entry* t_entry;

#ifdef DEBUG_GSML_TRACK_LAYER
	  cout << "x" << ": ";
#endif
	  t_entry = new Entry ( **src_entry );
	  
	  m_data->push_back (t_entry);
	}
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "done ..." << endl;
#endif
      
      selection = source_track.selection;
      
      // emit Signal
      SignalRegionChanged (0, source_track.length () );
	      
    } // end if (this != source_track)
}

void TrackLayer::Debug () const
{
  cout << "TrackLayer::debug:" << endl;
  
  cout << "  Name:         " << m_name << endl;
  cout << "  Entries:      " << m_data->size () << endl;
  cout << "  Fixated:      " << fixated << endl;
  cout << "  Write_Locked: " << write_locked << endl;
  cout << "  Write_Start:  " << write_start << endl;
  cout << "  Write_Length: " << write_length << endl;
  cout << "  Write_File:   " << write_file << endl;  
  
  cout << "  config:" << endl;
  cout << "    m_fileindex:       " << m_fileindex << endl;
  cout << "    m_temp_sampletype: " << m_temp_sampletype << endl;
  cout << "    m_temp_samplerate: " << m_temp_samplerate << endl;
  
  iterator cur_entry = m_data->begin ();
  for (; cur_entry != m_data->end (); ++cur_entry)
    {
      cout << "Entry " << cur_entry - m_data->begin () << ": ->Debug:" << endl;
      (*cur_entry)->Debug ();
    }
  
  cout << "End ..." << endl;
}

void TrackLayer::prepare_write (sample_index start)
{
  // is a write prepared ???
  if (fixated)
    {
      TrackLayer::ExceptFixated e;
      e.info = "Already fixated ... ";
      throw e;
    }
  if (write_locked)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Already prepared for write ... ";
      throw e;
    }
  
  // configure the temp-write-file
  
  write_file = new DataFileRAW (string (""),
				m_temp_sampletype,
				1,
				m_temp_samplerate, 0,
				true, true); // writeable and autodelete

  // register the tmep file (will generate the filename ...)
  m_fileindex->RegisterTempFile (write_file);

  // special case manual bind ... - cause we don't have a used entry
  // and it should not be destroyed during write!!! ... bind file to _me_
  write_file->Bind ();

  // set some stuff
  write_start = start;
  write_length = 0;
  write_locked = true;
}

// returns track length
GSML::sample_index TrackLayer::length () const
{
  sample_index length;

#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::length: " << endl;
#endif
  
  length = 0;
  
  iterator i;
  for (i = m_data->begin (); i != m_data->end (); i++)
    length += (*i)->length;
  
  return length;
};


TrackLayer::iterator TrackLayer::insert_entry (sample_index start, const Entry& n_entry)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::insert_entry: start_index: " << start << endl;
#endif
  
  TrackLayer::Entry* t_new_entry;
  
  // already fixated ?? (don't care about write locking ... - the internal should know what it does
  if (fixated)
    {   
      TrackLayer::ExceptLogicalError e;
      e.info = "Already fixated ... ";
      throw e;
    }
  
  iterator cur_entry = get_entry4index (start);
  
  // start in current vaild data entrys ...
  if (cur_entry != m_data->end () )
    {
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "  OK! - In the current vaild data ..." << endl;
#endif
      
      sample_index entry_index = get_start4entry (cur_entry);
      // NOT exactly between two entrys??
      if (entry_index < start)
	{
#ifdef DEBUG_GSML_TRACK_LAYER
	  cout << "  NOT on an entry boundary - I'll split it ..." << endl;
#endif
	  cur_entry = split_entry (cur_entry, start - entry_index);
	}
      t_new_entry = new TrackLayer::Entry;
      
      //copy entry
      *t_new_entry = n_entry;
      
      cur_entry = m_data->insert (cur_entry, t_new_entry);
    } // cur_entry != data->end ()
  // start behind current vaild data entrys ...
  else
    {
      sample_index cur_length = length (); 
      
      // gab between current end and start ??
      if (cur_length < start)
	{
	  t_new_entry = new TrackLayer::Entry;
	  
	  t_new_entry->length = start - cur_length;
	  
	  cur_entry = m_data->insert (cur_entry, t_new_entry);
	  ++cur_entry;
	}
      // OK create new entry here
      t_new_entry = new TrackLayer::Entry;
      
      //copy enntry
      *t_new_entry = n_entry;
      
      cout << "data->insert" << endl;
      cur_entry = m_data->insert (cur_entry, t_new_entry);
    } // end else ! cur_entry != m_data->end ()
  
  return cur_entry;
}

// used by internal routines to get the entrys real linear start
sample_index TrackLayer::get_start4entry (const iterator& entry) const
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::get_start4entry: " << endl;
#endif
  
  sample_index current_start = 0;
  iterator cur_entry = m_data->begin ();
  
  while (cur_entry != entry)
    {
      current_start += (*cur_entry)->length;
      cur_entry++;
    }
  
  return current_start;
}

// used by internal routines to get the entry and it's real linear-index
TrackLayer::iterator TrackLayer::get_entry4index (sample_index index) const
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::get_entry4index: index: " << index << endl;
#endif

 // find entry  
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "  while loop to find entry ..." << endl;
#endif

  iterator cur_entry = m_data->begin ();
  sample_index last_index = 0;
  
  while (cur_entry != m_data->end () && index >= last_index + (*cur_entry)->length)
    {
#ifdef DEBUG_GSML_TRACK_LAYER
      cout << "  tested: " << "xx"
	   << "; last_index " << last_index 
	   << "; this length: " << (*cur_entry)->length << endl;
#endif

      last_index += (*cur_entry)->length;
      cur_entry++;
    }
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "  while loop ended: n=" << "xx" << ", l_i: " << last_index << endl;
#endif

  return cur_entry;
}

// no fiaxte and write_locked check - cause it is private anyway ...
TrackLayer::iterator TrackLayer::split_entry (const iterator& split_entry, sample_index index_in_entry)
{
#ifdef DEBUG_GSML_TRACK_LAYER
  cout << "TrackLayer::split_entry: split_entry:" << "xx"
       << ", start_index: " << index_in_entry << endl;
#endif
  
  // test lenght of entry
  if ( (*split_entry)->length < index_in_entry)
    {
      TrackLayer::ExceptLogicalError e;
      e.info = "Entry too short - index_in_entry invaild ...";
      throw e;
    }
  
  Entry* new_entry;
  Entry* old_entry;
  
  iterator t_pos;
  
  new_entry = new Entry;
  
  old_entry = *split_entry;
  
  // copy entry
  *new_entry = *old_entry;
  
  // insert new entry
  t_pos = m_data->insert (split_entry + 1, new_entry);
  
  // edit values
  new_entry->start_in_file += index_in_entry;
  new_entry->length = old_entry->length - index_in_entry;
  
  old_entry->length = index_in_entry;
  
  return t_pos;
}
