/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   WavePlot architecture classes, interfaces and MinMaxAvg implementaion.
 *   This stuff stream and calculate the data for graphic views in the
 *   background.
 */

#include "WavePlot.hh"

using namespace std;
using namespace GSML;

/******************
 * WavePlotConfig *
 ******************/

WavePlotConfig::WavePlotConfig ()
{
  m_start = 0;
  m_plot_length = 1;
  m_sps = 1;
  m_view_start = 0;
  m_view_length = 1;
}

void WavePlotConfig::SetStart (sample_index n_start)
{
  m_start = n_start;
}

void WavePlotConfig::SetPlotLength (sample_index n_plot_length)
{
  m_plot_length = n_plot_length;
}

void WavePlotConfig::SetSamplesPerSample (sample_index n_sps)
{
  m_sps = n_sps;
}

void WavePlotConfig::SetViewStart (sample_index n_view_start)
{
  m_view_start = n_view_start;
}

void WavePlotConfig::SetViewLength (sample_index n_view_length)
{
  m_view_length = n_view_length;
}

sample_index WavePlotConfig::Start () const
{
  return m_start;
}

sample_index WavePlotConfig::Length () const
{
  return m_plot_length * m_sps;
}

sample_index WavePlotConfig::PlotLength () const
{
  return m_plot_length;
}

sample_index WavePlotConfig::SamplesPerSample () const
{
  return m_sps;
}

sample_index WavePlotConfig::ViewStart () const
{
  return m_view_start;
}

sample_index WavePlotConfig::ViewLength () const
{
  return m_view_length;
}

bool WavePlotConfig::operator == (const WavePlotConfig& src) const
{
  return (m_start == src.m_start && m_plot_length == src.m_plot_length &&
	  m_sps ==  src.m_sps  && m_view_start == src.m_view_start &&
	  m_view_length ==  src.m_view_length);
}

void WavePlotConfig::Debug () const
{
  cout << "WavePlotConfig::Debug:" << endl
       << "  Start:            " << m_start << endl
       << "  Length:           " << Length () << endl
       << "  PlotLength:       " << m_plot_length << endl
       << "  SamplesPerSample: " << m_sps << endl
       << "  ViewStart:        " << m_view_start << endl
       << "  ViewLength:       " << m_view_length << endl;
}


/********************
 * WavePlotProgress *
 ********************/

WavePlotProgress::WavePlotProgress ()
{
  m_length = 0;
}

sample_index WavePlotProgress::Length () const
{
  return m_length;
}

void WavePlotProgress::SetLength (sample_index n_length)
{
  set_length_impl (n_length);
}

bool WavePlotProgress::IsVaild (sample_index pos, sample_index length) const
{
  if (pos + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::IsVaild: wrong size";
    }
  for (sample_index i = pos; i < pos + length; ++ i)
    if (m_progress [i] != VAILD)
      return false;
  
  return true;
}

bool  WavePlotProgress::IsAllVaild () const
{
  for (sample_index i = 0; i < m_length; ++ i)
    if (m_progress [i] != VAILD)
      return false;
  
  return true;
}

sample_index WavePlotProgress::GetInvaildStart () const
{
  for (sample_index i = 0; i < m_length; ++ i)
    if (m_progress [i] != VAILD)
      return i;
  return m_length;
}

sample_index WavePlotProgress::GetInvaildEnd () const
{
  for (sample_index i = 1; i < m_length; ++ i)
    if (m_progress [i - 1] != VAILD && m_progress [i] == VAILD)
      return i;
  return m_length;
}

void WavePlotProgress::Vaild (sample_index start, sample_index length)
{
  // wrong params ??
  if (start + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::Vaild: wrong size";
    }
  
  for (sample_index i = start; i < start + length; ++ i)
    {
      m_progress [i] = VAILD;
    }
}

void WavePlotProgress::VaildAll ()
{
  for (sample_index i = 0; i < m_length; ++ i)
    {
      m_progress [i] = VAILD;
    }
}

void WavePlotProgress::Invaild (sample_index start, sample_index length)
{
  // wrong params ??
  if (start + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::Invaild: wrong size";
    }
  
  for (sample_index i = start; i < start + length; ++ i)
    {
      m_progress [i] = INVAILD;
    }
}

void WavePlotProgress::InvaildAll ()
{
  for (sample_index i = 0; i < m_length; ++ i)
    {
      m_progress [i] = INVAILD;
    }
}

void WavePlotProgress::Interpolate (sample_index start, sample_index length)
{
  // wrong params ??
  if (start + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::Interpolate: wrong size";
    }
  
  for (sample_index i = start; i < start + length; ++ i)
    {
      m_progress [i] = INTERPOLATED;
    }
}

void WavePlotProgress::Move (sample_index start, sample_index length, sample_index new_start)
{
  sample_index my_length = Length ();
  
  if (start + length > m_length || new_start + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::Move: wrong size";
    }
  
  move_impl (start, length, new_start);
}

void WavePlotProgress::Zero (sample_index start, sample_index length)
{
  if (start + length > m_length)
    {
      BasicExcept e;
      e.info = "WavePlotProgress::Zero: wrong size";
    }
  
  zero_impl (start, length);
}

void WavePlotProgress::Debug (bool verbose = false) const
{
  debug_impl (verbose);
}

void WavePlotProgress::set_length_impl (sample_index n_length)
{
  m_progress.resize (n_length);
  
  // if new_plot_length is greater -> fill with INVAILD:
  if (n_length > m_length)
    for (sample_index i = m_length; i < n_length; ++ i)
      m_progress [i] = INVAILD;
  
  m_length = n_length;
}

void WavePlotProgress::move_impl (sample_index start, sample_index length, sample_index new_start)
{
  vector<progress_state> t_progress;
  t_progress.resize (length);
  
  // copy
  for (sample_index i = 0; i < length; ++ i)
    t_progress [i] = m_progress [start + i];
  
  // back - copy
  for (sample_index i = 0; i < length; ++ i)
    m_progress [new_start + i] = t_progress [i];
  
  // invaild
  sample_index t_start = limit_min (start, new_start + length);
  sample_index t_end = start + length;
  
  for (sample_index i = t_start; i < t_end; ++ i)
    m_progress [i] = INVAILD;
}

void WavePlotProgress::zero_impl (sample_index start, sample_index length)
{
  // nothing to do
}

void WavePlotProgress::debug_impl (bool verbose) const
{
  cout << "WavePlotProgress::Debug:" << endl;
  cout << "  Length: " << m_length << endl;
}

/*******************
 * WavePlotManager *
 *******************/

#define LAST_LOOP_STAGE 2

WavePlotManager::WavePlotManager (Data* i_data)
{
  m_data = i_data;
  
  // config
  m_stream_chunk_size = 1024 * 4; // lower and higher values results in slower execution on my Athlon ... (more L1 & L2 cache misses ...)
  
  prepared = false;
  halt_level = 0;
  
  /// 2nd thread
  m_first_loop = true;
  m_loop_stage = 0;
  stop = false;
  
  // prepare layers
  LayerStuff stuff;
  stuff.layer = m_data -> CurrentLayerPtr ();
  layers.push_back (stuff);
  current_layer = layers.begin ();
  
  // connect signals
  m_data -> SignalCurrentLayerChanged.connect (slot (this, &WavePlotManager::SlotCurrentLayerChanged) );
  m_data -> SignalNewLayerPrepared.connect (slot (this, &WavePlotManager::SlotNewLayerPrepared) );
  m_data -> SignalNewLayerFixated.connect (slot (this, &WavePlotManager::SlotNewLayerFixated) );
  m_data -> SignalNewLayerCanceled.connect (slot (this, &WavePlotManager::SlotNewLayerCanceled) );
  
  Ready (false);
  usleep (250);
}

WavePlotManager::~WavePlotManager ()
{
  cout << "WavePlotManager::~WavePlotManage:" << endl;
  
  // first stop the other thread
  InternalWait ();
  
  iterator_l l;
  
  // find new_layer in our deque
  for (l = layers.begin (); l != layers.end (); ++ l)
    {
      iterator_p p = l -> plots.begin ();
      
      for (; p != l -> plots.end (); ++ p)
	{
	  cout << "BUG: WavePlot in Manager during dtor?!?: "
	       <<  p -> plot << " will be deleted ..." << endl;
	  del_helper2 (p -> plot);
	}
    }
}

Data* WavePlotManager::GetData ()
{
  return m_data;
}

DataLayer* WavePlotManager::CurrentLayerPtr ()
{
  return current_layer -> layer;
}

void WavePlotManager::SetCacheLayers (unsigned int new_max_layers)
{
  max_layers = new_max_layers;
  DeleteIfTooManyLayer ();
}

unsigned int WavePlotManager::CacheLayers ()
{
  return max_layers;
}

DataLayer* WavePlotManager::RegisterWavePlot (WavePlot* new_plot)
{
  //#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "RegisterWavePlot: new_plot: " << new_plot << endl;
  //#endif
  
  InternalWait ();

  iterator_l l = layers.begin ();
  
  for (; l != layers.end (); ++l)
    {
      PlotStuff stuff;
      stuff.type = typeid (new_plot).name ();
      
      if (l == current_layer) {
	stuff.plot = new_plot;
	stuff.plot->SignalConfigChanged.connect (slot (this, &WavePlotManager::SlotConfigChanged) );
      }
      else
	stuff.plot = 0;
      
      l->plots.push_back (stuff);
    }
  if (prepared)
    {
      PlotStuff stuff;
      stuff.type = typeid (new_plot) . name ();
      
      stuff.plot = new_plot -> Clone (prepared_layer.layer, false);
      stuff.plot -> SignalConfigChanged.connect (slot (this, &WavePlotManager::SlotConfigChanged) );
      
      prepared_layer.plots.push_back (stuff);
    }
  
  // m_loop_stage = 0; not needed - will emit an configure event anyway
  InternalContinue ();
  
  return current_layer -> layer;
}

bool WavePlotManager::UnRegisterWavePlot (WavePlot* old_plot)
{
  //#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "UnRegisterWavePlot:" << old_plot << endl;
  //#endif
  
  // search in current layer
  iterator_p p = current_layer -> plots.begin ();
  for (; p != current_layer -> plots.end (); ++ p)
    if (p->plot == old_plot)
      break;
  
  if (p == current_layer -> plots.end () )
    {
      cout << "UnRegisterWavePlot: plot not found (in current layer) ..." << endl;
      return false;
    }
  
  InternalWait ();
  
  // delete in all layers
  int delta = p - current_layer -> plots.begin ();
  
  for (iterator_l l = layers.begin (); l != layers.end (); ++ l)
    {
      p = l -> plots.begin () + delta;
      if (p -> plot != 0) {
	delete p -> plot;
      }
      l -> plots.erase (p);
    }
  if (prepared)
    {
      p = prepared_layer.plots.begin () + delta;
      if (p -> plot != 0) {
	delete p -> plot;
      }
      prepared_layer.plots.erase (p);
    }
  
  InternalContinue ();
  
  return true;
}

void WavePlotManager::WaitRequest ()
{
  cout << "WavePlotManager::WaitRequest" << endl;
  InternalWait ();
  ++ halt_level;

}

void WavePlotManager::ContinueRequest ()
{
  cout << "WavePlotManager::ContinueRequest" << endl;
  if (halt_level > 0)
    -- halt_level;
  
  InternalContinue ();
}

void WavePlotManager::Debug (bool verbose=false)
{
  cout << "WavePlotManager::Debug:" << endl;
  
  iterator_l l = layers.begin ();
  for (; l != layers.end (); ++ l)
    {
      cout << "Cache Layer: " << l - layers.begin ();
      if (l == current_layer)
	cout << " <-- !current!";
      cout << endl;
      cout << "  data_layer: " << l -> layer << endl;
      
      iterator_p p = l -> plots.begin ();
      for (; p != l -> plots.end (); ++ p)
	 {
	   cout << "    " << p - l -> plots.begin ()
		<< ": plot: " << p -> plot << ", type: " << p -> type
		<< endl;
	   if (verbose && p -> plot != 0)
	     p -> plot -> Debug ();
	 }
       
    }
  if (prepared)
    {
      iterator_p p = prepared_layer.plots.begin ();
      for (; p != prepared_layer.plots.end (); ++ p)
	 {
	   cout << "    " << p - l -> plots.begin ()
		<< ": plot: " << p -> plot << ", type: " << p -> type
		<< endl;
	   if (verbose && p -> plot != 0)
	     p -> plot -> Debug ();
	 }
    }
  else
    cout << "No prepared layer ..." << endl;
}

/*
 * private stuff
 */

void WavePlotManager::SlotCurrentLayerChanged (DataLayer* new_layer)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "SlotCurrentLayerChanged:" << endl;
#endif
  
  iterator_l l;
  
  // find new_layer in our deque
  for (l = layers.begin (); l != layers.end (); ++ l)
    if (l -> layer == new_layer)
      break;
  
  if (l == current_layer)
    {
      cout << "WavePlotManager: new_layer == current_layer !!! -> nothing to do" << endl;
      return;
    }
  
  InternalWait ();
  
  // do we have the new_layer in the cache??
  if (l == layers.end () )
    {
      cout << "layer not cached -> cloning" << endl;
      
      LayerStuff stuff;
      CloneLayer (&stuff, new_layer, false);
      layers.push_front (stuff);
    }
  
  // Debug (false);

  iterator_p p = current_layer -> plots.begin ();
  iterator_p p2 = l -> plots.begin ();
  
  // the new layer l MUST have exacly the same plot count and order ...
  for (; p != current_layer -> plots.end (); ++ p, ++ p2)
    {
      // do we have to create the plot??
      if (p2 -> plot == 0)
	{
	  cout << "plot did not exist: it will be created and cloned now ..." << endl;
	  p2 -> plot = p -> plot -> Clone (new_layer, false);
	  p2 -> plot -> SignalConfigChanged.connect (slot (this, &WavePlotManager::SlotConfigChanged) );
	}
      
      // check if config is up to date
      if (! (p -> plot -> GetConfig () == p2 -> plot -> GetConfig () ) )
	{
	  cout << "config has changed -> updateing ..." << endl;
	  p2 -> plot -> Config (p -> plot -> GetConfig () );
	}
      
      p -> plot -> UnsetCurrent (p2 -> plot);
      p2 -> plot -> SetCurrent ();
    }
  
  current_layer = l;
  
  m_loop_stage = 0;
  InternalContinue ();
}

void WavePlotManager::SlotNewLayerPrepared (DataLayer* new_layer)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "SlotNewLayerPrepared:" << endl;
#endif
  // bug check ;-) - if we are already connected...
  if (prepared) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlotManager::SlotNewLayerPrepared: I'm already prepared";
    }
  
  InternalWait ();
  
  prepared_layer.plots.clear ();
  CloneLayer (&prepared_layer, new_layer, true);
  
  m_loop_stage = 0;
  InternalContinue ();
  
  prepared = true;
}

void WavePlotManager::SlotNewLayerFixated ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "SlotNewLayerFixated:" << endl;
#endif
  // bug check ;-) - if we are prepared...
  if (!prepared) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlotManager::SlotNewLayerFixated: I'm not prepared";
    }
  
  InternalWait ();
  
  iterator_p p = current_layer -> plots.begin ();
  iterator_p p2 = prepared_layer.plots.begin ();
  
  // the new layer MUST have exactly the same plot count and order!
  for (; p != current_layer -> plots.end (); ++ p, ++ p2)
    {
      // check if config is up to date
      if (! (p -> plot -> GetConfig () == p2 -> plot -> GetConfig () ) )
	{
	  cout << "config has changed -> updateing ..." << endl;
	  p2 -> plot -> Config (p -> plot -> GetConfig () );
	}
      
      p -> plot -> UnsetCurrent (p2 -> plot);
      p2 -> plot -> SetCurrent ();
    }
  
  layers.push_back (prepared_layer);
  current_layer = layers.end () - 1;
  
  prepared_layer.plots.clear ();
  
  m_loop_stage = 0;
  InternalContinue ();
  
  prepared = false;
}

void WavePlotManager::SlotNewLayerCanceled ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "SlotNewLayerCanceled:" << endl;
#endif
  // bug check ;-) - if we are prepared...
  if (!prepared) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlotManager::SlotNewLayerCanceled: I'm not prepared";
    }
  
  iterator_p p = prepared_layer.plots.begin ();
  
  // the new layer MUST have exactly the same plot count and order!
  for (; p != prepared_layer.plots.end (); ++ p)
    del_helper2 (p -> plot);
  
  prepared_layer.plots.clear ();
  prepared = false;
}

void WavePlotManager::SlotConfigChanged (const WavePlotConfig& n_config, bool real_change)
{
#ifdef DEBUG_GSML_WAVE_PLOT_MANAGER
  cout << "WavePlotManager::SlotConfigChanged:" << endl;
#endif
  
  // pervormance optimized way to tell the Loop to restart ...
  if (halt_level == 0) {
    stop = true;
    m_loop_stage = 0;
    Continue ();
  }
}

// clone all WavePlots from current layer
void WavePlotManager::CloneLayer (LayerStuff* new_stuff, DataLayer* new_layer, bool is_prepared)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "CloneLayer:" << endl;
#endif
  
  new_stuff -> layer = new_layer;
  new_stuff -> plots.clear ();
  
  iterator_p p = current_layer -> plots.begin ();
  
  for (; p != current_layer -> plots.end (); ++ p)
    {
      PlotStuff stuff;
      stuff = *p;
      
      stuff.plot = stuff.plot -> Clone (new_layer, is_prepared);
      stuff.plot -> SignalConfigChanged.connect (slot (this, &WavePlotManager::SlotConfigChanged) );
      new_stuff -> plots.push_back (stuff);
    }
}

void WavePlotManager::DeleteIfTooManyLayer ()
{
  cout << "WavePlotManager::DeleteIfTooManyLayer:" << endl;
}

void WavePlotManager::InternalWait ()
{
  cout << "-->> InternalWait" << endl;
  stop = true;
  SyncWait ();
  cout << "<<--" << endl;
}

void WavePlotManager::InternalContinue ()
{
  cout << "-->> InternalContinue" << endl;
  
  if (halt_level == 0) {
    if (m_loop_stage >= LAST_LOOP_STAGE)
      m_loop_stage = 0;
    stop = false;
    cout << "DEBUG: calling Continue" << endl;
    Continue ();
  }
  else {
    cout << "DEBUG: WavePlotManager is halted ..." << endl;
  }
  cout << "<<--" << endl;
}

void WavePlotManager::Loop ()
{
  stop = false;
  
  // #ifdef DEBUG_GSML_WAVE_PLOT_MANAGER
  cout << endl << "---->> WavePlotManager::Loop ... (STAGE: "
       << m_loop_stage << ")" << endl << endl;
  //#endif
  
  if (m_first_loop)
    {
      ThreadHelper::SetPriority (5);
      m_first_loop = false;
    }
  
  sample_index t_chunk_size = m_stream_chunk_size;
  
  DataLayer* t_layer = current_layer -> layer;
  loop_tracks = t_layer -> Tracks ();
  
  // data buffers
  SampleBufferDefault t_buffer (t_chunk_size);
  
  // build list of invaild plots
#ifdef DEBUG_GSML_WAVE_PLOT_MANAGER
  cout << "clearing / resizing loop-plot lists for: " << loop_tracks << " tracks" << endl;
#endif
  
  loop_plots . resize (loop_tracks);
  for (track_index tr = 0; tr < loop_tracks; ++ tr)
    loop_plots [tr] . plots . resize (0);
  
  loop_intervals . Clear ();
  
  // build plot list
  
  cout << "building plot list ..." << endl;
  
  while (m_loop_stage < LAST_LOOP_STAGE && CreateListsForStage (m_loop_stage) == 0)
    ++ m_loop_stage;
  
  if (m_loop_stage >= LAST_LOOP_STAGE) {
    m_loop_stage = 0;
    Wait ();
    
    cout << endl << "<<---- EE NN DD WavePlotManager::Loop ALL DONE !!!" << endl << endl;
    return;
  }
  
  Timer t_timer;
  sample_index t_processed = 0;
  
  IntervalList <sample_index>::const_iterator t_cur_interval, t_end_interval;
  t_cur_interval = loop_intervals . begin ();
  t_end_interval = loop_intervals . end ();
  
  while (!stop && t_cur_interval != t_end_interval) {
    sample_index t_stream_pos = t_cur_interval -> start;
    sample_index t_interval_end = t_cur_interval -> end;
    
    cout << "streaming interval: " << t_stream_pos << " - " << t_interval_end << endl;
    
    // stream the interval
    while (!stop && t_stream_pos < t_interval_end)
      {
	sample_index t_chunk_end = t_stream_pos + t_chunk_size;
	
	for (track_index tr = 0; tr < loop_tracks; ++ tr)
	  {
	    // only stream if there is a plot in this _queue_
	    if (loop_plots [tr] . plots . size () > 0)
	      {
		t_buffer . Clear ();
		if (! (*t_layer) [tr] -> ReadData (t_stream_pos, t_chunk_size, &t_buffer) )
		  t_buffer . ZeroUnused ();
		
		t_processed += t_chunk_size;
		
		// stream
		for (l_plots_it i = loop_plots [tr] . plots . begin ();
		     i != loop_plots [tr] . plots . end ();
		     ++ i)
		  {
		    if (! i -> prepared) {
		      i -> plot -> PrepareStreamWrite (t_stream_pos);
		    i -> prepared = true;
		  }
		  i -> plot -> StreamWrite (t_chunk_size,  &t_buffer, 0);
		  
		  if (i->end <= t_chunk_end) {
		    i -> plot -> FixateStreamWrite ();
		    i = loop_plots [tr] . plots . erase (i);
		  }
		  } // end for all plots: stream
	      } // end if s.th. to stream
	  } // end for all tracks
	
	t_stream_pos += t_chunk_size;
      }
    ++ t_cur_interval;
  } // end while not end

  int t_time = limit_min (t_timer . Delta (), 1); // prevent div.zero below ...
  
  cout << "-->> Streamed " << t_processed << " samples in " << t_time / (1024 * 1024) << " s."
       << " => " << (double)t_processed / (double) t_time << " samples/s (x4 for MB ;-)" << endl;
  
  // fixate the plots that are still in the plot-list
  for (track_index tr = 0; tr < loop_tracks; ++ tr) {
    for (l_plots_it i = loop_plots [tr] . plots . begin ();
	 i != loop_plots [tr] . plots . end ();
	 ++ i) {
      if (i -> prepared)
	i -> plot -> FixateStreamWrite ();
    }
  }
  
  // clearing plot list ...
  for (track_index tr = 0; tr < loop_tracks; ++ tr)
    loop_plots [tr] . plots . resize (0);
  loop_plots . resize (0);
  
  if (stop) {
    m_loop_stage = 0;
    //#ifdef DEBUG_GSML_WAVE_PLOT_MANAGER
    cout << "---->> Loop has to exit on STOP-Request ..." << endl;
    //#endif
  }
  
  //#ifdef DEBUG_GSML_WAVE_PLOT_MANAGER
  cout << endl << "<<---- EE NN DD WavePlotManager::Loop ..." << endl << endl;
  //#endif
}

int WavePlotManager::CreateListsForStage (int stage)
{
  cout << "WavePlotManager::CreateListsForStage: " << stage << endl;
  int added_plots = 0;
  
  for (iterator_p p = current_layer -> plots . begin (); p != current_layer -> plots . end (); ++ p)
    {
      const WavePlotProgress& t_progress = p -> plot -> GetProgress ();
      const WavePlotConfig& t_config = p -> plot -> GetConfig ();
      
      // anything to do in this plot ??
      if (! t_progress . IsAllVaild () )
	{
	  track_index t_tr = p -> plot -> GetTrack ();
	  if (t_tr < loop_tracks)
	    {
	      cout << "s.th. invaild ..." << endl;
	      bool add = false;
	      
	      LoopPlotStuff t_p;
	      t_p.plot = p -> plot;
	      
	      sample_index t_pl_start = t_config . Start ();
	      sample_index t_pl_sps = t_config . SamplesPerSample ();
	      
	      sample_index t_pl_add_start;
	      sample_index t_pl_add_end;
	      
	      // only visible segments in stage 0
	      if (stage == 0) {
		t_pl_add_start = t_config . ViewStart ();
		t_pl_add_end = t_pl_add_start + t_config . ViewLength ();
		
		// only add if the viewed range is invaild ...
		if (! t_progress . IsVaild (t_pl_add_start, t_pl_add_end - t_pl_add_start) )
		  add = true;
	      }
	      else {
		t_pl_add_start = t_progress . GetInvaildStart ();
		t_pl_add_end = t_progress . GetInvaildEnd ();
		
		add = true;
	      }

	      if (add) {
		cout << "add: " << t_pl_add_start << " - " << t_pl_add_end << endl;
		
		// transform to real indexes
		t_pl_add_start = t_pl_add_start * t_pl_sps + t_pl_start;
		t_pl_add_end = t_pl_add_end * t_pl_sps + t_pl_start;
		
		t_p.start = t_pl_add_start;
		t_p.end = t_pl_add_end;
		
		cout << "add (real-index): " << t_p.start << " - " << t_p.end << endl;
		
		loop_plots [t_tr] . plots . push_back (t_p);
		loop_intervals . Add (t_pl_add_start, t_pl_add_end - t_pl_add_start);
		
		++ added_plots;
	      } // end if should be added
	    } // end track check
	  else
	    cout << "BUG BUG BUG: track out of range ..." << endl;
	} // end something invaild
      else
	cout << "all vaild" << endl;
    } // end for all plots
  
  cout << "WavePlotManager::CreateListsForStage: returning: " << added_plots << endl;
  return added_plots;
}

/************
 * WavePlot *
 ************/

WavePlot::WavePlot (WavePlotManager* i_manager, int i_track_no, WavePlotProgress* i_progress)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::WavePlot:" << endl;
#endif
  
  // init stuff
  m_manager = i_manager;
  m_track_no = i_track_no;
  m_progress = i_progress;
  
  m_layer = 0;
  
  registered = false;
  current = false;
  
  connected = false;
  external_stream = false;
  connected_track = 0;
  listening = false;
  
  // remeber: we mustn't connect yet - because we haven't registered ...
}

WavePlot::~WavePlot ()
{
  //#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::~WavePlot:" << endl;
  //#endif
}

void WavePlot::Config (const WavePlotConfig& n_config)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::Config:" << endl;
#endif

  // do last inits ...
  if (!registered)
    {
#ifdef DEBUG_GSML_WAVE_PLOT
      cout << "m_manager->RegisterWavePlot (" << this << ") ..." << endl;
#endif
      m_layer = m_manager->RegisterWavePlot (this);
      
      registered = true;
      current = true;
      
      connect_layer ();
      connect_track ();
    }
  
  if (m_config.Start () == n_config.Start () &&
      m_config.PlotLength () == n_config.PlotLength () &&
      m_config.SamplesPerSample () == n_config.SamplesPerSample () )
    {
#ifdef DEBUG_GSML_WAVE_PLOT
      cout << "WavePlot::Config: Only the ViewPoint has been changed ..." << endl;
#endif
      m_config = n_config;
      SignalConfigChanged (m_config, false);
    }
  else
    {
      // stop stream if we are currently get an external one
      // cout << "Config: waiting for external stream lock" << endl;
      SigC::Threads::MLock mlock (external_stream_lock);
      // cout << "Config: waiting for external stream lock: FINISHED" << endl;
      
      sample_index t_stream_pos = stream_start + stream_length;
      
      if (external_stream)
	internal_fixate ();
      
      // TODO: try to recycle via Manager
      
      m_config = n_config;
      
      m_progress->SetLength (m_config.PlotLength () );
      m_progress->InvaildAll ();
      
      // to sniff a current running stream
      if (external_stream)
      	internal_prepare_write (t_stream_pos, false);
      
      SignalConfigChanged (m_config, true);
      SignalProgressUpdated (0, m_config.PlotLength () );
    }
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::Config: END" << endl;
#endif
}

const WavePlotConfig& WavePlot::GetConfig ()  const
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::GetConfig:" << endl;
#endif
  
  return m_config;
}

const WavePlotProgress& WavePlot::GetProgress () const
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::GetProgress:" << endl;
#endif
  
  return *m_progress;
}

void WavePlot::CloneHelper (DataLayer* new_layer, WavePlot* old_plot, bool is_prepared)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::CloneHelper: layer: " << new_layer << ", old_plot: " << old_plot << endl;
#endif
  
  // cout << "WavePlot::CloneHelper: cloned object: " << this << endl;
  
  // maybe the next lines (without the last) aren't need with a copy-ctor ...
  m_config = old_plot->m_config;
  m_layer = new_layer;
  
  
  registered = true; // called from the inheritenced class from the Manager ...
  current = false;
  
  connect_layer ();
  connect_track ();
}

void WavePlot::UnsetCurrent (WavePlot* new_current)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::UnsetCurrent: new_current: " << new_current << endl;
#endif
  
  if (!current)
    cout << "BUG: WavePlot::UnsetCurrent: was not current ??? (continue anyway ...)" << endl;
  
  current = false;
  
  SignalUnsetCurrent.emit (new_current);
}

void WavePlot::SetCurrent ()
{
  //#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SetCurrent: " << endl;
  //#endif
  current = true;
  
  cout << "SetCurrent END" << endl;
}

DataLayer* WavePlot::GetDataLayer ()
{
  return m_layer;
}

TrackLayer* WavePlot::GetTrackLayer ()
{
  return connected_track;
}

track_index WavePlot::GetTrack ()
{
  return m_track_no;
}

void WavePlot::Debug () const
{
  cout << "WavePlot::Debug:" << endl;
  
  cout << "  m_manager:  " << m_manager << endl
       << "  m_layer:    " << m_layer << endl
       << "  registered: " << registered << endl
       << "  current:    " << current << endl
       << "  connected:  " << connected << endl
       << endl;
  
  if (connected)
    {
      cout << "    connected_track: " << connected_track
	   << "    listening:       " << listening
	   << "    passing:         " << passing
	   << "    insert_write:    " << insert_write
	   << endl;
    }
  
  m_config.Debug ();
  m_progress->Debug (false);
}

void WavePlot::PrepareStreamWrite (sample_index start)
{
  // cout << "PrepareStreamWrite: waiting for lock ..." << endl;
  SigC::Threads::MLock mlock (external_stream_lock);
  // cout << "PrepareStreamWrite: waiting for lock FINISHED" << endl;
  
  if (connected) {
    cout << "BUG: FX write in progress (can't be current ...)" << endl;
    return;
  }
  if (external_stream) {
    cout << "BUG: An extrernal stream is already in progress ..." << endl;
    return;
  }
  
  internal_prepare_write (start, false);
  external_stream = true;
}

void WavePlot::StreamWrite (sample_index length, SampleBufferDefault* source, sample_index buffer_index)
{
  // cout << "StreamWrite: waiting for lock ..." << endl;
  SigC::Threads::MLock mlock (external_stream_lock);
  // cout << "StreamWrite: waiting for lock FINISHED" << endl;
  
  if (!external_stream) {
    cout << "BUG: No external stream is prepated ?!?!" << endl;
    return;
  }
  
  internal_write (length, source, buffer_index);
}

// When this mehtod is changed, UPDATE Config also!!!
void WavePlot::FixateStreamWrite ()
{
  // cout << "FixateStreamWrite: waiting for lock ..." << endl;
  SigC::Threads::MLock mlock (external_stream_lock);
  // cout << "FixateStreamWrite: waiting for lock FINISHED" << endl;
  
  if (!external_stream) {
    cout << "BUG: No external stream is prepated ?!?!" << endl;
    return;
  }
  
  internal_fixate (); // insert_write, signal
  external_stream = false;
  
  // cout << "FixateStreamWrite: END" << endl;
}

void WavePlot::SlotTrackCreated (track_index track)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotTrackCreated:" << endl;
#endif

  // bug check ;-) - if we aren't already connected...
  if (!connected)// some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotTrackCreated: I'm not connected!!!!!!";
    }  
  
  // are we affected ???
  if (track <= m_track_no)
    {
      m_progress->InvaildAll ();
      
      // we have to reconnect to the new 
      disconnect_track ();
      connect_track ();
    }
}

void WavePlot::SlotTrackErased (track_index track)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotTrackErased:" << endl;
#endif

  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotTrackDeleted: I'm not connected!!!!!!";
    }
  
  // are we affected ???
  if (track <= m_track_no)
    {
      // the old track doe not exist anymore - so the connections are deleted by itselves ...
      if (track == m_track_no) // it is our track
	{
#ifdef DEBUG_GSML_WAVE_PLOT
	  cout << "special case: it is our track!!" << endl;
#endif
	  connected_track = 0;
	}
	  
      m_progress->InvaildAll ();
	  
      disconnect_track ();
      connect_track ();
    }
}

// Track_Layer

void WavePlot::SlotPreparedOverWrite (sample_index start)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotPreparedOverWrite: start: " << start << endl;
#endif
  
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotPreparedOverWrite: I'm not connected!!!!!!";
    }
  
  if (listening) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotPreparedOverWrite: write stream already prepared";
    }
  
  internal_prepare_write (start, true);
  
  listening = true;
  insert_write = false;
}

void WavePlot::SlotPreparedInsertWrite (sample_index start)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotPreparedInsertWrite: start: " << start << endl;
#endif
  
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotPreparedInsertWrite: I'm not connected!!!!!!";
    }
  
  if (listening) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotPreparedInsertWrite: write stream already prepared";
    }
  
  internal_prepare_write (start, true);
  
  listening = true;
  insert_write = true;
}

void WavePlot::SlotFixatedWrite ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotFixatedWrite:" << endl;
#endif
  
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotFixateWrite: I'm not connected!!!!!!";
    }
  
  if (!listening) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotFixateWrite: write stream not prepared";
    }
  
  internal_fixate (insert_write);
  
  passing = false;
  listening = false;
}

void WavePlot::SlotCanceledWrite ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotCanceledWrite:" << endl;
#endif
  
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotCancelWrite: I'm not connected!!!!!!";
    }
  
  if (!listening) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotCancelWrite: write stream not prepared";
    }
  
  listening = false;
}

void WavePlot::SlotWroteData (sample_index length, SBHandle source, sample_index buffer_start)
{
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::LayerSlotWroteData: I'm not connected!!!!!!";
    }
  
  // bug check ;-) - if there is no prepated write
  if (!listening) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotWroteData: no write stream prepared";
    }
  
  internal_write (length, source, buffer_start);
}

void WavePlot::SlotRegionRemoved (sample_index start, sample_index length)
{
//#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotRegionRemoved: start: " << start << ", length: " << length << endl;
//#endif
  
  // bug check ;-) - if we aren't already connected...
  if (connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotRegionMoved: I'm not connected!!!!!!";
    }
  
  sample_index t_start = m_config.Start ();
  sample_index t_length = m_config.Length ();
  sample_index t_end =  t_start + t_length;
      
  if (start >= t_end) // unaffected ??
    {
      cout << "unaffected - behind my area ..." << endl;
      return;
    }
  
  sample_index affected_length
    = limit_max (start + length, t_end) - start;
  
  cout << "affected_length: " << affected_length << endl;
  
  if (start + affected_length >= t_end) // nothing to reuse ??
    {
      cout << "all affected ..." << endl;
      m_progress->InvaildAll ();
      return;
    }
  
  cout << "some can be reused ..." << endl;
  
  // some can be reused
  sample_index t_sps = m_config.SamplesPerSample ();
  
  // adjust region
  sample_index move_start = limit_min (start + length, t_start + affected_length);
  sample_index move_length = t_end - move_start;
  sample_index new_start = limit_min (start, t_start);
  
  cout << "move: " << move_start << " + " << move_length << " -> " << new_start << endl;
  
  move_start -= t_start;
  new_start -= t_start;
  
  move_start /= t_sps;
  new_start /= t_sps;
  move_length /= t_sps;
  
  cout << "move: " << move_start << " + " << move_length << " -> " << new_start << endl;
  
  if (move_length < 0)
    StopInDebugger ();
  
  m_progress->Move (move_start, move_length, new_start);
}

void WavePlot::SlotRegionChanged (sample_index start, sample_index length)
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::SlotRegionChanged: start: " << start << ", length: " << length << endl;
#endif
  
  // bug check ;-) - if we aren't already connected...
  if (!connected) // some bug
    {
      BasicExcept e;
      e.info = "BUG!! - WavePlot::SlotRegionChanged: I'm not connected!!!!!!";
    }
  
  sample_index m_temp_start = m_config.Start ();
  sample_index m_temp_end = m_temp_start + m_config.Length ();
  
  if (start >= m_temp_end) // unaffected ??
    {
      cout << "unaffected - behind my area ..." << endl;
      return;
    }
  
  sample_index m_temp_sps = m_config.SamplesPerSample ();
  
  // adjust region
  sample_index change_start = limit_min (start, m_temp_start);
  sample_index change_end = limit_max (start + length, m_temp_end);
  sample_index change_length = change_end - change_start;
  
  change_start -= m_temp_start;
  change_start /= m_temp_sps;
  change_length /= m_temp_sps;
  
  cout << "invailding: " << change_start << " length " << change_length << endl;
  
  // update area - we have to reload this stuff in the second thread, later ...
  m_progress->Invaild (change_start, change_length);
}

void WavePlot::internal_prepare_write (sample_index start, bool do_buffer=true)
{
  // cout << "internal_prepare_write: start: " << start << ", do_buffer: " << do_buffer << endl;
  
  stream_buffered = do_buffer;
  
  if (do_buffer)
    prepare_buffer_impl ();
  
  m_between_write_buffer = new SampleBufferDefault (m_config.SamplesPerSample () );
  m_in_write_buffer = 0;
  
  sample_index t_c_start = m_config.Start ();
  sample_index t_c_sps = m_config.SamplesPerSample ();
  
  if (start < t_c_start)
    stream_first_alligned = t_c_start;
  else
    stream_first_alligned = start + start % t_c_sps;
  
  // cout << "stream_first_alligned: " << stream_first_alligned << endl;
  
  m_progress_pos = (stream_first_alligned - t_c_start) / t_c_sps;
  progress_signal_pos = m_progress_pos;
  // cout << "m_progress_pos: " << m_progress_pos << endl;
  
  stream_start = start;
  stream_length = 0;
}

void WavePlot::internal_write (sample_index length, SampleBuffer* source,
			       sample_index buffer_index)
{
  // cout << "internal_write: length: " << length << endl;
  
  sample_index t_cur_index = stream_start + stream_length;
  
  sample_index t_c_start = m_config.Start ();
  sample_index t_c_end = m_config.Start () + m_config.Length ();
  sample_index t_c_length = m_config.PlotLength ();
  
  // s.th. in our range
  if (t_cur_index + length > stream_first_alligned && t_cur_index < t_c_end)
    {
      // cout << "  in our range: cur: " << t_cur_index
      //	   << ", t_c_s: " << t_c_start << ", t_c_e: " << t_c_end << endl;
      
      // skip before first alligned
      sample_index t_processed = 0;
      sample_index t_c_sps = m_config.SamplesPerSample ();
      
      if (t_cur_index < stream_first_alligned)
	t_processed = stream_first_alligned - t_cur_index;
      
      // cout << "  processed after start allign limit: " << t_processed << endl;
      
      // stuff from last turn ??
      if (m_in_write_buffer > 0)
	{
	  // cout << "  s.th. in between write buffer" << endl;
	  sample_index t_fill = limit_max (t_c_sps - m_in_write_buffer, length - t_processed);
	  // cout << "  filling with: " << t_fill << endl;
	  
	  source->ReadData (buffer_index + t_processed, t_fill,
			    m_between_write_buffer->data + m_in_write_buffer);
	  m_in_write_buffer += t_fill;
	  m_between_write_buffer->SetUsedLength (m_in_write_buffer);
	  t_processed += t_fill;
	  
	  if (m_in_write_buffer == t_c_sps)
	    {
	      // cout << "use between buffer ... (for pos: " << m_progress_pos << ")" << endl;
	      process_data_impl (t_c_sps, m_between_write_buffer, 0, m_progress_pos);
	      m_in_write_buffer = 0;
	      ++m_progress_pos;
	    }
	  // cout << "  buffer not full enough ... (continue collecting in next turn)" << endl;
	}
      
      // bug ??
      if (t_processed > length)
	{
	  cout << "------->>>>>>> BB UU GG ?? :: VALUES CAN'T BE VAILD ?? !! ?? - Can they?" << endl;
	  cout << "  t_processed: " << t_processed << ", length: " << length << endl;
	  
	}
      
      // data left ??
      if (t_processed < length)
	{
	  // pass all that is in one big peace to the impl.
	  while (length - t_processed >= t_c_sps && m_progress_pos < t_c_length)
	    {
	      // cout << m_progress_pos << ", ";
	      process_data_impl (t_c_sps, source, buffer_index + t_processed, m_progress_pos);
	      t_processed += t_c_sps;
	      ++m_progress_pos;
	    }
	  
	  // save the rest for the next turn
	  if (t_processed < length && m_progress_pos < t_c_length)
	    {
	      sample_index t_save = limit_max (length - t_processed, t_c_sps);
	      // cout << "  some data left for the next turn: " << t_save << endl;
	      m_between_write_buffer->Clear ();
	      source->ReadData (buffer_index + t_processed, t_save, m_between_write_buffer->data);
	      m_between_write_buffer->SetUsedLength (t_save);
	      m_in_write_buffer = t_save;
	    }
	  
	  if (!stream_buffered)
	    {
	      sample_index t_v_start = progress_signal_pos;
	      sample_index t_v_length = m_progress_pos - progress_signal_pos;
	      
	      // cout << "signalling: " << t_v_start << " + " << t_v_length << endl;
	      
	      m_progress->Vaild (t_v_start, t_v_length);
	      SignalProgressUpdated (t_v_start, t_v_length);
	      progress_signal_pos = t_v_start + t_v_length;
	    }
	  
	}
    }
  // else
  //   cout << "  noting for us" << endl;
  
  stream_length += length;
}

void WavePlot::internal_fixate (bool was_insert=false)
{
  if (stream_buffered)
    fixate_buffer_impl ();
  
  del_helper2 (m_between_write_buffer);
  
  // TODO: some other things here and there
  //       e.g. update progress
}

void WavePlot::internal_cancel ()
{
  if (stream_buffered)
    cancel_buffer_impl ();
}

void WavePlot::connect_layer ()
{
  // connect to signals
  m_layer->SignalTrackCreated.connect (slot (this, &WavePlot::SlotTrackCreated) );
  m_layer->SignalTrackErased.connect (slot (this, &WavePlot::SlotTrackErased) );
}

void WavePlot::connect_track ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::connect_track:" << endl;
#endif

  // try to disconnect
  disconnect_track ();

  // if m_track is vaild -> connect
  if (m_layer->Tracks () > m_track_no)
    {
#ifdef DEBUG_GSML_WAVE_PLOT
      cout << "ok - connect to signals:" << endl;
#endif
      connected_track = (*m_layer)[m_track_no];
      
      connection_wrote_data = 
	connected_track->SignalWroteData.connect (slot (this, &WavePlot::SlotWroteData) );
      connection_region_removed =
	connected_track->SignalRegionRemoved.connect (slot (this, &WavePlot::SlotRegionRemoved) );
      connection_region_changed =
	connected_track->SignalRegionChanged.connect (slot (this, &WavePlot::SlotRegionChanged) );
      connection_prepared_overwrite =
	connected_track->SignalPreparedOverWrite.connect (slot (this, &WavePlot::SlotPreparedOverWrite) );
      connection_prepared_insertwrite =
	connected_track->SignalPreparedInsertWrite.connect (slot (this, &WavePlot::SlotPreparedInsertWrite) );
      connection_fixated_write =
	connected_track->SignalFixatedWrite.connect (slot (this, &WavePlot::SlotFixatedWrite) );
      connection_canceled_write =
	connected_track->SignalCanceledWrite.connect (slot (this, &WavePlot::SlotCanceledWrite) );
    }
  else
    {
#ifdef DEBUG_GSML_WAVE_PLOT
      cout << "hm?: my track is not there!:" << endl;
#endif
      connected_track = 0;
    }
}

void WavePlot::disconnect_track ()
{
#ifdef DEBUG_GSML_WAVE_PLOT
  cout << "WavePlot::disconnect_track:" << endl;
#endif
  
  if (connected_track != 0)
    {
      connection_wrote_data.disconnect ();
      connection_region_removed.disconnect ();
      connection_region_changed.disconnect ();
      connection_prepared_overwrite.disconnect ();
      connection_prepared_insertwrite.disconnect ();
      connection_fixated_write.disconnect ();
      connection_canceled_write.disconnect ();
    }
  connected_track = 0;
}

/*********************
 * *** WavePlots *** *
 *********************/

/*********************
 * MinMaxAvgWavePlot *
 *********************/

void MinMaxAvgWavePlot::Progress::SetLength (sample_index n_length)
{
  WavePlotProgress::set_length_impl (n_length);
  set_length_impl (n_length);
}

void MinMaxAvgWavePlot::Progress::Move (sample_index start, sample_index length, sample_index new_start)
{
  WavePlotProgress::move_impl (start, length, new_start);
  move_impl (start, length, new_start);
}

void MinMaxAvgWavePlot::Progress::Zero (sample_index start, sample_index length)
{
  WavePlotProgress::zero_impl (start, length);
  zero_impl (start, length);
}

void MinMaxAvgWavePlot::Progress::Debug (bool verbose = false) const
{
  WavePlotProgress::debug_impl (verbose);
  debug_impl (verbose);
}

void MinMaxAvgWavePlot::Progress::set_length_impl (sample_index n_length)
{
  min_data.resize (n_length);
  max_data.resize (n_length);
  neg_avg_data.resize (n_length);
  pos_avg_data.resize (n_length);
}

void MinMaxAvgWavePlot::Progress::move_impl (sample_index start, sample_index length, sample_index new_start)
{
  vector<pcm_default::value_type> t_min_data;
  vector<pcm_default::value_type> t_max_data;
  vector<pcm_default::value_type> t_neg_avg_data;
  vector<pcm_default::value_type> t_pos_avg_data;
  
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::Progress::move_impl:" << endl;
  cout << start << " + " << length << " -> " << new_start << endl;
#endif
  
  t_min_data.resize (length);
  t_max_data.resize (length);
  t_neg_avg_data.resize (length);
  t_pos_avg_data.resize (length);
      
  // copy
  for (int i = 0; i < length; ++i)
    {
      t_min_data [i] ( min_data [start + i] );
      t_max_data [i] ( max_data [start + i] );
      t_neg_avg_data [i] ( neg_avg_data [start + i] );
      t_pos_avg_data [i] ( pos_avg_data [start + i] );
    }
  
  // back - copy
  for (sample_index i = 0; i < length; ++i)
    {
      min_data [new_start + i] ( t_min_data [i] );
      max_data [new_start + i] ( t_max_data [i] );
      neg_avg_data [new_start + i] ( t_neg_avg_data [i] );
      pos_avg_data [new_start + i] ( t_pos_avg_data [i] );
    }
}

void MinMaxAvgWavePlot::Progress::zero_impl (sample_index start, sample_index length)
{
  for (int i = start; i < start + length; ++i)
    {
      min_data [i] (0);
      max_data [i] (0);
      neg_avg_data [i] (0);
      pos_avg_data [i] (0);
    }
}

void MinMaxAvgWavePlot::Progress::debug_impl (bool verbose) const
{
  if (verbose)
    {
      cout << "  Data: " << endl;
      
      for (sample_index i = 0; i < m_length; ++i)
	{
	  cout << "  " << i << ": " << IsUseable (i) << " ";
	  if ( IsUseable (i) )
	    { 
	      cout << "["
		   << min_data [i] . data
		   << ", " << max_data [i] . data
		   << ", " << neg_avg_data [i] . data
		   << ", " << pos_avg_data [i] . data
		   << "]"; 
	    }
	  cout << endl;
	}
      cout << endl;
    }
}

MinMaxAvgWavePlot::MinMaxAvgWavePlot (WavePlotManager* i_manager, int i_track) 
  : WavePlot (i_manager, i_track, &m_progress)
{
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::MinMaxAvgWavePlot:" << endl;
#endif
  
  min_working_data = &m_progress.min_data;
  max_working_data = &m_progress.max_data;
  neg_avg_working_data = &m_progress.neg_avg_data;
  pos_avg_working_data = &m_progress.pos_avg_data;
}

MinMaxAvgWavePlot::~MinMaxAvgWavePlot ()
{
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::~MinMaxAvgWavePlot:" << endl;
#endif
}

const MinMaxAvgWavePlot::Progress&  MinMaxAvgWavePlot::GetProgress () const
{
  return m_progress;
}

WavePlot* MinMaxAvgWavePlot::Clone (DataLayer* new_layer, bool is_prepared)
{
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::Clone: new_layer: " << new_layer << endl;
#endif
  MinMaxAvgWavePlot* t_plot = new MinMaxAvgWavePlot (m_manager, m_track_no);
  
  // cout << "MinMaxAvgWavePlot:Clone: cloned object: " << t_plot << endl;
  
  t_plot->m_progress = m_progress;
  
  t_plot->CloneHelper (new_layer, this, is_prepared);
  
  return t_plot;
}

void MinMaxAvgWavePlot::Debug ()
{
  cout << "MinMaxAvgWavePlot::Debug:" << endl;
  
  WavePlot::Debug ();
}

void MinMaxAvgWavePlot::prepare_buffer_impl ()
{
  //#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::prepare_buffer_impl:" << endl;
  //#endif
  
  min_temp_data = m_progress.min_data;
  max_temp_data = m_progress.max_data;
  neg_avg_temp_data = m_progress.neg_avg_data;
  pos_avg_temp_data = m_progress.pos_avg_data;
  
  min_working_data = &min_temp_data;
  max_working_data = &max_temp_data;
  neg_avg_working_data = &neg_avg_temp_data;
  pos_avg_working_data = &pos_avg_temp_data;
  
  buffering = true;
}

void MinMaxAvgWavePlot::fixate_buffer_impl ()
{
  //#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::fixate_buffer_impl:" << endl;
  //#endif

  if (buffering)
    {
      m_progress.min_data = min_temp_data;
      m_progress.max_data = max_temp_data;
      m_progress.neg_avg_data = neg_avg_temp_data;
      m_progress.pos_avg_data = pos_avg_temp_data;
      
      min_working_data = &m_progress.min_data;
      max_working_data = &m_progress.max_data;
      neg_avg_working_data = &m_progress.neg_avg_data;
      pos_avg_working_data = &m_progress.pos_avg_data;
      
      // mem saveing ;-)
      min_temp_data.resize (0);
      max_temp_data.resize (0);
      neg_avg_temp_data.resize (0);
      pos_avg_temp_data.resize (0);
    }
  else
    cout << "  BUG: was not buffered ?!?" << endl;
  
  buffering = false;
}

void MinMaxAvgWavePlot::cancel_buffer_impl ()
{
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::cancel_data:" << endl;
#endif
  
  // bug check ;-)
  if (buffering)
    {
      min_working_data = &m_progress.min_data;
      max_working_data = &m_progress.max_data;
      neg_avg_working_data = &m_progress.neg_avg_data;
      pos_avg_working_data = &m_progress.pos_avg_data;
      
      // mem saveing ;-)
      min_temp_data.resize (0);
      max_temp_data.resize (0);
      neg_avg_temp_data.resize (0);
      pos_avg_temp_data.resize (0);
    }  
  else
    cout << "  BUG: was not buffered ?!?" << endl;
  
  buffering = false;
}

void MinMaxAvgWavePlot::process_data_impl (sample_index length, const SampleBuffer* buffer,
					   sample_index buffer_index, sample_index progress_index)
{
#ifdef DEBUG_GSML_WAVE_PLOT_MIN_MAX
  cout << "MinMaxAvgWavePlot::process_data: length: " << length 
       << ", buffer_index: " << buffer_index << ", progress_index: " << progress_index << endl;
#endif
  
  if (length <= 0)
    {
      cout << "BUG BUG BUG: !!! MinMaxAvgWavePlot::process_data_impl called with negative/zero length" << endl;
      return;
    }
  
  SampleBufferDefault* temp_buffer;
  sample_index unused_length;
  
  // convert to our default pcm type
  temp_buffer = new SampleBufferDefault (length);
  buffer->ReadData (buffer_index, length, temp_buffer->data);
  
  // compute min/max
  
  pcm_default::value_type current_min;
  pcm_default::value_type current_max;
  
  current_min.data = temp_buffer->data [0];
  current_max.data = temp_buffer->data [0];
  
  for (sample_index i = 1; i < length; ++i)
    {
      current_min.data = limit_max (temp_buffer->data [i].data, current_min.data);
      current_max.data = limit_min (temp_buffer->data [i].data, current_max.data);
    }
  
  (*min_working_data) [progress_index] = current_min;
  (*max_working_data) [progress_index] = current_max;
  
  delete temp_buffer;
}
