/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   The Application class.
 */

#ifndef GSMP_APP_HH
#define GSMP_APP_HH

// GSMP Stuff
#include "Base.hh"
#include "GUI.hh"
#include "Pattern.hh"
#include "Window.hh"
#include "Project.hh"
#include "View.hh"
#include "Menu.hh"

#include "FindMemCorruption.hh"

namespace GSMP
{
  //using namespace SigC;
  
  class Manager : public Menu::MenuManager
  { // MenuManager wraper
    friend class App;
    friend class Project;
    friend class View;
  };
  
  class App : public Menu::MenuData
  {
  public:
    
    class Exception : public GSMP::Exception
    {
    public:
      Exception (const string& error);
    };
    
    class ProjectNotFound : public Exception
    {
    public:
      ProjectNotFound ();
    };
    
    class Preferences
    {
      friend class App;
    public:
      /* GSMP component system configuration */
      string default_view_type;

      /* default configurations for new projects */
      GSML::sample_rate default_sample_rate;
      int default_tracks;
      int default_max_datalayers;

      /* default GSML::DataFileIndex configuration */
      GSML::DataFileIndex::Config file_index_config;

      /* IO config */
      string default_mixer_node; // Factory item to Base "GSML::MixerNode"

      class MenuFilter
      {
      public:
	MenuFilter (const string& i_location, const string& i_title)
	  : location (i_location), title (i_title) {};
	string location;
	string title;
      };

      deque <MenuFilter> recommended_main_menu_l;    // Menu for MainWindows
      deque <MenuFilter> recommended_main_menu_r;    // Menu for MainWindows

      deque <MenuFilter> recommended_simple_menu_l;  // Menu for SimpleWindows
      deque <MenuFilter> recommended_simple_menu_r;  // Menu for SimpleWindows

      deque <MenuFilter> recommended_view_menu;      // Minimal contextmenu for Views
    protected:
    };   

    // TODO private:
    //      friend int main (int, char**);
    
    typedef deque <Project*>::iterator project_iterator;
    typedef deque <MainWindow* >::iterator main_window_iterator;
    typedef deque <SimpleWindow* >::iterator simple_window_iterator;
    
    App (int argc, char *argv[], const string& i_plugin_file, const string& i_toolkit_file, Manager* i_app_manager);
    // argc, argv are passed to GUI::Kit constructor
    // <my_plugin_file> is a string refering to a file 
    // containing the basic Plugins for the App:
    // <default_mainwin>, <default_simplewin>,

    ~App ();

    const Preferences& GetPreferences () {return m_preferences;}
    
    Project* CreateProject ();
    Project* CreateProject (GSML::sample_rate i_samplerate, int i_track_no,
			    int i_max_datalayers, const string& i_name);
    
    bool EraseProject (Project* del_prj);
    project_iterator EraseProject (const project_iterator& del_prj);

    project_iterator BeginProjects ();
    project_iterator EndProjects ();
    
    Pattern* CurrentPattern ();
    
    // Functions to pop up new Windows. String view_type determines which ViewImpl plugin will be used.
    // If view_type string is empty, default_view_type will be used. 

    MainWindow* CreateMainWindow (const string &view_type = "");
    MainWindow* CreateMainWindow (Project* project, const string &view_type = "");
    MainWindow* CreateMainWindow (SimpleWindow* copy_from);
    MainWindow* CreateMainWindow (View* instantiate_from);  // If View has parent win, it will be copied
    
    main_window_iterator BeginMainWindows ();
    main_window_iterator EndMainWindows ();

    SimpleWindow* CreateSimpleWindow (const string &view_type = "");
    SimpleWindow* CreateSimpleWindow (Project* project, const string &view_type = "");
    SimpleWindow* CreateSimpleWindow (SimpleWindow* copy_from);
    SimpleWindow* CreateSimpleWindow (View* instantiate_from);  // If View has parent win, it will be copied
    
    simple_window_iterator BeginSimpleWindows ();
    simple_window_iterator EndSimpleWindows ();

    GSMX_utils::Plugin::content* GetViewImplTypes ();
    
    GSML::DataFileIndex* GetDataFileIndex ();
    GSML::HWMixerNode* GetMixerNode ();
    
    //const GSMX_utils::Plugin::pluginset RegisterPluginFile (const string& file); 
    
    void Run ();
    void Init (GUI::InitScreen* init_screen); // load plugins, initialize IO etc (called by GUI kit)
    bool Exit ();

    SigC::Signal1<void, const GSMX_utils::Plugin::pluginset&> SignalPluginsRegistered;
    SigC::Signal1<void, Project*> SignalProjectCreated; // counterpart is in project
    
    SigC::Signal1<void, SimpleWindow*> SignalWindowCreated; // counterpart is in SimpleWindow
    SigC::Signal0<bool, GUIRequestMarshaller<bool> > SignalTerminateRequest; // someone wants to quit App, preventable by returning false.
    SigC::Signal0<void> SignalTerminated;    // We are exiting now

    string GuiComponentName (string component_name); // converts component name to current GUI plugin specific name 

    GUI::Kit* GetKit () {return m_kit;}
    
    Manager* GetManager () {return m_manager;}

  private:
    friend class Project;
    friend class View;
    friend class MainWindow;
    friend class SimpleWindow;

    void SlotMainWindowDeleted (MainWindow* the_window);
    void SlotSimpleWindowDeleted (MainWindow* the_window);

    GSML :: pcm_default pcm_type; // TODO: configure this...

    Preferences m_preferences;

    GSML::DataFileIndex* m_file_index;

    /* IO stuff */
    GSML::HWMixerNode* m_mixer_node;

    deque <Project*> m_projects;

    deque <MainWindow* > m_main_windows;
    deque <SimpleWindow* > m_simple_windows;

    Pattern* m_current_pattern;
    deque <NamedPattern*> m_named_patterns;
    Manager* m_manager;

    int unnamed_project_counter;

    GUI::Kit* m_kit;

    //  private:
    string gui_extension; // the GUI's id, appended to all visualisation plugins, when instantiated...

    // HACK:
    string m_plugin_file;

    FindMemCorruption m_fmc;
  };

} // end namespace GSMP

#endif // GSMP_APP_HH

