/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Importer / Exporter classes and low-level interfaces.
 */

#ifndef GSMP_EXTFILEIO_H
#define GSMP_EXTFILEIO_H

// GSML
#include "Basics.hh"
#include "SampleBuffer.hh"
#include "DataFile.hh"
#include "TrackLayer.hh"
#include "DataLayer.hh"
#include "IO_impl.hh"

// GSMP
#include "Base.hh"
#include "FXBase.hh"

/* brief Overview / Brainstorming ...
 *
 * This is the architecture for importing / exporting non-native files.
 *
 * The interface is defined by the ImportFile and ExportFile. This classes
 * are used by the Importer and Exporter which are GSMP Plugins that control
 * the import / export and to the GSMP generic FX things.
 *
 * TODO:
 *   Improve the Im- / Exporter to be configureable and be able to use
 * different Im- / ExportFile instances at once. (To be able to import
 * many files at once. E.g. select 6 raw-data files recorded with another
 * program ...
 *
 */

namespace GSMP {
  
  class Project;
  
  // interface for importable files
  class ImportFile
  {
  public:
    
    class Exception : public GSMP :: Exception
    {
    public:
      Exception (const string& error);
    };
    
    class Unprepared : public Exception
    {
    public:
      Unprepared (const string& desc);
    };
    
    virtual ~ImportFile () {};
    
    virtual string Extension () const = 0;
    
    virtual bool PrepareImport (const string& fname, bool temp_file_used) = 0;
    
    virtual GSML::track_index Tracks () = 0;
    virtual GSML::sample_rate Samplerate () = 0;
    virtual GSML::sample_index Length () = 0;
    
    virtual bool Import (GSML::track_index track, GSML::TrackLayer* layer) = 0;
    virtual void Close () = 0;
  };
  
  // interface for exportable files
  class ExportFile
  {
  public:
    
    class Exception : public GSMP :: Exception
    {
    public:
      Exception (const string& error);
    };
    
    virtual ~ExportFile() {};
    
    virtual bool Create (const string& fname, GSML::track_index tracks, GSML::sample_index rate,
			 GSML::sample_index length) = 0;
    virtual void Cancel () = 0;
    virtual void Close () = 0;
    
    virtual GSML::sample_index NextChunkSize () = 0;
    
    // the Write method is called with a linear track increment - makes the impl much easier ...
    virtual bool Write (GSML::track_index track, GSML::SampleBufferDefault* src, bool last_write) = 0;
  };
  
  // Currently the file im- and exporter dump the project 1:1 into a ExportFile object
  
  class FileImporter : public FXBase
  {
  public:
    FileImporter (const string& i_fname, ImportFile* i_obj);
    virtual ~FileImporter ();
    
    virtual bool Start_impl (Project* project, string& error_string);
    virtual void Canceled ();
    virtual void Loop ();
    
  protected:
    
    // states
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_length;
    
    GSML::track_index m_cur_track;
    bool m_prepared;
    
    // config
    string m_fname;
    ImportFile* m_obj;
  };
  
  class FileExporter : public FXBase
  {
  public:
    FileExporter (const string& fname, ExportFile* obj);
    virtual ~FileExporter ();
    
    virtual bool Start_impl (Project* project, string& error_string);
    virtual void Canceled ();
    virtual void Loop ();
  
  protected:
    
    // states
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_length;
    
    GSML::sample_index m_cur_pos;
    
    bool m_created;
    
    // config
    string m_fname;
    ExportFile* m_obj;
  };
  
  // Currently tge MixExporter dumps the first 2 channels into a ExportFile object
  
  class MixExporter : public FXBase, public GSML::MixerNode
  {
  public:
    MixExporter (const string& i_fname, ExportFile* i_obj);
    virtual ~MixExporter ();
    
    // custom interface
    
    virtual bool Start_impl (Project* project, string& error_string);
    virtual void Canceled ();
    virtual void Loop ();
  
    // mixer-node interface
    
    virtual GSML::track_index PlaybackChannels ();
    virtual GSML::track_index CaptureChannels ();
    
    virtual int PlaybackPeriodSize ();
    virtual int CapturePeriodSize ();
    
    virtual int PlaybackPeriods ();
    virtual int CapturePeriods ();
    
    virtual int PlaybackXRuns ();
    virtual int CaptureXRuns ();
    
    virtual void SetPlayback (bool state);
    virtual void SetCapture (bool state);
    
    virtual bool IsPlayback ();
    virtual bool IsCapture ();
    
    virtual void StartIO (GSML::Mixer& source);
    virtual void StopIO ();
    
    virtual bool IsIORunning ();
    
    virtual GSML::sample_index GetCurrentPos ();
    
    virtual void Debug ();
    
  protected:
    // states
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_length;
    
    int m_chunk_size;
    GSML::sample_index m_cur_pos;
    
    bool m_got_mixer_start;
    
    GSML::Mixer* m_mixer;
    GSML::DataSource* m_data_source;
    
    GSML::UniversalMixerParams m_pl_params;
    GSML::UniversalMixerParams m_ca_params;
    
    bool m_created;
    
    // config
    string m_fname;
    ExportFile* m_obj;
  };
  
} // end namespace GSMP

#endif // GSMP_EXTFILEIO_H
