/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Extended IO-Streaming, according Importer and OggVorbis / MP3
 *   IO-Stream classes.
 */

#ifndef GSMP_EXTFILEIO_IMPL_H
#define GSMP_EXTFILEIO_IMPL_H

// GSML
#include "DataFile.hh"

// GSMP
#include "ExtFileIO.hh"

#include "MADHelper.hh" // our mad-lib abstraction (mid-level interface)

#if EXT_FIO_OGGVORBIS == 1
// external
#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

#endif

namespace GSMP
{
  /* Documentation:
   *
   * The StreamIO and RandomAccessIO define the interface for the non-native file formates.
   * The compressed formates like OggVorbis and MP3 had a great impact on the current
   * implementation.
   *
   * StreamIO: Defines the interface for stream sources (sources where seeks are very
   *   inefficent).
   * RandomAccessIO: Defines the interface for sources that can be efficently accessed
   *   at random positions.
   *
   * StreamIO and RandomAccessIO does not have to be thread-safe*. Because we would need
   *   it in every single IO object :-(. I decided to move this into the FileStreamer.
   *   Since it should be used for all StreamIO objects (and locking is very important
   *   for it anyway) I decided to secure it. (See object chain below)
   *   * The objects will not be accessed from different threads at once. But if the
   *     implementations share static or other global data between instances (like
   *     the MADHelper) this HAS to be thread-safe!!!
   *
   *   !!! SO THREADED ACCESS TO OBJECTS BEFORE THE STREAMER ARE NOT ALLOWED !!!
   *
   * FileCache: A object that wraps a StreamIO source within a cache so that ramdonm
   *   access is not that painfully slow.
   * FileStreamer: Contains (dynamically managed) StreamIO objects so that seeks are
   *   avoided when different program-parts (like the graphic-views and IO) reads at
   *   different positions.
   *
   * DataFileAdapter: Adapts a StreamIO object to work as GSML DataFile. (needs a thread-
   *   safe source!!
   *
   * Normal object layout:
   *   DataFileAdapter -> FileStreamer -> FileCache -> StreamIO (e.g. OggVorbisStreamIO)
   *
   * Obejcts which get's another slave/child object in the ctor normally delete's it in
   *   the dtor - because normally all this stuff gets deleted via the GSML::DataFileIndex
   *   when undo-layer get's deleted or during program exit ... . These child-objects
   *   should normally be open!
   * 
   * (Speed? - Let's hope the GCC does got optimization .... ;-) And most of the time we
   * process more than 1000 frames at once anyway ...)
   */
  
  // interface for non-native file input backend
  class StreamIO
  {
  public:
    
    class Exception : public GSMP::Exception
    {
    public:
      Exception (const string& error);
    };
    
    virtual ~StreamIO () {};
    
    /* Control */
    virtual bool Open (const string& fname) = 0;
    virtual bool IsOpen () = 0;
    virtual bool Close () = 0;
    
    // clone - with the same state (open even when this instance is open)
    virtual StreamIO* Clone () = 0;
    
    /* Info */
    virtual GSML::sample_index Length () = 0;
    virtual GSML::track_index Tracks () = 0;
    virtual GSML::sample_rate Rate () = 0;
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index length, GSML::pcm_24_8bit::value_type* dest) = 0;
    virtual GSML::sample_index Seek (GSML::sample_index pos) = 0;
  };
  
  class RandomAccessIO
  {
  public:
    
    class Exception : public GSMP::Exception
    {
    public:
      Exception (const string& error);
    };
    
    class TrackInvaild : public Exception
    {
    public:
      TrackInvaild (const string& error);
    };
    
    virtual ~RandomAccessIO () {};
    
    /* Control */
    virtual bool Open (const string& fname) = 0;
    virtual bool IsOpen () = 0;
    virtual bool Close () = 0;
    
    // clone - with the same state (open even when this instance is open)
    virtual RandomAccessIO* Clone () = 0;
    
    /* Info */
    virtual GSML::sample_index Length () = 0;
    virtual GSML::track_index Tracks () = 0;
    virtual GSML::sample_rate Rate () = 0;
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index start, GSML::sample_index length,
				     GSML::track_index track, GSML::pcm_24_8bit::value_type* dest) = 0;
  };
  
  class FileCache : public RandomAccessIO
  {
  public:
    FileCache (StreamIO* i_src);
    virtual ~FileCache ();
    
    /* Control */
    virtual bool Open (const string& fname);
    virtual bool IsOpen ();
    virtual bool Close ();
    
    // clone - with neutral state (non-open even when this instance is open)
    virtual RandomAccessIO* Clone ();
    
    /* Info */
    virtual GSML::sample_index Length ();
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Rate ();
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index start, GSML::sample_index length,
				     GSML::track_index track, GSML::pcm_24_8bit::value_type* dest);
  private:
    
    GSML::sample_index m_length;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    
    GSML::SimpleCache <GSML::pcm_24_8bit::value_type, GSML::sample_index> m_cache;
    StreamIO* m_src;
  };
  
  class FileStreamer : public RandomAccessIO
  {
  public:
    FileStreamer (RandomAccessIO* i_src);
    virtual ~FileStreamer ();
    
    /* Control */
    virtual bool Open (const string& fname);
    virtual bool IsOpen ();
    virtual bool Close ();
    
    // clone - with neutral state (non-open even when this instance is open)
    virtual RandomAccessIO* Clone ();
    
    /* Info */
    virtual GSML::sample_index Length ();
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Rate ();
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index start, GSML::sample_index length,
				     GSML::track_index track, GSML::pcm_24_8bit::value_type* dest);
  private:
    
    // all the next are helpers for Read so they are globally locked
    inline void AddStream (RandomAccessIO* src);
    inline RandomAccessIO* BestStream (GSML::sample_index start, GSML::sample_index length);
    inline void AgeStreams (RandomAccessIO* except_stream = 0);
    inline void UpdateStream (RandomAccessIO* stream, GSML::sample_index start, GSML::sample_index length);
    
    // lock the stream read uses
    inline void LockStream (RandomAccessIO* stream);
    inline void UnlockStream (RandomAccessIO* stream);
    
    struct stream_stuff
    {
      RandomAccessIO* src;
      
      GSML::sample_index pos;
      GSML::sample_index cached_length;
      int age;
      
      bool locked;
      // no real lock needed becasue the tables are globally locked (any we can't iterate
      // over changing tables ...)
    };
    
    GSML::sample_index m_length;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    
    vector <stream_stuff> m_streams;
    
    SigC::Threads::Mutex m_global_lock;
    
    int m_max_streams;
    int m_max_age;
  };
  
  // Needs a THREAD-SAFE source!!!
  class DataFileAdapter : public GSML::DataFile
  {
  public:
    DataFileAdapter (RandomAccessIO* i_src, const string i_fname,  bool i_autodelete);
    virtual ~DataFileAdapter ();
    
    virtual void ReadData (GSML::sample_index start,
			   GSML::sample_index count,
			   GSML::track_index track, 
			   GSML::SampleBuffer* destination);
    
    virtual void WriteData (GSML::sample_index count,
			    GSML::track_index track, 
			    GSML::SampleBuffer* source,
			    GSML::sample_index buffer_index);
    
    virtual void FixateWrite ();
    
    virtual void SetFilename (string& new_filename);
    
  private:
    //SigC::Threads::Mutex m_lock;
    
    GSML::sample_index m_length;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    
    RandomAccessIO* m_src;
  };
  
  /* Generic Importer for StreamIO objects (impl. examples: OggVorbis and MP3)
   *
   * The inheritenced class only needs to implement the two pure virtual functions
   * Extension and InternalPrepare ...
   */
  
  class StreamIOImportFile : public ImportFile
  {
  public:
    StreamIOImportFile ();
    virtual ~StreamIOImportFile ();
    
    virtual string Extension () const = 0;
    
    virtual bool PrepareImport (const string& fname, bool temp_file_used);
    virtual void Close ();
    
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Samplerate ();
    virtual GSML::sample_index Length ();
    
    virtual bool Import (GSML::track_index track, GSML::TrackLayer* layer);
    
  protected:
    
    virtual StreamIO* InternalPrepare () = 0;
    
    // config
    string m_fname;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_length;
    
    // states
    bool m_prepared;

    DataFileAdapter* m_data_file;
    bool m_registerd_file;
  };
  
  /* Wav file implementation */
  
  class WavExportFile : public ExportFile
  {
  public:
    WavExportFile ();
    virtual ~WavExportFile ();
    
    virtual bool Create (const string& fname, GSML::track_index tracks, GSML::sample_index rate, GSML::sample_index length);
    virtual void Cancel ();
    virtual void Close ();
    
    virtual GSML::sample_index NextChunkSize ();
    virtual bool Write (GSML::track_index track, GSML::SampleBufferDefault* src, bool last_write);
    
  protected:
    
    // states
    
    int m_fd;
    GSML::pcm_16bit::value_type* interleaved_buffer;
    
    bool m_created_file;
    
    // config
    string m_filename;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_cur_length;
    GSML::sample_index m_length;
    
    int m_chunk_size;
  };
  
  class WavImportFile : public ImportFile
  {
  public:
    WavImportFile ();
    virtual ~WavImportFile ();
    
    virtual string Extension () const;
    
    virtual bool PrepareImport (const string& fname, bool temp_file_used);
    virtual void Close ();
    
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Samplerate ();
    virtual GSML::sample_index Length ();
    
    virtual bool Import (GSML::track_index track, GSML::TrackLayer* layer);
    
  protected:
    int m_fd;
    
    // config
    string m_filename;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    GSML::sample_index m_length;
    int m_head_length;
    
    // states
    bool m_prepared;

    GSML::DataFileRAW* m_data_file;
    bool m_registerd_file;
  };
  

#if EXT_FIO_OGGVORBIS == 1
  /* OggVorbis implementation !! ;-) */
  
  // the GSML realtime access implementation
  class OggVorbisStream : public StreamIO
  {
  public:
    OggVorbisStream ();
    virtual ~OggVorbisStream ();
    
    /* Control */
    virtual bool Open (const string& fname);
    virtual bool IsOpen ();
    virtual bool Close ();
    
    // clone - with the same state (open even when this instance is open)
    virtual StreamIO* Clone ();
    
    /* Info */
    virtual GSML::sample_index Length ();
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Rate ();
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index length, GSML::pcm_24_8bit::value_type* dest);
    virtual GSML::sample_index Seek (GSML::sample_index pos);
    
  protected:
    
    GSML::sample_index m_length;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    
    GSML::sample_index m_stream_pos;
    
    string m_fname;
    FILE* m_file;
    
    int m_vorbis_section;
    
    OggVorbis_File m_vorbis_file;
  };
  
  // the Importer
  class OggVorbisImportFile : public StreamIOImportFile
  {
  public:
    OggVorbisImportFile () {};
    virtual ~OggVorbisImportFile () {};
    
    virtual string Extension () const {
      return "ogg";
    }
    
  protected:
    virtual StreamIO* InternalPrepare () {
      return new OggVorbisStream ();
    }
  };

#endif // EXT_FIO_OGGVORBIS == 1

  /* MP3 implementation !! ;-) */

#if EXT_FIO_MP3 == 1

  class MP3Stream : public StreamIO
  {
  public:
    MP3Stream ();
    virtual ~MP3Stream ();
    
    /* Control */
    virtual bool Open (const string& fname);
    virtual bool IsOpen ();
    virtual bool Close ();
    
    // clone - with the same state (open even when this instance is open)
    virtual StreamIO* Clone ();
    
    /* Info */
    virtual GSML::sample_index Length ();
    virtual GSML::track_index Tracks ();
    virtual GSML::sample_rate Rate ();
    
    /* Operations */
    virtual GSML::sample_index Read (GSML::sample_index length, GSML::pcm_24_8bit::value_type* dest);
    virtual GSML::sample_index Seek (GSML::sample_index pos);
    
  protected:
    
    bool m_open;
    MADHelper m_mad;
    
    GSML::sample_index m_length;
    GSML::track_index m_tracks;
    GSML::sample_rate m_rate;
    
    GSML::sample_index m_stream_pos;
    
    string m_fname;
  };
  
  // the Importer
  class MP3ImportFile : public StreamIOImportFile
  {
  public:
    MP3ImportFile () {};
    virtual ~MP3ImportFile () {};
    
    virtual string Extension () const {
      return "mp3";
    }
    
  protected:
    virtual StreamIO* InternalPrepare () {
      return new MP3Stream ();
      
    }
  };

# endif // EXT_FIO_MP3 == 1

} // end namespace GSMP

#endif // GSMP_EXTFILEIO_IMPL_H
