/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Base class for FXs.
 */

#ifndef GSMP_FX_BASE
#define GSMP_FX_BASE

#include "Threads.hh"
#include "Data.hh"

namespace GSMP
{
  class Project;

  class FXBase : public GSMX_utils::Threads::Job
  {
  public:
    FXBase (const string& info, const string& comment);
    virtual ~FXBase () {}

    bool Start (Project* project, string& error_string);
    void Cancel ();

    bool PollProgressData (); // retruns true if still running

  private:

    double old_progress;

  protected:
    GSML::DataLayer* src_layer;
    GSML::DataLayer* dst_layer;

    void SetReadOnly (); // no dst_layer will be prepared on start-up

    virtual bool Start_impl (Project* project, string& error_string);

    // Communication with FX Thread
    virtual void Canceled () {}; // Here an implementation may do some cleanup etc..

    volatile double progress;
    volatile bool finished;

    Project* m_project;

    string m_info;
    string m_comment;
    GSML::Data::ProgressInfo::Info* info;
  private:
    GSMX_utils::Threads::Function0<int> DoCancel;
   
    // FX Thread Side
    int DoCancel__ ();

    bool read_only;
  };


  // FX

  class LinearMonoFXPlug : public SigC::Object
  { // Streamable FX
  public:
    GSML::pcm_24_8bit Stype;
    typedef GSML::pcm_24_8bit::value_type Sdata;

    struct Setup
    {
      bool generator;                     // = false
      bool do_declick;                    // not impl yet !
      GSML::sample_index declick_samples; // not impl yet !
      GSML::sample_index queue_prefill;   // = 0
    };

    Setup setup;

    LinearMonoFXPlug ();
    virtual ~LinearMonoFXPlug () {};

    // to be implemented by Derived Class

    // Called each time a new area is to be processed
    virtual void Prepare (Sdata* queue_prefill,                // Data for queue-Prefilling
			  GSML::sample_index area_length,     // Length of the Area
			  GSML::sample_index area_offset,     // Start in Area (may not be zero in RT IO !)
			  GSML::sample_index real_pos) {};    // real Position in Track

    virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset) = 0;
  };


  // Base Classes for some simple effects

  class LinearMonoFX : public FXBase
  { // Iterates all selected segments on all Tracks, automatically manages progressbar
  public:
    LinearMonoFX (const string& info, const string& comment);
    virtual ~LinearMonoFX ();

    void SetBlockSize (GSML::sample_index block_length);
    void SetPrefetch (GSML::sample_index prefetch_samples, bool silence_at_start);
    void SetPostfetch (GSML::sample_index postfetch_samples, bool silence_at_end);
    void SetGenerator (bool state); // Write-only FX
    void SetSelector (bool state); // Read-Only (auto selection etc)
    void SetDeclick (bool do_declick, GSML::sample_index declick_samples); // Level segment borders

  protected:
    virtual void PrepareFX (GSML::sample_index length) {}; // Called on new Area
    virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length) = 0;

    void FinishArea (); // Skip processing of remaing Data in Area
    GSML::sample_index Offset (); // returns current offset  
  private:
    class Area
    {
    public:
      GSML::TrackLayer* src_layer;
      GSML::TrackLayer* dst_layer;
      GSML::sample_index start;
      GSML::sample_index length;
    };

    virtual void Loop ();
    virtual bool Start_impl (Project* project, string& error_string);

    GSML::sample_index m_block_size;
    GSML::sample_index m_prefetch;
    GSML::sample_index m_postfetch;
    GSML::sample_index m_declick;
    bool generator;
    bool selector;
    bool declick;

    // Progress calculation stuff
    GSML::sample_index sum_length;
    GSML::sample_index processed_length;

    // Processing
    vector <Area> area_list;
    vector <Area>::const_iterator current_area;
    GSML::sample_index current_offset;
    GSML::SampleBufferDefault* m_sample_buffer;
  };

  class LinearMonoFXPlugWriter : public LinearMonoFX
  {
  public:
    LinearMonoFXPlugWriter (LinearMonoFXPlug* plug, const string& info, const string& comment);
    virtual ~LinearMonoFXPlugWriter ();

  protected:
    virtual void PrepareFX (GSML::sample_index length);
    virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

    LinearMonoFXPlug* m_plug;
  };

} // end namespace GSMP

#endif // GSMP_IO_HH
