/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *
 * Abstact (non-Widgetset depending) dynamic menu system.
 * Very, very messy indeed. To be COMPLETELY replaced by a more universal and coderfriendly
 * implementation till v. 0.1.x !
 *
 */

#ifndef GSMP_MENU_HH
#define GSMP_MENU_HH

#include <string>
#include <deque>
#include <map>
#include <algorithm>
#include <vector>

#include <sigc++/signal_system.h>

#include "Base.hh"
#include "Activator.hh"

#include "FindMemCorruption.hh"

// OVERALL TODO: wrap type specific stl iterators with class-scoped typedefs !
// OVERALL TODO: use more derived/generic base classes and generalized APIs for all kinds of Items / Finders !

/********************************************************************************************
 *
 * activator string format
 *
 * forbiden chars in activator strings:
 * '\n' '\0' and any other kind of control characters.
 *
 * reserved chars are
 * ':' '>' '$' '&' '='
 * these chars must be escaped with '$' (eg. "$:") when they should be 
 * treated as "normal" text chars; Hint: "$$" actualy means '$' 
 *  
 * format:
 * Location ':' Scope ':' Type [ ':' Property '=' value {'&' Property '=' value} ]
 *
 *
 *
 ********************************************************************************************

 ********************************************************************************************
 *
 * <Location> consists of a list of <Entity> strings:
 * 
 * <Entity> string format:
 *
 * name ['&' group ['&' priority]]
 *   name and group may be any string (note treatment of reserved chars !)
 *   if not applied, group is assigned "zero".
 *   priority is a number base ten, with or without '-' sign.
 *   Ths number should have max 4 digits. If not applied, priority 
 *   is assigned 0.
 *   Entities are first sorted after groups, (which are separated by seperator rows in 
 *   menus), then after their priority (where a lower value means higher priority), 
 *   then alphabeticaly after names.
 *   note that groups or prioritys for node-Entities (aka "submenus")
 *   which already exist, will be ignored. 
 *
 * the lists items are separated with a '>' char;
 * the last item in this list is refered to as "leave", the heading items are refered to as
 * nodes. Every Location must have at least one node. 
 *
 * Exceptions: DuplicateLocation, MalformedLocation
 *
 * No two same locations must apear whithin one MenuManager context.
 * If a new activator is registered with an existing Location,
 * a DuplicateLocation Exception will be thrown.
 * A Malformed Location string will result in throwing an MalformedLocation Exception. This
 * is also the case if there is no node (which means, there is no '>' character).
 * In case a node-Entity (aka "submenu") appears, which is already registered as a leave
 * or vice-versa, a dublicateLocation exception will be thrown.
 *
 *********************************************************************************************

 *********************************************************************************************
 * Scope:
 *
 * one out of "App", "Project", "View"
 *
 * Exceptions: InvalidScope
 *
 *********************************************************************************************

 *********************************************************************************************
 *
 * Types:
 *
 * "Item"
 *     normal menu item, will be represented by SimpleItem class. Appears
 *     in menus with the name of its last entity.
 * 
 *     optional Properties:
 *     "Icon"  filename
 *          Allows item to be displayed in Toolbars. Default is "", 
 *          which means no icon is used (so the item can not appear in toolbars)
 *     "State"  ( "on" | "off" )
 *          determines if a menu item is enabled or blocked. Default is "on"
 *
 * "Mutator"
 *     an item which allows its text / icon to be changed during execution.
 *     represented by class MutatorItem.
 *
 *     optional Properties:
 *     "Icon"  filename
 *          Allows item to be displayed in Toolbars. Default is "".
 *     "State"  ( "on" | "off" )
 *          determines if a menu item is enabled or blocked. Default is "on"
 *
 * "Checked"
 *      items which may be checked, represented by class CheckedItem.
 *
 *      optional Properties:
 *      "Checked"  ("checked" | "unchecked"). 
 *          Default is "unchecked".
 *
 *      "State"  ( "on" | "off" )
 *          as with Item. Default is "on"
 *
 *
 * "Selection"
 *      theese items offer a choice "one out of a group" to the user.
 *      eg. Waveplot selection. Represented by class SelectionItem.
 *
 *      required Properties:
 *      "Set" setname
 *         the selection will be aggregated by this set (represented by class SelectionSet)
 *          which ensures there is only one item in this set activated at a time.
 *
 *      optional Properties:
 *      "State"  ( "on" | "off" )
 *          determines if the item is enabled or blocked. Default is "on"
 *
 *
 *      HINT: BULLSHIT !! TODO : Correct Documentation for "Selected"
 *      "Selected" ( "yes" | "no")
 *          given "yes", this item will be selected by default. If there are more than one
 *          DelectionItems claiming to be selected by default, the menusystem chooses one.
 *          NOTE: The Item which is choosen to be selected will be automatically invoked
 *          each time a Project/View is registered - depending on Scope. With Scope="App"
 *          this Item will be invoked when added to the database. 
 *
 *
 *
 * To be implemented In far Future:
 * some kind of "history" list.
 *
 *
 *
 * optional Properties for all items:
 *
 * "Tooltip" short_description_string
 * "Shortcut" gtk_style_shortcut
 *
 * Default for both is <nil>.
 *
 *
 * Exceptions: UnknownType, MalformedProperty, InvalidPropertyValue, PropertyRequiered, 
 * GroupMismatch.
 * the SetMissmatch exception will be thrown in case two SelectionItems whithin
 * one Set have different scopes.
 * Unknown Properties will be silently ignored.
 *
 ***************************************************************************************

 ***************************************************************************************
 *
 * Examples:
 *
 * simple...
 *
 * File>Exit:App:Item
 *
 * ...and complex:
 *
 * stuff&nonsense>useless&something&-23>select me:Project:Selection:Tooltip=Oh, please click $& select me !!&Set=braindead&state=on
 * 
 *************************************************************************************** 

 ***************************************************************************************
 *
 * simple Location string format (for searching/filtering)
 *
 * List of Entity names, seperated by '>'. A simple Location string may consist of one
 * single Entity name. (In fact, class SimpleLocation is just a Specialisation of class 
 * SplitString (string, '>', true))
 *
 ***************************************************************************************/

namespace GSMP
{
  /* some forwards, all of these MUST be derived from MenuData */
  class App;
  class Project;
  class View;

  namespace Menu
  {

    /* Base Exception for Menu stuff */

    class Exception : public GSMP::Exception
    {
    public:
      Exception (const string& error);
    };

    /***** activator exception classes. See activator string description above for more info. *****/ 

    class MalformedLocation : public Exception
    {
    public:
      MalformedLocation (const string& the_location);
    };

    class DuplicateLocation : public Exception
    {
    public:
      DuplicateLocation (const string& the_location);
    };

    class MalformedProperty : public Exception
    {
    public:
      MalformedProperty (const string& the_property);
    };

    class InvalidPropertyValue : public Exception
    {
    public:
      InvalidPropertyValue (const string& the_property, const string& the_value);
    };

    class PropertyRequiered : public Exception
    {
    public:
      PropertyRequiered (const string& expected_property, const string& for_type);
    };

    class InvalidScope : public Exception
    {
    public:
      InvalidScope (const string& the_scope);
    };

    class UnknownType : public Exception
    {
    public:
      UnknownType (const string& the_type);
    };

    class SetMismatch : public Exception
    {
    public:
      SetMismatch (const string& the_set);
    };

    class IteratorException : public Exception
    {
    public:
      IteratorException ();
    };

    class UninitializedMenuData : public Exception
    {
    public:
      UninitializedMenuData ();
    };

    /***** Here are the classes that are parsed from a location string *******/

    class Entity
    {  // represents an entity.
    public:
      Entity (const string& init_entity);

      string name;
      string group;
      int sorting_priority;

      bool IsLess (const Entity& than) const; // comparism for sorting.

    private:
      string my_string;
      // creates an Entity which will always been sorted to bottom of group.
      Entity (const string& seperator_group, int ignored);
      friend class SeperatorItem;
    };



    class Location
    {  // represents a location
    public:
      Location (const string& path_string);

      deque <Entity>::const_iterator Root ();
      deque <Entity>::const_iterator Leave ();
    private:
      deque <Entity> entities;
      string my_string;
    };




    /*********************************** All item types ********************************/

    class SubMenuItem;
    class MenuManager;

    class Item
    { //base class
    public:
      Item (SubMenuItem* init_parent, const Entity& init_params);

      void Debug () const; // does some debugging output on stdout.
      virtual void DebugTypeSpecific () const = 0; // called via Debug (). Prints type specific debuging info. 

      string Location () const; // location string w.o groupings and priorities ("SimpleLocation string format")

      SubMenuItem*  my_parent;
      Entity my_params;
    };

    class LeaveItem : public Item, public SigC::Object
    { // an item that can be "activated" and has a reactor.
    public:

      enum Scope
	{
	  APP,
	  PROJECT,
	  VIEW
	};

      class Property
      { // represents a property;
      public:
	Property (const string& init_property);
	Property (const Property& copy);
	Property () {};
	string id;     // part before '='
	string value;  // part behind '='
      };

      LeaveItem (SubMenuItem* init_parent, const Entity& init_params, 
		 const Scope& init_scope, const string& init_properties);

      void DebugLeaveItem () const; //  called by higher level instances

      virtual void Activate (View* view) const = 0; // when user clicks in <view>, call this...

      Scope my_scope;
      string tooltip;
      string shortcut;

      string reactor_category;
      string reactor_name;
 
      map <string, Property> my_properties;

    private:
      void ParseProperties (const string& property_part); // called by constructor only.   
    };

    class SimpleItem : public LeaveItem
    { // normal menu item
    public:
      SimpleItem (SubMenuItem* init_parent, const Entity& init_params, 
		  const Scope& init_scope, const string& init_properties);

      bool IsEnabled (const View* here) const;
      void SetState (View* there, bool state) const;

      void BlockAll () const;     // temp. disabeling in every View
      void UnBlockAll () const;   // "restores" old state

      virtual void DebugTypeSpecific () const;
      virtual void Activate (View* view) const;
  
      void Invoke (View* view) {Activate (view);}

    private:
      friend class SimpleItemFinder;
      friend class MenuCollection;

      mutable SigC::Signal1 <void, const SimpleItem*> SignalEnabledChanged;

      bool default_state;
      mutable bool blocked;
    };


    class MutatorItem : public LeaveItem
    { // menu item which may change
    public:
      MutatorItem (SubMenuItem* init_parent, const Entity& init_params, 
		   const Scope& init_scope, const string& init_properties);
    };

    class CheckedItem : public LeaveItem
    { // item which may be "checked"
    public:
      CheckedItem (SubMenuItem* init_parent, const Entity& init_params, 
		   const Scope& init_scope, const string& init_properties);

      bool IsChecked (const View* here) const;
      bool SetCheck (View* there, bool state) const;
      void ToggleCheck (View* there) {Activate (there);} // result not returned

      bool IsEnabled (const View* here) const;
      void SetState (View* there, bool state) const;

      void BlockAll () const;     // temp. disabeling in every View
      void UnBlockAll () const;   // "restores" old state

      virtual void DebugTypeSpecific () const;
      virtual void Activate (View* view) const {SetCheck (view, !IsChecked (view));}


    private:
      friend class CheckedItemFinder;
      friend class MenuCollection;

      mutable SigC::Signal1 <void, const CheckedItem*> SignalCheckedChanged;
      mutable SigC::Signal1 <void, const CheckedItem*> SignalEnabledChanged;

      void SetCheck__ (View* here, bool state) const;

      bool default_checked;
      bool default_state;
      mutable bool blocked;

      StatusActivator* Parasite (View* in_view) const;
      void InsertParasite (View* in_view, StatusActivator* parasite) const;
      void RemoveParasite (View* from_view) const;
    };

    class SelectionItem : public LeaveItem
    { // a "one out of many" selectable item
    public:
      SelectionItem (MenuManager* i_mm, SubMenuItem* init_parent, const Entity& init_params, 
		 const Scope& init_scope, const string& init_properties);

      bool IsSelected (const View* here) const; 
      bool Select (View* view) const; // returns true on success

      virtual void DebugTypeSpecific () const;
      virtual void Activate (View* view) const {Select (view);}

    private:
      friend class SelectionItemFinder;
      friend class MenuCollection;
      friend class Set;

      string m_set;

      mutable SigC::Signal1 <void, const SelectionItem*> SignalSelectionChanged;
    };

    /* generic items */

    class SeperatorItem : public Item
    { // group seperator, for gui implementations only
    public:
      SeperatorItem (SubMenuItem* init_parent, const string& groupname);  

      virtual void DebugTypeSpecific () const;
      virtual void Activate (View* view);
    };


    class SubMenuItem : public Item
    {
    public:
      SubMenuItem (SubMenuItem* init_parent, const Entity& init_params);

      void Insert (Item* this_item); // also generates group seperators if nec.

      virtual void DebugTypeSpecific () const;

      map <string, Item*> my_named_items; // no seperators, indexed by name to find existing ones.
      deque <Item*> my_items;             // all items sorted including group seperators;
      map <string, int> my_groups;        // second pair value indicates how many items (including
                                          // the group seperator) are within that group
    };

    /********************* Marshalling Interface *****************************/

#   ifndef OBSOLETE_CODE 

    class Notifier
    {
    public:
      /* Signals, that will be emited every time an item has been added.
       * First pointer references the new item, the second points to the highest
       * submenu that is affected by that insertion. */

      SigC::Signal2 <void, SimpleItem*, SubMenuItem*> SimpleItemAdded;
      SigC::Signal2 <void, MutatorItem*, SubMenuItem*> MutatorItemAdded;
      SigC::Signal2 <void, CheckedItem*, SubMenuItem*> CheckedItemAdded;
      SigC::Signal2 <void, SelectionItem*, SubMenuItem*> SelectionItemAdded;

      /* State of SimpleItem changed, boolean parameter indicates new state */
      SigC::Signal2 <void, SimpleItem*, bool> SimpleItemStateChanged;
    };

#   endif

    /********************************************* Helper Objects **************************/

    class SplitString
    { // splits a string into substrings at each occurence of <c>. 
      // Ignores '$'-escaped characters
    public:
      SplitString (const string& s, char c, bool remove_escapes);
      
      inline deque<string>::const_iterator begin () const {return substrings.begin ();}
      inline deque<string>::const_iterator end () const {return substrings.end ();}
      inline deque<string>::size_type size () const {return substrings.size ();};

    protected:
      deque <string> substrings;
    };

    class SimpleLocation : public SplitString
    {
    public:
      SimpleLocation (const string& location) : SplitString (location, '>', true), m_string (location) {};
      const string m_string;

      SimpleLocation (const SimpleLocation& i_location) : SplitString (i_location.m_string, '>', true), m_string (i_location.m_string) {}; 
    };


    class IsAGreaterItem : public unary_function <Item*, bool>
    { //this is a usefull predicate for stl algorithms
    public:
      explicit IsAGreaterItem (Item* than) {compare_against = than;}
      inline bool operator () (const Item* i) const {return compare_against -> my_params.IsLess (i -> my_params);}

    private:
      Item* compare_against; // initialized by <than>
    };


    class MenuManager;
    class Set;

    class MenuData : public SigC::Object
    {
    protected:
      MenuData () {initialized = false;}
      virtual ~MenuData ();

      MenuData (const MenuData& old_data);
    private:
      friend class MenuManager;
      friend class SimpleItem;
      friend class CheckedItem;
      friend class MutatorItem;
      friend class SelectionItem;

      LeaveItem::Scope m_scope;
      bool initialized;

      void TestNewSet (string name, Set* set);

      mutable map <string, bool> states;
      mutable map <string, Set*> selections;
      mutable map <string, StatusActivator*> parasites; 

      SigC::Signal1<void, MenuData*> SignalDeleted;
    };


    /**************************** Gui Helpers ********************************/

    class Filter;


    class ItemPath
    { // Path through global tree
    public:
      ItemPath (const SubMenuItem& from, const Item& to);
      bool Matched () const {return (begin () != end ()); }

      typedef vector <int> :: const_iterator const_iterator;
      const_iterator begin () const {return subitem_no.begin ();}
      const_iterator end () const {return subitem_no.end ();}      
    private:
      vector <int> subitem_no;
    };


    class SimpleItemFinder : public SigC::Object
    { // Could be used for buttons etc
    public:
      SimpleItemFinder (const MenuManager& i_menumanager, const string& location);
      bool Matched () {return (m_item);};

      SigC::Signal0 <void> SignalMatched;
      SigC::Signal0 <void> SignalStateChanged;

      const SimpleItem* Get () {return m_item;} // May be 0 !
    private:
      void HandleSignals (const SimpleItem* ignored);

      const SimpleItem* m_item;
      const MenuManager& m_menumanager;
    };

    class CheckedItemFinder : public SigC::Object
    { // Could be used for checkboxes etc
    public:
      CheckedItemFinder (const MenuManager& i_menumanager, const string& location);
      bool Matched () {return (m_item);};

      SigC::Signal0 <void> SignalMatched;
      SigC::Signal0 <void> SignalCheckedChanged;
      SigC::Signal0 <void> SignalStateChanged;

      const CheckedItem* Get () {return m_item;} // May be 0 !
    private:
      void HandleSignalCh (const CheckedItem* ignored);
      void HandleSignalEn (const CheckedItem* ignored);

      const CheckedItem* m_item;
      const MenuManager& m_menumanager;
    };


    class SelectionItemFinder : public SigC::Object
    { // Could be used for radiobuttons etc
    public:
      SelectionItemFinder (const MenuManager& i_menumanager, const string& location);
      bool Matched () {return (m_item);};

      SigC::Signal0 <void> SignalMatched;
      SigC::Signal0 <void> SignalSelectionChanged;

      const SelectionItem* Get () {return m_item;} // May be 0 !
    private:
      void HandleSignals (const SelectionItem* ignored);

      const SelectionItem* m_item;
      const MenuManager& m_menumanager;
    };


    class MenuCollection : public SigC::Object
    { // this is a class for managing a collection of menus. The GUI may use this for creating a Menubar
      // (or a Part of a Menubar - idealy two collections are used, one for the menus to the left and one
      // for the menus to the right)
      // This class may even be usefull for a single menu, eg a context menu from multiple filters
    
    public:

      class Entry
      {
      public:
	const SubMenuItem* SubMenu () const {return m_submenu;}
	SubMenuItem* SubMenu () {return m_submenu;}

	const string& Title () const {return m_title;}
      private:
	friend class MenuCollection;

	Entry (const Filter& i_filter);
	SubMenuItem* m_submenu;
	string m_title;
      };

      typedef deque <Entry> :: const_iterator const_iterator;
      typedef deque <Entry> :: iterator iterator;


      int size () {return m_entries.size ();} // Number of Submenuitems
      const_iterator begin () const {return m_entries.begin ();}
      const_iterator end () const {return m_entries.end ();}
      
      iterator begin () {return m_entries.begin ();}
      iterator end () {return m_entries.end ();}
      


      SigC::Signal0 <void> SubMenusChanged;
      // first argument are times iterator begin () has to be incremented to get submenu.
      SigC::Signal3 <void, int, ItemPath*, const SimpleItem&> SignalSimpleItemAdded;
      SigC::Signal3 <void, int, ItemPath*, const SelectionItem&> SignalSelectionItemAdded;
      SigC::Signal3 <void, int, ItemPath*, const CheckedItem&> SignalCheckedItemAdded;

      SigC::Signal3 <void, int, ItemPath*, const SimpleItem&> SimpleItemStateChanged;
      SigC::Signal3 <void, int, ItemPath*, const SelectionItem&> SelectionChanged;
      SigC::Signal3 <void, int, ItemPath*, const CheckedItem&> SignalCheckedChanged;
      SigC::Signal3 <void, int, ItemPath*, const CheckedItem&> CheckedItemStateChanged;
      
      void Debug ();

      ~MenuCollection ();
    private:
      friend class MenuManager;
      MenuCollection (/*View* i_view,*/ const MenuManager& i_menumanager, deque <Filter> i_filters, bool append_unknown);

      // Private Slots
      void HandleSignalEC (const SimpleItem* item);
      void HandleSignalSC (const SelectionItem* item);
      void HandleSignalCI (const CheckedItem* item);
      void HandleSignalCS (const CheckedItem* item);


      bool append; // append_unknown
      deque <Entry> m_entries;
      deque <Filter> m_filters;
      //View* m_view;
      const MenuManager& m_menumanager;
    };

    class ToolBar
    { // Manages a "toolbar"
      ToolBar (View* init_view, deque <Filter> content, bool append_unknown );
    };


    class Filter
    {
    public:
      Filter (const string& location, const string& alias_title);
      bool Matched () const; // becomes true whenever matched a submenu

    private:
      friend class MenuCollection;
      friend class MenuCollection::Entry;
      friend class ToolBar;

      bool Try (const LeaveItem& match);      // try to match item, return true on success
      bool InSubTree (const LeaveItem& item); // Item in subtree ? (calls Try if not matched yet...)

      SimpleLocation my_location;
      string alias;
      SubMenuItem* my_submenu; // reference to node in ManuManager's global tree
    };



    /************************* Global MenuManaging *****************************/ 

    class Set
    {
      const SelectionItem* GetDefault () const {return default_selection;}
      const SelectionItem* GetCurent () const {return current_selection;}
    private:
      friend class MenuManager;
      friend class SelectionItem;
      friend class MenuData;

      Set (LeaveItem::Scope i_scope, const SelectionItem* i_default);
      Set (const Set&); // ctor, also activates default_selection

      void ChangeCurrentSelection (const SelectionItem* to_selection);

      LeaveItem::Scope m_scope;
      const SelectionItem* default_selection;
      const SelectionItem* current_selection;
    };

    class MenuManager : public SigC::Object
    {
    public:
      MenuCollection* CreateMenuCollection (/*View* init_view,*/ deque <Filter> i_filters, bool append_unknown);
      // if append_unknown is set, all the items which can not be sorted,
      // appear in menus to the right of the left_side menus
      // as if they where initialized with a 
      // Filter (<top level node name>, <top level node name> )

      ToolBar* CreateToolbar (View* init_view, deque <Filter> content, bool append_unknown );

      MenuManager ();

      class const_iterator
      { // Simple unidirectional iterator over all LeaveItems in global Tree;
      public:
	void operator ++ ();
	const LeaveItem* operator * ();

	const_iterator (const const_iterator&);

	bool operator == (MenuManager::const_iterator);
	bool operator != (MenuManager::const_iterator);

      private:
	friend class MenuManager;
	const_iterator (SubMenuItem* i_root);

	deque <deque <Item*> :: const_iterator> positions;
	SubMenuItem* m_root;
      };

      const_iterator begin () const;
      const_iterator end () const;


      /* Signals, that will be emited every time an item has been added.
       * First pointer references the new item, the second points to the highest
       * submenu that is affected by that insertion. */

      SigC::Signal2 <void, SimpleItem*, SubMenuItem*> SimpleItemAdded;
      SigC::Signal2 <void, MutatorItem*, SubMenuItem*> MutatorItemAdded;
      SigC::Signal2 <void, CheckedItem*, SubMenuItem*> CheckedItemAdded;
      SigC::Signal2 <void, SelectionItem*, SubMenuItem*> SelectionItemAdded;

      /* State of SimpleItem changed, boolean parameter indicates new state */
      //SigC::Signal2 <void, SimpleItem*, bool> SimpleItemStateChanged;

      class LeaveItemFactory
      {
      public:
	virtual LeaveItem* Create (SubMenuItem* init_parent, const Entity& init_params, 
				   const LeaveItem::Scope& init_scope, const string& init_properties) = 0;
	virtual void Signal (MenuManager* through_this, LeaveItem* about_that, SubMenuItem* first_affected) = 0;
      protected:
	LeaveItemFactory (MenuManager* i_mm) {m_mm = i_mm;}

	MenuManager* m_mm;
      };


#ifndef GSMP_TESTING
    protected:
#endif // Allow protected member access in test-environment

      void Add (const string& activator, const string& reactor_category, const string& reactor_name);
      
      void Delete (const string& activator);
      void Delete (const string& reactor_category, const string& reactor_item);

      void InitializeMenuData (App* new_app);
      void InitializeMenuData (Project* new_project);
      void InitializeMenuData (View* new_view);
      void InitializeMenuDataCopy (View* new_view);

      void DebugMenuManager () const;
      
    private:
      friend class Filter;
      friend class MenuCollection;
      friend class ToolBar;
      friend class MenuData;
      friend class SelectionItem;

      SubMenuItem* root; // this is the absolute root of the location tree;
      map <string, LeaveItem::Scope> known_scopes;
      map <string, LeaveItemFactory*> known_item_types;

      void InsertSet (string set_name, Set* the_set);
      SigC::Signal2 <void, string, Set*> SignalSetInserted;

      // personal Slots
      void MenuDataDeleted (MenuData* menu_data);
      map <string, Set*> all_sets; // global db
      vector <View*> all_views;

      FindMemCorruption m_fmc;
    };


  } // end namespace Menu;
} // end namespace Gsmp;

#endif //  GSMP_MENU_HH

