/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   The GSMP Project class.
 */

#ifndef GSMP_PROJECT
#define GSMP_PROJECT

#include <deque>

// GSML stuff:
#include "Data.hh"
#include "WavePlot.hh"
#include "IO.hh"
#include "IO_impl.hh"

// GSMP stuff:
#include "Menu.hh"
#include "Window.hh"
#include "FXMixerSource.hh"

#include "FindMemCorruption.hh"


namespace GSMP
{
  /* marshaller (for SigC++ signals) which cancels the signal-emitting process when a Slot retuns false,
   * and returns whether all Slots returned true if it got canceled by a Slot's false
   * (thanks to the nice iX (www.heise.de/ix) article about the SigC++ ;-)
   */
  template <class T> class GUIRequestMarshaller
  {
  public:
    typedef T InType;
    typedef T OutType;
    
    GUIRequestMarshaller () { m_value = true; }
    
    static T default_value () { return true; }
    
    // signal-emitting process gets canceled when we return true!
    bool marshal (const InType& in)
      {
	m_value = in;
	return !m_value;
      }
    
    OutType value () { return m_value; };
    
  private:
    OutType m_value;
  };
  
  class ParameterDialog;
  class ParameterDialogAction;
  class EnhParameterDialog;
  class DialogAction;
  
  class Project : public Menu::MenuData
    {
    public:

      class Exception : public GSMP::Exception
      {
      public:
	Exception (const string& error);
      };
    
      class ViewNotFound : public Exception
      {
      public:
	ViewNotFound ();
      };

    private:
      friend class App;

      Project (App* i_app, GSML::sample_rate i_sample_rate, int i_track_no,
	       int i_max_datalayers, const string& i_name);
      ~Project ();
    
    public:

      class MessageInfo
      { // manages Project specific message strings that may be displayed by GUI components
	// we use a deque instead of a queue here to allow iteration over buffered messages. 
      public:
	
	typedef deque <string>::iterator iterator;
	typedef deque <string>::const_iterator const_iterator;
	
	MessageInfo (int i_max_messages);

	const string& GetMessage ();      // get newest message
	void PutMessage (const string& message);
	
	const_iterator begin (); // iterator to oldest message stored in queue
	const_iterator end ();

	SigC::Signal0 <void> SignalUpdated;
      private:
	deque <string>   m_messages;
	int              max_messages; // maximum no of messages to be stored in queue
      };
      
      typedef deque <View*>::iterator view_iterator;

      App* GetApp () {return m_app;}
      const App* GetApp () const {return m_app;}

      GSML::WavePlotManager* GetWavePlotManager ();
      const GSML::WavePlotManager* GetWavePlotManager () const;

      MessageInfo message_info;

      // Projects Name
      const string& Name ();
      void SetName (const string& name);

      MainWindow* AnyMainWindow (bool always); // returns a MainWindow, where the Project is visible
                                    // if there is no MainWindow displaying a View of the Project, 
                                    // and <always> is true, a new MainWindow will be created.
      SimpleWindow* AnyWindow (bool always); // returns Main- or SimpleWindow as above. If there is none, 
                                  // and <always> is true, a new SimpleWindow is constructed

      GSML::UniversalMixerParams* m_pl_mixer_params; // Playback
      GSML::UniversalMixerParams* m_ca_mixer_params; // Capture

      GSML::Mixer* GetMixer ();

      int Views (); // # of Views
      int VisibleViews (); // # of visible Views
      int OrphanedViews (); // # of orphaned Views
      
      view_iterator BeginViews ();
      view_iterator EndViews ();

      view_iterator BeginVisibleViews ();
      view_iterator EndVisibleViews ();

      view_iterator BeginOrphanedViews ();
      view_iterator EndOrphanedViews ();

      // Functions to erase Views. Will also kill parent windows.

      bool EraseView (View* del_view);
      view_iterator EraseView (const view_iterator& del_view);

      // SigC::Signal1<void, View*> SignalViewCreated; // counterpart is in project
      // Check WindowCreated Sig instead !    

      GSML::DataSource& DataSource ();
      FXMixerSource& FXFilter ();
      GSML::SelectionFilterSource& SelectionFilter ();
      GSML::MultiplexerFilterSource& MultiplexerFilter ();
      GSML::MarkerFilterSource& MarkerFilter ();
      GSML::LoopFilterSource& LoopFilter (); 

      GSML::MixerSource* GetToplevelSource ();

      void DoFX (FXBase* fx);
      SigC::Handle <ParameterDialog, SigC::Scopes::Extend> FXConfigDialog (ParameterDialogAction* action, string title);
      SigC::Handle <EnhParameterDialog, SigC::Scopes::Extend> FXConfigDialog (DialogAction* action, string title);

      void Cancel (); // Cancels FX if Running

      // Conversion of pos into Time
      string Time (GSML::sample_index pos);
 
      // Functions, which move io_pos marker
      void StartPlaybackTracking (GSML::sample_index expected_latency, const string& stopitem);
      void StopPlaybackTracking ();

      // Time & Position Tracking
      GSML::sample_index PlaybackPos ();
      string PlaybackTime (); 
      SigC::Signal0 <void> SignalTimeRefresh;

      SigC::Signal0 <void> SignalNameChanged;
      SigC::Signal0 <bool, GUIRequestMarshaller<bool> > SignalEraseRequest;  // someone is going to kill this project, preventable by returning false.
      SigC::Signal0 <void> SignalErased; // project will be killed now

      // Wrapped GSML::Data Methods

      inline GSML::sample_index Length () const {return m_data . Length ();}
      inline GSML::track_index Tracks () const {return m_data . Tracks ();}
      
      inline GSML::DataLayer* CurrentLayerPtr () {return m_data . CurrentLayerPtr ();}
      inline int Layers () {return m_data . Layers ();}

      inline GSML::Data::iterator CurrentLayer () {return m_data . CurrentLayer ();}
      inline GSML::Data::iterator BeginLayer () {return m_data . BeginLayer ();}
      inline GSML::Data::iterator EndLayer () {return m_data . EndLayer ();}
      
      void SetCurrentLayer (GSML::Data::iterator new_layer);

      GSML::DataLayer* PrepareNewLayer ();
      void FixateNewLayer ();
      void CancelNewLayer ();

      inline GSML::DataFileIndex* GetDataFileIndex () {return m_data . GetDataFileIndex ();}
      inline GSML::PCMDatatype* TempSampletype () {return m_data . TempSampletype ();}
      inline GSML::sample_rate TempSamplerate () {return m_data . TempSamplerate ();}

      // NOT SUPPORTED YET !!
      // void SetTempSampletype (GSML::PCMDatatype* new_type);
      // void SetTempSamplerate (GSML::sample_rate new_rate);
      
      inline void SetMaxLayers (unsigned int new_max_layers) {m_data . SetMaxLayers (new_max_layers);}
      inline unsigned int MaxLayers () {return m_data . MaxLayers ();}
      inline bool IsPrepared () {return m_data . IsPrepared ();}
      
      inline bool IsModified () {return m_data . IsModified ();}
      inline void SetModified (bool new_state) {m_data . SetModified (new_state);}
                     
      inline GSML::Data::ProgressInfo& progress_info () {return m_data . progress_info;}
      
      inline SigC::Signal1<void, GSML::DataLayer*>& SignalCurrentLayerChanged () {
	return m_data . SignalCurrentLayerChanged;} // also emited afet a layer-fixate!
      inline SigC::Signal1<void, GSML::DataLayer*>& SignalNewLayerPrepared () {return m_data . SignalNewLayerPrepared;}
      inline SigC::Signal0<void>& SignalNewLayerFixated () {return m_data . SignalNewLayerFixated;}
      inline SigC::Signal0<void>& SignalNewLayerCanceled () {return m_data . SignalNewLayerCanceled;}

    private:
      friend class View;
      friend class MainWindow;

      GSML::Data m_data;
      GSML::DataLayer* prepared_layer;

      // try to copy an existing view with implementation of given type. If there is no such,
      // create new View with type. passing "" as view_impl_type means copiing any view,
      // if there is one.

      View* CreateView (const string& view_impl_type = "");
      View* CopyView (View* copy_from);

      void SetViewOphaned (View* view);
      void SetViewVisible (View* view);

      // Slots
      void SlotGUIRefresh (); // for Playback - tracking
      void SlotGUIRefreshFX (); // for FX-tracking
      void SlotInfoPushed (const GSML::Data::ProgressInfo::Info* info);
      void SlotInfoPopped ();

      void AdjustMixerParams (GSML::DataLayer* for_layer);

      GSML::WavePlotManager* m_wave_plot_manager;
      App*  m_app;

      deque <View*> m_views; // visible view are sorted before ophaned ones.
      view_iterator view_border; // first orphaned view

      view_iterator FindView (View* view); // searches m_views for view

      string m_name;
      GSML::Mixer* m_mixer;

      // Mixer sources for Playback;
      GSML::DataSource m_data_source;
      FXMixerSource m_fx_source;
      GSML::SelectionFilterSource m_selection_source;
      GSML::MultiplexerFilterSource m_multiplexer_source;
      GSML::MarkerFilterSource m_marker_source;
      GSML::LoopFilterSource m_loop_source;

      // Playback tracking stuff
      SigC::Connection slot_gui_refresh;
      GSML::sample_index m_latency;
      string m_stopitem;
      GSML::sample_index last_pos;
      string last_time; 

      // FX Tracking
      SigC::Connection slot_gui_refresh_fx;
      FXBase* m_fx;

      Menu::SimpleItemFinder cancel_button;

      FindMemCorruption m_fmc;
    };
}

#endif // GSMP_PROJECT
