/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Mixer dialog.
 */

#include "MixerDialogs.hh"

#include <sstream>
#include <iomanip>

#include <gtk--/adjustment.h>
#include <gtk--/separator.h> 

#include <math.h>

using namespace GSML;
using namespace GSMP;
using namespace Gtk;


/**************
 * MixerScale *
 **************/

const double mixer_range = exp (1.6 * log (2));

MixerScale::MixerScale ()
  : Gtk::VBox (false, 2)
{
  internal_changed = false;
  
  m_scale.set_usize (40, -1); // hack to set a proper x-size

  pcm_default t;
  dB_fs = t.dBfs ();
  
  m_scale.set_draw_value (false);
  m_scale.set_digits (5);
  m_adjustment = m_scale.get_adjustment ();
  m_adjustment -> set_upper (1);
  //m_adjustment -> set_lower (0);
  m_adjustment->value_changed.connect (slot (this, &MixerScale::SlotAdjustmentChanged) );
  
  pack_start (m_scale);
  pack_start (m_label, false, false);
}

MixerScale::~MixerScale ()
{
  cout << "YEAH: We get deleted ... (MixerScale)" << endl;
}

void MixerScale::SetValue (pcm_default::value_type value)
{
  internal_changed = true;
  
  double val = (double) value / dB_fs;
  val = pow (val / mixer_range, 1.0/3.0);

  //cout << "SetVal : " << val << endl;

  m_adjustment->set_value (1.0 - val);
  
  internal_changed = false;
}

void MixerScale::SlotAdjustmentChanged ()
{
  double value = pow (1.0 - m_adjustment->get_value (), 3.0) * mixer_range;

  if (0.98 < value && value < 1.02)
    value = 1;

  //cout << "Adjustment changed : " << value << endl;

  stringstream text;

  if (value > 0)
    {
      double db = 10* (log (value) / log (2));
      cout << " db: " << db << endl; 
  
      if (db > 0.0001)
	text << "+";
      if (db < 0)
	text << "-";

      db += 0.001;  
       int v = (int) (db);
      text << abs (v);

      if (-100 < v)
	{
	  db= (db - v) * 10;
	  int n = (int) db;
	  text << "." << abs (n);
	}
    }
  else
    text << "-inf";

  m_label.set_text (text.str () );
  
  if (!internal_changed)
    {
      pcm_default::value_type val = (pcm_default::value_type) {(long) (value * dB_fs)};
	    SignalChanged (val);
    }
}

/************
 * LEDArray *
 ************/

LEDArray::LEDArray (int i_width, int i_height, double i_max_perc)
{
  m_width = i_width;
  m_height = i_height;
  m_max_perc = i_max_perc;
  
  m_keep_peaks = false;
  m_do_falloff = false;
  
  size (m_width, m_height);
  
  m_last_perc = 0.0;
  m_peak_perc = 0.0;
  
  configured = false;
}

LEDArray::~LEDArray ()
{
}

void LEDArray::SetKeepPeaks (bool state)
{
  m_keep_peaks = state;
  
  if (m_keep_peaks == true)
    {
      m_peak_perc = 0.0;
      m_peak_delay = 1000000;
      m_peak_timer.Reset ();
    }
}

bool LEDArray::IsKeepPeaks ()
{
  return m_keep_peaks;
}

void LEDArray::SetPeakDelay (int time)
{
  m_peak_delay = time;
}

int LEDArray::PeakDelay ()
{
  return m_peak_delay;
}

void LEDArray::SetFallOff (bool state)
{
  m_do_falloff = state;
  if (m_do_falloff)
    m_falloff_step = 0.08;
}

bool LEDArray::IsFallOff ()
{
  return m_do_falloff;
}

void LEDArray::SetFallOffStep (double step)
{
  if (step < 0)
    step *= -1.0;
  m_falloff_step = step;
}

double LEDArray::FallOffStep ()
{
  return m_falloff_step;
}

void LEDArray::Clear ()
{
  m_last_perc = m_max_perc;
  m_peak_perc = 0.0;
  
  Draw (0.0);
}

void LEDArray::Draw (double perc)
{
  if (configured)
    {
      double perc_per_y = m_max_perc / m_height;
      
      for (int i = 0; i < m_height; ++i)
	{
	  int ypos = m_height - i - 1;
	  
	  double cur_perc = i * perc_per_y;
	  
	  if ( (i % 2) && (cur_perc < perc || (m_keep_peaks && (m_peak_perc >= cur_perc) && (m_peak_perc < cur_perc + (2 * perc_per_y) ) ) ) )
	    {
	      if (cur_perc <= 0.85)
		pixmap.draw_line (green_gc,
				  0, ypos,
				  m_width, ypos);
	      else
	      	pixmap.draw_line (red_gc,
				  0, ypos,
				  m_width, ypos);
	    }
	  else
	    pixmap.draw_line (back_gc,
			      0, ypos,
			      m_width, ypos);
	}
      
      int marker_pos = m_height - (int) (1.0 / perc_per_y);
      pixmap.draw_line (marker_gc,
			0, marker_pos,
			m_width, marker_pos);
      
      m_last_perc = perc;
      
      if (m_last_perc > m_peak_perc)
	{
	  m_peak_perc = m_last_perc;
	  m_peak_timer.Reset ();
	}
      
      if (m_keep_peaks && m_peak_timer.Delta () > m_peak_delay)
	{
	  if (m_do_falloff)
	    m_peak_perc = limit_min (m_peak_perc - m_falloff_step, 0.0);
	  else {
	    m_peak_perc = m_last_perc;
	    m_peak_timer.Reset ();
	  }
	}
      
      GdkRectangle update_rect;
      update_rect.x = 0;
      update_rect.y = 0;
      update_rect.width = m_width;
      update_rect.height = m_height;
      draw (&update_rect);
    }
  else
    cout << "?!?! Not configured??" << endl;
}

gint LEDArray::configure_event_impl (GdkEventConfigure* event)
{
  win = get_window ();
  visual = win.get_visual ();
  
  m_width = width ();
  m_height = height ();
  
  if (pixmap)
    pixmap.release();
  
  pixmap.create (win, m_width, m_height);
  
  if (!back_gc)
    {
      back_gc = Gdk_GC (pixmap);
      Gdk_Color color ("black");
      get_colormap ().alloc (color);
      back_gc.set_foreground (color);
    }
  if (!green_gc)
    {
      green_gc = Gdk_GC (pixmap);
      Gdk_Color color ("green");
      get_colormap ().alloc (color);
      green_gc.set_foreground (color);
    }
  if (!red_gc)
    {
      red_gc = Gdk_GC (pixmap);
      Gdk_Color color ("red");
      get_colormap ().alloc (color);
      red_gc.set_foreground (color);
    }
  if (!marker_gc)
    {
      marker_gc = Gdk_GC (pixmap);
      Gdk_Color color ("white");
      get_colormap ().alloc (color);
      marker_gc.set_foreground (color);
    }
  
  configured = true;
  
  Clear ();
  
  return true;
}

gint LEDArray::expose_event_impl (GdkEventExpose* event)
{
  win.draw_pixmap (back_gc,
  		   pixmap,
  		   event->area.x, event->area.y, // src
  		   event->area.x, event->area.y, // dest
  		   event->area.width, event->area.height);
  
  return true;
}

/**************************
 * MixerTrackMatrixWidget *
 **************************/

MixerTrackMatrixWidget::MixerTrackMatrixWidget (UniversalMixerParams* i_mixer_params, Mixer* i_mixer,
						track_index i_track, bool is_pl)
  : m_vu_meter (15, 150, 1.25)
{
  m_mixer_params = i_mixer_params;
  m_mixer = i_mixer;
  m_track = i_track;
  m_is_pl = is_pl;

  internal_changed = false;
  
  set_border_width (2);
  set_spacing (2);
  
  stringstream t_text; t_text << "Track: " << m_track;
  Gtk::Label*  t_label = manage (new Gtk::Label (t_text.str () ) );
  pack_start (*t_label, false);
  
  m_hbox.set_spacing (2);
  pack_start (m_hbox);
  
  m_scale.SetValue (m_mixer_params->TrackValue (m_track) );
  m_scale.SignalChanged.connect (slot (this, &MixerTrackMatrixWidget::SlotTrackScaleValueChanged) );
  m_hbox.pack_start (m_scale);
  
  m_vu_meter.SetKeepPeaks (true);
  m_vu_meter.SetFallOff (true);
  m_hbox.pack_start (m_vu_meter, false);
  
  SlotChannelsChanged (m_mixer_params->Channels (), false);
  
  m_mixer_params->SignalChannelsChanged.connect (bind (slot (this, &MixerTrackMatrixWidget::SlotChannelsChanged), true) );
  m_mixer_params->SignalTrackValueChanged.connect (slot (this, &MixerTrackMatrixWidget::SlotTrackValueChanged) );
  m_mixer_params->SignalMatrixValueChanged.connect (slot (this, &MixerTrackMatrixWidget::SlotMatrixValueChanged) );
  m_mixer_params->SignalTrackMuteChanged.connect (slot (this, &MixerTrackMatrixWidget::SlotTrackMuteChanged) );
}

MixerTrackMatrixWidget::~MixerTrackMatrixWidget ()
{
  while (m_scales.size () > 0)
    {
      cout << "deleting one MixerScale ..." << endl;
      del_helper2 (m_scales.back () );
      m_scales.pop_back ();
    }
  cout << "YEAH: We get deleted ... (MixerTrackMatrixWidget)" << endl;
}

void MixerTrackMatrixWidget::GUIRefresh ()
{
  if (m_mixer->IsIORunning () )
    {
      pcm_default type;
      pcm_default::value_type dBfs = type.dBfs ();
      Mixer::vu_meter_data t_data;
      if (m_is_pl) {
	t_data = m_mixer->PlaybackTrackVU (m_track);
      }
      else {
	t_data = m_mixer->CaptureTrackVU (m_track);
      }
      double value = (double) t_data.GetMaxAbs ();
      value /= (double) dBfs;
      
      m_vu_meter.Draw (value);
      
      m_io_was_running = true;
    }
  else
    if (m_io_was_running)
      {
	m_vu_meter.Clear ();
	m_io_was_running = false;
      }
}

void MixerTrackMatrixWidget::SlotChannelsChanged (track_index n_channels, bool do_show)
{
  for (track_index i = m_scales.size (); i < n_channels; ++i)
    {
      m_scales.push_back (new MixerScale () );
      m_scales.back ()->SetValue (m_mixer_params->MatrixValue (m_track, i) );
      m_scales.back ()->SignalChanged.connect (bind (slot (this, &MixerTrackMatrixWidget::SlotMatrixScaleValueChanged), i) );
      
      m_hbox.pack_start (* m_scales.back () );
    }
  for (track_index i = m_scales.size (); i > n_channels; --i)
    {
      del_helper2 (m_scales.back () );
      m_scales.pop_back ();
    }
  if (do_show)
    m_hbox.show_all ();
}

void MixerTrackMatrixWidget::SlotTrackValueChanged (track_index track, pcm_default::value_type value)
{
  if (track == m_track)
    m_scale.SetValue (value);
}

void MixerTrackMatrixWidget::SlotMatrixValueChanged (track_index track, track_index channel, pcm_default::value_type value)
{
  if (track == m_track)
    {
      if (channel < m_scales.size () )
	{
	  m_scales [channel]->SetValue (value);
	}
      else
	cout << "BUG: MixerTrackWidget: not enough channles in vector ?!" << endl;
    }
}

void MixerTrackMatrixWidget::SlotTrackMuteChanged (track_index track, bool n_state)
{
  if (!internal_changed && track == m_track) {
    /*
    if (m_is_pl)
      tbutton->set_active (n_state);
    else
    tbutton->set_active (!n_state);
    */
  }
}

void MixerTrackMatrixWidget::SlotTrackScaleValueChanged (pcm_default::value_type value)
{
  m_mixer_params->SetTrackValue (m_track, value);
}

void MixerTrackMatrixWidget::SlotMatrixScaleValueChanged (pcm_default::value_type value, track_index channel)
{
  m_mixer_params->SetMatrixValue (m_track, channel, value);
}


/**************************
 * MixerTrackMatrixDialog *
 **************************/

MixerTrackMatrixDialog::MixerTrackMatrixDialog (GSMP::Project* i_project, UniversalMixerParams* i_mixer_params, Mixer* i_mixer,
						track_index i_track, bool is_pl)
{
  m_mixer_params = i_mixer_params;
  m_mixer = i_mixer;
  m_track = i_track;
  m_is_pl = is_pl;
  m_project = i_project; 
  
  if (m_is_pl)
    set_title ("Playback Mixer Matrix");
  else
    set_title ("Capture Mixer Matrix");
  
  get_vbox ()->set_border_width (5);
  get_vbox ()->set_spacing (5);
  
  m_track_matrix = new MixerTrackMatrixWidget (m_mixer_params, m_mixer, m_track, m_is_pl);
  get_vbox ()->pack_start (*m_track_matrix);
  
  Gtk::Button* t_button = manage (new Gtk::Button ("Close") );
  t_button->clicked.connect (destroy.slot() );
  get_action_area ()->pack_start (*t_button);
  
  m_mixer_params->SignalTracksChanged.connect (slot (this, &MixerTrackMatrixDialog::SlotTracksChanged) );
  m_project -> GetApp () -> GetKit () -> SignalGUIRefresh.connect (slot (this, &MixerTrackMatrixDialog::SlotGUIRefresh));

  show_all ();
}

MixerTrackMatrixDialog::~MixerTrackMatrixDialog ()
{
  del_helper2 (m_track_matrix);
  cout << "YEAH: We get deleted ... (MixerTrackMatrixDialog)" << endl;
}

void MixerTrackMatrixDialog::SlotTracksChanged (track_index tracks)
{
  if (tracks < m_track)
    destroy.slot();
}

void MixerTrackMatrixDialog::SlotGUIRefresh ()
{
  m_track_matrix->GUIRefresh ();
}

/********************
 * MixerTrackWidget *
 ********************/

MixerTrackWidget::MixerTrackWidget (Project* i_project, UniversalMixerParams* i_mixer_params, Mixer* i_mixer,
				    track_index i_track, bool is_pl=true)
  : m_vu_meter (15, 150, 1.25)
{
  m_project = i_project;
  m_mixer_params = i_mixer_params;
  m_mixer = i_mixer;
  m_track = i_track;
  m_is_pl = is_pl;
  
  internal_changed = false;
  m_mixer_track_matrix_dialog = 0;
  
  set_border_width (5);
  set_spacing (5);
  
  stringstream t_text; t_text << "Tr: " << m_track;
  Gtk::Label*  t_label = manage (new Gtk::Label (t_text.str () ) );
  pack_start (*t_label, false);
  
  m_hbox.set_spacing (2);
  pack_start (m_hbox);
  
  m_scale.SetValue (m_mixer_params->TrackValue (m_track) );
  m_scale.SignalChanged.connect (slot (this, &MixerTrackWidget::SlotScaleValueChanged) );
  m_hbox.pack_start (m_scale);
  
  m_vu_meter.SetKeepPeaks (true);
  m_vu_meter.SetFallOff (true);
  m_hbox.pack_start (m_vu_meter, false);
  
  m_mute_button.remove ();
  if (m_is_pl) {
    m_mute_button.add_label ("Mute", 0.5, 0.5);
    m_mute_button.set_active (m_mixer_params->IsTrackMuted (m_track) );
  }
  else {
    m_mute_button.add_label ("Rec", 0.5, 0.5);
    m_mute_button.set_active (!m_mixer_params->IsTrackMuted (m_track) );
  }
  m_mute_button.toggled.connect (slot (this, &MixerTrackWidget::SlotMuteButtonToggled) );
  pack_start (m_mute_button, false); // no expand
  
  m_chm_button.add_label ("ChMix", 0.5, 0.5);
  m_chm_button.toggled.connect (slot (this, &MixerTrackWidget::SlotChmButtonToggled) );
  pack_start (m_chm_button, false); // no expand
  
  m_mixer_params->SignalTrackValueChanged.connect (slot (this, &MixerTrackWidget::SlotValueChanged) );
  m_mixer_params->SignalTrackMuteChanged.connect (slot (this, &MixerTrackWidget::SlotMuteChanged) );  
}

MixerTrackWidget::~MixerTrackWidget ()
{
  cout << "YEAH: We get deleted ... (MixerTrackWidget)" << endl;
}

void MixerTrackWidget::SlotValueChanged (track_index track, pcm_default::value_type value)
{
  if (track == m_track)
    m_scale.SetValue (value);
}

void MixerTrackWidget::SlotMuteChanged (track_index track, bool n_state)
{
  if (!internal_changed && track == m_track) {
    if (m_is_pl)
      m_mute_button.set_active (n_state);
    else
      m_mute_button.set_active (!n_state);
  }
}

void MixerTrackWidget::SlotScaleValueChanged (pcm_default::value_type value)
{
  m_mixer_params->SetTrackValue (m_track, value);
}

void MixerTrackWidget::SlotMuteButtonToggled ()
{
  bool state = m_mute_button.get_active ();
  
  internal_changed = true;
  if (m_is_pl)
      m_mixer_params->SetTrackMute (m_track, state);
    else
      m_mixer_params->SetTrackMute (m_track, !state);
  
  internal_changed = false;
}

void MixerTrackWidget::SlotChmButtonToggled ()
{
  if (m_chm_button.get_active () )
    {
      m_mixer_track_matrix_dialog = manage (new MixerTrackMatrixDialog (m_project, m_mixer_params, m_mixer, m_track, m_is_pl) );
      m_mixer_track_matrix_dialog->destroy.connect (slot (this, &MixerTrackWidget::SlotMTMDDeleted) );
    }
  else
    if (m_mixer_track_matrix_dialog)
      m_mixer_track_matrix_dialog->destroy ();
}

void MixerTrackWidget::SlotMTMDDeleted ()
{
  m_mixer_track_matrix_dialog = 0;
  m_chm_button.set_active (false);
}

void MixerTrackWidget::GUIRefresh ()
{
  if (m_mixer->IsIORunning () )
    {
      pcm_default type;
      pcm_default::value_type dBfs = type.dBfs ();
      Mixer::vu_meter_data t_data;
      if (m_is_pl) {
	t_data = m_mixer->PlaybackTrackVU (m_track);
      }
      else {
	t_data = m_mixer->CaptureTrackVU (m_track);
      }
      double value = (double) t_data.GetMaxAbs ();
      value /= (double) dBfs;
      
      m_vu_meter.Draw (value);
      
      m_io_was_running = true;
    }
  else
    if (m_io_was_running)
      {
	m_vu_meter.Clear ();
	m_io_was_running = false;
      }
}

/********************
 * MixerTrackDialog *
 ********************/

MixerTrackDialog::MixerTrackDialog (Project* i_project, UniversalMixerParams* i_mixer_params,
				    bool is_pl)
{
  m_project = i_project;
  m_mixer_params = i_mixer_params;
  m_is_pl = is_pl;
  
  if (m_is_pl)
    set_title ("Playback Track Mixer");
  else
    set_title ("Capture Track Mixer");
  
  get_vbox ()->set_border_width (5);
  get_vbox ()->set_spacing (5);
  
  get_vbox ()->pack_start (m_hbox);
  
  Gtk::Button* t_button = manage (new Gtk::Button ("Close") );
  t_button->clicked.connect (destroy.slot() );
  get_action_area ()->pack_start (*t_button);
  
  build_tracks ();
  
  m_mixer_params->SignalTracksChanged.connect (slot (this, &MixerTrackDialog::SlotTracksChanged) );
  
  m_project -> GetApp () -> GetKit () -> SignalGUIRefresh.connect (slot (this, &MixerTrackDialog::SlotGUIRefresh) );
  
  show_all ();
}

MixerTrackDialog::~MixerTrackDialog ()
{
  while (m_tracks.size () > 0)
    {
      del_helper2 (m_tracks.back () );
      m_tracks.pop_back ();
    }
  cout << "YEAH: We get deleted ... (MixerTrackDialog)" << endl;
}

void MixerTrackDialog::build_tracks ()
{
  track_index tracks = m_mixer_params->Tracks ();
  for (track_index tr = 0; tr < tracks; ++tr)
    {
      MixerTrackWidget* t_track = new MixerTrackWidget (m_project, m_mixer_params, m_project -> GetMixer (), tr, m_is_pl);
      m_hbox.pack_start (*t_track);
      m_tracks.push_back (t_track);
      
      if (tr + 1 < tracks) {
	Gtk::VSeparator* t_sep = manage (new Gtk::VSeparator);
	m_hbox.pack_start (*t_sep);
      }
    }
}

void MixerTrackDialog::SlotTracksChanged (track_index tracks)
{
  while (m_tracks.size () > tracks)
    {
      cout << "TrackChanged: deleting" << endl;
      del_helper2 (m_tracks.back () );
      m_tracks.pop_back ();
      
      m_hbox.children ().pop_back ();
    }
  
  while (m_tracks.size () < tracks)
  {
    cout << "TrackChanged: adding" << endl;
    
    if (m_tracks.size () > 0) {
      Gtk::VSeparator* t_sep = manage (new Gtk::VSeparator);
      m_hbox.pack_start (*t_sep);
    }
    
    MixerTrackWidget* t_track = new MixerTrackWidget (m_project, m_mixer_params, m_project -> GetMixer (), m_tracks.size (), m_is_pl);
    m_hbox.pack_start (*t_track);
    m_tracks.push_back (t_track);
    m_hbox.show_all ();
  }
}

void MixerTrackDialog::SlotGUIRefresh ()
{
  for (int i = 0; i < m_tracks.size (); ++i)
    m_tracks [i]->GUIRefresh ();
}

/**********************
 * MixerChannelWidget *
 **********************/

MixerChannelWidget::MixerChannelWidget (UniversalMixerParams* i_mixer_params, Mixer* i_mixer,
					track_index i_channel, bool is_pl=true)
  : m_vu_meter (15, 150, 1.25)
{
  m_mixer_params = i_mixer_params;
  m_mixer = i_mixer;
  m_channel = i_channel;
  m_is_pl = is_pl;
  
  internal_changed = false;
  
  set_border_width (5);
  set_spacing (5);
  
  stringstream t_text; t_text << "Ch: " << m_channel;
  Gtk::Label*  t_label = manage (new Gtk::Label (t_text.str () ) );
  pack_start (*t_label, false);
  
  m_hbox.set_spacing (2);
  pack_start (m_hbox);
  
  m_scale.SetValue (m_mixer_params->ChannelValue (m_channel) );
  m_scale.SignalChanged.connect (slot (this, &MixerChannelWidget::SlotScaleValueChanged) );
  m_hbox.pack_start (m_scale);
  
  m_vu_meter.SetKeepPeaks (true);
  m_vu_meter.SetFallOff (true);
  m_hbox.pack_start (m_vu_meter, false);
  
  m_mute_button.remove ();
  m_mute_button.add_label ("Mute", 0.5, 0.5);
  m_mute_button.set_active (m_mixer_params->IsChannelMuted (m_channel) );
  m_mute_button.toggled.connect (slot (this, &MixerChannelWidget::SlotMuteButtonToggled) );
  pack_start (m_mute_button, false); // no expand
  
  m_mixer_params->SignalChannelValueChanged.connect (slot (this, &MixerChannelWidget::SlotValueChanged) );
  m_mixer_params->SignalChannelMuteChanged.connect (slot (this, &MixerChannelWidget::SlotMuteChanged) );  
}

MixerChannelWidget::~MixerChannelWidget ()
{
  cout << "YEAH: We get deleted ... (MixerChannelWidget)" << endl;
}

void MixerChannelWidget::SlotValueChanged (track_index channel, pcm_default::value_type value)
{
  if (channel == m_channel)
    m_scale.SetValue (value);
}

void MixerChannelWidget::SlotMuteChanged (track_index channel, bool n_state)
{
  if (!internal_changed && channel == m_channel) {
    m_mute_button.set_active (n_state);
  }
}

void MixerChannelWidget::SlotScaleValueChanged (pcm_default::value_type value)
{
  m_mixer_params->SetChannelValue (m_channel, value);
}

void MixerChannelWidget::SlotMuteButtonToggled ()
{
  bool state = m_mute_button.get_active ();
  
  internal_changed = true;
  m_mixer_params->SetChannelMute (m_channel, state);
  internal_changed = false;
}

void MixerChannelWidget::GUIRefresh ()
{
  if (m_mixer->IsIORunning () )
    {
      pcm_default type;
      pcm_default::value_type dBfs = type.dBfs ();
      Mixer::vu_meter_data t_data;
      if (m_is_pl) {
	t_data = m_mixer->PlaybackChannelVU (m_channel);
      }
      else {
	t_data = m_mixer->CaptureChannelVU (m_channel);
      }
      double value = (double) t_data.GetMaxAbs ();
      value /= (double) dBfs;
      
      m_vu_meter.Draw (value);

      m_io_was_running = true;
    }
  else
    if (m_io_was_running)
      {
	m_vu_meter.Clear ();
	m_io_was_running = false;
      }
}

/**********************
 * MixerChannelDialog *
 **********************/

MixerChannelDialog::MixerChannelDialog (Project* i_project, UniversalMixerParams* i_mixer_params,
					bool is_pl)
{
  m_project = i_project;
  m_mixer_params = i_mixer_params;
  m_is_pl = is_pl;
  
  if (m_is_pl)
    set_title ("Playback Channel Mixer");
  else
    set_title ("Capture Channel Mixer");
  
  get_vbox ()->set_border_width (5);
  get_vbox ()->set_spacing (5);
  
  get_vbox ()->pack_start (m_hbox);
  
  Gtk::Button* t_button = manage (new Gtk::Button ("Close") );
  t_button->clicked.connect (destroy.slot() );
  get_action_area ()->pack_start (*t_button);
  
  build_channels ();
  
  m_mixer_params->SignalChannelsChanged.connect (slot (this, &MixerChannelDialog::SlotChannelsChanged) );
  m_project -> GetApp () -> GetKit () -> SignalGUIRefresh.connect (slot (this, &MixerChannelDialog::SlotGUIRefresh) );
  
  show_all ();
}

MixerChannelDialog::~MixerChannelDialog ()
{
  while (m_channels.size () > 0)
    {
      del_helper2 (m_channels.back () );
      m_channels.pop_back ();
    }
  cout << "YEAH: We get deleted ... (MixerChannelDialog)" << endl;
}

void MixerChannelDialog::build_channels ()
{
  track_index channels = m_mixer_params->Channels ();
  for (track_index ch = 0; ch < channels; ++ch)
    {
      MixerChannelWidget* t_channel = new MixerChannelWidget (m_mixer_params, m_project -> GetMixer (), ch, m_is_pl);
      m_hbox.pack_start (*t_channel);
      m_channels.push_back (t_channel);
      
      if (ch + 1 < channels) {
	Gtk::VSeparator* t_sep = manage (new Gtk::VSeparator);
	m_hbox.pack_start (*t_sep);
      }
    }
}

void MixerChannelDialog::SlotChannelsChanged (track_index channels)
{
  while (m_channels.size () > channels)
    {
      cout << "ChannelsChanged: deleting" << endl;
      del_helper2 (m_channels.back () );
      m_channels.pop_back ();
      
      m_hbox.children ().pop_back ();
    }
  
  while (m_channels.size () < channels)
  {
    cout << "ChannelsChanged: adding" << endl;
    
    if (m_channels.size () > 0) {
      Gtk::VSeparator* t_sep = manage (new Gtk::VSeparator);
      m_hbox.pack_start (*t_sep);
    }
    
    MixerChannelWidget* t_channel = new MixerChannelWidget (m_mixer_params, m_project -> GetMixer (), m_channels.size (), m_is_pl);
    m_hbox.pack_start (*t_channel);
    m_channels.push_back (t_channel);
    m_hbox.show_all ();
  }
}

void MixerChannelDialog::SlotGUIRefresh ()
{
  for (int i = 0; i < m_channels.size (); ++i)
    m_channels [i]->GUIRefresh ();
}


Gtk::Widget* PLChannelDialogX::Create ()
{
  return new MixerChannelDialog (m_project, m_project -> m_pl_mixer_params, true);
}


Gtk::Widget* CAChannelDialogX::Create ()
{
  return new MixerChannelDialog (m_project, m_project -> m_ca_mixer_params, false);
}


Gtk::Widget* PLTrackDialogX::Create ()
{
  return new MixerTrackDialog (m_project, m_project -> m_pl_mixer_params, true);
}

Gtk::Widget* CATrackDialogX::Create ()
{
  return new MixerTrackDialog (m_project, m_project -> m_ca_mixer_params, false);
}
