/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Normal graphic view. Will be replaces by a new overlay architecture.
 */

#include "NormalView.hh"

using namespace std;
using namespace GSML;
using namespace GSMP;

NormalTrackView::NormalTrackView (WavePlotTracer<MinMaxAvgWavePlot>* i_plot,
				  const WavePlotConfig* i_config,
				  NormalView* i_parent, int i_width = 450, bool i_master = false)
{
  m_waveplot = i_plot;
  m_config = i_config;
  m_parent = i_parent;
  m_master = i_master;
  
  m_current_plot = m_waveplot->Plot ();
  
  m_width = i_width;
  m_height = 50;
  
  set_usize (450, 50);
  
  size (m_width, m_height);
  
  m_start_delta = 0;
  m_internal_delta = 0;
  
  m_cache_length = m_width * SMOTH_SCROLL_FACTOR;
  
  m_use_temp_sel = false;

  refresh_needed = false;
  refresh_region.resize (m_cache_length);
  
  set_events (GDK_EXPOSURE_MASK
	      | GDK_LEAVE_NOTIFY_MASK
	      | GDK_BUTTON_PRESS_MASK
	      | GDK_BUTTON_RELEASE_MASK
	      | GDK_POINTER_MOTION_MASK
	      | GDK_POINTER_MOTION_HINT_MASK);
  
  m_waveplot->SignalWavePlotChanged.connect (slot (this, &NormalTrackView::SlotWavePlotChanged) );
  m_waveplot->SignalProgressUpdated.connect (slot (this, &NormalTrackView::SlotWavePlotUpdated) );
  
  m_current_sel = 0;
  m_current_markers = 0;
  
  connect_selection ();
  connect_markers ();

  DEBUG__ (GSMP, normal)
    cout << "NormalTrackView constructed" << endl;
}

NormalTrackView::~NormalTrackView ()
{
  DEBUG__ (GSMP, normal)
    cout << "NormalTrackView Deleted" << endl;
}

void NormalTrackView::SetStartDelta (int n_delta, bool force_redaw)
{
  /*
  cout << "n_delta: " << n_delta << endl;
  
  cout << "width: " << m_width
       << ", height: " << m_height
       << ", chache_length: " << m_cache_length
       << ", m_start_delta: " << m_start_delta
       << ", m_internal_delta: " << m_internal_delta << endl;
  */
  m_start_delta = n_delta;
  
  // adjust internal start / draw pixmap
  int t_offset = m_start_delta - m_internal_delta;
  //cout << "t_offset: " << t_offset << endl;
  
  // smoth-srolling / area cached
  if (!force_redaw && t_offset >= 0 && t_offset + m_width <= m_cache_length)
    {
      win.draw_pixmap (back_gc,
		       pixmap,
		       t_offset, 0, // src
		       0, 0, // dest
		       m_width, m_height);
    }
  else // area not cached
    {
      sample_index t_max_start = m_config->PlotLength () - m_cache_length;
      sample_index t_perf_start = limit_min (m_start_delta  - (m_cache_length - m_width) / 2, 0);
      /*
	cout << "max_start: " << t_max_start
	<< ", perf_start:" << t_perf_start << endl;
      */
      m_internal_delta = limit_max (t_perf_start, t_max_start);
      
      add_refresh_all ();
    }
  
  /*
  cout << "width: " << m_width
       << ", height: " << m_height
       << ", chache_length: " << m_cache_length
       << ", m_start_delta: " << m_start_delta
       << ", m_internal_delta: " << m_internal_delta << endl;
  */
}

void NormalTrackView::SetTempSelection (sample_index n_start, sample_index n_end)
{
  if (!m_use_temp_sel) // first call
    {
      m_use_temp_sel = true;
      m_temp_sel_start = n_start;
      m_temp_sel_end = n_end;
    }
  
  sample_index t_start = m_temp_sel_start;
  sample_index t_end = m_temp_sel_end;

  m_temp_sel_start = n_start;
  m_temp_sel_end = n_end;
  
  // optimal refresh
  if (m_temp_sel_start != t_start) // changes of start
    {
      sample_index t_r_start = min (m_temp_sel_start, t_start);
      sample_index t_r_end = max (m_temp_sel_start, t_start);
      
      add_refresh_from_real (t_r_start, t_r_end);
    }
  if (m_temp_sel_end != t_end) // changes of end
    {
      sample_index t_r_start = min (m_temp_sel_end, t_end);
      sample_index t_r_end = max (m_temp_sel_end, t_end);
      
      add_refresh_from_real (t_r_start, t_r_end);
    }
}

void NormalTrackView::RemoveTempSelection ()
{
  if (m_use_temp_sel)
    {
      m_use_temp_sel = false;
      
      add_refresh_from_real (m_temp_sel_start, m_temp_sel_end);
    }
  else
    cout << "DEBUG: temp_selction was not used?!?" << endl;
}

void NormalTrackView::SetRefresh ()
{
  add_refresh_all ();
}

int NormalTrackView::Width () const
{
  return m_width;
}

int NormalTrackView::Height () const
{
  return m_height;
}

void NormalTrackView::Debug () const
{
  cout << "NormalEnhWaveView::Debug:" << endl;
  cout << "  m_start_delta:    " <<  m_start_delta << endl
       << "  m_internal_delta: " << m_internal_delta << endl;
}

void NormalTrackView::add_refresh (int n_start, int n_end)
{
  SigC::Threads::MLock hack (refresh_lock);
  
  if (refresh_region.size () < m_cache_length)
    {
      cout << "BUG: add_refresh: refresh_region's size is too small ?!?" << endl;
      refresh_region.resize (m_cache_length);
    }
  
  if (n_start > m_cache_length || n_end > m_cache_length)
    cout << "BUG: add_refresh: values out ot range!"
	 << " (s: " << n_start << ", e: " << n_end << "ml: " << m_cache_length << endl;
  
  for (int i = n_start; i < n_end && i < m_cache_length; ++i)
    refresh_region [i] = true;
  refresh_needed = true;
}

void NormalTrackView::add_refresh_from_real (sample_index n_start, sample_index n_end)
{
  sample_index t_start = m_config->Start ();
  sample_index t_sps = m_config->SamplesPerSample ();
  
  //cout << "add_refresh_from_real: " << n_start << " - " << n_end << endl;
  
  t_start += m_internal_delta * t_sps; // start of cached region
  
  if (n_end > t_start && n_start < t_start + m_cache_length * t_sps)
  {
    sample_index t_r_start = limit_min (n_start - t_start, 0) / t_sps;
    sample_index t_r_end = limit_max ( (n_end- t_start) / t_sps, m_cache_length);
    
    t_r_end = limit_min (t_r_end, t_r_start + 1);
    
    //cout << "  r4r: will refresh: " << t_r_start << " - " << t_r_end << endl;
    add_refresh (t_r_start, t_r_end);
  }
  else
    {
      //cout << "add_refresh_from_real: not in range ..." << endl;
    }
}   

void NormalTrackView::add_refresh_all ()
{
  add_refresh (0, m_cache_length);
}

// no lock, because it is only used in the (locked) draw_wave method
void NormalTrackView::refresh_clear ()
{
  for (int i = 0; i < refresh_region.size (); ++i)
    refresh_region [i] = false;
  
  refresh_needed = false;
}


void NormalTrackView::realize_impl ()
{
  Gtk::DrawingArea::realize_impl ();  

  win = get_window ();
  if (!win)
    cout << "???????" << endl;

  m_width = width ();
  m_height = height ();
  m_cache_length = m_width * SMOTH_SCROLL_FACTOR;

  pixmap.create (win, m_cache_length, m_height);
}

gint NormalTrackView::configure_event_impl (GdkEventConfigure* event)
{
  cout << "NormalTrackView::configure_event_impl:" << endl;

  Gtk::DrawingArea::configure_event_impl (event);
  
  win = get_window ();
  if (!win)
    cout << "???????" << endl;
  
  int t_height = m_height;
  int t_width = m_width;
  
  m_height = height ();
  m_width = width ();
  
  // only proceed if the size changed
  if (m_height != t_height || m_width != t_width)
    {
      m_width = width ();
      m_height = height ();
      m_cache_length = m_width * SMOTH_SCROLL_FACTOR;

      pixmap.release();  
      pixmap.create (win, m_cache_length, m_height);

      {
	back_gc = Gdk_GC (pixmap);
	Gdk_Color color ("white");
	get_colormap ().alloc (color);
	back_gc.set_foreground (color);
      }
      {
	back_sel_gc = Gdk_GC (pixmap);
	Gdk_Color color ("light grey");
	get_colormap ().alloc (color);
	back_sel_gc.set_foreground (color);
      }
      {
	back_temp_sel_gc = Gdk_GC (pixmap);
	Gdk_Color color ("light blue");
	get_colormap ().alloc (color);
	back_temp_sel_gc.set_foreground (color);
      }
      {
	back_undef_gc = Gdk_GC (pixmap);
	Gdk_Color color ("orange");
	get_colormap ().alloc (color);
	back_undef_gc.set_foreground (color);
      }
      {
	wave_gc = Gdk_GC (pixmap);
	Gdk_Color color ("blue");
	get_colormap ().alloc (color);
	wave_gc.set_foreground (color);
      }
      {
	wave_sel_gc = Gdk_GC (pixmap);
	Gdk_Color color ("dark blue");
	get_colormap ().alloc (color);
	wave_sel_gc.set_foreground (color);
      }
      {
	wave_temp_sel_gc = Gdk_GC (pixmap);
	Gdk_Color color ("red");
	get_colormap ().alloc (color);
	wave_temp_sel_gc.set_foreground (color);
      }
      {
	grid_gc = Gdk_GC (pixmap);
	Gdk_Color color ("dark grey");
	get_colormap ().alloc (color);
	grid_gc.set_foreground (color);
      }
      {
	marker_gc = Gdk_GC (pixmap);
	Gdk_Color color ("dark red");
	get_colormap ().alloc (color);
	marker_gc.set_foreground (color);
      } 
      
      refresh_region.resize (m_cache_length);
      
      // only announce width changes ...
      if (m_width != t_width)
	m_parent->SlotChildConfigured (this, m_width);
  
      add_refresh_all ();
    }
  
  return true;
}

int NormalTrackView::expose_event_impl (GdkEventExpose* event)
{
  /*
  cout << "NormalTrackView::expose_event_impl: "
       << "x: " << event->area.x
       << ", y: " << event->area.y
       << ", w: " << event->area.width
       << ", h: " << event->area.height
       << endl;
  */
  
  int t_offset = m_start_delta - m_internal_delta;
  
  /*
  cout << "m_start_delta: " << m_start_delta
       << ", m_internal_delta: " << m_internal_delta
       << endl;
  
  cout << "t_offset: " << t_offset
       << ", m_cache_length: " << m_cache_length << endl;
  */
  win.draw_pixmap (back_gc,
  		   pixmap,
  		   event->area.x + t_offset, event->area.y, // src
  		   event->area.x, event->area.y, // dest
  		   event->area.width, event->area.height);
  
  return true;
}

// this handles the mouse button events to set selections ...
gint NormalTrackView::button_press_event_impl (GdkEventButton* event)
{
  // cout << "NormalTrackView::button_press_event_impl:" << endl;
  
  m_parent->SlotButtenPress (event);
  
  return true;
}

gint NormalTrackView::button_release_event_impl (GdkEventButton* event)
{
  // cout << "NormalTrackView::button_release_event_impl:" << endl;
  
  m_parent->SlotButtenRelease (event);
  
  return true;
}

gint NormalTrackView::motion_notify_event_impl (GdkEventMotion* event)
{
  // cout << "NormalTrackView::motion_notify_event_impl:" << endl;
  
  int x, y;
  GdkModifierType state;
  Gdk_Window window (event->window);
  
  if (event->is_hint)
    {
      window.get_pointer (x, y, state);
    }
  else
    {
      x = (int) event->x;
      y = (int) event->y;
      state = (GdkModifierType) event->state;
    }
  
  m_parent->SlotMotionNotify (event, state);
  
  return true;
}

void NormalTrackView::draw_wave ()
{
  // cout << "trackView::draw_wave:" << endl;
  
  SigC::Threads::MLock hack (refresh_lock);
  
  const MinMaxAvgWavePlot::Progress* t_progress = &(m_waveplot->Plot ()->GetProgress () );
  
  int t_width = limit_max (m_cache_length, t_progress->Length () );
  
  // limit the start and end of refresh
  sample_index t_start = 0;
  sample_index t_end = m_cache_length;
  
  for (int i = 0; i < m_cache_length && !refresh_region [i]; ++i)
    t_start = i;

  for (int i = m_cache_length - 1; i >= 0 && !refresh_region [i]; --i)
    t_end = i;
  
  
  // cout << "drawing (in cached pixmap): " << t_start << " - " << t_end << endl;
  
  // loop to first selection segment
  sample_index t_sps = m_config->SamplesPerSample ();
  sample_index t_real_sample = + m_config->Start () + (t_start + m_internal_delta) * t_sps;
  
  TrackLayer::Selection::const_iterator t_sel;
  TrackLayer::Selection::const_iterator t_sel_end;
  
  if (!m_current_sel)
    connect_selection ();
  
  if (m_current_sel)
    {
      t_sel = m_current_sel->begin ();
      t_sel_end = m_current_sel->end ();
      
      while (t_sel != t_sel_end && t_sel->end < t_real_sample) {
	//cout << " this segment ends before current sample: " << t_sel->end << endl;
	++t_sel;
      }
    }
  //else
  //  cout << "DEBUG: no selection available?! - right??" << endl;
  
  DataLayer::Markers::const_iterator t_marker;
  DataLayer::Markers::const_iterator t_marker_end;
  
  if (!m_current_markers)
    connect_markers ();
  
  if (m_current_markers)
    {
      t_marker = m_current_markers->begin ();
      t_marker_end = m_current_markers->end ();
      
      while (t_marker != t_marker_end && t_marker->position < t_real_sample) {
	//cout << " this marker is before current sample: name: "
	//     << t_marker->name << ", pos: " << t_marker->position << endl;
	++t_marker;
      }
    }
  //else
  //  cout << "DEBUG: no markers available?! - right??" << endl;
  
  prepare_waveform (t_progress);
  
  sample_index t_wave_plot_length = m_config->PlotLength ();
  
  sample_index t_v_grid_pos = t_start + (50 * t_sps - t_real_sample % (50 * t_sps) ) / t_sps;
  //cout << "first grid pos: " << t_v_grid_pos << ", my start: " << t_start << endl;
  
  for (sample_index i = t_start; i < t_end; ++i)
    {
      int plot_pos = i + m_internal_delta;
      sample_index t_real_sample_end = t_real_sample + t_sps;
      
      // has to be refreshed ??
      if (refresh_region [i])
	{
      	  bool t_temp_sel = false;
	  bool t_selected = false;
	  bool t_marker_here = false;
      
	  // temp selection ??
	  if (m_use_temp_sel && m_temp_sel_start <= t_real_sample && m_temp_sel_end >= t_real_sample_end)
	    {
	      t_selected = true;
	      t_temp_sel = true;
	    }
	  else if (m_current_sel) // real selection ??
	    {
	      if (t_sel != t_sel_end && t_sel->start <= t_real_sample)
		t_selected = true;
	    }
      
	  // marker ??
	  if (m_current_markers)
	    if (t_marker != t_marker_end && t_marker->position >= t_real_sample
		&& t_marker->position < t_real_sample_end)
	      t_marker_here = true;
      
	  // draw background
	  if (t_temp_sel)
	    pixmap.draw_line (back_temp_sel_gc, i, 0, i, m_height);
	  else if (t_selected)
	    pixmap.draw_line (back_sel_gc, i, 0, i, m_height);
	  else
	    pixmap.draw_line (back_gc, i, 0, i, m_height);
      
	  bool t_v_grid = false; //i == t_v_grid_pos;
      
	  draw_grid (i, grid_gc, t_v_grid);
	
	  if (t_v_grid)
	    t_v_grid_pos += 50;
      
	  if (plot_pos < t_wave_plot_length)
	    {
	      if ( t_progress->IsUseable (plot_pos) )
		{
		  if (t_temp_sel)
		    draw_waveform (i, wave_temp_sel_gc, plot_pos);
		  else if (t_selected)
		    draw_waveform (i, wave_sel_gc, plot_pos);
		  else
		    draw_waveform (i, wave_gc, plot_pos);
		}
	    }
	  else
	    {
	      cout << "BUG: EnhWaveView::draw_wave: plot_pos out of range!!!: " << endl
		   << "     pos: " << plot_pos << ", WP-length: " << t_wave_plot_length << endl;
	    }
      
	  /*
	    else
	    {
	    pixmap.draw_line (back_undef_gc, i, 0, i, m_height);
	    }
	  */
      
	  if (t_marker_here)
	    draw_marker (i, marker_gc);
	}
  
      // next selection segment
      if (m_current_sel && t_sel != t_sel_end)
	if (t_sel->end <= t_real_sample_end)
	  while (t_sel != t_sel_end && t_sel->end <= t_real_sample_end)
	    t_sel++;
      
      // next marker
      if (m_current_markers && t_marker != t_marker_end)
	if (t_marker->position < t_real_sample_end)
	  while (t_marker != t_marker_end && t_marker->position < t_real_sample_end)
	    t_marker++;
      
      t_real_sample += t_sps;
    } 
  
  if (m_master)
    {
      /*
      Gdk_Font t_font;
      t_font.load ("-adobe-helvetica-bold-r-normal-*-13-*-*-*-p-*-iso8859-1");
      
      Gdk_GC text_gc;
      {
	text_gc = Gdk_GC (pixmap);
	Gdk_Color color ("black");
	get_colormap ().alloc (color);
	text_gc.set_foreground (color);
      }
      
      string text;
      text = "ZFactor: " + itos ( (int) t_sps) + ", s_delta: " + itos (m_start_delta);
      pixmap.draw_string (t_font, text_gc, 5, 25, text);
      */
    }
  
  // something on the screen to redrawn ??
  
  int t_offset = m_start_delta - m_internal_delta;
  {
    int i = t_start;
    while (i < t_end)
      {
	// search to first refreshed sample
	while (i < t_end && ! refresh_region [i] )
	  ++i;
	
	// search to end of this block
	int j = i;
	while (j < t_end  && refresh_region [j] )
	  ++j;
	
	//cout << "refreshed segment: " << i << " - " << j << endl;
	
	int r_s = limit_min_max (t_start - t_offset, 0, m_width);
	int r_e = limit_min_max (t_end - t_offset, 0, m_width);
	int r_l = r_e - r_s;
	
	if (r_l > 0)
	  {
	    //cout << "redrawing: " << r_s << " - " << r_e << ": l: " << r_l << endl;
	    
	    GdkRectangle update_rect;
	    update_rect.x = r_s;
	    update_rect.y = 0;
	    update_rect.width = r_l;
	    update_rect.height = m_height;
	    draw (&update_rect);
	  
	  }
	//else
	//  cout << "NOS (Not On Screen)" << endl;
	i = j; // start at current segment's end
      }
  }
  refresh_clear ();
}

void NormalTrackView::prepare_waveform (const MinMaxAvgWavePlot::Progress* n_progress)
{
  pcm_default type;
  
  waveform_progress = n_progress;
  
  waveform_middle = m_height / 2;
  waveform_dBfs = type.dBfs ();
  waveform_vscale = waveform_dBfs / (waveform_middle - 10);  
}

void NormalTrackView::draw_waveform (int view_pos, Gdk_GC& gc, int plot_pos)
{
  // draw wave
  
  pcm_default::value_type min_v = waveform_progress->min_data [plot_pos];
  pcm_default::value_type max_v = waveform_progress->max_data [plot_pos];
  
  int min_value = min_v / waveform_vscale;
  int max_value = max_v / waveform_vscale;
  
  /* // fill wave to mid-line
  min_value = limit_max (min_value, 0);
  max_value = limit_min (max_value, 0);
  */
  pixmap.draw_line (gc,
		    view_pos, waveform_middle - min_value,
		    view_pos, waveform_middle - max_value);
}

void NormalTrackView::draw_grid (int view_pos, Gdk_GC& gc, bool draw_vert)
{
  pixmap.draw_point (gc, view_pos, waveform_middle -  waveform_dBfs / waveform_vscale);
  pixmap.draw_point (gc, view_pos, waveform_middle +  waveform_dBfs / waveform_vscale);
  if (draw_vert)
    pixmap.draw_line (gc,
		      view_pos, 0,
		      view_pos, m_height);
}

void NormalTrackView::draw_marker (int view_pos, Gdk_GC& gc)
{
  //cout << "drawwing marker ..." << endl;
  pixmap.draw_line (gc,
		    view_pos, 0,
		    view_pos, m_height);
}

void NormalTrackView::connect_selection ()
{
  if (m_current_sel)
    {
      connection_selection_selected.disconnect ();
      connection_selection_unselected.disconnect ();
      connection_selection_active_changed.disconnect ();
    }
  
  TrackLayer* t_track = m_current_plot->GetTrackLayer ();
  if (t_track)
    {
      m_current_sel = &(t_track->selection);
      
      connection_selection_selected 
	= m_current_sel->SignalSelected.connect (slot (this, &NormalTrackView::SlotSelectionSelected) );
      connection_selection_unselected
	= m_current_sel->SignalUnselected.connect (slot (this, &NormalTrackView::SlotSelectionUnselected) );
      connection_selection_active_changed
	= m_current_sel->SignalActiveChanged.connect (slot (this, &NormalTrackView::SlotSelectionActiveChanged) );
    }
  else
    {
      cout << "WARNING: Cant get Track ?!? (OK before first WavePlot::Config ...)" << endl;
      m_current_sel = 0;
    }
}

void NormalTrackView::connect_markers ()
{
  cout << "NormalTrackView::connect_markers:" << endl;
  if (m_current_markers)
    {
      cout << "  disconnecing:" << endl;
      connection_marker_added.disconnect ();
      connection_marker_deleted.disconnect ();
      connection_marker_changed.disconnect ();
    }
  
  DataLayer* t_layer = m_current_plot->GetDataLayer ();
  if (t_layer)
    {
      m_current_markers = &(t_layer->markers);
      
      cout << "  connecting:" << endl;
      connection_marker_added
	= m_current_markers->SignalAdded.connect (slot (this, &NormalTrackView::SlotMarkerAdded) );
      connection_marker_deleted
	= m_current_markers->SignalDeleted.connect (slot (this, &NormalTrackView::SlotMarkerDeleted) );
      connection_marker_changed
	= m_current_markers->SignalChanged.connect (slot (this, &NormalTrackView::SlotMarkerChanged) );
    }
  else
     {
      cout << "WARNING: Cant get Layer?!? (OK before first WavePlot::Config ...)" << endl;
      m_current_markers = 0;
     }
}

void NormalTrackView::SlotGUIRefresh ()
{
  // cout << "NormalTrackView::SlotGUIRefresh:" << endl;
  
  if (refresh_needed)
    {
      // cout << "NormalTrackView::SlotGUIRefresh: -> refreshing ..." << endl;
      draw_wave ();
    }
}

void NormalTrackView::SlotSelectionSelected (sample_index start, sample_index length)
{
  //cout << "NormalTrackView::SlotSelectionSelected:" << endl;
  
  add_refresh_from_real (start, start + length);
}

void NormalTrackView::SlotSelectionUnselected (sample_index start, sample_index length)
{
  //cout << "NormalTrackView::SlotSelectionUnselected:" << endl;
  
  add_refresh_from_real (start, start + length);
}

void NormalTrackView::SlotSelectionActiveChanged (bool n_state)
{
  //cout << "NormalTrackView::SlotSelectionActiveChanged:" << endl;
  add_refresh_all ();
}

void NormalTrackView::SlotMarkerAdded (const string& n_name, sample_index n_pos)
{
  //cout << "NormalTrackView::SlotMarkerAdded:" << endl;
  add_refresh_from_real (n_pos, n_pos + 1);
}
 
void NormalTrackView::SlotMarkerDeleted (const string& n_name, sample_index n_pos)
{
  //cout << "NormalTrackView::SlotMarkerDeleted:" << endl;
  add_refresh_from_real (n_pos, n_pos + 1);
}
  
void NormalTrackView::SlotMarkerChanged (const string& n_name, sample_index o_pos, sample_index n_pos)
{
  //cout << "NormalTrackView::SlotMarkerChanged:" << endl;
  add_refresh_from_real (o_pos, o_pos + 1);
  add_refresh_from_real (n_pos, n_pos + 1);
}

void NormalTrackView::SlotWavePlotChanged (MinMaxAvgWavePlot* n_plot)
{
  // cout << "NormalTrackView::SlotWavePlotChanged:" << endl;
  
  m_current_plot = n_plot;
  
  connect_selection ();
  connect_markers ();
  
  add_refresh_all ();
}

void NormalTrackView::SlotWavePlotUpdated (sample_index start, sample_index length)
{
  // cout << "NormalTrackView::SlotWavePlotUpdated: start: " << start << ", length: " << length << endl;
  if (length > 0) // we get 0 if more changed (in a too complex way ...)
    {
      if (start <= m_internal_delta + m_cache_length && start + length > m_internal_delta)
	{
	  // cout << "update in cached region ..." << endl;
	  sample_index t_start = limit_min (start - m_internal_delta, 0);
	  sample_index t_end = limit_max (start + length - m_internal_delta, m_width * SMOTH_SCROLL_FACTOR);
	  
	  add_refresh (t_start, t_end);
	}
      //else
      //  cout << "update in not cached region ..." << endl;
    }
  else
    {
      cout << "debug. got length 0 -> do nothing ..." << endl;
    }
}

NormalView::NormalView (View* i_view)
  : ViewImpl (i_view), m_adjustment (0, 0, 0)
{
  Project* t_project = m_view->GetProject ();
  
  m_manager = t_project->GetWavePlotManager ();
  m_timeout = & (t_project->GetApp ()->GetKit ()->SignalGUIRefresh);
  
  initialize ();
}

NormalView::NormalView (View* i_view, const NormalView& src)
  : ViewImpl (i_view), m_adjustment (0, 0, 0)
{
  Project* t_project = m_view->GetProject ();
  
  m_manager = t_project->GetWavePlotManager ();
  m_timeout = &(t_project->GetApp ()->GetKit ()->SignalGUIRefresh);
  
  initialize ();
  
  set_zoom (src.m_sps);
  m_auto_zoom_all = src.m_auto_zoom_all;
  set_start (src.m_start);
}

NormalView::NormalView (const NormalView& src)
  : ViewImpl (src.m_view), m_adjustment (0, 0, 0)
{
  m_manager = src.m_manager;
  m_timeout = src.m_timeout;
  
  initialize ();
  
  m_auto_zoom_all = src.m_auto_zoom_all;
  set_start (src.m_start);
  set_zoom (src.m_sps);
}

NormalView::~NormalView ()
{
  // TODO: nicer clean up
  
  m_manager -> WaitRequest ();
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); ++i)
    {
      del_helper2 (i -> view);
      del_helper2 (i -> track_info); 
      del_helper2 (i -> plot);
    }
  m_manager -> ContinueRequest ();
  
  delete m_menu;
  delete menu_popup;
}

void NormalView::SetStart (sample_index n_start)
{
  m_auto_zoom_all = false;
  
  set_start (n_start);
}

sample_index NormalView::Start () const
{
  return m_start;
}

void NormalView::SetZoom (sample_index n_zoom_factor)
{
  m_auto_zoom_all = false;
  set_zoom (limit_min (n_zoom_factor, 1) );
}

void NormalView::SetZoomAll ()
{
  sample_index t_length = m_manager->CurrentLayerPtr ()->Length ();
  
  m_auto_zoom_all = true;
  
  set_start (0);
  set_zoom (limit_min (m_layer_length / m_view_width, 1) );
}

int NormalView::Zoom () const
{
  return m_sps;
}

sample_index NormalView::Width () const
{
  return m_view_width;
}

sample_index NormalView::Length () const
{
  return m_view_length;
}

sample_index NormalView::Pointer () const
{
  return m_pointer_pos;
}

void NormalView::SetTrackVisible (track_index track_no, bool status)
{
  // TODO:
}

bool NormalView::IsTrackVisible (track_index track_no) const
{
  // TODO:
}

ViewImpl* NormalView::Clone (View* n_view)
{
  NormalView* t_view = new NormalView (n_view, *this);
  return t_view;
}

void NormalView::initialize ()
{
  m_start = 0;
  m_layer_length = m_manager->CurrentLayerPtr () -> Length ();
  m_plot_start = 0;
  m_plot_length = 1;
  m_cache_length = 1;
  
  m_view_width = 450;
  m_view_length = m_view_width * m_sps;
  
  m_sps = 1;
  m_view_start_delta = 0;
  
  m_pointer_pos = 0;
  
  m_auto_zoom_all = false;
  m_autoscroll = false;
  
  reconfig_plots (m_plot_start, m_sps); // create a vaild WavePlotConfig (especially the length ...)
  
  cout << "m_plot_length:" << m_plot_length << ", is: " << m_config.PlotLength () << endl;
  
  m_adjustment.changed.connect ( slot (this, &NormalView::SlotAdjustmentChanged) );
  m_adjustment.value_changed.connect ( slot (this, &NormalView::SlotAdjustmentValueChanged) );

  track_index t_tracks = m_manager->CurrentLayerPtr ()->Tracks ();

  m_scrollbar.set_adjustment (m_adjustment);
  
  App* t_app = m_view -> GetProject () -> GetApp ();
  deque<App::Preferences::MenuFilter>::const_iterator i = t_app -> GetPreferences ().recommended_view_menu.begin ();
  deque<App::Preferences::MenuFilter>::const_iterator j = t_app -> GetPreferences ().recommended_view_menu.end ();

  deque <GSMP::Menu::Filter> t_filters;
  for (i; i != j; ++i)
    t_filters.push_back (Menu::Filter ((*i) . location, (*i) . title));
  m_menu = t_app -> GetManager () -> CreateMenuCollection (t_filters, false);
  menu_popup = new MenuGtk (m_view, m_menu);

  for (track_index tr = 0; tr < t_tracks; ++tr)
    {
      cout << "adding track view: " << tr << endl;
      view_stuff t_stuff;
      
      t_stuff.plot = new WavePlotTracer<MinMaxAvgWavePlot> (m_manager, tr);
      t_stuff.plot->Config (m_config);
      t_stuff.view = new NormalTrackView (t_stuff.plot, &m_config, this, m_view_width, tr == 0);
      t_stuff.track_info = new TrackInfo (m_view -> GetProject (), tr);

      m_views.push_back (t_stuff);
    }
    
  m_manager->GetData ()->SignalCurrentLayerChanged.connect (slot (this, &NormalView::SlotLayerChanged) );
  m_timeout->connect (slot (this, &NormalView::SlotGUIRefresh) );
  
  rebuild_table ();
}

void NormalView::clear_table ()
{
  track_index t_tracks = m_manager->CurrentLayerPtr ()->Tracks ();

  remove (m_hruler);
  remove (m_scrollbar);

  cout << "REMOVE: " << t_tracks << " Tracks, " <<  m_views.size () << " Views"<< endl;

  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    {
      view_stuff& t_stuff = (*i);
      remove (*t_stuff.track_info);
      remove (*t_stuff.view);
   }
}

void NormalView::rebuild_table ()
{
  track_index t_tracks = m_manager->CurrentLayerPtr ()->Tracks ();

  cout << "REBUILD: " << t_tracks << " Tracks, " <<  m_views.size () << " Views"<< endl;

  resize (t_tracks + 2, 2);

  attach (m_hruler, 1, 2, 0, 1, (GTK_FILL|GTK_EXPAND), 0);
  attach (m_scrollbar, 1, 2, t_tracks + 1, t_tracks + 2, (GTK_FILL|GTK_EXPAND), 0);

  int y = 1;

  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    {
      view_stuff& t_stuff = (*i);      
      attach (*t_stuff.track_info, 0, 1, y, y + 1, 0, (GTK_FILL|GTK_EXPAND));
      attach (*t_stuff.view, 1, 2, y, y + 1);
      ++y;
    }

  show_all ();
}


void NormalView::set_zoom (int n_sps)
{
  if (m_sps != n_sps)
    {
      sample_index t_middle = m_start + m_view_length / 2;
      m_sps = n_sps;
      reconfig_plots (m_plot_start, m_sps);
      
      set_start (t_middle - m_view_length / 2);
      update_adjustment ();
      SignalZoomChanged (m_sps);
    }
}

void NormalView::set_start (sample_index n_start, bool adjustment_update = true,
			       bool force_plot_reconf = false)
{
  //cout << "NormalView::set_start: " << n_start << endl;
  
  n_start = limit_min (limit_max (n_start, m_layer_length - m_view_length), 0);
  
  sample_index t_delta = n_start - m_plot_start; 
  
  //cout << "  current: plot_start: " << m_plot_start << ", new start: " << n_start
  //     << ", delta: " << t_delta
  //     << endl << "  current view start delta: " << m_view_start_delta << endl;
  
  // in plot range: easy scrolling (by delta):
  if (!force_plot_reconf && t_delta >= 0 && t_delta / m_sps + m_view_width <= m_plot_length)
    {
      m_view_start_delta = t_delta / m_sps;
      set_views_start_delta ();
    }
  else
    {
      // not in range: reconfig plots (slow)
      //cout << "  can't easy scroll ..." << endl;
	
      sample_index t_cache_length = m_view_width * SCROLL_CACHE_FACTOR * m_sps;
      sample_index t_view_start_delta = m_view_width * m_sps;
	
      sample_index t_plot_start;
      // on which side should we place the caches larger part (scroll direction)
      if (t_delta > 0) // scroll right
	t_plot_start = limit_min (n_start - t_view_start_delta, 0);
      else // scroll left
	t_plot_start = limit_min (n_start - t_cache_length + t_view_start_delta, 0);
      
      //cout << "  new plot-start: " << t_plot_start << endl;
	
      m_view_start_delta = (n_start - t_plot_start) / m_sps;
      //cout << "  new m_view_start_delta: " << m_view_start_delta << endl;
      
      set_views_start_delta ();
      reconfig_plots (t_plot_start, m_sps);
    }
  
  m_start = m_plot_start + m_view_start_delta * m_sps; // correct rounding errors
  
  if (adjustment_update)
    m_adjustment.set_value (m_start);
  
  SignalStartChanged (m_start);
}

void NormalView::reconfig_plots (sample_index n_start, sample_index n_sps)
{
  // cout << "NormalView::reconfig_plots:" << endl;
  
  m_plot_start = n_start;
  m_sps = n_sps;
  
  // new optimization hack ;-) (not only mem, but also zero generation time saving ...)
  m_plot_length = limit_min (limit_max (m_view_width * SCROLL_CACHE_FACTOR,
					(m_layer_length - m_plot_start) / m_sps),
			     m_view_width * SMOTH_SCROLL_FACTOR);
  /*
  if (m_plot_length != m_view_width * SCROLL_CACHE_FACTOR)
    cout << "  Newest optimization hack: m_plot_length limited: " << m_plot_length << endl;
  */
  m_cache_length = m_plot_length * n_sps;

  m_config.SetStart (m_plot_start);
  m_config.SetPlotLength (m_plot_length);
  m_config.SetSamplesPerSample (m_sps);
  m_config.SetViewStart (m_view_start_delta);
  m_config.SetViewLength (m_view_width);
  
  // maybe this must be updated ...
  m_view_length = m_view_width * m_sps;
  m_cache_length = m_plot_length * m_sps;

  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i->plot->Config (m_config);
}

void NormalView::set_views_start_delta ()
{
  // cout << "NormalView::set_views_start_delta: " << n_view_start_delta << endl;
  m_config.SetViewStart (m_view_start_delta);
  
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    {
      i->plot->Config (m_config);
      i->view->SetStartDelta (m_view_start_delta);
    }
}

void NormalView::debug_internals ()
{
  cout << "NormalView::debug_internals:" << endl
       << "m_start:            " << m_start << endl
       << "m_layer_length:     " << m_layer_length << endl
       << "m_plot_start:       " << m_plot_start << endl
       << "m_plot_length:      " << m_plot_length << endl
       << "m_cache_length:     " << m_cache_length << endl
       << "m_view_width:       " << m_view_width << endl
       << "m_view_length:      " << m_view_length << endl
       << "m_sps:              " << m_sps << endl
       << "m_view_start_delta: " << m_view_start_delta << endl
       << "m_auto_zoom_all:    " << m_auto_zoom_all << endl
       << "m_pointer_pos:      " << m_pointer_pos << endl;
}

void NormalView::update_adjustment ()
{
  m_adjustment.set_page_size (m_view_length);
  m_adjustment.set_page_increment (m_view_length / 2);
  m_adjustment.set_step_increment (m_view_length / 10);
  m_adjustment.set_upper (m_layer_length);
  
  m_hruler.set_range (m_start,
  		       m_start + m_view_length,
		       m_start,
		       m_layer_length);
}

void NormalView::start_selection (sample_index n_current)
{
  selection_start = n_current;
  selecting = true;
  selected = false;
}

void NormalView::set_selection (sample_index n_current)
{
  if (!selecting)
    {
      selection_start = n_current;
      selection_current = n_current;
      selecting = true; 
    }
  selection_current = n_current;
  selected = true;
  
  sample_index t_sel_start = min (selection_start, selection_current);
  sample_index t_sel_end = max (selection_start, selection_current);
  
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i->view->SetTempSelection (t_sel_start, t_sel_end);
}

void NormalView::end_normal_selection ()
{
  sample_index t_sel_start = min (selection_start, selection_current);
  sample_index t_sel_end = max (selection_start, selection_current);
  
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i->view->RemoveTempSelection ();
  
  DataLayer* t_layer = m_manager->CurrentLayerPtr ();
  
  cout << "clearing: " << endl;
  for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
    (*t_layer) [tr]->selection.Clear ();
  
  cout << "selecting: " << t_sel_start << " - " << t_sel_end << endl;
  for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
    (*t_layer) [tr]->selection.Select (t_sel_start, t_sel_end - t_sel_start);
  
  //for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
  //  (*t_layer) [tr]->selection.Debug ();
      
  selecting = false;
}

void NormalView::end_state_selection (bool plus_minus)
{
  sample_index t_sel_start = min (selection_start, selection_current);
  sample_index t_sel_end = max (selection_start, selection_current);
  
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i->view->RemoveTempSelection ();
  
  DataLayer* t_layer = m_manager->CurrentLayerPtr ();
  
  if (plus_minus)
    {
      cout << "selecting: " << t_sel_start << " - " << t_sel_end << endl;
      for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
	(*t_layer) [tr] ->selection.Select (t_sel_start, t_sel_end - t_sel_start);
    }
  else
    {
      cout << "unselecting: " << t_sel_start << " - " << t_sel_end << endl;
      for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
	(*t_layer) [tr] ->selection.Unselect (t_sel_start, t_sel_end - t_sel_start);
    }
  
  //for (track_index tr = 0; tr < t_layer->Tracks (); tr++)
  //  (*t_layer) [tr] ->selection.Debug ();
  
  selecting = false;
}

void NormalView::end_no_selection_but_marker ()
{
  DataLayer* t_layer = m_manager->CurrentLayerPtr ();
  
  t_layer->markers.SetMarker ("cur_pos", selection_start);
  
  selecting = false;
}

void NormalView::SlotChildConfigured (NormalTrackView* view, int n_width)
{
  cout << "NormalView::SlotChildConfigured: " << n_width << endl;
  
  // cout << "setting new values..." << endl;
  m_view_width = n_width;
  m_view_length = m_view_width * m_sps;
  
  // check for auto-zoom all
  if (m_auto_zoom_all) {
    SetZoomAll ();
  }
  else {
    set_start (m_start, true, true);
    update_adjustment ();
  }
  
  view->SetStartDelta (m_view_start_delta, true);
}

void NormalView::SlotMotionNotify (GdkEventMotion* event, GdkModifierType state)
{
  //cout << "NormalView::SlotMotionNotify:" << endl;
  
  m_hruler.motion_notify_event.emit (event);
  
  sample_index t_cursor_pos;
  t_cursor_pos = limit_min_max (m_start + ( (sample_index) event->x * m_sps),
				0,
				m_layer_length);
  
  if (state & GDK_BUTTON1_MASK)
    {
      if (event->x < 0)
	{
	  m_autoscroll = true;
	  m_autoscroll_delta = (int) (event->x) * AUTO_SCOLL_ACCEL;
	}
      else if (event->x > m_view_width)
	{
	  m_autoscroll = true;
	  m_autoscroll_delta = (int) (event->x - m_view_width) * AUTO_SCOLL_ACCEL;
	}
      else
	m_autoscroll = false;
      
      set_selection (t_cursor_pos);
    }
  
  m_pointer_pos = t_cursor_pos;
  SignalPointerMoved (m_pointer_pos);
}

void NormalView::SlotButtenPress (GdkEventButton* event)
{
  cout << "NormalView::SlotButtenPress:" << endl;
  
  sample_index t_cursor_pos;
  t_cursor_pos = limit_min_max (m_start + ( (sample_index) event->x * m_sps),
				0,
				m_layer_length);
  
  switch (event->button)
    {
    case 0x1:
      {
	start_selection (t_cursor_pos);
      } 
      break;
    case 0x2:
      {
 	//temp_sel.mark_a = cursor_pos;
      }
      break;
    case 0x3:
      {
	menu_popup -> popup (event->button, event->time);
      }
      break;
    default:
      cout << "NormalView::SlotButtonPress: unknown event ...??" << endl;
    }
  m_pointer_pos = t_cursor_pos;
  SignalPointerMoved (m_pointer_pos);
}

void NormalView::SlotButtenRelease (GdkEventButton* event)
{
  cout << "NormalView::SlotButtenRelease:" << endl;
  
  sample_index t_cursor_pos;
  t_cursor_pos = limit_min_max (m_start + ( (sample_index) event->x * m_sps),
				0,
				m_layer_length);
  switch (event->button)
    {
    case 0x1:
      {
	// modifiers ...
	if (selected)
	  {
	    if ( (event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK ) // removew
	      {
		end_state_selection (false);
	      }
	    else if ( (event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK ) // add
	      {
		end_state_selection (true);
	      }
	    else // normal
	      {
		end_normal_selection ();
	      }
	  }
	else
	  {
	    end_no_selection_but_marker ();
	  }
      } 
      break;
    case 0x2:
      {
 	//temp_sel.mark_a = cursor_pos;
      }
      break;
    case 0x3:
      {
	// menu_popup.popup (event->button, event->time);
      }
      break;
    default:
      cout << "NormalView::SlotButtonRelease: unknown event ...??" << endl;
    }
  if (m_autoscroll)
    m_autoscroll = false;
  
  m_pointer_pos = t_cursor_pos;
  SignalPointerMoved (m_pointer_pos);
}

// this is normally from a 2nd thread !! - keep that in mind - especially
// when you use a not threadsafe GUI like GTK-- ...
void NormalView::SlotAdjustmentChanged ()
{
  /*
  cout << "NormalView::SlotAdjustmentChanged:" << endl;
  
  cout << "NormalViewChanged:" << endl;
  cout << "value: " << m_adjustment->get_value ()
       << ", lower: " << m_adjustment->get_lower ()
       << ", upper: " << m_adjustment->get_upper ()
       << ", step_inc: " << m_adjustment->get_step_increment ()
       << ", page_inc: " << m_adjustment->get_page_increment ()
       << ", page_size: " << m_adjustment->get_page_size ()
       << endl;
  */
}

void NormalView::SlotAdjustmentValueChanged ()
{
  sample_index t_start = (sample_index) m_adjustment.get_value ();
  if (t_start != m_start)
    set_start (t_start, false);
  
  m_hruler.set_range (m_start,
  		       m_start + m_view_length,
  		       m_start,
  		       m_layer_length);
}

void NormalView::SlotLayerChanged (DataLayer* n_layer)
{
  cout << "NormalView::SlotLayerChanged:" << endl;
  m_layer_length = n_layer->Length ();
  
  update_adjustment ();
  
  // check for auto-zoom all
  if (m_auto_zoom_all)
    {
      SetZoomAll ();
    }
  
  // maybe adapt start (if layer shorter and zoom at end ...)
  set_start (m_start);
  
  cout << "adjusting view count ..." << endl;
  DataLayer* t_layer = n_layer;
  track_index t_track_count = t_layer->Tracks ();
  
  if (m_views.size () != t_track_count)
    {
      clear_table ();
      
      m_manager -> WaitRequest ();
      while (m_views.size () > t_track_count)
	{
	  del_helper2 (m_views.back ().view);
	  del_helper2 (m_views.back ().track_info); 
	  del_helper2 (m_views.back ().plot);
	  
	  m_views.pop_back ();
	}
  
      if (m_views.size () < t_track_count)
	for (track_index tr = m_views.size (); tr < t_track_count; tr++)
	  {
	    cout << "adding track view: " << tr << endl;
	    view_stuff t_stuff;
	    cout << "  createing Plot ..." << endl;
	    t_stuff.plot = new WavePlotTracer<MinMaxAvgWavePlot> (m_manager, tr);
	    cout << "  configuring Plot ..." << endl;
	    t_stuff.plot->Config (m_config);
	    cout << "  createing View ..." << endl;
	    t_stuff.view = new NormalTrackView (t_stuff.plot, &m_config, this, m_view_width, tr == 0);
	
	    cout << "  creating: TrackInfo ..." << endl;
	    t_stuff.track_info = new TrackInfo (m_view -> GetProject (), tr);
	
	    m_views.push_back (t_stuff);
	
	    //t_stuff.view->SetStartDelta (m_view_start_delta);
	
	    //cout << "  showing View ..." << endl;
	    //t_stuff.view->show ();
	  }
      
      rebuild_table ();
      
      m_manager -> ContinueRequest ();
    }

  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i-> track_info -> SetDataLayer (n_layer);
}

void NormalView::SlotGUIRefresh ()
{
  if (m_autoscroll)
    {
      set_start (limit_min_max (m_start + m_autoscroll_delta * m_sps,
				0,
				limit_min (m_layer_length - m_view_length, 0) ) );
      
      // set temp selection
      if (m_autoscroll_delta < 0)
	set_selection (m_start);
      else
	set_selection (m_start + m_view_length);
    }
  
  for (deque<view_stuff>::iterator i = m_views.begin (); i != m_views.end (); i++)
    i->view->SlotGUIRefresh ();
}
