/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Normal graphic view. Will be replaces by a new overlay architecture.
 */

#ifndef GSMP_NORMAL_VIEW
#define GSMP_NORMAL_VIEW

// GSML stuff:

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "WavePlot.hh"

// stl containers
#include <vector>
#include <deque>

// GSMP
#include "ToolkitBasics.hh"
#include "TrackInfo.hh" 

// gtk--
#include <gtk--/drawingarea.h> 
#include <gtk--/box.h>
#include <gtk--/ruler.h> 
#include <gtk--/scrollbar.h>
#include <gtk--/adjustment.h>

#define SCROLL_CACHE_FACTOR 10
#define SMOTH_SCROLL_FACTOR 4
#define AUTO_SCOLL_ACCEL 6

namespace GSMP
{
  using namespace GSML;

  class NormalView; // forward

  // sub class for a single track
  class NormalTrackView : public Gtk::DrawingArea
  {
  public:
    NormalTrackView (WavePlotTracer<MinMaxAvgWavePlot>* i_plot, const WavePlotConfig* i_config,
		     NormalView* i_parent, int i_width, bool i_master = false); // master will draw additional graphic stuff (marker heads ...)
    ~NormalTrackView ();
  
    void SetStartDelta (int n_delta, bool force_redaw = false);
  
    void SetTempSelection (sample_index n_start, sample_index n_end);
    void RemoveTempSelection ();
  
    void SetRefresh (); // only temp. debug hack (context menu call ...)
  
    int Width () const;
    int Height () const;
  
    void SlotGUIRefresh ();
    
    void Debug () const;
    
  private:
  
    void add_refresh (int n_start, int n_end);
    void add_refresh_from_real (sample_index start, sample_index length);
  
    void add_refresh_all ();
    void refresh_clear ();
  
    virtual void draw_wave (); // internal to draw the wave
    
    const MinMaxAvgWavePlot::Progress* waveform_progress;
    int waveform_middle;
    int waveform_vscale;
    int waveform_dBfs;
  
    virtual void prepare_waveform (const MinMaxAvgWavePlot::Progress* n_progress);
    virtual void draw_waveform (int view_pos, Gdk_GC& gc, int plot_pos);
    virtual void draw_grid (int view_pos, Gdk_GC& gc, bool draw_vert);
    virtual void draw_marker (int view_pos, Gdk_GC& gc);
  
    void connect_selection ();
    void connect_markers ();
  
    // Signal Slots:
    void SlotWavePlotChanged (MinMaxAvgWavePlot* n_plot);
    void SlotWavePlotUpdated (sample_index start, sample_index length);
  
    SigC::Connection connection_selection_selected;
    void SlotSelectionSelected (sample_index, sample_index);
    SigC::Connection connection_selection_unselected;
    void SlotSelectionUnselected (sample_index, sample_index);
    SigC::Connection connection_selection_active_changed;
    void SlotSelectionActiveChanged (bool n_state);
  
    SigC::Connection connection_marker_added;
    void SlotMarkerAdded (const string& n_name, sample_index n_pos);
    SigC::Connection connection_marker_deleted;
    void SlotMarkerDeleted (const string& n_name, sample_index n_pos);
    SigC::Connection connection_marker_changed;
    void SlotMarkerChanged (const string& n_name, sample_index o_pos, sample_index n_pos);
  
    // gtk-- event-impl overwrites ...
    virtual void realize_impl ();
    virtual gint configure_event_impl (GdkEventConfigure* event);
    virtual gint expose_event_impl (GdkEventExpose* event);
    virtual gint button_press_event_impl (GdkEventButton* event);
    virtual gint button_release_event_impl (GdkEventButton* event);
    virtual gint motion_notify_event_impl (GdkEventMotion* event);
  
    // Gtk-- stuff
    Gdk_Window win;
  
    Gdk_Pixmap pixmap; // the cached image data
  
    Gdk_GC back_gc;  // the normal background
    Gdk_GC back_sel_gc;  //  background for selections
    Gdk_GC back_temp_sel_gc;  //  background for temp selections
    Gdk_GC back_undef_gc; // undef (not yet loaded) background
    Gdk_GC wave_gc;  // for the wave
    Gdk_GC wave_sel_gc;  // for the wave
    Gdk_GC wave_temp_sel_gc;  // for the wave
    Gdk_GC grid_gc;  // for the grid
    Gdk_GC marker_gc;  // for the markers
  
    // internal states
    int m_width; // current width and height of our window
    int m_height;
  
    int m_start_delta; // the start in plot we should display
    int m_internal_delta; // our start in plot of cache pixmap
  
    int m_cache_length; // the length of our cache pixmap
  
    sample_index m_temp_sel_start;
    sample_index m_temp_sel_end;
    bool m_use_temp_sel;
  
    SigC::Threads::Mutex refresh_lock;
    bool refresh_needed;
    vector <bool> refresh_region;
  
    // GSML Stuff
    WavePlotTracer<MinMaxAvgWavePlot>* m_waveplot;
    const WavePlotConfig* m_config;
    NormalView* m_parent;
    bool m_master;
  
    MinMaxAvgWavePlot* m_current_plot;
    TrackLayer::Selection* m_current_sel;
    DataLayer::Markers* m_current_markers;
  };
     
  class NormalView : public Gtk::Table, public ViewImpl
  {
  public:
    NormalView (View* i_view);
    NormalView (View* i_view, const NormalView& src);
    NormalView (const NormalView& src);
    virtual ~NormalView ();
  
    virtual void SetStart (sample_index n_start);
    virtual sample_index Start () const;
    
    virtual void SetZoom (sample_index n_zoom_factor);
    virtual void SetZoomAll ();
    virtual int Zoom () const;
  
    virtual sample_index Width () const;
    virtual sample_index Length () const;
  
    virtual void SetTrackVisible (track_index track_no, bool status);
    virtual bool IsTrackVisible (track_index track_no) const;
    
    virtual sample_index Pointer () const;
    
    virtual ViewImpl* Clone (View* n_view);
  
  protected:
    friend class NormalTrackView;
  
    /********************
     * internal helpers *
     ********************/
    void initialize ();

    void clear_table ();
    void rebuild_table ();
    
    void set_zoom (int n_sps);
    void set_start (sample_index n_start, bool adjustment_update = true, bool force_plot_reconf = false);
    void reconfig_plots (sample_index n_start, sample_index n_sps);
    void set_views_start_delta ();
    void update_adjustment ();
  
    void start_selection (sample_index n_current);
    void set_selection (sample_index n_current);
    void end_normal_selection ();
    void end_state_selection (bool plus_minus);
    void end_no_selection_but_marker ();
  
    void debug_internals ();

    /***********************************
     * "signal" handlers / connections *
     ***********************************/
    // some child callbacks
    void SlotChildConfigured (NormalTrackView* view, int n_width);
    void SlotButtenPress (GdkEventButton* event);
    void SlotButtenRelease (GdkEventButton* event);
    void SlotMotionNotify (GdkEventMotion* event, GdkModifierType state);
  
    // Data
    void SlotLayerChanged (DataLayer* n_layer);
  
    void SlotAdjustmentChanged ();
    void SlotAdjustmentValueChanged ();
  
    void SlotGUIRefresh (); // clock signal for delayed GUI refresh 

#ifdef OLD_HACK_MENU  
    // gtk-- callbacks (hacks)
    void slot_context_zoom_out ();
    void slot_context_zoom_in ();
    void slot_context_zoom_one2one ();
    void slot_context_zoom_all ();
    void slot_context_zoom_sel ();
  
    void slot_context_select_all ();
    void slot_context_select_clear ();
    void slot_context_select_invert ();
  
    void slot_context_debug_redraw ();
    void slot_context_debug_reload ();
    void slot_context_debug_dump_internal ();
#endif

    /*****************
     * my properties *
     *****************/
  
    // internal (important) states
    WavePlotConfig m_config;
  
    // all values we need
    sample_index m_start;        // our display start
    sample_index m_layer_length; // whole DataLayer length
    sample_index m_plot_start;   // plot's start
    sample_index m_plot_length;  // plot's length
    sample_index m_cache_length; // real length of plot's
    int m_view_width;            // view's width
    sample_index m_view_length;  // view's real length
    int m_sps;                   // out zoom factor
    
    sample_index m_pointer_pos;  // current mouse pointer pos
    
    int m_view_start_delta;      // current view start delta in the plot
    bool m_auto_zoom_all;        // sould be auto-track on the whole track
    
    // selection states
    bool selecting;
    bool selected;
    sample_index selection_start;
    sample_index selection_current;
  
    // autoscrolling states
    bool m_autoscroll;
    int m_autoscroll_delta;
  
    // GSML configs
    class view_stuff
    {
    public:
      TrackInfo* track_info;
      NormalTrackView* view;
      WavePlotTracer<MinMaxAvgWavePlot>* plot;
    };
  
    deque<view_stuff> m_views;
  
    WavePlotManager* m_manager;
    Signal0<void>* m_timeout;
  
    // gtk-- gui
    //    Gtk::Table view_table;
  
    Gtk::HRuler m_hruler;
    Gtk::HScrollbar m_scrollbar;
    Gtk::Adjustment m_adjustment;
  
    GSMP::Menu::MenuCollection* m_menu;
    MenuGtk* menu_popup;
  };

} // namespace GSMP
#endif // GSMP_NORMAL_VIEW
