/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include "PluginInterface.hh"
#include "Data.hh"
#include "Activator.hh"
#include "Project.hh"
#include "math.h"

#include "ParameterDialog.hh"
#include "PreHearFX.hh"
#include "ValuesGeneric.hh"

class FullScaleDB : public GSMP::LinearMonoFX
{
public:
  FullScaleDB ();

private:
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);
};

class SimpleSinusGen : public GSMP::LinearMonoFXPlug
{
public:
  SimpleSinusGen (double i_f, double i_vol);
  void SetFrequency (double* f);
  void SetVol (double* vol);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  double period;
  double inc;
  double m_vol;
};

class SimpleSquareGen : public GSMP::LinearMonoFXPlug
{
public:
  SimpleSquareGen (double i_f, double i_vol);
  void SetFrequency (double* f);
  void SetVol (double* vol);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  double period;
  double inc;
  GSML::pcm_24_8bit::value_type min_vol;
  GSML::pcm_24_8bit::value_type max_vol;
};

class SimpleSawGen : public GSMP::LinearMonoFXPlug
{
public:
  SimpleSawGen (double i_f, double i_vol);
  void SetFrequency (double* f);
  void SetVol (double* vol);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  double period;
  double inc;
  double m_vol;
};


class PulseGen : public GSMP::LinearMonoFXPlug
{
public:
  PulseGen (GSML::sample_index T, double vol);
  void SetPeriod (GSML::sample_index* T);
  void SetVol (double* vol);

  void Flush () {c_pos = 0;}

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  GSML::pcm_24_8bit::value_type ampl;
  GSML::sample_index c_pos;
  GSML::sample_index T;
};


/*********************************************************************************************************/

START_PLUGIN_INTERFACE;
PLUGIN_ITEM("Function","FsdbFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Development>Ffullscale dB:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","SinFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Generators>sinus:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","SquareFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-27-10","Filters>Generators>square:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","SawFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-27-10","Filters>Generators>sawtooth:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","PulseFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Generators>pulse:Project:Item:State=on", NULL);
END_PLUGIN_INTERFACE;


class FsdbFXPlugin : public GSMP::Menu::Activator
{
public:
  FsdbFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    m_project -> DoFX (new FullScaleDB);
  }
};

class SinFXPlugin : public GSMP::Menu::Activator
{
public:
  SinFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Sinus Generator"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <SimpleSinusGen>, public GSMP::DialogAction
  {
  public:
    virtual SimpleSinusGen* Create ()
    {
      return new SimpleSinusGen (f, v);
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("sinus"), string ("generated sinus")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (level, string (""));
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      level -> Set (level -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      level = new GSMP::SingleVolume (0, 1, 1, v, string ("Level"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 300.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <SimpleSinusGen> (m_project, *this);
      prehear -> SetInteractiveDouble (&SimpleSinusGen::SetVol, &v, *level);
      prehear -> SetInteractiveDouble (&SimpleSinusGen::SetFrequency, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete level;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* level;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <SimpleSinusGen>* prehear;

    GSMP::Project* m_project;

    double f;
    double v;
  };
};


class SquareFXPlugin : public GSMP::Menu::Activator
{
public:
  SquareFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Square Generator"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <SimpleSquareGen>, public GSMP::DialogAction
  {
  public:
    virtual SimpleSquareGen* Create ()
    {
      return new SimpleSquareGen (f, v);
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("square"), string ("generated square")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (level, string (""));
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      level -> Set (level -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      level = new GSMP::SingleVolume (0, 1, 1, v, string ("Level"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 300.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <SimpleSquareGen> (m_project, *this);
      prehear -> SetInteractiveDouble (&SimpleSquareGen::SetVol, &v, *level);
      prehear -> SetInteractiveDouble (&SimpleSquareGen::SetFrequency, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete level;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* level;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <SimpleSquareGen>* prehear;

    GSMP::Project* m_project;

    double f;
    double v;
  };
};


class SawFXPlugin : public GSMP::Menu::Activator
{
public:
  SawFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Sawtooth Generator"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <SimpleSawGen>, public GSMP::DialogAction
  {
  public:
    virtual SimpleSawGen* Create ()
    {
      return new SimpleSawGen (f, v);
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("sawtooth"), string ("generated sawtooth")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (level, string (""));
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      level -> Set (level -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      level = new GSMP::SingleVolume (0, 1, 1, v, string ("Level"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 300.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <SimpleSawGen> (m_project, *this);
      prehear -> SetInteractiveDouble (&SimpleSawGen::SetVol, &v, *level);
      prehear -> SetInteractiveDouble (&SimpleSawGen::SetFrequency, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete level;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* level;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <SimpleSawGen>* prehear;

    GSMP::Project* m_project;

    double f;
    double v;
  };
};


class PulseFXPlugin : public GSMP::Menu::Activator
{
public:
  PulseFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Pulse generator"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <PulseGen>, public GSMP::DialogAction
  {
  public:
    virtual PulseGen* Create ()
    {
      PulseGen* t_p = new PulseGen (T, v);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("Pulse"), string ("pulse gen")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (mT, string (""));
      dialog -> AddValue (mv, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      mv -> Set (mv -> Default ());
      mT -> Set (mT -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      mv = new GSMP::SingleVolume (0.0, 2.0, 0.25, v, string ("Amplitude"));
      mT = new GSMP::SingleShortTime (4, 44100, 2200, T, 44100, string ("Period"), 1, true);

      prehear = new GSMP::PrehearButton <PulseGen> (m_project, *this);
      prehear -> SetInteractiveDouble (&PulseGen::SetVol, &v, *mv);
      prehear -> SetInteractiveIndex (&PulseGen::SetPeriod, &T, *mT);
    }

    ~Action ()
    {
      delete prehear;
      delete mT;
      delete mv;
    }

    // Dialog Elements
    GSMP::SingleVolume* mv;
    GSMP::SingleShortTime* mT;
    GSMP::PrehearButton <PulseGen>* prehear;

    GSMP::Project* m_project;

    double v;
    GSML::sample_index T;
  };
};

PLUGIN_FACTORY(Function,FsdbFXPlugin,GSMP::Menu::Activator,FsdbFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,SinFXPlugin,GSMP::Menu::Activator,SinFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,SquareFXPlugin,GSMP::Menu::Activator,SquareFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,SawFXPlugin,GSMP::Menu::Activator,SawFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,PulseFXPlugin,GSMP::Menu::Activator,PulseFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));

/**********************************************************************************************************/

using namespace GSML;

FullScaleDB::FullScaleDB ()
  : LinearMonoFX (string ("+- 0 db"), string ("overwrite +- 0 db"))
{
  SetGenerator (true);
}

void FullScaleDB::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit type;

  pcm_24_8bit::value_type* dst_data = dst -> data;
  for (sample_index x = 0; x < length; x ++)
    dst_data [x] = type.dBfs ();
}


SimpleSinusGen::SimpleSinusGen (double  i_f, double i_vol)
{
  setup.generator = true;
  SetVol (&i_vol);
  SetFrequency (&i_f);
}

void SimpleSinusGen::SetFrequency (double* f)
{
  inc = *f;
}

void SimpleSinusGen::SetVol (double* vol)
{
  m_vol = ((double) Stype.dBfs ()) * (*vol);
}


void SimpleSinusGen::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  period = 0;
}


void SimpleSinusGen::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  for (sample_index x = 0; x < length; x ++)
    {
      double y = sin (period) * m_vol;
      data [x] (y);
      period += inc;
    }

  while (period > M_PI)
    period -= (2 * M_PI);
}



SimpleSquareGen::SimpleSquareGen (double  i_f, double i_vol)
{
  setup.generator = true;
  SetVol (&i_vol);
  SetFrequency (&i_f);
}

void SimpleSquareGen::SetFrequency (double* f)
{
  inc = *f;
}

void SimpleSquareGen::SetVol (double* vol)
{
  double t_vol = ((double) Stype.dBfs ()) * (*vol);
  max_vol (t_vol);
  t_vol = -t_vol;
  min_vol (t_vol);
  cout  << "   min/max " << min_vol.data << " " << max_vol.data << endl;
}


void SimpleSquareGen::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  period = 0;
}


void SimpleSquareGen::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  for (sample_index x = 0; x < length; x ++)
    {
      data [x] = (period > 0.0) ? min_vol : max_vol;
      period += inc;

      if (period > M_PI)
	period -= (2 * M_PI);
    }
}


SimpleSawGen::SimpleSawGen (double  i_f, double i_vol)
{
  setup.generator = true;
  SetVol (&i_vol);
  SetFrequency (&i_f);
}

void SimpleSawGen::SetFrequency (double* f)
{
  inc = *f;
}

void SimpleSawGen::SetVol (double* vol)
{
  m_vol = ((double) Stype.dBfs ()) * ((*vol) / M_PI);
}


void SimpleSawGen::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  period = 0;
}


void SimpleSawGen::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  for (sample_index x = 0; x < length; x ++)
    {
      double y = period * m_vol;
      data [x] (y);
      period += inc;

      if (period > M_PI)
	period -= (2 * M_PI);
    }
}




PulseGen::PulseGen (sample_index period, double vol)
{
  pcm_24_8bit type;
  setup . generator = true;
  vol *= type.dBfs ();
  ampl (vol);
  T = period;
  c_pos = 0;
}

void PulseGen::SetPeriod (GSML::sample_index* n_T)
{
  T = *n_T;
}

void PulseGen::SetVol (double* vol)
{
  double t_vol = *vol;
  pcm_24_8bit type;
  t_vol *= type.dBfs ();
  ampl (t_vol);
}

void PulseGen::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  Flush ();
}

void PulseGen::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  pcm_24_8bit::value_type zero;
  zero (0);

  for (sample_index x = 0; x < length; x ++)
    {
      data [x] = (c_pos % T == 0) ? ampl : zero;
      c_pos ++;
    }
}
