/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include "PluginInterface.hh"
#include "Data.hh"
#include "Activator.hh"
#include "Project.hh"
#include "ParameterDialog.hh"

#include "IIRFilters.hh"

#include "PreHearFX.hh"
#include "ValuesGeneric.hh"
#include "math.h"

#define LOCKFX SigC::Threads::MLock lazylock (lock)

class Limiter : public GSMP::LinearMonoFX
{
public:
  Limiter (double threshold, GSML::sample_index attack, GSML::sample_index decay, GSML::sample_index fall_off);

private:
  virtual void PrepareFX (GSML::sample_index length);
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  GSML::sample_index shape_pos;
  GSML::sample_index next_above_limit;

  GSML::pcm_24_8bit::value_type thresh;
  GSML::sample_index m_attack;
  GSML::sample_index m_decay;
  GSML::sample_index m_fall_off;

  double conversion_factor;
};

class OldCompressor : public GSMP::LinearMonoFX
{
public:
  OldCompressor (double gain, double level, double harmonics_level, double trim_gain);

private:
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  GSML::pcm_24_8bit::value_type m_gain;
  GSML::pcm_24_8bit::value_type m_trim_gain;

  GSML::pcm_24_8bit::value_type f_coef;
  GSML::pcm_24_8bit::value_type df_coef;

  GSML::pcm_24_8bit::value_type minus10db;
};


class Compressor : public GSMP::LinearMonoFXPlug
{
public:
  Compressor ();
  virtual ~Compressor () {};

  void SetInputGain (double* gain);

  GraphEq* GetDetectEQ ();
  void SetUseEq (bool* state);

  void SetDetectionIntervall (GSML::sample_index* i);
  void SetPeakMode (bool* mode);

  void SetThreshold (double* thresh);
  void SetSoften (double* soft);
  void SetRatio (double* ratio);
  void SetAmp (double* amp);

  void SetAttack (GSML::sample_index* i);
  void SetRelease (GSML::sample_index* i);


  // for Easy Realtime Config
  void EqSetLevel_31  (double* vol) {detect_eq.SetLevel (0, *vol);}
  void EqSetLevel_62  (double* vol) {detect_eq.SetLevel (1, *vol);}
  void EqSetLevel_125 (double* vol) {detect_eq.SetLevel (2, *vol);}
  void EqSetLevel_250 (double* vol) {detect_eq.SetLevel (3, *vol);}
  void EqSetLevel_500 (double* vol) {detect_eq.SetLevel (4, *vol);}
  void EqSetLevel_1k  (double* vol) {detect_eq.SetLevel (5, *vol);}
  void EqSetLevel_2k  (double* vol) {detect_eq.SetLevel (6, *vol);}
  void EqSetLevel_4k  (double* vol) {detect_eq.SetLevel (7, *vol);}
  void EqSetLevel_8k  (double* vol) {detect_eq.SetLevel (8, *vol);}
  void EqSetLevel_16k (double* vol) {detect_eq.SetLevel (9, *vol);}

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  LevelDetector& Detector () {if (peak_mode) return m_peak_detect; return m_avg_detect;}

  PeakLevelDetector m_peak_detect;
  AvgLevelDetector m_avg_detect;
  LinearShaper m_shaper;

  GraphEq detect_eq;
  Sdata* detect_eq_buffer;
  GSML::sample_index eq_buffer_length;

  volatile bool peak_mode;
  volatile bool use_eq;

  GSML::pcm_24_8bit::value_type ingain;
  GSML::pcm_24_8bit::value_type thresh;
  GSML::pcm_24_8bit::value_type knee_start;
  GSML::pcm_24_8bit::value_type knee_end;
  GSMP::pcm_24_8bit::value_type inv_knee_length;
  GSML::pcm_24_8bit::value_type ratio;
  GSML::pcm_24_8bit::value_type outgain;

  double d_thresh;
  double d_knee;

  GSML::sample_index m_attack;
  GSML::sample_index m_release;
};

class NoiseGate : public GSMP::LinearMonoFXPlug
{
public:
  NoiseGate ();

  void SetDetectionIntervall (GSML::sample_index* i);
  void SetPeakMode (bool* mode);

  void SetThreshold (double* thresh);

  void SetAttack (GSML::sample_index* i);
  void SetHold (GSML::sample_index* i);
  void SetRelease (GSML::sample_index* i);

  GraphEq* GetDetectEQ ();
  void SetUseEq (bool* state);

  // for Easy Realtime Config
  void EqSetLevel_31  (double* vol) {detect_eq.SetLevel (0, *vol);}
  void EqSetLevel_62  (double* vol) {detect_eq.SetLevel (1, *vol);}
  void EqSetLevel_125 (double* vol) {detect_eq.SetLevel (2, *vol);}
  void EqSetLevel_250 (double* vol) {detect_eq.SetLevel (3, *vol);}
  void EqSetLevel_500 (double* vol) {detect_eq.SetLevel (4, *vol);}
  void EqSetLevel_1k  (double* vol) {detect_eq.SetLevel (5, *vol);}
  void EqSetLevel_2k  (double* vol) {detect_eq.SetLevel (6, *vol);}
  void EqSetLevel_4k  (double* vol) {detect_eq.SetLevel (7, *vol);}
  void EqSetLevel_8k  (double* vol) {detect_eq.SetLevel (8, *vol);}
  void EqSetLevel_16k (double* vol) {detect_eq.SetLevel (9, *vol);}

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  LevelDetector& Detector () {if (peak_mode) return m_peak_detect; return m_avg_detect;}

  PeakLevelDetector m_peak_detect;
  AvgLevelDetector m_avg_detect;
  LinearShaper m_shaper;

  GraphEq detect_eq;
  Sdata* detect_eq_buffer;
  GSML::sample_index eq_buffer_length;

  volatile bool peak_mode;
  volatile bool use_eq;

  GSML::pcm_24_8bit::value_type thresh;

  GSML::sample_index m_attack;
  GSML::sample_index m_release;
  GSML::sample_index m_hold;
  GSML::sample_index hold_time_decayed;
};


class DeClick : public GSMP::LinearMonoFXPlug
{
public:
  DeClick (GSML::sample_index max_delay_data);

  void SetThreshold (double* thresh);

  void SetAttack (GSML::sample_index* i);
  void SetHold (GSML::sample_index* i);
  void SetRelease (GSML::sample_index* i);

  void SetDamp (double *d);
  void SetDuck (double *d);

private:
  SigC::Threads::Mutex lock;

  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  LinearShaper m_shaper;
  RingBuffer m_buffer;

  GSML::pcm_24_8bit::value_type thresh;

  GSML::pcm_24_8bit::value_type damp_factor;
  GSML::pcm_24_8bit::value_type duck_factor;

  GSML::pcm_24_8bit::value_type delta_val_current_pos;
  GSML::pcm_24_8bit::value_type delta_val_m_delay;

  GSML::sample_index m_attack;
  GSML::sample_index m_release;
  GSML::sample_index m_hold;
  GSML::sample_index hold_time_decayed;
  GSML::sample_index max_delay;

  RingBuffer::StaticAccess* m_delay;
  RingBuffer::StaticAccess* current_pos;
};


/*********************************************************************************************************/

START_PLUGIN_INTERFACE;
//PLUGIN_ITEM("Function","LimiterFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Dynamics&G&2>Limit:Project:Item:State=on", NULL);
//PLUGIN_ITEM("Function","CompressFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Dynamics&G&2>Compress (old):Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","CompressorPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Dynamics&G&2>Compress:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","NoiseGatePlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Dynamics&G&2>Noise gate:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","DeClickPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-09","Filters>Dynamics&G&2>de-click:Project:Item:State=on", NULL);
END_PLUGIN_INTERFACE;

class LimiterFXPlugin : public GSMP::Menu::Activator
{
public:
  LimiterFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("Limiter"));
    t_dialog -> AddDouble (0, 0.95, 0.75, &(t_action -> tresh), string ("Threshold"));
    t_dialog -> AddDouble (0, 1000, 25, &(t_action -> attack), string ("Attack"));
    t_dialog -> AddDouble (0, 10000, 150, &(t_action -> decay), string ("Decay"));
    t_dialog -> AddDouble (0, 10000, 200, &(t_action -> fall), string ("Fall off"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      return new Limiter (tresh, attack, decay, fall);
    }

    double tresh;
    double attack;
    double decay;
    double fall;  
  };
};


class CompressFXPlugin : public GSMP::Menu::Activator
{
public:
  CompressFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("OldCompressor"));
    t_dialog -> AddDouble (0, 2, 1, &(t_action -> gain), string ("Pre gain"));
    t_dialog -> AddDouble (0, 1, 0.5, &(t_action -> lev), string ("Level"));
    t_dialog -> AddDouble (0, 1, 0, &(t_action -> harm), string ("Harmonics"));
    t_dialog -> AddDouble (0, 1, 1, &(t_action -> trim), string ("Trim gain"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      return new OldCompressor (gain, lev, harm, trim);
    }

    double gain;
    double lev;
    double harm;
    double trim;  
  };
};



class CompressorPlugin : public GSMP::Menu::Activator
{
public:
  CompressorPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Compressor"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::LinearMonoFXPlugFactory <Compressor>, public GSMP::DialogAction
  {
  public:
    virtual Compressor* Create ()
    {
      Compressor* t_com = new Compressor;

      t_com -> SetInputGain (&igain);
      t_com -> SetDetectionIntervall (&cint);
      t_com -> SetPeakMode (&peak);
      t_com -> SetUseEq (&ueq);
      t_com -> SetThreshold (&th);
      t_com -> SetSoften (&kn);
      t_com -> SetRatio (&rat);
      t_com -> SetAmp (&ampl);
      t_com -> SetAttack (&att);
      t_com -> SetRelease (&rel);

      GraphEq* t_eq = t_com -> GetDetectEQ ();

      t_eq -> SetLevel_31  (&v31);
      t_eq -> SetLevel_62  (&v62);
      t_eq -> SetLevel_125 (&v125);
      t_eq -> SetLevel_250 (&v250);
      t_eq -> SetLevel_500 (&v500);
      t_eq -> SetLevel_1k  (&v1k);
      t_eq -> SetLevel_2k  (&v2k);
      t_eq -> SetLevel_4k  (&v4k);
      t_eq -> SetLevel_8k  (&v8k);
      t_eq -> SetLevel_16k (&v16k);

      return t_com;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("compressing"), string ("Compressor")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (pregain, string ("Input / Source"));

      dialog -> AddValue (calcl, string ("Level Detection"));
      dialog -> AddValue (mode, string ("Level Detection"));
      dialog -> AddValue (eq, string ("Level Detection"));

      dialog -> AddValue (thresh, string ("Transformation"));
      dialog -> AddValue (ratio, string ("Transformation"));
      dialog -> AddValue (knee, string ("Transformation"));
      dialog -> AddValue (amp, string ("Transformation"));

      dialog -> AddValue (attack, string ("Shaping"));
      dialog -> AddValue (release, string ("Shaping"));

      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      pregain -> Set (pregain -> Default ());
      mode -> Set (mode -> Default ());
      eq -> Set (eq -> Default ());
      calcl -> Set (calcl -> Default ());
      knee -> Set (knee -> Default ());
      thresh -> Set (thresh -> Default ());
      ratio -> Set (ratio -> Default ());
      amp -> Set (amp -> Default ());
      attack -> Set (attack -> Default ());
      release -> Set (release -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      pregain = new GSMP::SingleVolume (0, 2, 1, igain, string ("Input Gain"));
      mode = new GSMP::SingleCheck (false, peak, string ("Peak Mode"));
      eq = new GSMP::SingleCheck (false, ueq, string ("pre Eq"));
      calcl = new GSMP::SingleShortTime (20, 2000, 200, cint, 44100, string ("Intervall"));
      knee = new GSMP::SingleVolume (1, exp (0.7 * log (2)), exp (0.2 * log (2)), kn, string ("Soften knee"));
      thresh = new GSMP::SingleVolume (0, 1, 0.5, th, string ("Threshold"));
      ratio = new GSMP::SingleVolume (0, 1, 0.5, rat, string ("Ratio"));
      amp = new GSMP::SingleVolume (0, 2, 1, ampl, string ("Output Amplification"));
      attack = new GSMP::SingleShortTime (4, 9000, 400, att, 44100, string ("Attack"), 1, false);
      release = new GSMP::SingleShortTime (400, 80000, 4000, rel, 44100, string ("Release"), 1, false);


      // Pre Eq
      vol31 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v31, string ("31 Hz"));
      vol62 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v62, string ("62 Hz"));
      vol125 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v125, string ("125 Hz"));
      vol250 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v250, string ("250 Hz"));
      vol500 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v500, string ("500 Hz"));
      vol1k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v1k, string ("1 kHz"));
      vol2k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v2k, string ("2 kHz"));
      vol4k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v4k, string ("4 kHz"));
      vol8k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v8k, string ("8 kHz"));
      vol16k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v16k, string ("16 kHz"));

      //init Prehear button
      prehear = new GSMP::PrehearButton <Compressor> (m_project, *this);
      prehear -> SetInteractiveDouble (&Compressor::SetInputGain, &igain, *pregain);
      prehear -> SetInteractiveDouble (&Compressor::SetThreshold, &th, *thresh);
      prehear -> SetInteractiveDouble (&Compressor::SetSoften, &kn, *knee);
      prehear -> SetInteractiveDouble (&Compressor::SetRatio, &rat, *ratio);
      prehear -> SetInteractiveDouble (&Compressor::SetAmp, &ampl, *amp);
      prehear -> SetInteractiveIndex (&Compressor::SetDetectionIntervall, &cint, *calcl);
      prehear -> SetInteractiveIndex (&Compressor::SetAttack, &att, *attack);
      prehear -> SetInteractiveIndex (&Compressor::SetRelease, &rel, *release);
      prehear -> SetInteractiveBool (&Compressor::SetPeakMode, &peak, *mode);
      prehear -> SetInteractiveBool (&Compressor::SetUseEq, &ueq, *eq);

      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_31, &v31, *vol31);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_62, &v62, *vol62);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_125, &v125, *vol125);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_250, &v250, *vol250);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_500, &v500, *vol500);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_1k, &v1k, *vol1k);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_2k, &v2k, *vol2k);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_4k, &v4k, *vol4k);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_8k, &v8k, *vol8k);
      prehear -> SetInteractiveDouble (&Compressor::EqSetLevel_16k, &v16k, *vol16k);

      eq -> ValueChanged.connect (slot (this, &Action::SlotPreEqChanged));
    }

    ~Action ()
    {
      delete prehear;
      
      delete pregain;
      delete mode;
      delete eq;
      delete calcl;
      delete knee;
      delete ratio;
      delete amp;
      delete thresh;
      delete attack;
      delete release;

      delete vol31;
      delete vol62;
      delete vol125;
      delete vol250;
      delete vol500;
      delete vol1k;
      delete vol2k;
      delete vol4k;
      delete vol8k;
      delete vol16k;
    }

    // Dialog Elements
    GSMP::SingleVolume* pregain;
    GSMP::SingleCheck* mode;
    GSMP::SingleCheck* eq;
    GSMP::SingleShortTime* calcl;
    GSMP::SingleVolume* knee;
    GSMP::SingleVolume* thresh;
    GSMP::SingleVolume* ratio;
    GSMP::SingleVolume* amp;
    GSMP::SingleShortTime* attack;
    GSMP::SingleShortTime* release;

    GSMP::PrehearButton <Compressor>* prehear;

    GSMP::Project* m_project;

    // FX Values
    double igain;
    double kn;
    double th;
    double rat;
    double ampl;
    bool peak;
    bool ueq;
    GSML::sample_index cint;
    GSML::sample_index att;
    GSML::sample_index rel;


    // PRE EQ =====================

    class PreEqAction : public GSMP::DialogAction
    {
    public:
      PreEqAction (CompressorPlugin::Action* i_action)
      {
	m_action = i_action;
      }

      virtual void Launch () { }

      virtual bool IsEnabled ()
      {
	return true;
      }

      virtual void Init (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> AddValue (m_action -> vol31, string (""));
	dialog -> AddValue (m_action -> vol62, string (""));
	dialog -> AddValue (m_action -> vol125, string (""));
	dialog -> AddValue (m_action -> vol250, string (""));
	dialog -> AddValue (m_action -> vol500, string (""));
	dialog -> AddValue (m_action -> vol1k, string (""));
	dialog -> AddValue (m_action -> vol2k, string (""));
	dialog -> AddValue (m_action -> vol4k, string (""));
	dialog -> AddValue (m_action -> vol8k, string (""));
	dialog -> AddValue (m_action -> vol16k, string (""));
      }

      virtual void Reset (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> Reset ("");

	m_action -> vol31 -> Set (m_action -> vol31 -> Default ());
	m_action -> vol62 -> Set (m_action -> vol62 -> Default ());
	m_action -> vol125 -> Set (m_action -> vol125 -> Default ());
	m_action -> vol250 -> Set (m_action -> vol250 -> Default ());
	m_action -> vol500 -> Set (m_action -> vol500 -> Default ());
	m_action -> vol1k -> Set (m_action -> vol1k -> Default ());
	m_action -> vol2k -> Set (m_action -> vol2k -> Default ());
	m_action -> vol4k -> Set (m_action -> vol4k -> Default ());
	m_action -> vol8k -> Set (m_action -> vol8k -> Default ());
	m_action -> vol16k -> Set (m_action -> vol16k -> Default ());
      
	Init (dialog);
	dialog -> Show ();
      }

      CompressorPlugin::Action*  m_action;
    };

    GSMP::SingleVolume* vol31;
    GSMP::SingleVolume* vol62;
    GSMP::SingleVolume* vol125;
    GSMP::SingleVolume* vol250;
    GSMP::SingleVolume* vol500;
    GSMP::SingleVolume* vol1k;
    GSMP::SingleVolume* vol2k;
    GSMP::SingleVolume* vol4k;
    GSMP::SingleVolume* vol8k;
    GSMP::SingleVolume* vol16k;

    double v31;
    double v62;
    double v125;
    double v250;
    double v500;
    double v1k;
    double v2k;
    double v4k;
    double v8k;
    double v16k;

    void SlotPreEqChanged ()
    {
      if (eq -> State ())
	{
	  PreEqAction* t_action = new PreEqAction (this);
	  SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
	  t_dialog = m_project -> FXConfigDialog (t_action, string ("Pre EQ for Compressor"));
	  t_dialog -> Show ();
	}
    }

  };
};









class NoiseGatePlugin : public GSMP::Menu::Activator
{
public:
  NoiseGatePlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Noise Gate"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::LinearMonoFXPlugFactory <NoiseGate>, public GSMP::DialogAction
  {
  public:
    virtual NoiseGate* Create ()
    {
      NoiseGate* t_gat = new NoiseGate;

      t_gat -> SetDetectionIntervall (&cint);
      t_gat -> SetPeakMode (&peak);
      t_gat -> SetUseEq (&ueq);
      t_gat -> SetThreshold (&th);
      t_gat -> SetAttack (&att);
      t_gat -> SetHold (&hol);
      t_gat -> SetRelease (&rel);

      GraphEq* t_eq = t_gat -> GetDetectEQ ();

      t_eq -> SetLevel_31  (&v31);
      t_eq -> SetLevel_62  (&v62);
      t_eq -> SetLevel_125 (&v125);
      t_eq -> SetLevel_250 (&v250);
      t_eq -> SetLevel_500 (&v500);
      t_eq -> SetLevel_1k  (&v1k);
      t_eq -> SetLevel_2k  (&v2k);
      t_eq -> SetLevel_4k  (&v4k);
      t_eq -> SetLevel_8k  (&v8k);
      t_eq -> SetLevel_16k (&v16k);

      return t_gat;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("gating"), string ("Noise gate")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (calcl, string ("Level Detection"));
      dialog -> AddValue (mode, string ("Level Detection"));
      dialog -> AddValue (eq, string ("Level Detection"));

      dialog -> AddValue (thresh, string ("Envelope"));
      dialog -> AddValue (attack, string ("Envelope"));
      dialog -> AddValue (hold, string ("Envelope"));
      dialog -> AddValue (release, string ("Envelope"));

      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      mode -> Set (mode -> Default ());
      eq -> Set (eq -> Default ());
      calcl -> Set (calcl -> Default ());
      thresh -> Set (thresh -> Default ());
      attack -> Set (attack -> Default ());
      hold -> Set (hold -> Default ());
      release -> Set (release -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      mode = new GSMP::SingleCheck (false, peak, string ("Peak Mode"));
      eq = new GSMP::SingleCheck (false, ueq, string ("pre Eq"));
      calcl = new GSMP::SingleShortTime (5, 1000, 100, cint, 44100, string ("Intervall"));
      thresh = new GSMP::SingleVolume (0, 1, 0.5, th, string ("Threshold"));
      attack = new GSMP::SingleShortTime (2, 2000, 200, att, 44100, string ("Attack"), 1, false);
      hold = new GSMP::SingleShortTime (0, 10000, 1000, hol, 44100, string ("Hold"), 1, true);
      release = new GSMP::SingleShortTime (400, 80000, 4000, rel, 44100, string ("Release"), 1, false);

      // Pre Eq
      vol31 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v31, string ("31 Hz"));
      vol62 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v62, string ("62 Hz"));
      vol125 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v125, string ("125 Hz"));
      vol250 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v250, string ("250 Hz"));
      vol500 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v500, string ("500 Hz"));
      vol1k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v1k, string ("1 kHz"));
      vol2k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v2k, string ("2 kHz"));
      vol4k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v4k, string ("4 kHz"));
      vol8k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v8k, string ("8 kHz"));
      vol16k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v16k, string ("16 kHz"));

      //init Prehear button
      prehear = new GSMP::PrehearButton <NoiseGate> (m_project, *this);
      prehear -> SetInteractiveDouble (&NoiseGate::SetThreshold, &th, *thresh);
      prehear -> SetInteractiveIndex (&NoiseGate::SetDetectionIntervall, &cint, *calcl);
      prehear -> SetInteractiveIndex (&NoiseGate::SetAttack, &att, *attack);
      prehear -> SetInteractiveIndex (&NoiseGate::SetHold, &hol, *hold);
      prehear -> SetInteractiveIndex (&NoiseGate::SetRelease, &rel, *release);
      prehear -> SetInteractiveBool (&NoiseGate::SetPeakMode, &peak, *mode);
      prehear -> SetInteractiveBool (&NoiseGate::SetUseEq, &ueq, *eq);

      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_31, &v31, *vol31);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_62, &v62, *vol62);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_125, &v125, *vol125);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_250, &v250, *vol250);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_500, &v500, *vol500);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_1k, &v1k, *vol1k);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_2k, &v2k, *vol2k);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_4k, &v4k, *vol4k);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_8k, &v8k, *vol8k);
      prehear -> SetInteractiveDouble (&NoiseGate::EqSetLevel_16k, &v16k, *vol16k);

      eq -> ValueChanged.connect (slot (this, &Action::SlotPreEqChanged));
    }

    ~Action ()
    {
      delete prehear;
      
      delete mode;
      delete eq;
      delete calcl;
      delete thresh;
      delete attack;
      delete release;
      delete hold;

      delete vol31;
      delete vol62;
      delete vol125;
      delete vol250;
      delete vol500;
      delete vol1k;
      delete vol2k;
      delete vol4k;
      delete vol8k;
      delete vol16k;
    }

    // Dialog Elements
    GSMP::SingleCheck* mode;
    GSMP::SingleCheck* eq;
    GSMP::SingleShortTime* calcl;
    GSMP::SingleVolume* thresh;
    GSMP::SingleShortTime* attack;
    GSMP::SingleShortTime* hold;
    GSMP::SingleShortTime* release;

    GSMP::PrehearButton <NoiseGate>* prehear;

    GSMP::Project* m_project;

    // FX Values
    double th;
    bool peak;
    bool ueq;
    GSML::sample_index cint;
    GSML::sample_index att;
    GSML::sample_index rel;
    GSML::sample_index hol;

    // PRE EQ =====================

    class PreEqAction : public GSMP::DialogAction
    {
    public:
      PreEqAction (NoiseGatePlugin::Action* i_action)
      {
	m_action = i_action;
      }

      virtual void Launch () { }

      virtual bool IsEnabled ()
      {
	return true;
      }

      virtual void Init (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> AddValue (m_action -> vol31, string (""));
	dialog -> AddValue (m_action -> vol62, string (""));
	dialog -> AddValue (m_action -> vol125, string (""));
	dialog -> AddValue (m_action -> vol250, string (""));
	dialog -> AddValue (m_action -> vol500, string (""));
	dialog -> AddValue (m_action -> vol1k, string (""));
	dialog -> AddValue (m_action -> vol2k, string (""));
	dialog -> AddValue (m_action -> vol4k, string (""));
	dialog -> AddValue (m_action -> vol8k, string (""));
	dialog -> AddValue (m_action -> vol16k, string (""));
      }

      virtual void Reset (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> Reset ("");

	m_action -> vol31 -> Set (m_action -> vol31 -> Default ());
	m_action -> vol62 -> Set (m_action -> vol62 -> Default ());
	m_action -> vol125 -> Set (m_action -> vol125 -> Default ());
	m_action -> vol250 -> Set (m_action -> vol250 -> Default ());
	m_action -> vol500 -> Set (m_action -> vol500 -> Default ());
	m_action -> vol1k -> Set (m_action -> vol1k -> Default ());
	m_action -> vol2k -> Set (m_action -> vol2k -> Default ());
	m_action -> vol4k -> Set (m_action -> vol4k -> Default ());
	m_action -> vol8k -> Set (m_action -> vol8k -> Default ());
	m_action -> vol16k -> Set (m_action -> vol16k -> Default ());
      
	Init (dialog);
	dialog -> Show ();
      }

      NoiseGatePlugin::Action*  m_action;
    };

    GSMP::SingleVolume* vol31;
    GSMP::SingleVolume* vol62;
    GSMP::SingleVolume* vol125;
    GSMP::SingleVolume* vol250;
    GSMP::SingleVolume* vol500;
    GSMP::SingleVolume* vol1k;
    GSMP::SingleVolume* vol2k;
    GSMP::SingleVolume* vol4k;
    GSMP::SingleVolume* vol8k;
    GSMP::SingleVolume* vol16k;

    double v31;
    double v62;
    double v125;
    double v250;
    double v500;
    double v1k;
    double v2k;
    double v4k;
    double v8k;
    double v16k;

    void SlotPreEqChanged ()
    {
      if (eq -> State ())
	{
	  PreEqAction* t_action = new PreEqAction (this);
	  SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
	  t_dialog = m_project -> FXConfigDialog (t_action, string ("Pre EQ for Noise Gate"));
	  t_dialog -> Show ();
	}
    }

  };
};





class DeClickPlugin : public GSMP::Menu::Activator
{
public:
  DeClickPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("De-Click"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::LinearMonoFXPlugFactory <DeClick>, public GSMP::DialogAction
  {
  public:
    virtual DeClick* Create ()
    {
      DeClick* t_dc = new DeClick (200);
 
      t_dc -> SetThreshold (&th);
      t_dc -> SetAttack (&att);
      t_dc -> SetHold (&hol);
      t_dc -> SetRelease (&rel);
      t_dc -> SetDamp (&dmp);
      t_dc -> SetDuck (&dck);

      return t_dc;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("declicking"), string ("De-Click")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (thresh, string ("Transformation"));
      dialog -> AddValue (damp, string ("Transformation"));
      dialog -> AddValue (duck, string ("Transformation"));
      dialog -> AddValue (attack, string ("Envelope"));
      dialog -> AddValue (hold, string ("Envelope"));
      dialog -> AddValue (release, string ("Envelope"));

      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      thresh -> Set (thresh -> Default ());
      attack -> Set (attack -> Default ());
      hold -> Set (hold -> Default ());
      release -> Set (release -> Default ());
      duck -> Set (duck -> Default ());
      damp -> Set (damp -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      thresh = new GSMP::SingleVolume (0, 0.5, 0.3, th, string ("Threshold"));
      duck = new GSMP::SingleVolume (0, 1, 0.5, dck, string ("Duck"));
      damp = new GSMP::SingleFrequency (0.0, 2.0 * M_PI, M_PI, dmp, 44100, string ("High damp"));
      attack = new GSMP::SingleShortTime (2, 200, 5, att, 44100, string ("Attack"), 1, false);
      hold = new GSMP::SingleShortTime (0, 10000, 1000, hol, 44100, string ("Hold"), 1, true);
      release = new GSMP::SingleShortTime (100, 20000, 500, rel, 44100, string ("Release"), 1, false);

      //init Prehear button
      prehear = new GSMP::PrehearButton <DeClick> (m_project, *this);
      prehear -> SetInteractiveDouble (&DeClick::SetThreshold, &th, *thresh);
      prehear -> SetInteractiveDouble (&DeClick::SetDamp, &dmp, *damp);
      prehear -> SetInteractiveDouble (&DeClick::SetDuck, &dck, *duck);
      prehear -> SetInteractiveIndex (&DeClick::SetAttack, &att, *attack);
      prehear -> SetInteractiveIndex (&DeClick::SetHold, &hol, *hold);
      prehear -> SetInteractiveIndex (&DeClick::SetRelease, &rel, *release);
    }

    ~Action ()
    {
      delete prehear;
      

      delete thresh;
      delete attack;
      delete release;
      delete hold;
      delete duck;
      delete damp;
    }

    GSMP::SingleVolume* thresh;
    GSMP::SingleShortTime* attack;
    GSMP::SingleShortTime* hold;
    GSMP::SingleShortTime* release;
    GSMP::SingleVolume* duck;
    GSMP::SingleFrequency* damp;

    GSMP::PrehearButton <DeClick>* prehear;

    GSMP::Project* m_project;

    // FX Values
    GSML::sample_index att;
    GSML::sample_index rel;
    GSML::sample_index hol;
    double th;
    double dmp;
    double dck;
  };
};



PLUGIN_FACTORY(Function,LimiterFXPlugin,GSMP::Menu::Activator,LimiterFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,CompressFXPlugin,GSMP::Menu::Activator,CompressFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,CompressorPlugin,GSMP::Menu::Activator,CompressorPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,NoiseGatePlugin,GSMP::Menu::Activator,NoiseGatePlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,DeClickPlugin,GSMP::Menu::Activator,DeClickPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));


/**********************************************************************************************************/


using namespace GSML;

Limiter::Limiter (double threshold, GSML::sample_index attack, GSML::sample_index decay, GSML::sample_index fall_off)
  : LinearMonoFX (string ("limiting"), string ("limiter"))
{
  GSML::pcm_24_8bit type;

  conversion_factor= type.dBfs ();
  threshold *= conversion_factor;
  thresh (threshold);

  m_attack = attack;
  m_decay = decay;
  m_fall_off = fall_off;

  SetPostfetch (attack, false);
}

void Limiter::PrepareFX (GSML::sample_index length)
{
  shape_pos = -1;
  next_above_limit = -1;
}

void Limiter::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;


  for (sample_index x = 0; x < length; x ++)
    {
      if (abs (src_data [x]) >= thresh)
	{ // above threshold
	  dst_data [x] = src_data [x];
	  shape_pos = 0;
	  next_above_limit = -1;
	}

      else if (shape_pos >= 0 && shape_pos < m_decay)
	{ // in decay delay
	  dst_data [x] = src_data [x];
	  ++ shape_pos;
	  next_above_limit = -1;
	}

      else
	{
	  if (next_above_limit < 0)
	    { // scan attack range for next sample above threshold
	      for (sample_index i = 0; i < m_attack; i++)
		if (abs (src_data [x + i]) >= thresh)
		  {
		    next_above_limit = i;
		    i = m_attack;
		  }
	    }
	  else
	    -- next_above_limit;
	       
	  double factor = 0;
	
	  if (next_above_limit > 0)
	    factor = ((double) (m_attack - next_above_limit) / (double) m_attack);
	  
	  if (shape_pos > m_decay + m_fall_off)
	    shape_pos = -1; // fall off finished

	  if (shape_pos >= m_decay)
	    { // in fall off 
	      double t_factor = ((double) (m_decay + m_fall_off - shape_pos) / (double) m_fall_off);

	      if (t_factor > factor)
		factor = t_factor;

	      ++ shape_pos;
	    }

	  //factor *= conversion_factor;
	  factor *= src_data [x];
	  dst_data [x] (factor);
	}
    }
}



OldCompressor::OldCompressor (double gain, double level, double harmonics_level, double trim_gain)
  : LinearMonoFX (string ("compressing"), string ("compressor"))
{
  GSML::pcm_24_8bit type;
  double t_factor = type.dBfs ();

  gain = t_factor * gain;
  level = t_factor *  level;
  harmonics_level = t_factor * harmonics_level;
  trim_gain = t_factor * trim_gain;
  t_factor /= 2;

  f_coef (-level);
  df_coef (-harmonics_level);
  m_gain (gain);
  m_trim_gain (trim_gain);
 
  minus10db (t_factor);

  SetPostfetch (3, false);
}

void OldCompressor::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  pcm_24_8bit::value_type t_zero;
  pcm_24_8bit type;
  t_zero (0);

  for (sample_index x = 0; x < length; x ++)
    {
      bool sign = false;

      // Harmonic Precompression

      pcm_24_8bit::value_type t_delta_sample = src_data [x + 1] - src_data [x];
      if (t_delta_sample < 0)
	{
	  t_delta_sample = t_zero - t_delta_sample;
	  sign = true;
	}

      if (t_delta_sample.data > type.dBfs ())
	t_delta_sample = type.dBfs ();

      t_delta_sample = df_coef * ((t_delta_sample * t_delta_sample * t_delta_sample) - (t_delta_sample * t_delta_sample));

      pcm_24_8bit::value_type t_sample;

      if (sign)
	t_sample = src_data [x] + t_delta_sample;
      else
	t_sample = src_data [x] - t_delta_sample;
 
      // Compression

      sign = false;
      t_sample = t_sample * m_gain;

      if (t_sample < 0)
	{
	  t_sample = t_zero - t_sample;
	  sign = true;
	}

      if (t_sample.data > type.dBfs ())
	t_sample = type.dBfs ();

      t_sample = (f_coef * ((t_sample * t_sample * t_sample) - (t_sample * t_sample))) + t_sample;
      t_sample = t_sample * m_trim_gain;

      if (sign)
	dst_data [x] = t_zero - t_sample;
      else
	dst_data [x] = t_sample;
    }
}

Compressor::Compressor ()
  : m_peak_detect (2001, 2001), m_avg_detect (2001, 2001), m_shaper (1, 50, 50)
{
  pcm_24_8bit type;
  peak_mode = false;
  thresh = type.dBfs ();
  knee_start = thresh;
  knee_end = thresh;
  outgain = thresh;
  ratio (0);

  d_thresh = 1;
  d_knee = 1;

  eq_buffer_length = 0;
  detect_eq_buffer = 0;
  use_eq = false;

  m_attack = 50;
  m_release = 50;
}

void Compressor::SetInputGain (double* gain)
{
  pcm_24_8bit type;
  double t_gain = *gain * ((double) type.dBfs ());
  ingain (t_gain);
}


void Compressor::SetDetectionIntervall (GSML::sample_index* i)
{
  m_peak_detect.SetLength (*i);
  m_avg_detect.SetLength (*i);
}

void Compressor::SetPeakMode (bool* mode)
{
  peak_mode = *mode;
}

GraphEq* Compressor::GetDetectEQ ()
{
  return &detect_eq;
}


void Compressor::SetUseEq (bool* state)
{
  use_eq = *state;
}

void Compressor::SetThreshold (double* thresh)
{
  d_thresh = *thresh;
  m_shaper.SetUp (d_thresh, m_attack);
  m_shaper.SetDown (d_thresh, m_release);

  pcm_24_8bit type;
  double t_thresh = d_thresh * ((double) type.dBfs ());
  this -> thresh (t_thresh);
  t_thresh *= d_knee;
  knee_end (t_thresh);
  knee_start = this -> thresh + this -> thresh - knee_end;
  double t_inv_knee_length = (1.0 / (d_knee - 1.0)) * ((double) type.dBfs ());
  inv_knee_length (t_inv_knee_length);
}

void Compressor::SetSoften (double* soft)
{
  d_knee = *soft;
  pcm_24_8bit type;
  double t_thresh = d_thresh * ((double) type.dBfs ());
  t_thresh *= d_knee;
  knee_end (t_thresh);
  knee_start = thresh + thresh - knee_end;
  double t_inv_knee_length = (1.0 / (d_knee - 1.0)) * ((double) type.dBfs ());
  inv_knee_length (t_inv_knee_length);
}

void Compressor::SetRatio (double* ratio)
{
  pcm_24_8bit type;
  double t_ratio = *ratio * ((double) type.dBfs ());
  this -> ratio (t_ratio); 
}

void Compressor::SetAmp (double* amp)
{
  pcm_24_8bit type;
  double t_amp = *amp * ((double) type.dBfs ());
  outgain (t_amp);
}


void Compressor::SetAttack (GSML::sample_index* i)
{
  m_attack = *i;
  m_shaper.SetUp (d_thresh, m_attack);
}

void Compressor::SetRelease (GSML::sample_index* i)
{
  m_release = *i;
  m_shaper.SetDown (d_thresh, m_release);
}


void Compressor::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LevelDetector& t_detect = Detector ();
  pcm_24_8bit type;

  pcm_24_8bit::value_type offset = thresh * (type.dBfs () - ratio);

  if (use_eq)
    {
      if (eq_buffer_length < length)
	{ // resize Array
	  if (detect_eq_buffer)
	    delete detect_eq_buffer;
	  detect_eq_buffer = new Sdata [length];
	  eq_buffer_length = length;
	}

      memcpy (detect_eq_buffer, data, length * sizeof (pcm_24_8bit::value_type));
      detect_eq.ComputeBlock (detect_eq_buffer, length, current_offset);
    }

  Sdata* detect_data = (use_eq) ? detect_eq_buffer : data;

  for (int x = 0; x < length; x++)
    { 
      pcm_24_8bit::value_type t_level = m_shaper (t_detect (ingain * detect_data [x]));

      if (t_level.data < knee_start.data)
	t_level = type.dBfs ();
      else
	{ // Compressed dynamic area
	  double t_quotient = ((double) offset) / ((double) t_level);
	  t_quotient *= type.dBfs ();
	  pcm_24_8bit::value_type t_adj_level;
	  t_adj_level (t_quotient);
	  t_adj_level = t_adj_level + ratio;

	  if (t_level.data < knee_end.data)
	    { // softening
	      pcm_24_8bit::value_type t_posfact = (knee_end - t_level) * inv_knee_length;
	      t_level = t_posfact + (t_adj_level * (type.dBfs () - t_posfact));
	    }
	  else
	    t_level = t_adj_level;
	}

      data [x] = data [x] * outgain * t_level;
    }
}

void Compressor::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_peak_detect.Flush ();
  m_avg_detect.Flush ();
}











NoiseGate::NoiseGate ()
  : m_peak_detect (2001, 2001), m_avg_detect (2001, 2001), m_shaper (1, 50, 50)
{
  pcm_24_8bit type;
  peak_mode = false;

  eq_buffer_length = 0;
  detect_eq_buffer = 0;
  use_eq = false;

  GSML::pcm_24_8bit::value_type thresh = type.dBfs ();

  GSML::sample_index m_hold = 0;
  GSML::sample_index hold_time_decayed = 0;

  m_attack = 50;
  m_release = 50;
}

void NoiseGate::SetDetectionIntervall (GSML::sample_index* i)
{
  m_peak_detect.SetLength (*i);
  m_avg_detect.SetLength (*i);
}

void NoiseGate::SetPeakMode (bool* mode)
{
  peak_mode = *mode;
}

GraphEq* NoiseGate::GetDetectEQ ()
{
  return &detect_eq;
}

void NoiseGate::SetUseEq (bool* state)
{
  use_eq = *state;
}

void NoiseGate::SetThreshold (double* thresh)
{
  pcm_24_8bit type;
  double t_thresh = *thresh * ((double) type.dBfs ());
  this -> thresh (t_thresh);
}

void NoiseGate::SetAttack (GSML::sample_index* i)
{
  m_attack = *i;
  m_shaper.SetUp (1.0, m_attack);
}

void NoiseGate::SetHold (GSML::sample_index* i)
{
  m_hold = *i;
}

void NoiseGate::SetRelease (GSML::sample_index* i)
{
  m_release = *i;
  m_shaper.SetDown (1.0, m_release);
}


void NoiseGate::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LevelDetector& t_detect = Detector ();
  pcm_24_8bit type;

  if (use_eq)
    {
      if (eq_buffer_length < length)
	{ // resize Array
	  if (detect_eq_buffer)
	    delete detect_eq_buffer;
	  detect_eq_buffer = new Sdata [length];
	  eq_buffer_length = length;
	}

      memcpy (detect_eq_buffer, data, length * sizeof (pcm_24_8bit::value_type));
      detect_eq.ComputeBlock (detect_eq_buffer, length, current_offset);
    }

  Sdata* detect_data = (use_eq) ? detect_eq_buffer : data;

  for (int x = 0; x < length; x++)
    { 
      pcm_24_8bit::value_type t_level = t_detect (detect_data [x]);

      if (t_level.data >= thresh.data)
	{
	  t_level = type.dBfs ();
	  hold_time_decayed = 0;
	}
      else if (hold_time_decayed < m_hold)
	t_level = type.dBfs ();
      else
	t_level (0);

      data [x] = data [x] * m_shaper (t_level);

      hold_time_decayed++;
    }
}


void NoiseGate::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_peak_detect.Flush ();
  m_avg_detect.Flush ();
}





DeClick::DeClick (GSML::sample_index max_delay_data)
  : m_buffer (max_delay_data + 1), m_shaper (1.0, 50, 50)
{
  pcm_24_8bit type;

  m_attack = 50;
  m_release = 50;
  max_delay = max_delay_data;

  m_hold = 100;
  hold_time_decayed = 0;

  thresh = type.dBfs ();

  damp_factor (0);
  duck_factor (0);

  m_delay = m_buffer.CreateAccess (min (m_attack, max_delay));
  current_pos = m_buffer.CreateAccess (0);

  delta_val_current_pos (0);
  delta_val_m_delay (0);
}


void DeClick::SetThreshold (double* thr)
{
  pcm_24_8bit type;
  double t_thresh = *thr;
  t_thresh *= type.dBfs ();
  thresh (t_thresh);
}

void DeClick::SetAttack (GSML::sample_index* i)
{
  LOCKFX;
  m_shaper.SetUp (1.0, *i);
  m_delay -> SetOffset (min (*i, max_delay));
}

void DeClick::SetHold (GSML::sample_index* i)
{
  m_hold = *i;
}
void DeClick::SetRelease (GSML::sample_index* i)
{
  LOCKFX;
  m_shaper.SetDown (1.0, *i);
}

void DeClick::SetDamp (double *d)
{
  pcm_24_8bit type;
  double t_damp = *d;
  t_damp *= type.dBfs ();
  t_damp = 1.0 - (t_damp / (2.0 * M_PI));
  damp_factor (t_damp);
}

void DeClick::SetDuck (double *d)
{
  pcm_24_8bit type;
  double t_duck = 1.0 - *d;
  t_duck *= type.dBfs ();
  duck_factor (t_duck);
}


void DeClick::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;

  pcm_24_8bit type;

  cout << thresh.data << " " << type.dBfs () << endl;

  for (int x = 0; x < length; x++)
    { 
      pcm_24_8bit::value_type t_level = data [x] - delta_val_current_pos;
      t_level.data = abs (t_level.data);
      m_buffer [current_pos] = data [x];
      delta_val_current_pos = data [x];

      if (t_level.data >= thresh.data)
	{
	  t_level = type.dBfs ();
	  hold_time_decayed = 0;
	}
      else if (hold_time_decayed < m_hold)
	t_level = type.dBfs ();
      else
	t_level (0);

      pcm_24_8bit::value_type t_envelope = m_shaper (t_level);
      pcm_24_8bit::value_type t_data = m_buffer [m_delay];

      data [x] = (type.dBfs () - (t_envelope * duck_factor)) * (t_data - (t_envelope * damp_factor * delta_val_m_delay));

      delta_val_m_delay = t_data;

      ++hold_time_decayed;
      ++m_buffer;
    }
}

void DeClick::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  pcm_24_8bit type;
  m_buffer.Flush ();
  m_shaper.ForceSet (type.dBfs ());
  hold_time_decayed = 0;
  delta_val_m_delay (0);
  delta_val_current_pos (0);
}
