/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include "PluginInterface.hh"
#include "Data.hh"
#include "Activator.hh"
#include "Project.hh"
#include "ParameterDialog.hh"

#include "math.h"
#include "IIRFilters.hh"

#include "PreHearFX.hh"
#include "ValuesGeneric.hh"


#define LOCKFX SigC::Threads::MLock lazylock (lock)

class IIRLowPass : public GSMP::LinearMonoFXPlug
{
public:
  IIRLowPass ();

  void SetF (double* frequency);
  void SetD (double* resonance);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Biquad m_biquad;
  SigC::Threads::Mutex lock;

  double d;
  double omega;
};

class IIRHighPass : public GSMP::LinearMonoFXPlug
{
public:
  IIRHighPass ();

  void SetF (double* frequency);
  void SetD (double* resonance);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Biquad m_biquad;
  SigC::Threads::Mutex lock;

  double d;
  double omega;
};

class IIRBandPass : public GSMP::LinearMonoFXPlug
{
public:
  IIRBandPass ();

  void SetF (double* frequency);
  void SetQ (double* quality);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Biquad m_biquad;
  SigC::Threads::Mutex lock;

  double q;
  double omega;
};


class IIRBandBlock : public GSMP::LinearMonoFXPlug
{
public:
  IIRBandBlock ();

  void SetF (double* frequency);
  void SetQ (double* quality);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Biquad m_biquad;
  SigC::Threads::Mutex lock;

  double q;
  double omega;
};

class IIRBiquadGeneric : public GSMP::LinearMonoFXPlug
{
public:
  IIRBiquadGeneric ();

  void SetA1 (double* a1);
  void SetA2 (double* a2);
  void SetB0 (double* b0);
  void SetB1 (double* b1);
  void SetB2 (double* b2);

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Biquad m_biquad;
  SigC::Threads::Mutex lock;
};



class GenericAllPass : public GSMP::LinearMonoFXPlug
{
public:
  GenericAllPass ();

  void SetG (double* g);
  void SetT (GSML::sample_index* T);
private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  IIRAllPass m_all_pass;
};

class GenericComb : public GSMP::LinearMonoFX
{
public:
  GenericComb (GSML::sample_index period, double c, double damp);

private:
  virtual void PrepareFX (GSML::sample_index length);
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  IIRComb m_comb;
};

class Schroeder : public GSMP::LinearMonoFX
{
public:
  Schroeder (double reverb_time, double balance, double high_damp);
  ~Schroeder ();
private:
  virtual void PrepareFX (GSML::sample_index length);
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  GSML::pcm_24_8bit::value_type dir_fact;
  GSML::pcm_24_8bit::value_type rev_fact;

  IIRComb* combs [15];

  IIRAllPass p1;
  IIRAllPass p2;
};


class Gardner : public GSMP::LinearMonoFX
{ // a "very big room" reverb implementation after gardner
public:
  class Bridge
  {
  public:
    GSML::sample_index pre_delay;
    GSML::sample_index bridge_delay;
    GSML::sample_index post_delay;
    GSML::sample_index sub_bridge1_delay; // sum of sub_bridge_delays must NOT be greater
    GSML::sample_index sub_bridge2_delay; // than bridge_delay !
    double g_bridge;
    double g_sub_bridge1;
    double g_sub_bridge2;
    double punch_out;
  };

  Gardner (Bridge b1, Bridge b2, Bridge g3, double high_damp, double feed_back);
  ~Gardner ();
private:
  virtual void PrepareFX (GSML::sample_index length);
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  RingBuffer* m_buffer;
  IIRFirstOrder* m_feedback;
  GSML::pcm_24_8bit::value_type feedback_level;

  class Patch
  {
  public:
    Patch (Bridge& bridge, GSML::sample_index start_pos);

    IIRAllPass master_pass;
    IIRAllPass sub1;
    IIRAllPass sub2;

    GSML::sample_index master_patch_pos; // patch pos in global ringbuffer
    GSML::sample_index patch_in1; // patch into local ringbuffer
    GSML::sample_index patch_in2; // patch into local ringbuffer
  };

  Patch* patches[3];

  GSML::pcm_24_8bit::value_type outs [3];
  GSML::sample_index out_pos [3];
};

class LargeRev : public GSMP::LinearMonoFX
{
public:
  LargeRev (double dir_gain,
	    GSML::sample_index refl_delay, int refls, double refl_feedback, double refl_damp, double refl_gain, double mix_in_reverb,
	    GSML::sample_index rev_delay, double rev_feedback, double rev_damp, double rev_gain);

private:
  virtual void PrepareFX (GSML::sample_index length);
  virtual void ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length);

  Reverberator reverb;
};





class MediumRev :  public GSMP::LinearMonoFXPlug
{
public:
  MediumRev ();

  void SetDirectGain (double* dir_gain);
  void SetReverbGain (double* rev_gain);

  void SetDelay1 (GSML::sample_index* delay1);
  void SetDelay2 (GSML::sample_index* delay2);

  void SetFeedBack1 (double* fb1);
  void SetFeedBack2 (double* fb2);

  void SetHighDamp (double* high_damp);

  GraphEq* GetPreEQ ();
  void SetUseEq (bool* state);

  // for Easy Realtime Config
  void EqSetLevel_31  (double* vol) {pre_eq.SetLevel (0, *vol);}
  void EqSetLevel_62  (double* vol) {pre_eq.SetLevel (1, *vol);}
  void EqSetLevel_125 (double* vol) {pre_eq.SetLevel (2, *vol);}
  void EqSetLevel_250 (double* vol) {pre_eq.SetLevel (3, *vol);}
  void EqSetLevel_500 (double* vol) {pre_eq.SetLevel (4, *vol);}
  void EqSetLevel_1k  (double* vol) {pre_eq.SetLevel (5, *vol);}
  void EqSetLevel_2k  (double* vol) {pre_eq.SetLevel (6, *vol);}
  void EqSetLevel_4k  (double* vol) {pre_eq.SetLevel (7, *vol);}
  void EqSetLevel_8k  (double* vol) {pre_eq.SetLevel (8, *vol);}
  void EqSetLevel_16k (double* vol) {pre_eq.SetLevel (9, *vol);}

private:
  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);

  Reverberator reverb;

  GraphEq pre_eq;
  Sdata* pre_eq_buffer;
  GSML::sample_index eq_buffer_length;
  volatile bool use_eq;

  double d_gain;
  double r_gain;

  CrossInserter* d1;
  CrossInserter* d2;

  CrossInserter* fb1;
  double fb2; // Feedback 2 will be done via reverberator Feedback

  double damp;
};



/*********************************************************************************************************/

START_PLUGIN_INTERFACE;
PLUGIN_ITEM("Function","IIRLowFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Frequency&G&1>IIR low pass filter:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","IIRHighFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Frequency&G&1>IIR high pass filter:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","IIRBandFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Frequency&G&1>IIR band pass filter:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","IIRAntiBandFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Frequency&G&1>IIR band block filter:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","IIRBiquadGenericFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Generic>IIR biquad filter:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","IIRGenericAllPassFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Generic>IIR all pass:Project:Item:State=on", NULL);
//PLUGIN_ITEM("Function","IIRGenericCombFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Generic>IIR comb filter:Project:Item:State=on", NULL);
//PLUGIN_ITEM("Function","GardnerFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Reverb&K>Gardner:Project:Item:State=on", NULL);
//PLUGIN_ITEM("Function","SchroederFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Reverb&K>Schroeder:Project:Item:State=on", NULL);
//PLUGIN_ITEM("Function","LargeRevFXPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Reverb&K>Large size room&B:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","MediumRevPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Reverb&K>Medium size room&B&-1:Project:Item:State=on", NULL);
PLUGIN_ITEM("Function","GraphEqPlugin", "Valentin Ziegler <valentin.ziegler@student.hu-berlin.de>","2001-08-10","Filters>Frequency&G&1>Graphical EQ&A:Project:Item:State=on", NULL);

END_PLUGIN_INTERFACE;

class IIRLowFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRLowFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("IIR lowpass"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <IIRLowPass>, public GSMP::DialogAction
  {
  public:
    virtual IIRLowPass* Create ()
    {
      IIRLowPass* t_p = new IIRLowPass;
      t_p -> SetF (&f);
      t_p -> SetD (&d);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR lowpass"), string ("IIR lowpass")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (res, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      res -> Set (res -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      res = new GSMP::SingleVolume (0.25, exp (3.5 * log (2)), 1, d, string ("Resonance"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 600.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <IIRLowPass> (m_project, *this);
      prehear -> SetInteractiveDouble (&IIRLowPass::SetD, &d, *res);
      prehear -> SetInteractiveDouble (&IIRLowPass::SetF, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete res;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* res;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <IIRLowPass>* prehear;

    GSMP::Project* m_project;

    double f;
    double d;
  };
};



class IIRHighFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRHighFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("IIR highpass"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <IIRHighPass>, public GSMP::DialogAction
  {
  public:
    virtual IIRHighPass* Create ()
    {
      IIRHighPass* t_p = new IIRHighPass;
      t_p -> SetF (&f);
      t_p -> SetD (&d);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR highpass"), string ("IIR highpass")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (res, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      res -> Set (res -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      res = new GSMP::SingleVolume (0.25, exp (3.5 * log (2)), 1, d, string ("Resonance"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 600.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <IIRHighPass> (m_project, *this);
      prehear -> SetInteractiveDouble (&IIRHighPass::SetD, &d, *res);
      prehear -> SetInteractiveDouble (&IIRHighPass::SetF, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete res;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* res;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <IIRHighPass>* prehear;

    GSMP::Project* m_project;

    double f;
    double d;
  };
};


class IIRBandFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRBandFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("IIR bandpass"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <IIRBandPass>, public GSMP::DialogAction
  {
  public:
    virtual IIRBandPass* Create ()
    {
      IIRBandPass* t_p = new IIRBandPass;
      t_p -> SetF (&f);
      t_p -> SetQ (&q);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR bandpass"), string ("IIR bandpass")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (qua, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      qua -> Set (qua -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      qua = new GSMP::SingleVolume (0.25, exp (3.5 * log (2)), 1, q, string ("Quality"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 600.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <IIRBandPass> (m_project, *this);
      prehear -> SetInteractiveDouble (&IIRBandPass::SetQ, &q, *qua);
      prehear -> SetInteractiveDouble (&IIRBandPass::SetF, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete qua;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* qua;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <IIRBandPass>* prehear;

    GSMP::Project* m_project;

    double f;
    double q;
  };
};


class IIRAntiBandFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRAntiBandFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("IIR bandblock"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <IIRBandBlock>, public GSMP::DialogAction
  {
  public:
    virtual IIRBandBlock* Create ()
    {
      IIRBandBlock* t_p = new IIRBandBlock;
      t_p -> SetF (&f);
      t_p -> SetQ (&q);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR bandblock"), string ("IIR bandblock")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (frequency, string (""));
      dialog -> AddValue (qua, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      qua -> Set (qua -> Default ());
      frequency -> Set (frequency -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      qua = new GSMP::SingleVolume (0.25, exp (3.5 * log (2)), 1, q, string ("Quality"));
      frequency = new GSMP::SingleFrequency (2.0 * M_PI * 30.0 / 44100.0,
					     2.0 * M_PI * 15000.0 / 44100.0,
					     2.0 * M_PI * 600.0 / 44100.0,
					     f, 44100, string ("Frequency"));

      prehear = new GSMP::PrehearButton <IIRBandBlock> (m_project, *this);
      prehear -> SetInteractiveDouble (&IIRBandBlock::SetQ, &q, *qua);
      prehear -> SetInteractiveDouble (&IIRBandBlock::SetF, &f, *frequency);
    }

    ~Action ()
    {
      delete prehear;
      delete qua;
      delete frequency;
    }

    // Dialog Elements
    GSMP::SingleVolume* qua;
    GSMP::SingleFrequency* frequency;
    GSMP::PrehearButton <IIRBandBlock>* prehear;

    GSMP::Project* m_project;

    double f;
    double q;
  };
};


class IIRBiquadGenericFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRBiquadGenericFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Generic IIR biquad filter"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <IIRBiquadGeneric>, public GSMP::DialogAction
  {
  public:
    virtual IIRBiquadGeneric* Create ()
    {
      IIRBiquadGeneric* t_bq = new IIRBiquadGeneric;

      t_bq -> SetA1 (&a1);
      t_bq -> SetA2 (&a2);
      t_bq -> SetB0 (&b0);
      t_bq -> SetB1 (&b1);
      t_bq -> SetB2 (&b2);
      return t_bq;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR biquad"), string ("generic biquad IIR filter")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (va1, string ("Recursion"));
      dialog -> AddValue (va2, string ("Recursion"));
      dialog -> AddValue (vb0, string ("Mix"));
      dialog -> AddValue (vb1, string ("Mix"));
      dialog -> AddValue (vb2, string ("Mix"));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      va1 -> Set (va1 -> Default ());
      va2 -> Set (va2 -> Default ());
      vb0 -> Set (vb0 -> Default ());
      vb1 -> Set (vb1 -> Default ());
      vb2 -> Set (vb2 -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      // initialaize values
      va1 = new GSMP::SingleVolume (-2.0, 2.0, 0.0, a1, string ("t - T"));
      va2 = new GSMP::SingleVolume (-1.0, 1.0, 0.0, a2, string ("t - 2T"));
      vb0 = new GSMP::SingleVolume (-4.0, 4.0, 1.0, b0, string ("t"));
      vb1 = new GSMP::SingleVolume (-4.0, 4.0, 0.0, b1, string ("t - T"));
      vb2 = new GSMP::SingleVolume (-4.0, 4.0, 0.0, b2, string ("t - 2T"));

      //init Prehear button
      prehear = new GSMP::PrehearButton <IIRBiquadGeneric> (m_project, *this);
      prehear -> SetInteractiveDouble (&IIRBiquadGeneric::SetA1, &a1, *va1);
      prehear -> SetInteractiveDouble (&IIRBiquadGeneric::SetA2, &a2, *va2);
      prehear -> SetInteractiveDouble (&IIRBiquadGeneric::SetB0, &b0, *vb0);
      prehear -> SetInteractiveDouble (&IIRBiquadGeneric::SetB1, &b1, *vb1);
      prehear -> SetInteractiveDouble (&IIRBiquadGeneric::SetB2, &b2, *vb2);
    }

    ~Action ()
    {
      delete prehear;
      delete va1;
      delete va2;
      delete vb0;
      delete vb1;
      delete vb2;
    }

    // Dialog Elements
    GSMP::SingleVolume* va1;
    GSMP::SingleVolume* va2;
    GSMP::SingleVolume* vb0;
    GSMP::SingleVolume* vb1;
    GSMP::SingleVolume* vb2;

    GSMP::PrehearButton <IIRBiquadGeneric>* prehear;

    GSMP::Project* m_project;

    // FX Values
    double a1;
    double a2;
    double b0;
    double b1;
    double b2;
  };
};



class IIRGenericAllPassFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRGenericAllPassFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Generic IIR Allpass"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <GenericAllPass>, public GSMP::DialogAction
  {
  public:
    virtual GenericAllPass* Create ()
    {
      GenericAllPass* t_p = new GenericAllPass;
      t_p -> SetT (&T);
      t_p -> SetG (&g);
      return t_p;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("IIR Allpass"), string ("generic IIR allpass")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (mT, string (""));
      dialog -> AddValue (mg, string (""));
      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      mg -> Set (mg -> Default ());
      mT -> Set (mT -> Default ());

      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      mg = new GSMP::SingleVolume (0.0, 1.0, 0.25, g, string ("g"));
      mT = new GSMP::SingleShortTime (5, 4410, 44, T, 44100, string ("Period"), 1, true);

      prehear = new GSMP::PrehearButton <GenericAllPass> (m_project, *this);
      prehear -> SetInteractiveDouble (&GenericAllPass::SetG, &g, *mg);
      prehear -> SetInteractiveIndex (&GenericAllPass::SetT, &T, *mT);
    }

    ~Action ()
    {
      delete prehear;
      delete mT;
      delete mg;
    }

    // Dialog Elements
    GSMP::SingleVolume* mg;
    GSMP::SingleShortTime* mT;
    GSMP::PrehearButton <GenericAllPass>* prehear;

    GSMP::Project* m_project;

    double g;
    GSML::sample_index T;
  };
};




// class IIRGenericAllPassFXPlugin : public GSMP::Menu::Activator
// {
// public:
//   IIRGenericAllPassFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
//     : Activator (i_app, i_project, i_view) {}
  
//   virtual void Execute ()
//   {
//     Action* t_action = new Action;
//     SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("all pass filter"));
//     t_dialog -> AddDouble (5, 500, 10, &(t_action -> t), string ("Period time"));
//     t_dialog -> AddDouble (0.1, 0.9, 0.5, &(t_action -> g), string ("G Factor"));
//     t_dialog -> Show ();
//   }

// private:
//   class Action : public GSMP::ParameterDialogAction
//   {
//   public:
//     virtual GSMP::FXBase* Build ()
//     {
//       return new GenericAllPass (g, t);
//     }

//     double t;
//     double g;
//   };
// };



class IIRGenericCombFXPlugin : public GSMP::Menu::Activator
{
public:
  IIRGenericCombFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("comb filter"));
    t_dialog -> AddDouble (3, 500, 10, &(t_action -> t), string ("Period time"));
    t_dialog -> AddDouble (0.1, 0.9, 0.5, &(t_action -> c), string ("Recursion factor"));
    t_dialog -> AddDouble (0.1, 0.9, 0.5, &(t_action -> d), string ("High damp"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      return new GenericComb (t, c, d);
    }

    double t;
    double c;
    double d;
  };
};


class SchroederFXPlugin : public GSMP::Menu::Activator
{
public:
  SchroederFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("Schroeder reverb"));
    t_dialog -> AddDouble (0, 1, 0.75, &(t_action -> t), string ("Reverb time"));
    t_dialog -> AddDouble (0.1, 0.9, 0.5, &(t_action -> b), string ("Balance"));
    t_dialog -> AddDouble (0.1, 0.9, 0.5, &(t_action -> d), string ("High damp"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      return new Schroeder (t, b, d);
    }

    double t;
    double b;
    double d;
  };
};



class GardnerFXPlugin : public GSMP::Menu::Activator
{
public:
  GardnerFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("Gardner reverb"));
    t_dialog -> AddDouble (0, 0.99, 0.95, &(t_action -> fb), string ("feed_back"));
    t_dialog -> AddDouble (0, 1, 0.3, &(t_action -> hd), string ("high_damp"));
    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      Gardner::Bridge b1;
      Gardner::Bridge b2;
      Gardner::Bridge b3;

      b1.pre_delay = 5;
      b1.bridge_delay = 940;
      b1.post_delay = 170;
      b1.sub_bridge1_delay = 350;
      b1.sub_bridge2_delay = 520;
      b1.g_bridge = 0.4;
      b1.g_sub_bridge1 = 0.3;
      b1.g_sub_bridge2 = 0.3;
      b1.punch_out = 0.34;


      b2.pre_delay = 740;
      b2.bridge_delay = 3820;
      b2.post_delay = 1360;
      b2.sub_bridge1_delay = 2720;
      b2.sub_bridge2_delay = 10;
      b2.g_bridge = 0.5;
      b2.g_sub_bridge1 = 0.25;
      b2.g_sub_bridge2 = 0;
      b2.punch_out = 0.34;

      b3.pre_delay = 130;
      b3.bridge_delay = 5280;
      b3.post_delay = 5;
      b3.sub_bridge1_delay = 3340;
      b3.sub_bridge2_delay = 1320;
      b3.g_bridge = 0.5;
      b3.g_sub_bridge1 = 0.25;
      b3.g_sub_bridge2 = 0.25;
      b3.punch_out = 0.14;

      return new Gardner (b1, b2, b3, hd, fb);
    }

    double fb;
    double hd;
  };
};



class LargeRevFXPlugin : public GSMP::Menu::Activator
{
public:
  LargeRevFXPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action;
    SigC::Handle <GSMP::ParameterDialog, SigC::Scopes::Extend> t_dialog = m_project -> FXConfigDialog (t_action, string ("Large room reverb"));
    t_dialog -> AddCount (0, 10, 0, &(t_action -> refls), string ("Early Reflections"));
    t_dialog -> AddDouble (100, 44100, 500, &(t_action -> refldel), string ("Reflections delay"));
    t_dialog -> AddDouble (0, 1, 0.5, &(t_action -> reflfb), string ("Reflections feedback"));
    t_dialog -> AddDouble (0, 1, 0.3, &(t_action -> reflhd), string ("Reflections high damp"));
    t_dialog -> AddDouble (0, 1, 0.5, &(t_action -> reflg), string ("Reflections mix"));
    t_dialog -> AddDouble (0, 1, 0.3, &(t_action -> reflm), string ("Reflections cross mix into reverb"));

    t_dialog -> AddDouble (100, 44100, 200, &(t_action -> revdel), string ("Reverb delay (after refl delay)"));
    t_dialog -> AddDouble (0, 1, 0.95, &(t_action -> revfb), string ("Reverb feedback"));
    t_dialog -> AddDouble (0, 1, 0.5, &(t_action -> revhd), string ("Reverb high damp"));
    t_dialog -> AddDouble (0, 1, 0.8, &(t_action -> revg), string ("Reverb mix"));

    t_dialog -> AddDouble (0, 1, 0.7, &(t_action -> dg), string ("Original signal mix"));

    t_dialog -> Show ();
  }

private:
  class Action : public GSMP::ParameterDialogAction
  {
  public:
    virtual GSMP::FXBase* Build ()
    {
      return new LargeRev (dg, (GSML::sample_index) refldel, refls, reflfb, reflhd, reflg, reflm, 
			   (GSML::sample_index) (revdel+ refldel), revfb, revhd, revg);
    }

    int refls;
    double refldel;
    double reflfb;
    double reflhd;
    double reflg;
    double reflm;

    double revdel;
    double revfb;
    double revhd;
    double revg;

    double dg;
  };
};


class MediumRevPlugin : public GSMP::Menu::Activator
{
public:
  MediumRevPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Medium Room Reverb"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <MediumRev>, public GSMP::DialogAction
  {
  public:
    virtual MediumRev* Create ()
    {
      MediumRev* t_rev = new MediumRev;

      t_rev -> SetDirectGain (&dg);
      t_rev -> SetReverbGain (&rg);
      t_rev -> SetDelay1 (&d1);
      t_rev -> SetDelay2 (&d2);
      t_rev -> SetFeedBack1 (&fb1);
      t_rev -> SetFeedBack2 (&fb2);
      t_rev -> SetHighDamp (&hd);

      GraphEq* t_eq = t_rev -> GetPreEQ ();

      t_eq -> SetLevel_31  (&v31);
      t_eq -> SetLevel_62  (&v62);
      t_eq -> SetLevel_125 (&v125);
      t_eq -> SetLevel_250 (&v250);
      t_eq -> SetLevel_500 (&v500);
      t_eq -> SetLevel_1k  (&v1k);
      t_eq -> SetLevel_2k  (&v2k);
      t_eq -> SetLevel_4k  (&v4k);
      t_eq -> SetLevel_8k  (&v8k);
      t_eq -> SetLevel_16k (&v16k);

      return t_rev;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("reverbating"), string ("Medium room reverb")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (del1, string ("Input"));
      dialog -> AddValue (del2, string ("Input"));
      dialog -> AddValue (eq, string ("Input"));

      dialog -> AddValue (feedb1, string ("Parameters"));
      dialog -> AddValue (feedb2, string ("Parameters"));
      dialog -> AddValue (damp, string ("parameters"));

      dialog -> AddValue (dgain, string ("Mix"));
      dialog -> AddValue (rgain, string ("Mix"));

      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      del1 -> Set (del1 -> Default ());
      del2 -> Set (del2 -> Default ());
      eq -> Set (eq -> Default ());

      feedb1 -> Set (feedb1 -> Default ());
      feedb2 -> Set (feedb2 -> Default ());
      damp -> Set (damp -> Default ());
      dgain -> Set (dgain -> Default ());
      rgain -> Set (rgain -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      // initialaize values
      feedb1 = new GSMP::SingleVolume (0.1, 0.95, 0.84, fb1, string ("Feedback 1"));
      feedb2 = new GSMP::SingleVolume (0.1, 0.95, 0.76, fb2, string ("Feedback 2"));
      dgain = new GSMP::SingleVolume (0.0, 2.0, 1.0, dg, string ("Direct"));
      rgain = new GSMP::SingleVolume (0.0, 2.0, 0.5, rg, string ("Reverb"));

      eq = new GSMP::SingleCheck (false, ueq, string ("pre reverb Eq"));

      damp = new GSMP::SingleFrequency (0.0, 2.0 * M_PI, M_PI / 2.0, hd, 44100, string ("High damp"));

      del1 = new GSMP::SingleShortTime (10, 22050, 1500, d1, 44100, string ("Delay 1"), 1, false);
      del2 = new GSMP::SingleShortTime (10, 22050, 4410, d2, 44100, string ("Delay 2"), 1, false);

      // Pre Eq
      vol31 = new GSMP::SingleVolume (0.0, 2.0, 0.0, v31, string ("31 Hz"));
      vol62 = new GSMP::SingleVolume (0.0, 2.0, 0.3, v62, string ("62 Hz"));
      vol125 = new GSMP::SingleVolume (0.0, 2.0, 0.7, v125, string ("125 Hz"));
      vol250 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v250, string ("250 Hz"));
      vol500 = new GSMP::SingleVolume (0.0, 2.0, 0.7, v500, string ("500 Hz"));
      vol1k = new GSMP::SingleVolume (0.0, 2.0, 0.3, v1k, string ("1 kHz"));
      vol2k = new GSMP::SingleVolume (0.0, 2.0, 0.0, v2k, string ("2 kHz"));
      vol4k = new GSMP::SingleVolume (0.0, 2.0, 0.0, v4k, string ("4 kHz"));
      vol8k = new GSMP::SingleVolume (0.0, 2.0, 0.0, v8k, string ("8 kHz"));
      vol16k = new GSMP::SingleVolume (0.0, 2.0, 0.0, v16k, string ("16 kHz"));


      //init Prehear button
      prehear = new GSMP::PrehearButton <MediumRev> (m_project, *this);
      prehear -> SetInteractiveDouble (&MediumRev::SetFeedBack1, &fb1, *feedb1);
      prehear -> SetInteractiveDouble (&MediumRev::SetFeedBack2, &fb2, *feedb2);
      prehear -> SetInteractiveDouble (&MediumRev::SetDirectGain, &dg, *dgain);
      prehear -> SetInteractiveDouble (&MediumRev::SetReverbGain, &rg, *rgain);
      prehear -> SetInteractiveDouble (&MediumRev::SetHighDamp, &hd, *damp);
      prehear -> SetInteractiveIndex (&MediumRev::SetDelay1, &d1, *del1);
      prehear -> SetInteractiveIndex (&MediumRev::SetDelay2, &d2, *del2);

      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_31, &v31, *vol31);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_62, &v62, *vol62);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_125, &v125, *vol125);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_250, &v250, *vol250);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_500, &v500, *vol500);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_1k, &v1k, *vol1k);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_2k, &v2k, *vol2k);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_4k, &v4k, *vol4k);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_8k, &v8k, *vol8k);
      prehear -> SetInteractiveDouble (&MediumRev::EqSetLevel_16k, &v16k, *vol16k);

      eq -> ValueChanged.connect (slot (this, &Action::SlotPreEqChanged));
    }

    ~Action ()
    {
      delete prehear;

      delete del1;
      delete del2;
      delete feedb1;
      delete feedb2;
      delete damp;
      delete dgain;
      delete rgain;

      delete vol31;
      delete vol62;
      delete vol125;
      delete vol250;
      delete vol500;
      delete vol1k;
      delete vol2k;
      delete vol4k;
      delete vol8k;
      delete vol16k;

    }

    // Dialog Elements
    GSMP::SingleShortTime* del1;
    GSMP::SingleShortTime* del2;
    GSMP::SingleVolume* feedb1;
    GSMP::SingleVolume* feedb2;
    GSMP::SingleFrequency* damp;
    GSMP::SingleVolume* dgain;
    GSMP::SingleVolume* rgain;
    GSMP::SingleCheck* eq;

    GSMP::PrehearButton <MediumRev>* prehear;

    GSMP::Project* m_project;

    // FX Values
    double dg;
    double rg;
    double fb1;
    double fb2;
    double hd;

    bool ueq;

    GSML::sample_index d1;
    GSML::sample_index d2;

    // PRE EQ =====================

    class PreEqAction : public GSMP::DialogAction
    {
    public:
      PreEqAction (MediumRevPlugin::Action* i_action)
      {
	m_action = i_action;
      }

      virtual void Launch () { }

      virtual bool IsEnabled ()
      {
	return true;
      }

      virtual void Init (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> AddValue (m_action -> vol31, string (""));
	dialog -> AddValue (m_action -> vol62, string (""));
	dialog -> AddValue (m_action -> vol125, string (""));
	dialog -> AddValue (m_action -> vol250, string (""));
	dialog -> AddValue (m_action -> vol500, string (""));
	dialog -> AddValue (m_action -> vol1k, string (""));
	dialog -> AddValue (m_action -> vol2k, string (""));
	dialog -> AddValue (m_action -> vol4k, string (""));
	dialog -> AddValue (m_action -> vol8k, string (""));
	dialog -> AddValue (m_action -> vol16k, string (""));
      }

      virtual void Reset (GSMP::EnhParameterDialog* dialog)
      {
	dialog -> Reset ("");

	m_action -> vol31 -> Set (m_action -> vol31 -> Default ());
	m_action -> vol62 -> Set (m_action -> vol62 -> Default ());
	m_action -> vol125 -> Set (m_action -> vol125 -> Default ());
	m_action -> vol250 -> Set (m_action -> vol250 -> Default ());
	m_action -> vol500 -> Set (m_action -> vol500 -> Default ());
	m_action -> vol1k -> Set (m_action -> vol1k -> Default ());
	m_action -> vol2k -> Set (m_action -> vol2k -> Default ());
	m_action -> vol4k -> Set (m_action -> vol4k -> Default ());
	m_action -> vol8k -> Set (m_action -> vol8k -> Default ());
	m_action -> vol16k -> Set (m_action -> vol16k -> Default ());
      
	Init (dialog);
	dialog -> Show ();
      }

      MediumRevPlugin::Action*  m_action;
    };

    GSMP::SingleVolume* vol31;
    GSMP::SingleVolume* vol62;
    GSMP::SingleVolume* vol125;
    GSMP::SingleVolume* vol250;
    GSMP::SingleVolume* vol500;
    GSMP::SingleVolume* vol1k;
    GSMP::SingleVolume* vol2k;
    GSMP::SingleVolume* vol4k;
    GSMP::SingleVolume* vol8k;
    GSMP::SingleVolume* vol16k;

    double v31;
    double v62;
    double v125;
    double v250;
    double v500;
    double v1k;
    double v2k;
    double v4k;
    double v8k;
    double v16k;

    void SlotPreEqChanged ()
    {
      if (eq -> State ())
	{
	  PreEqAction* t_action = new PreEqAction (this);
	  SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
	  t_dialog = m_project -> FXConfigDialog (t_action, string ("Pre EQ for MediumRev"));
	  t_dialog -> Show ();
	}
    }

  };
};


class GraphEqPlugin : public GSMP::Menu::Activator
{
public:
  GraphEqPlugin (GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view)
    : Activator (i_app, i_project, i_view) {}
  
  virtual void Execute ()
  {
    Action* t_action = new Action (m_project);
    SigC::Handle <GSMP::EnhParameterDialog, SigC::Scopes::Extend> t_dialog;
    t_dialog = m_project -> FXConfigDialog (t_action, string ("Graphical Eq"));
    t_dialog -> Show ();
  }

private:

  class Action : public GSMP::LinearMonoFXPlugFactory <GraphEq>, public GSMP::DialogAction
  {
  public:
    virtual GraphEq* Create ()
    {
      GraphEq* t_eq = new GraphEq;

      t_eq -> SetLevel_31  (&v31);
      t_eq -> SetLevel_62  (&v62);
      t_eq -> SetLevel_125 (&v125);
      t_eq -> SetLevel_250 (&v250);
      t_eq -> SetLevel_500 (&v500);
      t_eq -> SetLevel_1k  (&v1k);
      t_eq -> SetLevel_2k  (&v2k);
      t_eq -> SetLevel_4k  (&v4k);
      t_eq -> SetLevel_8k  (&v8k);
      t_eq -> SetLevel_16k (&v16k);

      return t_eq;
    }

    virtual void Launch ()
    {
      m_project -> DoFX (new GSMP::LinearMonoFXPlugWriter (Create (), string ("EQ"), string ("Graphical EQ")));
    }

    virtual bool IsEnabled ()
    {
      return true;
    }

    virtual void Init (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> AddValue (vol31, string (""));
      dialog -> AddValue (vol62, string (""));
      dialog -> AddValue (vol125, string (""));
      dialog -> AddValue (vol250, string (""));
      dialog -> AddValue (vol500, string (""));
      dialog -> AddValue (vol1k, string (""));
      dialog -> AddValue (vol2k, string (""));
      dialog -> AddValue (vol4k, string (""));
      dialog -> AddValue (vol8k, string (""));
      dialog -> AddValue (vol16k, string (""));

      dialog -> AddValue (prehear, string (""));
    }

    virtual void Reset (GSMP::EnhParameterDialog* dialog)
    {
      dialog -> Reset ("");

      vol31 -> Set (vol31 -> Default ());
      vol62 -> Set (vol62 -> Default ());
      vol125 -> Set (vol125 -> Default ());
      vol250 -> Set (vol250 -> Default ());
      vol500 -> Set (vol500 -> Default ());
      vol1k -> Set (vol1k -> Default ());
      vol2k -> Set (vol2k -> Default ());
      vol4k -> Set (vol4k -> Default ());
      vol8k -> Set (vol8k -> Default ());
      vol16k -> Set (vol16k -> Default ());
      
      Init (dialog);
      dialog -> Show ();
    }

    Action (GSMP::Project* i_project)
    {
      m_project = i_project;

      // initialaize values
      vol31 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v31, string ("31 Hz"));
      vol62 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v62, string ("62 Hz"));
      vol125 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v125, string ("125 Hz"));
      vol250 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v250, string ("250 Hz"));
      vol500 = new GSMP::SingleVolume (0.0, 2.0, 1.0, v500, string ("500 Hz"));
      vol1k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v1k, string ("1 kHz"));
      vol2k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v2k, string ("2 kHz"));
      vol4k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v4k, string ("4 kHz"));
      vol8k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v8k, string ("8 kHz"));
      vol16k = new GSMP::SingleVolume (0.0, 2.0, 1.0, v16k, string ("16 kHz"));

      //init Prehear button
      prehear = new GSMP::PrehearButton <GraphEq> (m_project, *this);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_31, &v31, *vol31);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_62, &v62, *vol62);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_125, &v125, *vol125);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_250, &v250, *vol250);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_500, &v500, *vol500);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_1k, &v1k, *vol1k);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_2k, &v2k, *vol2k);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_4k, &v4k, *vol4k);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_8k, &v8k, *vol8k);
      prehear -> SetInteractiveDouble (&GraphEq::SetLevel_16k, &v16k, *vol16k);
    }

    ~Action ()
    {
      delete prehear;
      delete vol31;
      delete vol62;
      delete vol125;
      delete vol250;
      delete vol500;
      delete vol1k;
      delete vol2k;
      delete vol4k;
      delete vol8k;
      delete vol16k;
    }

    // Dialog Elements
    GSMP::SingleVolume* vol31;
    GSMP::SingleVolume* vol62;
    GSMP::SingleVolume* vol125;
    GSMP::SingleVolume* vol250;
    GSMP::SingleVolume* vol500;
    GSMP::SingleVolume* vol1k;
    GSMP::SingleVolume* vol2k;
    GSMP::SingleVolume* vol4k;
    GSMP::SingleVolume* vol8k;
    GSMP::SingleVolume* vol16k;

    GSMP::PrehearButton <GraphEq>* prehear;

    GSMP::Project* m_project;

    // FX Values
    double v31;
    double v62;
    double v125;
    double v250;
    double v500;
    double v1k;
    double v2k;
    double v4k;
    double v8k;
    double v16k;
  };
};


PLUGIN_FACTORY(Function,GraphEqPlugin,GSMP::Menu::Activator,GraphEqPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRLowFXPlugin,GSMP::Menu::Activator,IIRLowFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRHighFXPlugin,GSMP::Menu::Activator,IIRHighFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRBandFXPlugin,GSMP::Menu::Activator,IIRBandFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRAntiBandFXPlugin,GSMP::Menu::Activator,IIRAntiBandFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRBiquadGenericFXPlugin,GSMP::Menu::Activator,IIRBiquadGenericFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRGenericAllPassFXPlugin,GSMP::Menu::Activator,IIRGenericAllPassFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,IIRGenericCombFXPlugin,GSMP::Menu::Activator,IIRGenericCombFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,SchroederFXPlugin,GSMP::Menu::Activator,SchroederFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,GardnerFXPlugin,GSMP::Menu::Activator,GardnerFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,LargeRevFXPlugin,GSMP::Menu::Activator,LargeRevFXPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));
PLUGIN_FACTORY(Function,MediumRevPlugin,GSMP::Menu::Activator,MediumRevPlugin,(GSMP::App* i_app, GSMP::Project* i_project, GSMP::View* i_view),(i_app, i_project, i_view));

/**********************************************************************************************************/


using namespace GSML;


IIRLowPass::IIRLowPass ()
{
  double t_omega = 0.5;

  d = 1.0;
  SetF (&t_omega);
}

void IIRLowPass::SetF (double* frequency)
{
  omega = *frequency;

  double a2 = (d * sin (omega) - 2) / (d * sin (omega) + 2);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 - a1 - a2) / 4;
  double b1 = 2 * b0;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2;
  //  if (fabs (a1) > 1.98)
  //  cout << "  A1 instable .... X(";
  //if (a2 < -0.98 || a2 > (0.98 - fabs (a1)))
  //  cout << "  A2 instable .... X(";

  cout << endl;

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRLowPass::SetD (double* resonance)
{
  d = 1.0 / *resonance;

  double a2 = (d * sin (omega) - 2) / (d * sin (omega) + 2);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 - a1 - a2) / 4;
  double b1 = 2 * b0;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2 << endl;  

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRLowPass::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_biquad (data [x]);
}

void IIRLowPass::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_biquad.Flush ();
}



IIRHighPass::IIRHighPass ()
{
  double t_omega = 0.5;

  d = 1.0;
  SetF (&t_omega);
}

void IIRHighPass::SetF (double* frequency)
{
  omega = *frequency;

  double a2 = (d * sin (omega) - 2) / (d * sin (omega) + 2);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 + a1 - a2) / 4;
  double b1 = - 2 * b0;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2;


  cout << endl;

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRHighPass::SetD (double* resonance)
{
  d = 1.0 / *resonance;

  double a2 = (d * sin (omega) - 2) / (d * sin (omega) + 2);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 + a1 - a2) / 4;
  double b1 = - 2 * b0;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2 << endl;  

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRHighPass::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_biquad (data [x]);
}

void IIRHighPass::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_biquad.Flush ();
}


IIRBandPass::IIRBandPass ()
{
  double t_omega = 0.5;

  q = 1.0;
  SetF (&t_omega);
}

void IIRBandPass::SetF (double* frequency)
{
  omega = *frequency;

  double a2 = (tan (omega/(2*q)) - 1) / (tan (omega/(2*q)) + 1);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 + a2) / 2;
  double b1 = 0;
  double b2 = - b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2;

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRBandPass::SetQ (double* quality)
{
  q = *quality;

  double a2 = (tan (omega/(2*q)) - 1) / (tan (omega/(2*q)) + 1);
  double a1 = (1 - a2) * cos (omega);
  double b0 = (1 + a2) / 2;
  double b1 = 0;
  double b2 = - b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2 << endl;  

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRBandPass::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_biquad (data [x]);
}

void IIRBandPass::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_biquad.Flush ();
}





IIRBandBlock::IIRBandBlock ()
{
  double t_omega = 0.5;

  q = 1.0;
  SetF (&t_omega);
}

void IIRBandBlock::SetF (double* frequency)
{
  omega = *frequency;

  double a2 = (tan (omega/(2*q)) - 1) / (tan (omega/(2*q)) + 1);
  double b0 = (1 - a2) / 2;
  double a1 = 2* b0 * cos (omega);
  double b1 = -a1;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2;

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRBandBlock::SetQ (double* quality)
{
  q = *quality;

  double a2 = (tan (omega/(2*q)) - 1) / (tan (omega/(2*q)) + 1);
  double b0 = (1 - a2) / 2;
  double a1 = 2* b0 * cos (omega);
  double b1 = -a1;
  double b2 = b0;

  cout << a1 << " " << a2 << "   " << b0 << " " << b1 << " " << b2 << endl;  

  LOCKFX;

  m_biquad.A1 (a1);
  m_biquad.A2 (a2);
  m_biquad.B0 (b0);
  m_biquad.B1 (b1);
  m_biquad.B2 (b2);
}

void IIRBandBlock::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_biquad (data [x]);
}

void IIRBandBlock::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_biquad.Flush ();
}




IIRBiquadGeneric::IIRBiquadGeneric ()
{

}

void IIRBiquadGeneric::SetA1 (double* a1)
{
  LOCKFX;
  m_biquad . A1 (*a1);
}

void IIRBiquadGeneric::SetA2 (double* a2)
{
  LOCKFX;
  m_biquad . A2 (*a2);
}
void IIRBiquadGeneric::SetB0 (double* b0)
{
  LOCKFX;
  m_biquad . B0 (*b0);
}

void IIRBiquadGeneric::SetB1 (double* b1)
{
  LOCKFX;
  m_biquad . B1 (*b1);
}

void IIRBiquadGeneric::SetB2 (double* b2)
{
  LOCKFX;
  m_biquad . B2 (*b2);
}


void IIRBiquadGeneric::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  LOCKFX;
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_biquad (data [x]);
}

void IIRBiquadGeneric::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_biquad.Flush ();
}


GenericAllPass::GenericAllPass ()
  : m_all_pass (0.0, 4412)
{

}

void GenericAllPass::SetG (double *g)
{
  m_all_pass . SetGFactor (*g);
}

void GenericAllPass::SetT (GSML::sample_index* T)
{
  m_all_pass . SetPeriod (*T);
}

void GenericAllPass::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  m_all_pass.Flush ();
}

void GenericAllPass::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  for (sample_index x = 0; x < length; x ++)
    data [x] = m_all_pass (data [x]);
}



GenericComb::GenericComb (sample_index period, double c, double damp)
  : LinearMonoFX (string ("IIR comb filter"), string ("generic IIR comb filter")),  m_comb (period, c, damp)
{
}

void GenericComb::PrepareFX (GSML::sample_index length)
{
  m_comb.Flush ();
}

void GenericComb::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  for (sample_index x = 0; x < length; x ++)
    dst_data [x] = m_comb (src_data [x]);
}


Schroeder::Schroeder (double reverb_time, double balance, double high_damp)
  : LinearMonoFX (string ("Schroeder reverb"), string ("Schroeder reverb")), p1 (0.7, 220), p2 (0.7, 75)
{
  pcm_24_8bit type;

  double t_dir = balance;
  double t_rev = (1 - balance) / 7.5;
  t_dir *= type.dBfs ();
  t_rev *= type.dBfs ();

  dir_fact (t_dir);
  rev_fact (t_rev);

  //sample_index comb_times [4] = {1323, 1543, 1764, 1985};
  //{150, 280, 723, 985};
  reverb_time *= 0.04;

  sample_index comb_times [15];
  int x;
  for (int i = 0, x = 2; i < 15; i++, x += i)
    comb_times [i] = 2000 - (8 * i);

  for (int i = 0; i < 15; i++)
    {
      double g = 0.98; //(0.95 + reverb_time) * pow (((double) comb_times [0]) / ((double) comb_times [i]), 0.1); //pow (10, ((double) comb_times [i]) / reverb_time);
      combs [i] = new IIRComb (comb_times [i], g, high_damp);
    }
}

Schroeder::~Schroeder ()
{
  for (int i = 0; i < 15; i++)
    delete combs [i];
}

void Schroeder::PrepareFX (GSML::sample_index length)
{
 for (int i = 0; i < 15; i++)
   combs [i] -> Flush ();

 p1.Flush ();
 p2.Flush ();
}

void Schroeder::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  for (sample_index x = 0; x < length; x ++)
    {
      pcm_24_8bit::value_type t_val;
      t_val (0);
      for (int i = 0; i < 15; i++)
	{
	  IIRComb& t_comb = *(combs [i]);
	  t_val = t_val + t_comb (src_data [x]);
	}

      t_val = p2 (p1 (t_val));
      dst_data [x] = (dir_fact * src_data [x]) + (rev_fact * t_val);
    }
}


Gardner::Gardner (Bridge b1, Bridge b2, Bridge b3, double high_damp, double feed_back)
  : LinearMonoFX (string ("Gardner reverb"), string ("Gardner reverb"))
{
  sample_index t_pos = 0;
  pcm_24_8bit type;

  for (int i = 0; i < 3; i++)
    {
      Bridge& t_bridge = (i == 0) ? b1 : ((i == 1) ? b2 : b3);
      patches [i] = new Patch (t_bridge, t_pos);
      t_pos += t_bridge.pre_delay + t_bridge.post_delay + t_bridge.bridge_delay;
      t_bridge.punch_out *= type.dBfs ();
      outs [i] (t_bridge.punch_out);
      out_pos [i] = t_pos;
    }

  feed_back *= type.dBfs ();
  feedback_level (feed_back);

  m_buffer = new RingBuffer (t_pos + 1);
  m_feedback = new IIRFirstOrder (high_damp, 1 - high_damp);
}

Gardner::~Gardner ()
{
  for (int i = 0; i < 3; i++)
    delete patches [i];
  
  delete m_buffer;
  delete m_feedback;
}

Gardner::Patch::Patch (Bridge& bridge, sample_index start_pos)
  : master_pass (bridge.g_bridge, bridge.bridge_delay),
    sub1 (bridge.g_sub_bridge1, bridge.sub_bridge1_delay),
    sub2 (bridge.g_sub_bridge2, bridge.sub_bridge2_delay)
{
  master_patch_pos = start_pos + bridge.pre_delay;
  patch_in1 = bridge.sub_bridge1_delay;
  patch_in2 = bridge.sub_bridge2_delay;
}


void Gardner::PrepareFX (GSML::sample_index length)
{
  m_buffer -> Flush ();
  m_feedback -> Flush ();

  for (int i; i < 3; i++)
    {
      patches [i] -> master_pass.Flush ();
      patches [i] -> sub1.Flush ();
      patches [i] -> sub1.Flush ();
    }
}

void Gardner::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  //  pcm_24_8bit type;
  //double t_double = type.dBfs ();
  //t_double /= 3;
  //pcm_24_8bit::value_type onedivthree;
  //onedivthree (t_double);

  for (sample_index x = 0; x < length; x ++)
    {
      pcm_24_8bit::value_type t_result;
      t_result (0);

      for (int i = 0; i < 3; i++)
	{
	  IIRAllPass& t_pass = (patches[i] -> master_pass);
	  IIRAllPass& t_subpass1 = (patches[i] -> sub1);
	  IIRAllPass& t_subpass2 = (patches[i] -> sub2);

	  t_pass [patches [i] -> patch_in1] = t_subpass1 (t_pass [patches [i] -> patch_in1]);
	  t_pass [patches [i] -> patch_in2] = t_subpass2 (t_pass [patches [i] -> patch_in2]);

	  (*m_buffer) [patches [i] -> master_patch_pos] = t_pass ((*m_buffer) [patches [i] -> master_patch_pos]);
	  t_result = t_result + (outs [i] * (*m_buffer) [out_pos [i]]);
	}

      dst_data [x] = t_result;

      pcm_24_8bit::value_type t_feedback;
      //t_feedback (0);
      //for (int i = 2; i < 3; i++)
      t_feedback = /*t_feedback +*/ (*m_buffer) [out_pos /*[i]*/ [2]];
      //t_feedback = t_feedback * onedivthree;
      (*m_buffer) [(GSML::sample_index) 0] = (feedback_level * (*m_feedback) (t_feedback)) + src_data [x];

      ++(*m_buffer);
    }
}


LargeRev::LargeRev (double dir_gain,
		    sample_index refl_delay, int refls, double refl_feedback, double refl_damp, double refl_gain, double mix_in_reverb,
		    sample_index rev_delay, double rev_feedback, double rev_damp, double rev_gain)
  : LinearMonoFX (string ("reverbating"), string ("Large room reverb")), reverb (rev_delay + 1, 12450)
{
  Reverberator::InConf t_inconf;
  
  // Set Reverb feeding
  t_inconf.insert_pos = 0;
  t_inconf.grab_pos = rev_delay;
  t_inconf.insert_gain = 1;
  t_inconf.fb_gain = (refls > 1) ? refl_feedback : 0;
  t_inconf.fb_damp = refl_damp;

  reverb.InsInput (t_inconf);

  // Set predelay for early reflections
  for (int i = 0; i > refls; i++)
    {
      if (i == 0)
	{
	  t_inconf.grab_pos = refl_delay;
	  t_inconf.insert_gain = mix_in_reverb;
	  t_inconf.fb_gain = refl_feedback;

	  reverb.InsInput (t_inconf);
	}
      else if (i > 1) // 1 has already been set with reverb feeding
	{
	  t_inconf.grab_pos = rev_delay - (GSML::sample_index) (pow (0.8, (double) i));
	  reverb.InsInput (t_inconf);
	}
    }

  // Setting reverb bridges
  Reverberator::BridgeConf t_conf;

  t_conf.lower = 5;
  t_conf.higher = 385;
  t_conf.i_g = 0.3;
  reverb.InsBridge (t_conf);

  t_conf.lower = 386;
  t_conf.higher = 887;
  t_conf.i_g = 0.3;
  reverb.InsBridge (t_conf);

  t_conf.lower = 1813;
  t_conf.higher = 5650;
  t_conf.i_g = 0.5;
  reverb.InsBridge (t_conf);

  t_conf.lower = 1815;
  t_conf.higher = 4547;
  t_conf.i_g = 0.25;
  reverb.InsBridge (t_conf);

  t_conf.lower = 7149;
  t_conf.higher = 12441;
  t_conf.i_g = 0.5;
  reverb.InsBridge (t_conf);

  t_conf.lower = 7150;
  t_conf.higher = 10501;
  t_conf.i_g = 0.25;
  reverb.InsBridge (t_conf);

  t_conf.lower = 10505;
  t_conf.higher = 11823;
  t_conf.i_g = 0.25;
  reverb.InsBridge (t_conf);

  Reverberator::OutConf t_outconf;

  t_outconf.pos = 1063;
  t_outconf.gain = 0.48;
  reverb.InsOutput (t_outconf);

  t_outconf.pos = 7017;
  t_outconf.gain = 0.2;
  reverb.InsOutput (t_outconf);

  t_outconf.pos = 12445;
  t_outconf.gain = 0.2;
  reverb.InsOutput (t_outconf);

  reverb.SetFeedBack (rev_feedback, rev_damp);
  reverb.SetMix (dir_gain, refl_gain, rev_gain);
}

void LargeRev::PrepareFX (GSML::sample_index length)
{
  reverb.Flush ();
}

void LargeRev::ComputeBlock (SampleBufferDefault* src, SampleBufferDefault* dst, sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  for (sample_index x = 0; x < length; x ++)
    dst_data [x] = reverb (src_data [x]);
}




MediumRev::MediumRev ()
  : reverb (22051, 13188)
{
  d_gain = 1;
  r_gain = 1;

  Reverberator::InConf t_inconf;
  
  // Set Reverb feeding
  t_inconf.insert_pos = 0;
  t_inconf.grab_pos = 4000;
  t_inconf.insert_gain = 1.0;
  t_inconf.fb_gain = 0;
  t_inconf.fb_damp = 0;

  d1 = reverb.InsInput (t_inconf);

  t_inconf.grab_pos = 4000;
  t_inconf.insert_pos = 6703;
  t_inconf.insert_gain = 0.75;
  d2 = reverb.InsInput (t_inconf);

  // Setting reverb bridges
  Reverberator::BridgeConf t_conf;

  t_conf.lower = 5;
  t_conf.higher = 366;
  t_conf.i_g = 0.7;
  reverb.InsBridge (t_conf);

  t_conf.lower = 367;
  t_conf.higher = 1336;
  t_conf.i_g = 0.5;
  reverb.InsBridge (t_conf);

  t_conf.lower = 4;
  t_conf.higher = 1544;
  t_conf.i_g = 0.3;
  reverb.InsBridge (t_conf);

  t_conf.lower = 1764;
  t_conf.higher = 3087;
  t_conf.i_g = 0.5;
  reverb.InsBridge (t_conf);

  t_conf.lower = 6705;
  t_conf.higher = 7135;
  t_conf.i_g = 0.6;
  reverb.InsBridge (t_conf);

  t_conf.lower = 6704;
  t_conf.higher = 8423;
  t_conf.i_g = 0.3;
  reverb.InsBridge (t_conf);

  Reverberator::OutConf t_outconf;

  t_outconf.pos = 1545;
  t_outconf.gain = 0.5;
  reverb.InsOutput (t_outconf);

  t_outconf.pos = 6042;
  t_outconf.gain = 0.5;
  reverb.InsOutput (t_outconf);

  t_outconf.pos = 8500;
  t_outconf.gain = 0.5;
  reverb.InsOutput (t_outconf);

  Reverberator::RegainConf t_regainconf;
  t_regainconf.pos = 6050;
  t_regainconf.gain = 0.5;
  fb1 = reverb.InsRegain (t_regainconf);

  fb2 = 0.5;
  damp = 0.5;

  reverb.SetFeedBack (fb2, damp);
  reverb.SetMix (d_gain, 0, r_gain);

  eq_buffer_length = 0;
  pre_eq_buffer = 0;
  use_eq = false;
}

GraphEq* MediumRev::GetPreEQ ()
{
  return &pre_eq;
}


void MediumRev::SetUseEq (bool* state)
{
  use_eq = *state;
}

void MediumRev::SetDirectGain (double* dir_gain)
{
  d_gain = *dir_gain;
  reverb.SetMix (d_gain, 0, r_gain);
}

void MediumRev::SetReverbGain (double* rev_gain)
{
  r_gain = *rev_gain;
  reverb.SetMix (d_gain, 0, r_gain);
}

void MediumRev::SetDelay1 (GSML::sample_index* delay1)
{
  d1 -> SetSrcOfs (*delay1);
}

void MediumRev::SetDelay2 (GSML::sample_index* delay2)
{
  d2 -> SetSrcOfs (*delay2);
}

void MediumRev::SetFeedBack1 (double* feedb1)
{
  double t_fb = (*feedb1) - 1.0;
  fb1 -> SetGain (t_fb);
}

void MediumRev::SetFeedBack2 (double* feedb2)
{
  fb2 = *feedb2;
  reverb.SetFeedBack (fb2, damp);
}

void MediumRev::SetHighDamp (double* high_damp)
{
  damp = *high_damp;
  damp = 1.0 - (damp / (2.0 * M_PI));

  reverb.SetFeedBack (fb2, damp);
}

void MediumRev::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  reverb.Flush ();
}


void MediumRev::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  if (use_eq)
    {
      if (eq_buffer_length < length)
	{ // resize Array
	  if (pre_eq_buffer)
	    delete pre_eq_buffer;
	  pre_eq_buffer = new Sdata [length];
	  eq_buffer_length = length;
	}

      memcpy (pre_eq_buffer, data, length * sizeof (pcm_24_8bit::value_type));
      pre_eq.ComputeBlock (pre_eq_buffer, length, current_offset);

      for (sample_index x = 0; x < length; x ++)
	data [x] = reverb (data [x], pre_eq_buffer [x]);
    }
  else
    for (sample_index x = 0; x < length; x ++)
      data [x] = reverb (data [x]);
}


