/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include "IIRFilters.hh"

using namespace GSML;

RingBuffer::StaticAccess::StaticAccess (RingBuffer* i_parent)
{
  m_parent = i_parent;
}

void RingBuffer::StaticAccess::SetOffset (GSML::sample_index n_pos)
{
  m_pos = (m_parent -> current_pos + n_pos) % m_parent -> T;
}

RingBuffer::RingBuffer (GSML::sample_index period)
{
  buffer = new pcm_24_8bit::value_type [period];
  T = period;
  Flush ();
}

void RingBuffer::Flush ()
{
  current_pos = 0;
  for (int i = 0; i < T; i++)
    (buffer [i]) (0);
}


pcm_24_8bit::value_type& RingBuffer::Sample (sample_index offset)
{
  return buffer [(current_pos + offset) % T];
}

void RingBuffer::operator ++ ()
{
  --current_pos;
  if (current_pos < 0)
    current_pos = T - 1;

  int t_end = m_access_list.size ();
  for (int i = 0; i < t_end; i++)
    {
      int& t_pos = m_access_list [i] -> m_pos;
      t_pos --;
      if (t_pos < 0)
	t_pos = T - 1;
    }
}


RingBuffer::StaticAccess* RingBuffer::CreateAccess (GSML::sample_index i_pos)
{
  StaticAccess* r_access = new StaticAccess (this);
  r_access -> SetOffset (i_pos);
  m_access_list.push_back (r_access);
  return r_access;
}

void RingBuffer::DeleteAccess (StaticAccess* access)
{
  for (fx_vector <StaticAccess*>::iterator i = m_access_list.begin (); i != m_access_list.end (); i++)
    if ((*i) == access)
      {
	delete (*i);
	m_access_list.erase (i);
	return;
      }
}


RingBuffer::~RingBuffer ()
{
  delete buffer;

  for (fx_vector <StaticAccess*>::iterator i = m_access_list.begin (); i != m_access_list.end (); i++)
    delete (*i);
}


LevelDetector::LevelDetector (GSML::sample_index max_length, GSML::sample_index i_length)
  : m_buffer (max_length)
{
  m_max_length = max_length - 1;
  m_length = min (m_max_length, i_length);
}

void LevelDetector::SetLength (GSML::sample_index n_length)
{
  m_length = min (m_max_length, n_length);
  SetLength_impl (m_length);
}


PeakLevelDetector::PeakLevelDetector (sample_index max_length, sample_index i_length)
  : LevelDetector (max_length, i_length)
{
  active_peak = 0;
  active_vol (0);
}

void PeakLevelDetector::Flush ()
{
  m_buffer.Flush ();
  active_peak = 0;
  active_vol (0);
}

pcm_24_8bit::value_type PeakLevelDetector::operator () (pcm_24_8bit::value_type x)
{
  if (active_vol.data < abs (x.data))
    {
      active_vol.data = abs (x.data);
      active_peak = 0;
    }

  ++m_buffer;
  m_buffer [(GSML::sample_index) 0] = x;
  active_peak++;

  if (active_peak > m_length)
    {
      active_vol (0);
      active_peak = 0;
      for (sample_index i = 0; i < m_length; i++)
	if (abs (m_buffer [i].data) >= active_vol.data)
	  {
	    active_vol.data = abs (m_buffer [i].data);
	    active_peak = i;
	  }
    }

  return active_vol;
}


AvgLevelDetector::AvgLevelDetector (GSML::sample_index max_length, GSML::sample_index i_length)
  : LevelDetector (max_length, i_length)
{
  active_sum (0);

  old_length = i_length;

  GSML::pcm_24_8bit type;
  double t_invlength = i_length;
  t_invlength = 2 / t_invlength;
  t_invlength *= type.dBfs ();
  inv_length (t_invlength);
}

void AvgLevelDetector::Flush ()
{
  active_sum (0);
  m_buffer.Flush ();
}

pcm_24_8bit::value_type AvgLevelDetector::operator () (pcm_24_8bit::value_type x)
{
  m_buffer [(GSML::sample_index) 0].data = abs (x.data);
  active_sum = active_sum + m_buffer [(GSML::sample_index) 0] - m_buffer [m_length];
  ++ m_buffer;
  return active_sum * inv_length;
}

void AvgLevelDetector::SetLength_impl (sample_index n_length)
{
  GSML::pcm_24_8bit type;
  double t_invlength = n_length;
  t_invlength = 2 / t_invlength;
  t_invlength *= type.dBfs ();
  inv_length (t_invlength);

  if (n_length > old_length)
    for (int x = old_length; x < n_length; x++)
      active_sum = active_sum + m_buffer [x];

  else if (n_length < old_length)
    for (int x = n_length; x < old_length; x++)
      active_sum = active_sum - m_buffer [x];
   
  old_length = n_length;
}

LinearShaper::LinearShaper (double diff, GSML::sample_index up_length, GSML::sample_index down_length)
{
  SetUp (diff, up_length);
  SetDown (diff, down_length);
  current (0);
}

void LinearShaper::SetUp (double diff, GSML::sample_index up_length)
{
  GSML::pcm_24_8bit type;
  double t_val = type.dBfs ();
  t_val *= diff / ((double) (up_length - 1));
  climb (t_val);
}

void LinearShaper::SetDown (double diff, GSML::sample_index down_length)
{
  GSML::pcm_24_8bit type;
  double t_val = type.dBfs ();
  t_val *= diff / ((double) (down_length - 1));
  fall (t_val);
}

void LinearShaper::ForceSet (GSML::pcm_24_8bit::value_type x)
{
  current = x;
}

GSML::pcm_24_8bit::value_type LinearShaper::operator () (GSML::pcm_24_8bit::value_type x)
{
  if (x.data < current)
    current.data = max (x.data, (current - fall).data);
  else if (x.data > current)
    current.data = min (x.data, (current + climb).data);

  return current;
}



IIRFirstOrder::IIRFirstOrder (double i_d, double dc_correct)
{
  pcm_24_8bit type;
  i_d *= type.dBfs ();
  dc_correct *= type.dBfs ();

  d (i_d);
  dc (dc_correct);

  Flush ();
}

void IIRFirstOrder::Flush ()
{
  z_1 (0);
}

pcm_24_8bit::value_type IIRFirstOrder::operator () (pcm_24_8bit::value_type x)
{
  pcm_24_8bit::value_type t_val = (x * dc) + (z_1 * d);
  z_1 = t_val;
  return t_val;
}


IIRComb::IIRComb (sample_index period, double c_factor, double high_damp)
  : m_buffer (period - 1), m_damp (high_damp, 1 - high_damp)
{
  pcm_24_8bit type;
  c_factor *= type.dBfs ();
  c (c_factor);
}

void IIRComb::Flush ()
{
  m_buffer . Flush ();
  m_damp . Flush ();
}

pcm_24_8bit::value_type IIRComb::operator () (pcm_24_8bit::value_type x)
{
  pcm_24_8bit::value_type r_val = m_damp (m_buffer [(GSML::sample_index) 0]);
  m_buffer [(GSML::sample_index) 0] = x + (c * r_val);
  ++m_buffer;
  return r_val;
}


IIRAllPass::IIRAllPass (double g_factor, GSML::sample_index max_delay)
  : m_buffer (max_delay)
{
  SetGFactor (g_factor);
  T = max_delay - 1;

}

void IIRAllPass::SetGFactor (double g_factor)
{
  pcm_24_8bit type;
  g_factor *= type.dBfs ();
  g (g_factor);
}

void IIRAllPass::SetPeriod (GSML::sample_index n_delay)
{
  T = n_delay;
}

void IIRAllPass::Flush ()
{
  m_buffer.Flush ();
}

// pcm_24_8bit::value_type IIRAllPass::operator () (pcm_24_8bit::value_type x)
// {
//   pcm_24_8bit::value_type t_val = (m_buffer [T] - (g * x));
//   m_buffer [(GSML::sample_index) 0] = (t_val * g) + x;
//   ++m_buffer;
//   return t_val;
// }




IIRabc::IIRabc (double i_a, double i_b, double i_c)
{
  t_0 = 0;
  t_1 = 2;
  t_2 = 1;
  Flush ();

  pcm_24_8bit type;
  i_a *= 2 * type.dBfs ();
  i_b *= 2 * type.dBfs ();
  i_c *= 2 * type.dBfs ();
  a (i_a);
  b (i_b);
  c (i_c);

  Flush ();
}

void IIRabc::Flush ()
{
  for (int i = 0; i < 3; i++)
    {
      xt [i] (0);
      yt [i] (0);
    }
}

pcm_24_8bit::value_type IIRabc::operator () (pcm_24_8bit::value_type x)
{
  ++t_0;
  if (t_0 == 3)
    t_0 = 0;

  ++t_1;
  if (t_1 == 3)
    t_1 = 0;

  ++t_2;
  if (t_2 == 3)
    t_2 = 0;

  xt [t_0] = x;
  yt [t_0] = ( a * ( xt [t_0] - xt [t_2] ) )  +  ( c * yt [t_1] ) - ( b * yt [t_2] );

  return yt [t_0];
}


Biquad::Biquad ()
{
  GSML::pcm_24_8bit type;

  a1 (0);
  a2 (0);
  b0 = type.dBfs ();
  b1 (0);
  b2 (0);

  m_a1 = 0.0;
  m_a2 = 0.0;
  m_b0 = 1.0;
  m_b1 = 0.0;
  m_b2 = 0.0;

  Flush ();
}

void Biquad::Flush ()
{
  z_1 (0);
  z_2 (0);
}

void Biquad::A1 (double n_a1)
{
  pcm_24_8bit type;
  m_a1 = n_a1;
  n_a1 *= type.dBfs ();
  a1 (n_a1);
}

void Biquad::A2 (double n_a2)
{
  pcm_24_8bit type;
  m_a2 = n_a2;
  n_a2 *= type.dBfs ();
  a2 (n_a2);
}

void Biquad::B0 (double n_b0)
{
  pcm_24_8bit type;
  m_b0 = n_b0;
  n_b0 *= type.dBfs ();
  b0 (n_b0);
}

void Biquad::B1 (double n_b1)
{
  pcm_24_8bit type;
  m_b1 = n_b1;
  n_b1 *= type.dBfs ();
  b1 (n_b1);
}

void Biquad::B2 (double n_b2)
{
  pcm_24_8bit type;
  m_b2 = n_b2;
  n_b2 *= type.dBfs ();
  b2 (n_b2);
}

ReverberatorElement::ReverberatorElement (RingBuffer &i_buffer, sample_index lower, sample_index higher, double i_g)
  : m_buffer (i_buffer)
{
  pcm_24_8bit type;
  i_g *= type.dBfs ();
  g (i_g);
  l = m_buffer.CreateAccess (lower);
  h = m_buffer.CreateAccess (higher);
}

void ReverberatorElement::Calc ()
{
  pcm_24_8bit::value_type x = m_buffer [l];
  m_buffer [h] = m_buffer [h] - (g * x);
  m_buffer [l] = (m_buffer [h] * g) + x;
}


FeedBackElement::FeedBackElement (RingBuffer& i_buffer, GSML::sample_index lower, GSML::sample_index higher, double fb, double high_damp)
  : m_buffer (i_buffer), m_damping (high_damp, 1 - high_damp)
{
  pcm_24_8bit type;
  fb *= type.dBfs ();
  this -> fb (fb);

  l = m_buffer.CreateAccess (lower);
  h = m_buffer.CreateAccess (higher);
}

void FeedBackElement::Flush ()
{
  m_damping.Flush ();
}

void FeedBackElement::Calc ()
{
  m_buffer [l] = m_buffer [l] + fb * m_damping (m_buffer [h]);
}

 
CrossInserter::CrossInserter (RingBuffer & src_buffer, GSML::sample_index src_ofs, RingBuffer& dst_buffer, GSML::sample_index dst_ofs, double gain)
  : src (src_buffer), dst (dst_buffer)
{
  d = dst.CreateAccess (dst_ofs);
  s = src.CreateAccess (src_ofs);
  SetGain (gain);
}

void CrossInserter::SetGain (double gain)
{
  pcm_24_8bit type;
  gain *= type.dBfs ();
  g (gain);
}

void CrossInserter::SetSrcOfs (GSML::sample_index src_ofs)
{
  s -> SetOffset (src_ofs);
}

void CrossInserter::Calc ()
{
  dst [d] = dst [d] + g * src [s];
}


Reverberator::Reverberator (GSML::sample_index pre_delay_length, GSML::sample_index reverb_queue_length)
  : delay (pre_delay_length), reverb (reverb_queue_length)
{
  cout << "Reverberator::Reverberator" << endl;

  dir (0);
  early (0);
  rev (0);
  reverb_length = reverb_queue_length - 1;
  reverb_fb = new FeedBackElement (reverb, 0, reverb_length, 0, 0);

  current_d_pos = delay.CreateAccess (0);
  current_r_pos = reverb.CreateAccess (0);
}

Reverberator::~Reverberator ()
{
  delete reverb_fb;
  for (fx_vector <FeedBackElement*>::iterator i = feedbacks.begin (); i != feedbacks.end (); i++)
    delete (*i);
  for (fx_vector <CrossInserter*>::iterator i = inserts.begin (); i != inserts.end (); i++)
    delete (*i);
  for (fx_vector <ReverberatorElement*>::iterator i = reverberators.begin (); i != reverberators.end (); i++)
    delete (*i);

}

void Reverberator::InsBridge (BridgeConf& conf)
{
  reverberators.push_back (new ReverberatorElement (reverb, conf.lower, conf.higher, conf.i_g));
}

CrossInserter* Reverberator::InsInput (InConf& conf)
{
  CrossInserter* r_insert = new CrossInserter (delay, conf.grab_pos, reverb, conf.insert_pos, conf.insert_gain); 
  inserts.push_back (r_insert);
  feedbacks.push_back (new FeedBackElement (delay, 0, conf.grab_pos, conf.fb_gain, conf.fb_damp));
  return r_insert;
}

void Reverberator::InsOutput (OutConf& conf)
{
  OutConf_ conf_;
  conf_.pos = conf.pos;
  pcm_24_8bit type;
  double t_c = conf.gain * type.dBfs ();
  conf_.c (t_c);
  outs.push_back (conf_);
}

CrossInserter* Reverberator::InsRegain (RegainConf& conf)
{
  CrossInserter* r_insert = new CrossInserter (reverb, conf.pos, reverb, conf.pos, conf.gain - 1.0);
  inserts.push_back (r_insert);
  return r_insert;
}

void Reverberator::SetFeedBack (double level, double damp)
{
  delete reverb_fb;
  reverb_fb = new FeedBackElement (reverb, 0, reverb_length, level, damp);
}

void Reverberator::SetMix (double direct_gain, double early_gain, double rev_gain)
{
  pcm_24_8bit type;
  direct_gain *= type.dBfs ();
  early_gain *= type.dBfs ();
  rev_gain *= type.dBfs ();

  dir (direct_gain);
  early (early_gain);
  rev (rev_gain);
}

void Reverberator::Flush ()
{
  cout << "Reverberator::Flush" << endl;
  delay.Flush ();
  cout << "Reverberator::Flush" << endl;
  reverb.Flush ();
  cout << "Reverberator::Flush" << endl;
  reverb_fb -> Flush ();
  cout << "Reverberator::Flush" << endl;

  for (fx_vector <FeedBackElement*>::iterator i = feedbacks.begin (); i != feedbacks.end (); i++)
    (*i) -> Flush ();  

  cout << "Reverberator::Flush" << endl;
}

pcm_24_8bit::value_type Reverberator::operator () (pcm_24_8bit::value_type direct, pcm_24_8bit::value_type rev_)
{
  pcm_24_8bit::value_type f_e;
  pcm_24_8bit::value_type f_r;

  ++delay;
  ++reverb;

  delay [current_d_pos] (0); // We will sum x after feedbacks
  reverb [current_r_pos] (0);
  reverb_fb -> Calc ();

  for (fx_vector <FeedBackElement*>::iterator i = feedbacks.begin (); i != feedbacks.end (); i++)
    (*i) -> Calc ();

  f_e = delay [current_d_pos];
  delay [current_d_pos] = f_e + rev_;

  for (fx_vector <CrossInserter*>::iterator i = inserts.begin (); i != inserts.end (); i++)
    (*i) -> Calc ();
  for (fx_vector <ReverberatorElement*>::iterator i = reverberators.begin (); i != reverberators.end (); i++)
    (*i) -> Calc ();

  f_r (0);
  for (fx_vector <OutConf_>::iterator i = outs.begin (); i != outs.end (); i++)
    {
      f_r = f_r + ((i -> c) * reverb [i -> pos]);
    }

  return (dir * direct) + (f_e * early) + (f_r * rev);
}



GraphEq::GraphEq ()
{
  for (int i = 0; i < 10; i++)
    SetLevel (i, 1.0);

  // IIR Band filter setup, thanks to Motorola inc. !
  bands [0] = new IIRabc (.000787462865, 0.498425074, 0.998415336);
  bands [1] = new IIRabc (0.00157244917, 0.496855102, 0.996616209);
  bands [2] = new IIRabc (0.00316016172, 0.493679677, 0.993522095);
  bands [3] = new IIRabc (0.00628062774, 0.487438745, 0.986812425);
  bands [4] = new IIRabc (0.0124054279,  0.475189144, 0.972915729);
  bands [5] = new IIRabc (0.0242181804,  0.451479639, 0.941937749);
  bands [6] = new IIRabc (0.0461841095,  0.407631781, 0.871031797);
  bands [7] = new IIRabc (0.0845577687,  0.330884463, 0.699566951);
  bands [8] = new IIRabc (0.1198464,     0.2601072,   0.3176087);
  bands [9] = new IIRabc (0.159603,      0.1800994,  -0.4435172);
}

GraphEq::~GraphEq ()
{
  for (int i = 0; i < 10; i++)
    delete bands [i];
}


void GraphEq::SetLevel (int band, double vol)
{
  GSML::pcm_24_8bit type;
  vol *= type.dBfs ();
  levels [band] (vol);
}


void GraphEq::ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset)
{
  for (int t = 0; t < length; t++)
    {
      pcm_24_8bit::value_type t_in = data [t];
      pcm_24_8bit::value_type t_out;
      t_out (0);

      for (int i = 0; i < 10; i++)
	t_out = t_out + (levels [i] * ( (*(bands [i])) (t_in) ) );

      data [t] = t_out;
    }
}

void GraphEq::Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos)
{
  for (int i = 0; i < 10; i++)
    bands [i] -> Flush ();
}
