/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include <deque>
#include <vector>
#include "typedefs.hh"

#define fx_vector vector

class RingBuffer
{
public:

  class StaticAccess
  { // fast RW acess to a constant offset. Slows down RingBuffer::operator ++; 
  public:
    void SetOffset (GSML::sample_index n_pos);
  private:
    friend class RingBuffer;

    StaticAccess (RingBuffer* i_parent);
    ~StaticAccess () {};

    int m_pos;
    RingBuffer* m_parent;    
  };

  RingBuffer (GSML::sample_index period);
  ~RingBuffer ();


  void Flush (); 

  GSML::pcm_24_8bit::value_type& Sample (GSML::sample_index offset);
  GSML::pcm_24_8bit::value_type& operator [] (GSML::sample_index offset) {return Sample (offset);}
  GSML::pcm_24_8bit::value_type& operator [] (StaticAccess* pos) {return buffer [pos -> m_pos];}

  void operator ++ ();

  StaticAccess* CreateAccess (GSML::sample_index i_pos);
  void DeleteAccess (StaticAccess* access);

protected:
  friend class StaticAccess;

  fx_vector <StaticAccess*> m_access_list;

  GSML::pcm_24_8bit::value_type* buffer;
  GSML::sample_index T;
  int current_pos;
};

class IIRFirstOrder
{
public:
  IIRFirstOrder (double i_d, double dc_correct);

  void Flush (); 
  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);
private:
  GSML::pcm_24_8bit::value_type d;
  GSML::pcm_24_8bit::value_type dc;
  GSML::pcm_24_8bit::value_type z_1;
};

class LevelDetector
{
public:
  LevelDetector (GSML::sample_index max_length, GSML::sample_index i_length);

  void SetLength (GSML::sample_index n_length);

  virtual void Flush () = 0;
  virtual GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x) = 0;
protected:
  virtual void SetLength_impl (GSML::sample_index n_length) {};

  RingBuffer m_buffer;
  GSML::sample_index m_length;
  GSML::sample_index m_max_length;
};

class PeakLevelDetector : public LevelDetector
{
public:
  PeakLevelDetector (GSML::sample_index max_length, GSML::sample_index i_length);

  virtual void Flush ();
  virtual GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);
private:
  GSML::sample_index active_peak;
  GSML::pcm_24_8bit::value_type active_vol;
};

class AvgLevelDetector : public LevelDetector
{
public:
  AvgLevelDetector (GSML::sample_index max_length, GSML::sample_index i_length);

  virtual void Flush ();
  virtual GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);
private:
  virtual void SetLength_impl (GSML::sample_index n_length);

  GSML::pcm_24_8bit::value_type active_sum; // = sum of ringbuffer
  GSML::pcm_24_8bit::value_type inv_length; // = 2 / length;
  GSML::sample_index old_length;
};


class LinearShaper
{ // only for short intervals not greater than a few seconds !
public: 
  LinearShaper (double diff, GSML::sample_index up_length, GSML::sample_index down_length);

  void SetUp (double diff, GSML::sample_index up_length);
  void SetDown (double diff, GSML::sample_index down_length);

  void ForceSet (GSML::pcm_24_8bit::value_type x);
  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);
private:
  GSML::pcm_24_8bit::value_type climb;
  GSML::pcm_24_8bit::value_type fall;
  GSML::pcm_24_8bit::value_type current;
};



class IIRComb
{
public:
  IIRComb (GSML::sample_index period, double c_factor, double high_damp); // min period = 2 !!

  void Flush (); 
  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);
private:
  GSML::pcm_24_8bit::value_type c;
  RingBuffer m_buffer;
  IIRFirstOrder m_damp;
};


class IIRAllPass
{
public:
  IIRAllPass (double g_factor, GSML::sample_index max_delay);

  void SetGFactor (double g_factor);
  void SetPeriod (GSML::sample_index n_delay);
  void Flush ();

  // FX Patch in
  GSML::pcm_24_8bit::value_type& operator [] (GSML::sample_index offset) {return m_buffer [offset];}

  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x)
  {
    using namespace GSML;
    pcm_24_8bit::value_type t_val = (m_buffer [T] - (g * x));
    m_buffer [(GSML::sample_index) 0] = (t_val * g) + x;
    ++m_buffer;
    return t_val;
  }

private:
  GSML::pcm_24_8bit::value_type g;
  GSML::sample_index T;

  RingBuffer m_buffer;
};


class IIRabc
{ // an alpha-beta-gama filter as used in graphical eqs
  // transformation is:
  // y(t) = 2 * ( a*( x(t) - x(t-2) ) + c*y(t-1) - b*y(t-2) )

public:
  IIRabc (double a, double b, double c);
  void Flush ();

  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x);

private:
  GSML::pcm_24_8bit::value_type xt [3];
  GSML::pcm_24_8bit::value_type yt [3];

  int t_0;
  int t_1;
  int t_2;

  GSML::pcm_24_8bit::value_type a; // if fact 2*a
  GSML::pcm_24_8bit::value_type b; // if fact 2*b
  GSML::pcm_24_8bit::value_type c; // if fact 2*c
};



class Biquad
{ // A biquad IIR filter
public:
  Biquad ();

  void A1 (double n_a1);
  void A2 (double n_a2);
  void B0 (double n_b0);
  void B1 (double n_b1);
  void B2 (double n_b2);

  double A1 () {return m_a1;}
  double A2 () {return m_a2;}
  double B0 () {return m_b0;}
  double B1 () {return m_b1;}
  double B2 () {return m_b2;}

  void Flush ();

  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x)
  {
    GSML::pcm_24_8bit::value_type t_sample = x + (a1 * z_1) + (a2 * z_2);
    x = (t_sample * b0) + (b1 * z_1) + (b2 * z_2);
    z_2 = z_1;
    z_1 = t_sample;
    return x;
  }

protected:

  GSML::pcm_24_8bit::value_type a1;
  GSML::pcm_24_8bit::value_type a2;
  GSML::pcm_24_8bit::value_type b0;
  GSML::pcm_24_8bit::value_type b1;
  GSML::pcm_24_8bit::value_type b2;

  double m_a1;
  double m_a2;
  double m_b0;
  double m_b1;
  double m_b2;

  GSML::pcm_24_8bit::value_type z_1; // last sample
  GSML::pcm_24_8bit::value_type z_2; // sample before last sample
};




class ReverberatorElement
{ // An allpass operator unit to plug on an delay queue 
public:
  ReverberatorElement (RingBuffer &i_buffer, GSML::sample_index lower, GSML::sample_index higher, double i_g);

  void Calc ();
private:
  GSML::pcm_24_8bit::value_type g;
  RingBuffer::StaticAccess* l;
  RingBuffer::StaticAccess* h;

  RingBuffer& m_buffer;
};

class FeedBackElement
{ // an Feedback operator
public:
  FeedBackElement (RingBuffer& i_buffer, GSML::sample_index lower, GSML::sample_index higher, double fb, double high_damp);

  void Flush ();
  void Calc ();

private:
  RingBuffer& m_buffer;

  GSML::pcm_24_8bit::value_type fb;
  RingBuffer::StaticAccess* l;
  RingBuffer::StaticAccess* h;

  IIRFirstOrder m_damping;
};

class CrossInserter
{ // inserts from one Ringbuffer into another;
public:
  CrossInserter (RingBuffer & src_buffer, GSML::sample_index src_ofs, RingBuffer& dst_buffer, GSML::sample_index dst_ofs, double gain);

  void SetGain (double gain);
  void SetSrcOfs (GSML::sample_index src_ofs);

  void Calc ();
private:
  RingBuffer& src;
  RingBuffer& dst;

  GSML::pcm_24_8bit::value_type g;
  RingBuffer::StaticAccess* s;
  RingBuffer::StaticAccess* d;
};

class Reverberator
{ // A generic Reverberator
public:
  class BridgeConf
  {
  public:
    GSML::sample_index lower;
    GSML::sample_index higher;
    double i_g;
  };

  class OutConf
  {
  public:
    GSML::sample_index pos;
    double gain;
  };


  class RegainConf
  {
  public:
    GSML::sample_index pos;
    double gain;
  };

  class InConf
  {
  public:
    GSML::sample_index insert_pos; // insertion in reverb queue
    GSML::sample_index grab_pos; // from early reflections delay
    double insert_gain; // gain into reverb queue
    double fb_gain; // gain in fb queue (and to early_out)
    double fb_damp; // high_damp;
  };

  Reverberator (GSML::sample_index pre_delay_length, GSML::sample_index reverb_queue_length);
  ~Reverberator ();

  void InsBridge (BridgeConf& conf);
  CrossInserter* InsInput (InConf& conf);
  void InsOutput (OutConf& conf);
  CrossInserter* InsRegain (RegainConf& conf);

  void SetFeedBack (double level, double damp);
  void SetMix (double direct_gain, double early_gain, double rev_gain);

  void Flush ();

  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type x) {return (*this) (x, x);}
  GSML::pcm_24_8bit::value_type operator () (GSML::pcm_24_8bit::value_type direct, GSML::pcm_24_8bit::value_type rev);

private:
  class OutConf_
  {
  public:
    GSML::pcm_24_8bit::value_type c;
    GSML::sample_index pos;
  };

  GSML::pcm_24_8bit::value_type dir;
  GSML::pcm_24_8bit::value_type early;
  GSML::pcm_24_8bit::value_type rev;

  RingBuffer delay;
  RingBuffer reverb;

  GSML::sample_index reverb_length;

  fx_vector <FeedBackElement*> feedbacks; // delay>delay
  fx_vector <CrossInserter*> inserts; // delay>reverb and reverb>reverb
  fx_vector <ReverberatorElement*> reverberators;
  fx_vector <OutConf_> outs;

  FeedBackElement* reverb_fb;

  RingBuffer::StaticAccess* current_r_pos;
  RingBuffer::StaticAccess* current_d_pos;
};


//// Hack !
#include "FXBase.hh"

class GraphEq : public GSMP::LinearMonoFXPlug
{
public:
  GraphEq ();
  virtual ~GraphEq ();

  void SetLevel_31  (double* vol) {SetLevel (0, *vol);}
  void SetLevel_62  (double* vol) {SetLevel (1, *vol);}
  void SetLevel_125 (double* vol) {SetLevel (2, *vol);}
  void SetLevel_250 (double* vol) {SetLevel (3, *vol);}
  void SetLevel_500 (double* vol) {SetLevel (4, *vol);}
  void SetLevel_1k  (double* vol) {SetLevel (5, *vol);}
  void SetLevel_2k  (double* vol) {SetLevel (6, *vol);}
  void SetLevel_4k  (double* vol) {SetLevel (7, *vol);}
  void SetLevel_8k  (double* vol) {SetLevel (8, *vol);}
  void SetLevel_16k (double* vol) {SetLevel (9, *vol);}

  void SetLevel (int band, double vol);

  virtual void ComputeBlock (Sdata* data, GSML::sample_index length, GSML::sample_index current_offset);
  virtual void Prepare (Sdata* queue_prefill, GSML::sample_index area_length, GSML::sample_index area_offset, GSML::sample_index real_pos);
private:

  IIRabc* bands [10];
  GSML::pcm_24_8bit::value_type levels [10];
};
