/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Importer / Exporter classes and low-level interfaces.
 */

#include "ExtFileIO.hh"

#include "App.hh"
#include "Project.hh"

#include <unistd.h> // link

using namespace GSML;
using namespace GSMP;

ImportFile :: Exception :: Exception (const string& error)
  : GSMP :: Exception (string ("Ext. File IO: ") + error) {};

ImportFile :: Unprepared :: Unprepared (const string& error)
  : Exception (string ("Ext. File IO: ") + error) {};

/****************
 * FileImporter *
 ****************/

FileImporter::FileImporter (const string& i_fname, ImportFile* i_obj)
  : FXBase ("importing", "import")
{
  m_fname = i_fname;
  m_obj = i_obj;
  
  m_prepared = false;
}

FileImporter::~FileImporter ()
{
  cout << "OK !!: ~FileImporter: we get deleted" << endl;
  delete m_obj;
}

bool FileImporter::Start_impl (Project* project, string& error_string)
{
  bool t_temp_file_used = true;
  // build temp file name
  string t_fname;
  
  // GSML method for temp files
  FindUniqueName (t_fname, m_fname + "-GSMP-temp-", "~");
  
  cout << "t_fname (for hard-link): " << t_fname << endl;

  if (link (m_fname . c_str (), t_fname . c_str () ) < 0)
    {
      m_project -> message_info.PutMessage ("unable to create hard-link (wrong permissions?)");
      t_fname = m_fname;
      t_temp_file_used = false;
    }
  
  if (!m_obj->PrepareImport (t_fname, t_temp_file_used) ) {
    error_string = "ImportFile object doesn't recognize the file";
    return false;
  }
  
  m_tracks = m_obj -> Tracks ();
  m_rate = m_obj -> Samplerate ();
  if (m_rate != 44100)
    {
      cout << "WARNING m_rate != 44100 -> NOT supported. Using 44100." << endl;
      m_rate = 44100;
    }
  
  m_length = m_obj -> Length ();
  
  cout << "child found: tracks: " << m_tracks
       << ", rate: " << m_rate
       << ", length: " << m_length << endl;
  
  m_cur_track = 0;
  m_prepared = true;
  
  cout << "FileImporter::Start_impl:" << endl;
  
  return true;
}

void FileImporter::Canceled ()
{
  cout << "FileImporter::Canceled: NOT IMPLEMENTED ....." << endl;
}

void FileImporter::Loop ()
{
  cout << "FileImporter::Loop: ..." << endl;
  if (m_prepared) {
    if (m_cur_track < m_tracks)
      {
	if (m_cur_track == 0)
	  dst_layer->CreateNEmptyTracks (m_tracks);
	
	if (!m_obj->Import (m_cur_track, (*dst_layer) [m_cur_track] ) )
	  cout << "Import: error importing track: " << m_cur_track << endl;
	
	++m_cur_track;
	progress = (double) m_cur_track / (double) m_tracks;
      }
    else
      {
	string prj_name;
	string::size_type idx = m_fname.rfind ('/');
	if (idx != string::npos)
	  prj_name = m_fname.substr (idx + 1);
	else
	  prj_name = m_fname;
	
	m_project->SetName (prj_name);
	
	m_obj->Close ();
	cout << "import finished -> waiting" << endl;
	finished = true;
	Wait ();
      }
  }
  else {
    cout << "BUG: Hu?? We didn't get prepared???!!" << endl;
    Wait ();
  }
}

/****************
 * FileExporter *
 ****************/

FileExporter::FileExporter (const string& i_fname, ExportFile* i_obj)
  : FXBase ("exporting", "export")
{
  m_fname = i_fname;
  m_obj = i_obj;
  
  m_created = false;
  
  SetReadOnly ();
}

FileExporter::~FileExporter ()
{
  cout << "OK !!: ~FileExporter: we get deleted" << endl;
  delete m_obj;
}

bool FileExporter::Start_impl (Project* project, string& error_string)
{
  bool t_created = false;
  
  // check if file is in index
  if (m_project -> GetApp () -> GetDataFileIndex () -> IsInIndex (m_fname) )
    {
      error_string = "the file is used by GSMP - export not possible";
      return false;
    }
  
  cout << "file not in index ..." << endl;
  
  struct stat t_stat;
  if (stat (m_fname . c_str (), &t_stat) < 0)
    cout << "file doesn't exist n.th. to do" << endl;
  else
    {
      cout << "file exist -> unlinking ..." << endl;
      if (unlink (m_fname . c_str () ) < 0)
	{
	  error_string = "error deleteing the file (wrong permissions?)";
	  return false;
	}
    }
  
  t_created = true;
  
  if (t_created)
    {
      m_tracks = src_layer->Tracks ();
      m_rate = src_layer->TempSamplerate ();
      if (m_rate != 44100)
	{
	  cout << "WARNING m_rate != 44100 -> NOT supported. Resetting." << endl;
	  m_rate = 44100;
	}
      
      m_length = src_layer->Length ();
      
      cout << "creating file" << endl;
      
      if (m_obj->Create (m_fname, m_tracks, m_rate, m_length) )
	{
	  m_created = true;
	  m_cur_pos = 0;
	}
      else {
	error_string = "file creation was not possible";
	return false;
      }
    }
  
  return t_created;
}

void FileExporter::Canceled ()
{
  m_obj->Cancel ();
}

void FileExporter::Loop ()
{
  //cout << "FileExporter::Loop: ..." << endl;
  
  if (m_created) {
    sample_index t_to_read = m_length - m_cur_pos;
    if (t_to_read > 0)
      {
	sample_index t_chunk_size = m_obj->NextChunkSize ();
	SampleBufferDefault t_buffer (t_chunk_size);
	
	t_to_read = limit_max (t_to_read, t_chunk_size);
	
	for (track_index tr = 0; tr < m_tracks; ++tr)
	  {
	    //cout << " track: " << tr << ", t_to_read: " << t_to_read << endl;
	    t_buffer.Clear ();
	    if (! (*src_layer) [tr]->ReadData (m_cur_pos, t_to_read, &t_buffer) )
	      t_buffer.ZeroUnused ();
	    
	    m_obj->Write (tr, &t_buffer, t_to_read != t_chunk_size);
	  }
	m_cur_pos += t_to_read;
	progress = (double)m_cur_pos / (double)m_length;
      }
    else
      {
	m_obj->Close ();
	cout << "export finished -> waiting" << endl;
	finished = true;
	Wait ();
      }
  }
  else {
    cout << "BUG: Hu?? No file got created???!!" << endl;
    Wait ();
  }
}

/***************
 * MixExporter *
 ***************/

MixExporter::MixExporter (const string& i_fname, ExportFile* i_obj)
  : FXBase ("exporting mix", "mix-export")
{
  m_fname = i_fname;
  m_obj = i_obj;
  
  m_created = false;
  
  m_mixer = 0;
  m_data_source = 0;
 
  m_got_mixer_start = false;
  
  SetReadOnly ();
}

MixExporter::~MixExporter ()
{
  cout << "OK !!: ~MixExporter: we get deleted" << endl;
  del_helper2 (m_mixer);
  del_helper2 (m_data_source);
  
  delete m_obj;
}

bool MixExporter::Start_impl (Project* project, string& error_string)
{
  m_tracks = 2; // TODO: make configable
  m_rate = src_layer->TempSamplerate ();
  if (m_rate != 44100)
    {
      cout << "WARNING m_rate != 44100 -> NOT supported. Resetting." << endl;
      m_rate = 44100;
    }
  
  m_length = src_layer->Length ();
  
  cout << "creating file" << endl;
  
  if (m_obj->Create (m_fname, m_tracks, m_rate, m_length) )
    {
      m_created = true;
      m_cur_pos = 0;
    }
  else {
    error_string = "file creation was not possible";
    return false;
  }
  
  // needs a fixed chunk size for the mixer ...
  m_chunk_size = m_obj->NextChunkSize ();
  cout << "m_chunk_size: " << m_chunk_size << endl;
  
  m_pl_params = *(m_project->m_pl_mixer_params);
  m_ca_params = *(m_project->m_ca_mixer_params); // although we are pl-only, we need vaild ca params
  
  m_data_source = new DataSource (src_layer, 0);
  
  m_mixer = new Mixer ();
  m_mixer->SetPlayback (true);
  m_mixer->SetCapture (false);
  
  m_mixer->SetNode (this);
  
  cout << "MixExporter: starting Mixer" << endl;
  m_mixer->StartIO (m_data_source, &m_pl_params, &m_ca_params);
  
  cout << "returning ..." << endl;
  return true;
}

void MixExporter::Canceled ()
{
  // 1st stop mixer
  m_mixer->IOFinishedNotification ();
  
  m_obj->Cancel ();
}

void MixExporter::Loop ()
{
  cout << "MixExporter::Loop: ..." << endl;
  if (m_created) {
    if (m_got_mixer_start)
      {
	sample_index t_to_read = m_length - m_cur_pos;
	if (t_to_read > 0)
	  {
	    sample_index t_chunk_size = m_obj->NextChunkSize ();
	    if (t_chunk_size != m_chunk_size)
	      {
		cout << "child-file-obj chunk_size changed during export - unsupported!" << endl
		     << "  (mixer can't change this dynamically ...)! -> resetted" << endl;
		t_chunk_size = m_chunk_size;
	      }
	    
	    SampleBufferDefault t_buffer (t_chunk_size);
	    
	    t_to_read = limit_max (t_to_read, t_chunk_size);
	    
	    for (track_index tr = 0; tr < m_tracks; ++tr)
	      {
		cout << " track: " << tr << ", t_to_read: " << t_to_read << endl;
		t_buffer.Clear ();
		if (! m_mixer->Read (tr, &t_buffer) )
		  t_buffer.ZeroUnused ();
		
		m_obj->Write (tr, &t_buffer, t_to_read != t_chunk_size);
	      }
	    
	    m_cur_pos += t_to_read;
	    progress = (double) m_cur_pos / (double) m_length;
	  }
	else
	  {
	    m_mixer->IOFinishedNotification ();
	    m_obj->Close ();
	    cout << "export finished -> waiting" << endl;
	    finished = true;
	    Wait ();
	  }
      } // end if got_mixer_start
    else
      {
	cout <<" Waiting for Mixer to start" << endl;
	usleep (50000);
      }
  }
  else {
    if (!m_created)
      cout << "BUG: Hu?? No file got created???!!" << endl;
    cout << "  -> waiting ..." << endl;
    Wait ();
  }
}

track_index MixExporter::PlaybackChannels ()
{
  return m_tracks;
}

track_index MixExporter::CaptureChannels ()
{
  return 0;
}

int MixExporter::PlaybackPeriodSize ()
{
  return m_chunk_size;
}

int MixExporter::CapturePeriodSize ()
{
  return 0;
}

int MixExporter::PlaybackPeriods ()
{
  return 1;
}

int MixExporter::CapturePeriods ()
{
  return 0;
}

int MixExporter::PlaybackXRuns ()
{
  return 0;
}

int MixExporter::CaptureXRuns ()
{
  return 0;
}

void MixExporter::SetPlayback (bool state)
{
  // hm? - not used (we are playback only everytime)
}

void MixExporter::SetCapture (bool state)
{
  // hm? - not used (we are playback only everytime)
}

bool MixExporter::IsPlayback ()
{
  return true;
}

bool MixExporter::IsCapture ()
{
  return false;
}

void MixExporter::StartIO (Mixer& source)
{
  cout << "MixExporter::StartIO" << endl;
  if ( (void*) &source != (void*) this)
    cout << "MixExporter::StartIO: source != this ...!?!?" << endl;
  m_got_mixer_start = true;
}

void MixExporter::StopIO ()
{
  // m_mixer->IOFinishedNotification ();
  
  cout << "Got StopIO ?!?! - How can that be??" << endl;
}

bool MixExporter::IsIORunning ()
{
  return m_got_mixer_start;
}

sample_index MixExporter::GetCurrentPos ()
{
  return m_cur_pos;
}

void MixExporter::Debug ()
{
  cout << "MixExporter::Debug called ..." << endl;
}
