/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Extended IO-Streaming, according Importer and OggVorbis / MP3
 *   IO-Stream classes.
 */

#include "ExtFileIO_impl.hh"

using namespace GSML;
using namespace GSMP;

/******************
 * RandomAccessIO *
 ******************/

RandomAccessIO::Exception::Exception (const string& error)
  : GSMP::Exception (error) {}


RandomAccessIO::TrackInvaild::TrackInvaild (const string& error)
  : Exception (error) {}


/*************
 * FileCache *
 *************/

FileCache::FileCache (StreamIO* i_src)
{
  m_src = i_src;
  
  m_length = m_src -> Length ();
  m_tracks = m_src -> Tracks ();
  m_rate = m_src -> Rate ();
  
  m_cache . SetBoundElements (m_tracks);
}

FileCache::~FileCache ()
{
  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "~FileCache:" << endl;
  
  if (m_src -> IsOpen () ) {
    DEBUG__(GSMP_FILE_STREAMS, detail)
      cout << "~FileCache: closing source" << endl;
    m_src -> Close ();
  }
  delete m_src;
}

bool FileCache::Open (const string& fname)
{
  if (! m_src -> Open (fname) )
    return false;
  
  m_length = m_src -> Length ();
  m_tracks = m_src -> Tracks ();
  m_rate = m_src -> Rate ();
  
  m_cache . SetBoundElements (m_tracks);
  
  return true;
}

bool FileCache::IsOpen ()
{
  return m_src -> IsOpen ();
}

bool FileCache::Close ()
{
  return m_src -> Close ();
}
    
RandomAccessIO* FileCache::Clone ()
{
  StreamIO* t_src = m_src -> Clone ();
  return new FileCache (t_src);
}
    
sample_index FileCache::Length ()
{
  return m_length;
}

track_index FileCache::Tracks ()
{
  return m_tracks;
}

sample_rate FileCache::Rate ()
{
  return m_rate;
}

sample_index FileCache::Read (sample_index start, sample_index length,
			      track_index track, pcm_24_8bit::value_type* dest)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "FileCache::Read: " << start << " + " << length << ", tr: " << track << endl;
  
  if (track > m_tracks)
    throw TrackInvaild ("the specified track is invaild");
  
  m_cache . SetCachedRange (start, length);
  
  // fill cache
  if (m_cache . vaild_start > 0) // at the front
    {
      DEBUG__(GSMP_FILE_STREAMS, low_level)
	cout << "FileCache::Read reading front: " << m_cache . vaild_start << endl;
      
      m_src -> Seek (start);
      m_src -> Read (m_cache . vaild_start, m_cache . data);
    }
  
  if (m_cache . vaild_start + m_cache . vaild_length < length) // at the end
    {
      sample_index t_vaild_end = m_cache . vaild_start + m_cache . vaild_length;
      sample_index t_length = length - t_vaild_end;
      
      DEBUG__(GSMP_FILE_STREAMS, low_level)
	cout << "FileCache::Read reading back: "
	     << t_vaild_end << " + " << t_length << endl;
      
      m_src -> Seek (start + t_vaild_end);
      m_src -> Read (t_length, m_cache . data + t_vaild_end * m_tracks);
    }
  
  // update, because we have read all, now.
  m_cache . vaild_start = 0;
  m_cache . vaild_length = length;
  
  // deinterleace/copy the data
  
  //cout << "FileCache::Read deinterleave / copy data" << endl;
  //cout << "  m_tracks: " << m_tracks << endl;
  for (int i = 0; i < length; ++ i) {
    dest [i] =  m_cache . data [ i * m_tracks + track ];
  }
  
  return length;
}

/****************
 * FileStreamer *
 ****************/

FileStreamer::FileStreamer (RandomAccessIO* i_src)
{
  m_max_streams = 10;
  m_max_age = 1200;
  
  m_length = i_src -> Length ();
  m_tracks = i_src -> Tracks ();
  m_rate = i_src -> Rate ();
  
  AddStream (i_src);
}

FileStreamer::~FileStreamer ()
{
  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "~FileStreamer:" << endl;
  
  while (m_streams . size () > 0) {
    
    DEBUG__(GSMP_FILE_STREAMS, detail)
      cout <<  "~FileStreamer: closing / deleting stream ..." << endl;
    
    m_streams . back () . src -> Close ();
    delete m_streams . back () . src;
    m_streams . pop_back ();
  }
}

bool FileStreamer::Open (const string& fname)
{
  // TODO: implement re-open
  return false;
}

bool FileStreamer::IsOpen ()
{
  return m_streams [0] . src -> IsOpen ();
}

bool FileStreamer::Close ()
{
  // TODO: implement close ...
  return false;
}

RandomAccessIO* FileStreamer::Clone ()
{
  RandomAccessIO* t_src = m_streams . back () . src -> Clone ();
  return new FileStreamer (t_src);
}
    
sample_index FileStreamer::Length ()
{
  return m_length;
}

track_index FileStreamer::Tracks ()
{
  return m_tracks;
}

sample_rate FileStreamer::Rate ()
{
  return m_rate;
}

sample_index FileStreamer::Read (sample_index start, sample_index length,
				 track_index track, pcm_24_8bit::value_type* dest)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "FileStreamer::Read: " << start << " + " << length << ", tr: " << track << endl;
  
  RandomAccessIO* t_src = 0;
  {
    SigC::Threads::MLock lock (m_global_lock);
    
    DEBUG__(GSMP_FILE_STREAMS, low_level)
      cout << "  globally locked" << endl;
    
    t_src = BestStream (start, length);
    AgeStreams (t_src); // except t_src - will be overwritten in update-stream beloew - anyway
    LockStream (t_src);
  }
  
  DEBUG__(GSMP_FILE_STREAMS, low_level)
      cout << "  reading from source" << endl;
  
  sample_index t_length = t_src -> Read (start, length, track, dest);
  
  {
    SigC::Threads::MLock lock (m_global_lock);
    DEBUG__(GSMP_FILE_STREAMS, low_level)
      cout << "  globally locked" << endl;
    
    UpdateStream (t_src, start, length);
    UnlockStream (t_src);
  }
  
  return t_length;
}

void FileStreamer::AddStream (RandomAccessIO* src)
{
  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "FileStreamer::AddStream:" << endl;
  
  stream_stuff s;
  s . src = src;
  s . age = 0;
  s . pos = 0;
  s . cached_length = 0;
  s . locked = false;
  
  m_streams . push_back (s);
}

RandomAccessIO* FileStreamer::BestStream (sample_index start, sample_index length)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "FileStreamer::BestStream:" << endl;
  
  RandomAccessIO* t_src = 0;
  
  bool t_found = false;
  
  // search the best stream ...
  
  for (int i = 0; !t_found && i < m_streams . size (); ++ i)
    {
      if (!m_streams [i] . locked) // do not use locked streams
	{
	  // cached (??) or continues positions
	  sample_index startA = m_streams [i] . pos - m_streams [i] . cached_length;
	  sample_index startB = m_streams [i] . pos;
	  
	  if (start == startA || start == startB)
	    {
	      DEBUG__(GSMP_FILE_STREAMS, low_level)
		cout << "found stream" << endl;
	      
	      t_src = m_streams [i] . src;
	      t_found = true;
	    }
	  else
	    {
	      DEBUG__(GSMP_FILE_STREAMS, low_level)
		cout << "not found: age: " << m_streams [i] . age
		     << ", startA: " << startA
		     << ", startB: " << startB << endl;
	    }
	} // end if not locked
    } // end for all streams
  
  if (!t_found)
    {
      DEBUG__(GSMP_FILE_STREAMS, detail)
	cout << "not found ... " << endl;

      if (m_streams . size () <= m_max_streams)
	{
	  DEBUG__(GSMP_FILE_STREAMS, detail)
	    cout << "  creating new stream" << endl;
	  
	  t_src = m_streams [0] . src -> Clone ();
	  AddStream (t_src);
	}
      else
	{
	  DEBUG__(GSMP_FILE_STREAMS, detail)
	    cout << "  max streams reached - searching oldest one ..." << endl;
	  
	  bool t_first_unlocked = true;
	  int t_age = 0;
	  
	  for (int i = 0; i < m_streams . size (); ++ i)
	    {
	      if (!m_streams [i] . locked) // do not use locked streams
		{
		  int age = m_streams [i] . age;
		  DEBUG__(GSMP_FILE_STREAMS, detail)
		    cout << "  age: " << age << endl;
		  
		  if (t_first_unlocked)
		    {
		      t_src = m_streams [i] . src;
		      t_age = age;
		      t_first_unlocked = false;
		    }
		  else if (age > t_age) {
		    t_src = m_streams [i] . src;
		    t_age = age;
		  }
		  t_found = true;
		} // end not locked
	    } // end for streams
	} // end else oldest stream
    } // end if not found
  
  if (!t_found)
    {
      DEBUG__(GSMP_FILE_STREAMS, detail) {
	cout << "  no stream found (all must be locked ?!?)" << endl;
	cout << "  creating new stream" << endl;
      }
      t_src = m_streams [0] . src -> Clone ();
      AddStream (t_src);
    }
  
  return t_src;
}

void FileStreamer::AgeStreams (RandomAccessIO* except_stream = 0)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "FileStreamer::AgeStreams:" << endl;
  
  vector <stream_stuff> :: iterator i = m_streams . begin ();
  vector <stream_stuff> :: iterator i_end = m_streams . end ();
  while (i != i_end)
    {
      if (i-> src == except_stream)
	++ i;
      else
	{
	  ++ i -> age;
	  
	  if (i -> age <= m_max_age)
	    ++ i;
	  else if (!i -> locked) {
	    DEBUG__(GSMP_FILE_STREAMS, detail)
	      cout << "  AgeStreams: stream reached max age and is not locked -> deleting ..." << endl;
	    
	    i -> src -> Close ();
	    delete i -> src;
	    
	    i = m_streams . erase (i);
	    i_end = m_streams . end ();
	  }
	} // end not except stream
    } // end while loop
}

void FileStreamer::UpdateStream (RandomAccessIO* stream, sample_index start,
				 sample_index length)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
  cout << "FileStreamer::UpdateStream:" << endl;
  
  for (int i = 0; i < m_streams . size (); ++ i)
    {
      if (m_streams [i] . src == stream)
	{
	  m_streams [i] . age = 0;
	  m_streams [i] . pos = start + length;
	  m_streams [i] . cached_length = length;
	  return;
	}
    }
  cout << "BUG: FileStreamer::UpdateStream should never get here ?!?" << endl;
}

void FileStreamer::LockStream (RandomAccessIO* stream)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "FileStreamer::LockStream: " << stream << endl;
  
  for (int i = 0; i < m_streams . size (); ++ i)
    {
      if (m_streams [i] . src == stream)
	{
	  if (m_streams [i] . locked)
	    cout << "BUG: FileStreamer::LockStream stream already locked!" << endl;
	  m_streams [i] . locked = true;
	  return;
	}
    }
  cout << "BUG: FileStreamer::LockStream should never get here ?!?" << endl;
}

void FileStreamer::UnlockStream (RandomAccessIO* stream)
{
DEBUG__(GSMP_FILE_STREAMS, low_level)
  cout << "FileStreamer::UnlockStream: " << stream << endl;
  
  for (int i = 0; i < m_streams . size (); ++ i)
    {
      if (m_streams [i] . src == stream)
	{
	  m_streams [i] . locked = false;
	  return;
	}
    }
  cout << "BUG: FileStreamer::UnlockStream should never get here ?!?" << endl;
}

/*******************
 * DataFileAdapter *
 *******************/

DataFileAdapter::DataFileAdapter (RandomAccessIO* i_src, const string i_fname,  bool i_autodelete)
  : DataFile (i_fname, false, i_autodelete) // writeable, autodelete
{
  m_src = i_src;
  
  m_length = i_src -> Length ();
  m_tracks = i_src -> Tracks ();
  m_rate = i_src -> Rate ();
}

DataFileAdapter::~DataFileAdapter ()
{
  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "~DataFileAdapter:" << endl;
  
  if (m_src -> IsOpen () ) {
    DEBUG__(GSMP_FILE_STREAMS, detail)
      cout << "~DataFileAdapter: closing source" << endl;
    
    m_src -> Close ();
  }
  delete m_src;
}

void DataFileAdapter::ReadData (sample_index start,
				sample_index count,
				track_index track, 
				SampleBuffer* destination)
{
  //SigC::Threads::MLock lock (m_lock);
  
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "DataFileAdapter::ReadData: " << start << " + " << count << ", tr: " <<track << endl;
  
  if (track > m_tracks)
    {
      DataFile::DataFile::ExceptLogicalError e;
      e.filename = m_filename;
      e.info = "Critical: not enough tracks (was: ";
      e.info += track;
      e.info += ", have: ";
      e.info += m_tracks;
      e.info += ")";
      throw e;
    }
  
  // introduces some mem-coping ... (let's hope gcc fit's it into modern CPU's cache)
  pcm_24_8bit::value_type data [count];
  
  //cout << " reading" << endl;
  m_src -> Read (start, count, track, data);
  destination -> WriteData (count, data);
}

void DataFileAdapter::WriteData (sample_index count,
				 track_index track, 
				 SampleBuffer* source,
				 sample_index buffer_index)
{
  ExceptNotWriteable e;
  e.info = "Files are not writeable via this class";
  throw e;
}

void DataFileAdapter::FixateWrite ()
{
  ExceptNotWriteable e;
  e.info = "Files are not writeable via this class";
  throw e;
}

void DataFileAdapter::SetFilename (string& new_filename)
{
  ExceptNotWriteable e;
  e.info = "Files are not writeable via this class";
  throw e;
}

/**********************
 * StreamIOImportFile *
 **********************/

StreamIOImportFile::StreamIOImportFile ()
{
  m_prepared = false;
  
  m_data_file = 0;
  m_registerd_file = false;
}

StreamIOImportFile::~StreamIOImportFile ()
{
  if (m_data_file != 0 && !m_registerd_file)
    {
      cout << "WARNING: file not used / registerd -> deleting ... - RIGHT ???" << endl;
      delete m_data_file;
    }
}

bool StreamIOImportFile::PrepareImport (const string& fname, bool temp_file_used)
{
  m_fname = fname;
  
  StreamIO* t_stream = InternalPrepare ();
  
  if (! t_stream -> Open (m_fname) )
    {
      cout << "StreamIOImportFile: StreamIO can't open the file ..." << endl;
      delete t_stream;
      return false;
    }
  
  m_tracks = t_stream -> Tracks ();
  m_rate = t_stream -> Rate ();
  m_length = t_stream -> Length ();
  
  cout << "extracted: tracks: " << m_tracks
       << ", rate: " << m_rate
       << ", length: " << m_length << endl;
  
  // plug the top-level helpers together
  FileCache* t_cache = new FileCache (t_stream);
  FileStreamer* t_streamer = new FileStreamer (t_cache);
  m_data_file = new DataFileAdapter (t_streamer, m_fname, temp_file_used);
  
  m_prepared = true;
  
  cout << "OggVorbisImportFile::PrepareImport: finished" << endl;
  return true;
}

void StreamIOImportFile::Close ()
{
  // TODO: clean some other stuff
  m_prepared = false;
}
    
GSML::track_index StreamIOImportFile::Tracks ()
{
  return m_tracks;
}

GSML::sample_rate StreamIOImportFile::Samplerate ()
{
  return m_rate;
}

GSML::sample_index StreamIOImportFile::Length ()
{
  return m_length;
}
    
bool StreamIOImportFile::Import (GSML::track_index track, GSML::TrackLayer* layer)
{
  if (!m_registerd_file) {
    layer -> GetDataFileIndex () -> RegisterFile (m_data_file);
    m_registerd_file = true;
  }
  
  cout << "creating and inserting raw-entry for track: " << track << endl;
  TrackLayer::Entry t_entry (m_data_file);
  t_entry.length = m_length;
  t_entry.start_in_file = 0;
  t_entry.track_in_file = track;
  
  layer -> InsertRawEntry (0, t_entry);
}

/*************************
 * Stuff 4 M$ Wave Files *
 *************************/

/* Definitions for Micro$oft WAVE format
 * Thanks to the ALSA people ! (taken from aplay ;-)
 */

#define COMPOSE_ID(a,b,c,d)	((a) | ((b)<<8) | ((c)<<16) | ((d)<<24))
#define LE_SHORT(v)		(v)
#define LE_INT(v)		(v)
#define BE_SHORT(v)		bswap_16(v)
#define BE_INT(v)		bswap_32(v)

#define WAV_RIFF		COMPOSE_ID('R','I','F','F')
#define WAV_WAVE		COMPOSE_ID('W','A','V','E')
#define WAV_FMT			COMPOSE_ID('f','m','t',' ')
#define WAV_DATA		COMPOSE_ID('d','a','t','a')
#define WAV_PCM_CODE		1

/* it's in chunks like .voc and AMIGA iff, but my source say there
   are in only in this combination, so I combined them in one header;
   it works on all WAVE-file I have
*/
typedef struct {
  u_int magic;		/* 'RIFF' */
  u_int length;		/* filelen */
  u_int type;		/* 'WAVE' */
} WaveHeader;

typedef struct {
  u_short format;	/* should be 1 for PCM-code */
  u_short modus;	/* 1 Mono, 2 Stereo */
  u_int sample_fq;	/* frequence of sample */
  u_int byte_p_sec;
  u_short byte_p_spl;	/* samplesize; 1 or 2 bytes */
  u_short bit_p_spl;	/* 8, 12 or 16 bit */
} WaveFmtBody;

typedef struct {
  u_int type;		/* 'data' */
  u_int length;		/* samplecount */
} WaveChunkHeader;

/*****************
 * WavExportFile *
 *****************/

WavExportFile::WavExportFile ()
{
  m_created_file = false;
  m_fd = 0;
  interleaved_buffer = 0;
}

WavExportFile::~WavExportFile ()
{
  if (m_fd != 0)
    close (m_fd);
  del_helper2 (interleaved_buffer);
}

bool WavExportFile::Create (const string& fname, track_index tracks, sample_index rate,
			    sample_index length)
{
  m_filename = fname;
  m_tracks = tracks;
  m_rate = rate;
  m_length = length;
  
  m_cur_length = 0;
  
  if ( (m_fd = open ( m_filename.c_str (), O_WRONLY | O_CREAT | O_TRUNC, FILE_CREATE_MODE) ) < 0)
    return false;
  
  m_created_file = true;
  
  // write wav header
  WaveHeader h;
  WaveFmtBody f;
  WaveChunkHeader cf, cd;

  int bits;
  u_int tmp;
  u_short tmp2;
  
  bits = 16;
  
  h.magic = WAV_RIFF;
  tmp = length + sizeof (WaveHeader) + sizeof (WaveChunkHeader) + sizeof (WaveFmtBody) + sizeof (WaveChunkHeader) - 8;
  h.length = LE_INT(tmp);
  h.type = WAV_WAVE;
  
  cf.type = WAV_FMT;
  cf.length = LE_INT(16);

  f.format = LE_INT(WAV_PCM_CODE);
  f.modus = LE_SHORT(tracks);;
  f.sample_fq = LE_INT(m_rate);

  tmp2 = tracks * ((bits + 7) / 8);
  f.byte_p_spl = LE_SHORT(tmp2);
  tmp2 = tmp2 * m_rate;
  f.byte_p_sec = LE_SHORT(tmp2);
  
  f.bit_p_spl = LE_SHORT(bits);
  
  cd.type = WAV_DATA;
  cd.length = LE_INT(length * 2 * m_tracks);

  if (write (m_fd, &h, sizeof (WaveHeader) ) != sizeof (WaveHeader) ||
      write (m_fd, &cf, sizeof (WaveChunkHeader) ) != sizeof (WaveChunkHeader) ||
      write (m_fd, &f, sizeof (WaveFmtBody) ) != sizeof (WaveFmtBody) ||
      write (m_fd, &cd, sizeof (WaveChunkHeader) ) != sizeof (WaveChunkHeader) )
    {
      return false;
    }

  m_cur_length = 0;
  m_chunk_size = 1024 * 4;
  
  interleaved_buffer = new pcm_16bit::value_type [m_chunk_size * m_tracks];
  
  return true;
}

void WavExportFile::Cancel ()
{
  if (m_fd != 0)
    close (m_fd);

  if (m_created_file)
    unlink (m_filename.c_str () );
}

void WavExportFile::Close ()
{
  if (m_fd != 0)
    close (m_fd);
  
  m_fd = 0;
  m_created_file = false;
}

sample_index WavExportFile::NextChunkSize ()
{
  return m_chunk_size;
}

bool WavExportFile::Write (track_index track, SampleBufferDefault* src, bool last_write)
{
  sample_index t_length = limit_max (src->UsedLength (), m_chunk_size);
				     
  for (int i = 0; i < t_length; ++i)
    interleaved_buffer [i * m_tracks + track] ( pcm_24_8bit::Convert2int16 ( src->data [i] ) );
  
  bool t_ret;
  // write on last track write
  if (track == m_tracks - 1) {
    t_ret = write (m_fd, interleaved_buffer, t_length * m_tracks * 2) >= 0;
    m_cur_length += m_chunk_size;
  }
  
  return t_ret;
}

/*****************
 * WavImportFile *
 *****************/

WavImportFile::WavImportFile ()
{
  m_fd = 0;
  m_prepared = false;
  
  m_data_file = 0;
  m_registerd_file = false;
}

WavImportFile::~WavImportFile ()
{
  if (m_fd != 0)
    close (m_fd);
  
  if (m_data_file != 0 && !m_registerd_file)
    {
      cout << "WARNING: file not used / registerd -> deleting ... - RIGHT ???" << endl;
      delete m_data_file;
    }
}

string WavImportFile::Extension () const
{
  return "wav";
}

bool WavImportFile::PrepareImport (const string& fname, bool temp_file_used)
{
  m_filename = fname;
  
  if ( (m_fd = open ( m_filename.c_str (), O_RDONLY ) ) < 0) {
    cout << "erorr: open the file: " << m_filename << endl;
    return false;
  }
  
  WaveHeader h;
  WaveFmtBody f;
  WaveChunkHeader cf, cd;
  
  if (read (m_fd, &h, sizeof(WaveHeader)) != sizeof(WaveHeader) ||
      read (m_fd, &cf, sizeof(WaveChunkHeader)) != sizeof(WaveChunkHeader) ||
      read (m_fd, &f, sizeof(WaveFmtBody)) != sizeof(WaveFmtBody) ||
      read (m_fd, &cd, sizeof(WaveChunkHeader)) != sizeof(WaveChunkHeader) )
    {
      cout << "error: read error" << endl;
      close (m_fd);
      return false;
    }
  
  int bits;
  u_int tmp;
  u_short tmp2;
  
  bits = 16;
  
  if (h.magic != WAV_RIFF || h.type != WAV_WAVE || cf.type != WAV_FMT)
    {
      cout << "error: header seems not to be a wav file" << endl;
      close (m_fd);
      return false;
    }
  if ( LE_SHORT(f.format) != WAV_PCM_CODE )
    {
      cout << "error: not pcm coded" << endl;
      close (m_fd);
      return false;
    }
  
  m_tracks = LE_SHORT(f.modus);
  m_rate = LE_SHORT(f.sample_fq);
  
  cout << "raw length: " << LE_INT(cd.length) << endl;
  m_length = LE_INT(cd.length) / 2 / m_tracks;
  
  m_head_length = sizeof(WaveHeader) + sizeof(WaveChunkHeader)
    + sizeof(WaveFmtBody) + sizeof(WaveChunkHeader);
  
  cout << "extracted: tracks: " << m_tracks
       << ", rate: " << m_rate
       << ", length: " << m_length << endl
       << "  head_length (leading_trash): " << m_head_length << endl;
  
  pcm_16bit* temp_type = new pcm_16bit; // mem leak - TODO: rework types ...
  m_data_file = new DataFileRAW (m_filename, temp_type, m_tracks, m_rate,
				 m_head_length, false, temp_file_used); // writeable, autodelete
  
  m_prepared = true;
  
  cout << "WavImportFile::PrepareImport: finished" << endl;
  return true;
}

void WavImportFile::Close ()
{
  if (m_fd != 0)
    close (m_fd);
  
  m_fd = 0;
  m_prepared = false;
}

track_index WavImportFile::Tracks ()
{
  if (!m_prepared)
    {
      throw Unprepared ("unprepared");
    }
  return m_tracks;
}

sample_rate WavImportFile::Samplerate ()
{
  if (!m_prepared)
    {
      throw Unprepared ("unprepared");
    }
  return m_rate;
}

sample_index WavImportFile::Length ()
{
  return m_length;
}

bool WavImportFile::Import (track_index track, TrackLayer* layer)
{
  if (!m_registerd_file) {
    layer->GetDataFileIndex () -> RegisterFile (m_data_file);
    m_registerd_file = true;
  }
  
  cout << "creating and inserting raw-entry for track: " << track << endl;
  TrackLayer::Entry t_entry (m_data_file);
  t_entry.length = m_length;
  t_entry.start_in_file = 0;
  t_entry.track_in_file = track;
  
  layer -> InsertRawEntry (0, t_entry);
}

/*******************
 * OggVorbisStream *
 *******************/

#if EXT_FIO_OGGVORBIS == 1

OggVorbisStream::OggVorbisStream ()
{
  m_file = 0;
  m_stream_pos = 0;
}

OggVorbisStream::~OggVorbisStream ()
{
  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "~OggVorbisStream:" << endl;
  
  if (m_file != 0) {
    DEBUG__(GSMP_FILE_STREAMS, detail)
      cout << "~OggVorbisStream: closing OggVorbis file" << endl;
    
    ov_clear (&m_vorbis_file);
    //fclose (m_file);
    m_file = 0;
  }
}

bool OggVorbisStream::Open (const string& fname)
{
  if (m_file != 0) {
    DEBUG__(GSMP_FILE_STREAMS, normal)
    cout << "File already opened" << endl;
    return false;
  }
  
  m_fname = fname;
  
  cout << "OggVorbisStream::Open: " << m_fname << endl;
  
  if ( (m_file = fopen (m_fname . c_str (), "r") ) < 0) {
    DEBUG__(GSMP_FILE_STREAMS, normal)
      cout << "  can't open file" << endl;
    
    m_file = 0;
    return false;
  }
  
  cout << "m_file: " << m_file << endl;
  
  if (ov_open (m_file, &m_vorbis_file, NULL, -1) < 0) {
    DEBUG__(GSMP_FILE_STREAMS, normal)
      cout << "Input does not appear to be an Ogg bitstream" << endl;
    
    fclose (m_file);
    return false;
  }
  
  cout << "m_file: " << m_file << endl;
  
  if (! ov_seekable (&m_vorbis_file) ) {
    DEBUG__(GSMP_FILE_STREAMS, normal)
      cout << "Input is not seekable" << endl;
    
    ov_clear (&m_vorbis_file);
    fclose (m_file);
    //m_file = 0;
    return false;
  }
  
  // TODO: have we to free this ??
  vorbis_info* vi = ov_info (&m_vorbis_file, -1);
  
  m_tracks = vi -> channels;
  m_rate = vi -> rate;
  
  ogg_int64_t pcm_total = ov_pcm_total (&m_vorbis_file, -1);
  m_length = pcm_total;

  DEBUG__(GSMP_FILE_STREAMS, detail)
    cout << "recognized: " << m_tracks << " tracks, " << m_length << " samples." << endl;
  
  // if we are here, the file is open
  
  return true;
}

bool OggVorbisStream::IsOpen ()
{
  return m_file != 0;
}

bool OggVorbisStream::Close ()
{
  if (m_file == 0)
    return false;
  
  cout << "m_file: " << m_file << endl;
  ov_clear (&m_vorbis_file);
  cout << "m_file: " << m_file << endl;
  //fclose (m_file);
  m_file = 0;
  
  return true;
}

StreamIO* OggVorbisStream::Clone ()
{
  StreamIO* t_stream = new OggVorbisStream ();
  if (IsOpen () )
    if (! t_stream -> Open (m_fname) )
      cout << "BUG: OggVorbisStream: cant open cloned stream although I'm open??" << endl;
  
  return t_stream;
}

sample_index OggVorbisStream::Length ()
{
  if (IsOpen () )
    return m_length;
  else
    return 0;
}

track_index OggVorbisStream::Tracks ()
{
  if (IsOpen () )
    return m_tracks;
  else
    return 0;
}

sample_rate OggVorbisStream::Rate ()
{
  if (IsOpen () )
    return m_rate;
  else
    return 0;
}

sample_index OggVorbisStream::Read (sample_index length, pcm_24_8bit::value_type* dest)
{
  DEBUG__(GSMP_FILE_STREAMS, low_level)
    cout << "OggVorbisStream::Read: " << length << endl;
  
  sample_index t_processed = 0;
  
  if (m_stream_pos < m_length)
    {
      sample_index t_length = limit_max (length, m_length - m_stream_pos);
      
      int ret = 0;
      int read_bytes = 0;
      int needed_bytes = t_length * m_tracks * 2;
      
      pcm_16bit::value_type t_data [t_length * m_tracks];
      
      // Ogg/Vorbis seems to do the processing in a main-loop (whatever) and returns
      // only the bytes it has in this decodeing loop ...
      do
	{
	  //cout << "OggVorbisStream::Read: ov_read" << endl;
	  ret = ov_read (&m_vorbis_file, (char*) (t_data ) + read_bytes,
			 needed_bytes - read_bytes,
			 0, 2, 1, & m_vorbis_section);
	  
	  if (ret < 0)
	    cout << "OggVorbisStream::Read: Error: " << ret << "(ignored ...)" << endl;
	  
	  if (ret > 0)
	    read_bytes += ret;
	}
      while (ret > 0 && read_bytes < needed_bytes);
      
      // copy and scale pcm data
      //cout << "OggVorbisStream::Read: scale and copy data" << endl;
      for (sample_index i = 0; i < t_length * m_tracks; ++ i)
	dest [i] ( t_data [i] << 8);
      
      t_processed += t_length;
      m_stream_pos += t_length;
    }
  
  if (t_processed < length) {
    DEBUG__(GSMP_FILE_STREAMS, detail)
      cout << "Vorbis read: not all read -> zeroing ..." << endl;
    
    for (sample_index i = t_processed * m_tracks; i < length * m_tracks; ++ i)
	dest [i] ( 0 );
    
    t_processed = length;
  }
  
  return t_processed;
}

sample_index OggVorbisStream::Seek (sample_index pos)
{
  if (pos != m_stream_pos)
    {
      int ret = 0;
      
      ret = ov_pcm_seek (&m_vorbis_file, pos);
      if (ret < 0) {
	cout << "OggVorbis::Seek: error during search (pos: "
	     << pos << "): " << ret << "(ignored ...)" << endl;
      }
      
      m_stream_pos = pos;
    }
  
  return m_stream_pos;
}

#endif // EXT_FIO_OGGVORBIS == 1

/*************
 * MP3Stream *
 *************/

#if EXT_FIO_MP3 == 1

MP3Stream::MP3Stream ()
{
  m_open = false;
  m_stream_pos = 0;
}

MP3Stream::~MP3Stream ()
{
  if (m_open)
    m_mad . Close ();
}
    
bool MP3Stream::Open (const string& fname)
{
  if (m_open)
    return false;
  
  m_fname = fname;
  m_open = m_mad . Open (m_fname);
  
  if (m_open) {
    m_tracks = m_mad . Tracks ();
    m_rate = m_mad . Rate ();
    m_length = m_mad . Length ();
  }
  
  return m_open;
}

bool MP3Stream::IsOpen ()
{
  return m_open;
}

bool MP3Stream::Close ()
{
  if (!m_open)
    return false;
  
  bool t_ret = m_mad . Close ();
  m_open = false;
  return t_ret;
}
    
StreamIO* MP3Stream::Clone ()
{
  MP3Stream* t_stream = new MP3Stream ();
  if (IsOpen () )
    {
      t_stream -> m_fname = m_fname;
      t_stream -> m_open = t_stream -> m_mad . Open (m_fname, &m_mad);
      
      if (t_stream -> m_open) {
	t_stream -> m_tracks = t_stream -> m_mad . Tracks ();
	t_stream -> m_rate = t_stream -> m_mad . Rate ();
	t_stream -> m_length = t_stream -> m_mad . Length ();
      }
      else {
	cout << "BUG: MP3Stream: cant open cloned stream although I'm open??" << endl;
      }
    }
  
  return t_stream;
}

sample_index MP3Stream::Length ()
{
  if (IsOpen () )
    return m_length; 
  else
    return 0;
}

track_index MP3Stream::Tracks ()
{
  if (IsOpen () )
    return m_tracks;
  else
    return 0;
}

sample_rate MP3Stream::Rate ()
{
  if (IsOpen () )
    return m_rate;
  else
    return 0;
}
    
sample_index MP3Stream::Read (sample_index length, pcm_24_8bit::value_type* dest)
{
  //cout << "MP3Stream::Read: " << length << endl;
  
  mad_fixed_t t_buffer [length * m_tracks];
  
  m_mad . Read (length, t_buffer);
  for (int i = 0; i < length * m_tracks; ++ i)
    dest [i] ( t_buffer [i] >> 4 );
  
  m_stream_pos += length;
  
  return length;
}

sample_index MP3Stream::Seek (sample_index pos)
{
  if (pos != m_stream_pos)
    {
      m_mad . Seek (pos);
      m_stream_pos = pos;
    }
  
  return m_stream_pos;
}

# endif // EXT_FIO_MP3 == 1
