/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Base class for FXs.
 */

#include <string>
#include <vector>
#include "DataLayer.hh"
#include "Data.hh"

#include "Project.hh"

using namespace GSMP;
using namespace GSML;

FXBase::FXBase (const string& info, const string& comment)
{
  Ready (false);
  finished = false;
  progress = old_progress = 0;
  m_info = info;
  m_comment = comment;
  read_only = false;
}

bool FXBase::Start (Project* project, string& error_string)
{
  m_project = project;
  src_layer = m_project -> CurrentLayerPtr ();
  if (!read_only) 
    {
      dst_layer = m_project -> PrepareNewLayer ();
      dst_layer -> Comment (m_comment);
    }

  if (!Start_impl (project, error_string))
    {
      if (!read_only)
	m_project -> CancelNewLayer ();

      return false;
    }

  info = m_project -> progress_info () . PushInfo (m_info);

  DoCancel.Connect (this, &FXBase::DoCancel__);

  Continue ();

  return true;
}

bool FXBase::PollProgressData ()
{
  if (progress != old_progress)
    info -> SetProgress (progress);

  if (finished)
    {
      m_project -> progress_info () . PopInfo ();

      if (!read_only)
	m_project -> FixateNewLayer ();

      return false;
    }
  return true;
}

bool FXBase::Start_impl (Project* project, string& error_string)
{
  return true;
}

void FXBase::Cancel ()
{
  if (!finished) // TODO: make this saver :-(
    DoCancel (); // -> prevent main Loop exiting
  
  m_project -> progress_info () . PopInfo ();
  if (!read_only)
    m_project -> CancelNewLayer ();
}


int FXBase::DoCancel__ ()
{
  Canceled ();
  finished = true;
  Wait ();
}


void FXBase::SetReadOnly ()
{
  read_only = true;
}



LinearMonoFXPlug::LinearMonoFXPlug ()
{
  setup.generator = false;
  setup.do_declick = true;     // not impl yet !
  setup.declick_samples = 10; // not impl yet !
  setup.queue_prefill = 0;
}



LinearMonoFX::LinearMonoFX (const string& info, const string& comment)
  : FXBase (info, comment)
{
  m_block_size = 4096;
  m_prefetch = 0;
  m_postfetch = 0;
  m_declick = 20;
  generator = false;
  selector = false;
  declick = true;

  processed_length = 0;
  sum_length = 0;
  current_offset = 0;
  m_sample_buffer = 0;
}

LinearMonoFX::~LinearMonoFX ()
{
  if (m_sample_buffer)
    delete m_sample_buffer;
}


void LinearMonoFX::SetGenerator (bool state)
{
  generator = state;
}

void LinearMonoFX::SetPostfetch (GSML::sample_index postfetch_samples, bool silence_at_end)
{
  m_postfetch = postfetch_samples;
  if (silence_at_end)
    cout << "Warning ! LinearMonoFX::SetPostfetch silence_at_end not implemented !" << endl;
}



sample_index LinearMonoFX::Offset ()
{
  return current_offset;
}

void LinearMonoFX::Loop ()
{
  const Area& t_area = *current_area;

  if (current_offset < t_area.length)
    {
      progress = ((double) processed_length) / ((double) sum_length); 

      if (current_offset == 0)
	{
	  PrepareFX (t_area.length);
	  t_area.dst_layer -> PrepareOverWrite (t_area.start);
	}

      sample_index t_block = min (m_block_size, t_area.length - current_offset);

      // fetch & execute & writeback
      if (!generator)
	{
	  m_sample_buffer -> Clear ();
	  if (! t_area.src_layer -> ReadData (t_area.start + current_offset, t_block + m_postfetch, m_sample_buffer))
	    m_sample_buffer -> ZeroUnused ();
	}

      SampleBufferDefault* dst = new SampleBufferDefault (t_block);
      dst -> SetUsedLength (t_block);
      ComputeBlock (m_sample_buffer, dst, t_block);
      t_area.dst_layer -> WriteData (t_block, manage (dst), 0);

      current_offset += t_block;
      processed_length += t_block;
    }
  else
    {
      t_area.dst_layer -> FixateWrite ();
      if (++current_area == area_list.end ())
	{
	  finished = true;
	  Wait ();
	}
      else
	current_offset = 0;
    }
}


bool LinearMonoFX::Start_impl  (Project* project, string& error_string)
{
  DEBUG__ (GSMP, low_level)
    cout << "LinearMonoFX: generating area list" << endl;

  DataLayer::iterator i_dst = dst_layer -> Begin ();
  for (DataLayer::iterator i = src_layer -> Begin (); i != src_layer -> End (); i++, i_dst++)
    if ((*i) -> selection . IsActive ())
      for (TrackLayer::Selection::const_iterator j = (*i) -> selection . begin (); j != (*i) -> selection . end (); j++)
	{
	  Area t_area;
	  t_area.src_layer = *i;
	  t_area.dst_layer = *i_dst;
	  t_area.start = j -> start;
	  t_area.length = (j -> end) - t_area.start;

	  sum_length += t_area.length;
	  area_list.push_back (t_area);
	}

  if (area_list.size () == 0)
    {
//       DEBUG__ (GSMP, detail)
// 	cout << "No active Selections found, processing whole tracks" << endl;

//       i_dst = dst_layer -> Begin ();
//       for (DataLayer::iterator i = src_layer -> Begin (); i != src_layer -> End (); i++, i_dst++)
// 	if ((*i) -> selection . IsActive () && (*i) -> Length () > 0)
// 	  {
// 	    Area t_area;
// 	    t_area.src_layer = *i;
// 	    t_area.dst_layer = *i_dst;
// 	    t_area.start = 0;
// 	    t_area.length = (*i) -> Length ();

// 	    sum_length += t_area.length;
// 	    area_list.push_back (t_area);
// 	  }
//     }

//   if (area_list.size () == 0 && sum_length == 0)
//   {
      DEBUG__ (GSMP, detail)
	cout << "Sorry, nothing to do !?" << endl;

      error_string = "no data to work with.";
      return false;
    }

  current_area = area_list.begin ();
  m_sample_buffer = new SampleBufferDefault (m_block_size + m_postfetch);
  return true;
}




LinearMonoFXPlugWriter::LinearMonoFXPlugWriter (LinearMonoFXPlug* plug, const string& info, const string& comment)
  : LinearMonoFX (info, comment)
{
  m_plug = plug;

  if (m_plug -> setup.generator)
    SetGenerator (true);
}

LinearMonoFXPlugWriter:: ~LinearMonoFXPlugWriter ()
{
  delete m_plug;
}

void LinearMonoFXPlugWriter::PrepareFX (GSML::sample_index length)
{ // Not very much finished, yet !!!
  m_plug -> Prepare (0, length, 0, 0);
}

void LinearMonoFXPlugWriter::ComputeBlock (GSML::SampleBufferDefault* src, GSML::SampleBufferDefault* dst, GSML::sample_index length)
{
  pcm_24_8bit::value_type* src_data = src -> data;
  pcm_24_8bit::value_type* dst_data = dst -> data;

  if (! m_plug -> setup.generator)
    for (int i = 0; i < length; i++)
      dst_data [i] = src_data [i];

  m_plug -> ComputeBlock (dst_data, length, Offset ());
}
