/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Class for real-time FX prehearing.
 */

#include "FXMixerSource.hh"

using namespace GSML;
using namespace GSMP;
using namespace SigC::Threads;

GSML::track_index PlugHandle::Track ()
{
  return m_track;
}


FXMixerSource::FXMixerSource (GSML::DataLayer* i_layer)
{
  SetDataLayer (i_layer);
  m_capture_tracks = 0;
}

FXMixerSource::~FXMixerSource ()
{
  UnplugAll ();
}


void FXMixerSource::SetDataLayer (GSML::DataLayer* layer)
{
  m_layer = layer;
  m_playback_tracks = m_layer -> Tracks ();
  BuildTrackData ();
}


PlugHandle* FXMixerSource::AttachPlug (LinearMonoFXPlug* plug, GSML::track_index to_track)
{
  PlugHandle* r_plug = new PlugHandle;
  r_plug -> key = current_key++;
  r_plug -> m_track = to_track;
  r_plug -> fx = plug;

  MLock lazy (m_data_);

  m_data [to_track] . m_plugs [r_plug -> key] = plug;
  RebuildPlugVector (to_track);

  return r_plug;
}

void FXMixerSource::DetachPlug (PlugHandle* plug)
{
  MLock lazy (m_data_);

  map <int, LinearMonoFXPlug*>& t_map = m_data [plug -> m_track] . m_plugs;

  map <int, LinearMonoFXPlug*>::iterator i = t_map . find (plug -> key);
  if (i == t_map . end ())
    {
      cout << "=========Exception========" << endl;

      MixerSource::Except e;
      throw e;
    }
  t_map . erase (i);
  RebuildPlugVector (plug -> m_track);
  delete plug;
}

sample_index FXMixerSource::MinStart () const
{
  return 0;
};

sample_index FXMixerSource::MaxEnd () const
{
  return m_layer -> Length ();
};

bool FXMixerSource::IsInputable () const
{
  return false;
};
    
void FXMixerSource::StartO_impl ()
{
  BuildTables (0);
  MLock lazy (m_data_);
  BuildSelection ();
  // Set readpos 0 !!
};

void FXMixerSource::StartIO_impl ()
{ 
  MixerSource::Except e;
  throw e;
};

void FXMixerSource::StopIO_impl ()
{ 
  MLock lazy (m_data_);
  UnplugAll ();
  ClearTables ();
};
    
void FXMixerSource::SetReadPos_impl (track_index track, sample_index n_pos, bool is_master)
{
  {
    MLock lazy (m_data_);

    if (n_pos < CurrentReadPosition (track, is_master))
      m_data [track].current_segment = m_data [track].selection -> begin ();
 
    m_data [track].did_fx = false;
  }

  CurrentReadPosition (track, is_master) = n_pos;
};

void FXMixerSource::SetWritePos_impl (track_index track, sample_index n_pos, bool is_master)
{
  MixerSource::Except e;
  throw e;
};
    
bool FXMixerSource::Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
			    bool is_master)
{
  sample_index current_pos = CurrentReadPosition (track, is_master);
  sample_index old_length = length;

  if (!is_master)
    {
      sample_index fx_start = dest -> UsedLength ();
      if (! ((*m_layer) [track] -> ReadData (current_pos, length, dest)) )
	{
	  dest -> Zero (length - (dest -> UsedLength () - fx_start));
	}

      pcm_24_8bit::value_type* fx_data = dest -> data;

      MLock lazy (m_data_);
      track_data& t_track_data = m_data [track];

      while (length > 0)
	{ 
	  // skip previos selection segments
	  GSML::TrackLayer::Selection::const_iterator t_seg = t_track_data.current_segment;
	  while (t_seg != t_track_data.end && t_seg -> end <= current_pos)
	    t_seg++;

	  sample_index chunk_length;
	  if (t_seg == t_track_data.end)
	    chunk_length = length;
	  else if (t_seg -> start > current_pos)
	    { // before selection
	      chunk_length = min (length, t_seg -> start - current_pos);
	      t_track_data.did_fx = false;
	    }
	  else
	    { // Over selection

	      if (!t_track_data.did_fx)
		{ // Prepare fx
		  GSML::sample_index area_length = (t_seg -> end) - (t_seg -> start);
		  GSML::sample_index area_offset = current_pos - (t_seg -> start);
		  GSML::sample_index real_pos = current_pos;

		  for (vector <LinearMonoFXPlug*>::iterator c_fx =  t_track_data.plug_vector.begin ();
		       c_fx != t_track_data.plug_vector.end (); c_fx++)
		    (*c_fx) -> Prepare ((GSMP::LinearMonoFXPlug::Sdata*) 0, area_length, area_offset, real_pos);

		  t_track_data.did_fx = true;
		}

	      chunk_length = min (length, t_seg -> end - current_pos);
	      for (vector <LinearMonoFXPlug*>::iterator c_fx =  t_track_data.plug_vector.begin ();
		   c_fx != t_track_data.plug_vector.end (); c_fx++)
		(*c_fx) -> ComputeBlock (fx_data + fx_start, chunk_length, current_pos);
	    }

	  length -= chunk_length;
	  fx_start += chunk_length;
	  t_track_data.current_segment = t_seg;
	  current_pos += chunk_length;
	}
    }
  else
    current_pos += old_length;

  CurrentReadPosition (track, is_master) += old_length;
  return (current_pos < MaxEnd () );
};

bool FXMixerSource::Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
			     sample_index buffer_index, bool is_master)
{ 
  MixerSource::Except e;
  throw e;
};

void FXMixerSource::Debug_impl () const
{
  cout << "Debug !" << endl;
};


void FXMixerSource::RebuildPlugVector (GSML::track_index track)
{
  vector <LinearMonoFXPlug*>& t_vector = m_data [track].plug_vector;
  map <int, LinearMonoFXPlug*>& t_map = m_data [track].m_plugs;

  t_vector.clear ();
  for (map <int, LinearMonoFXPlug*>::iterator i = t_map.begin (); i != t_map.end (); i++)
    t_vector.push_back (i -> second);
}

void FXMixerSource::BuildTrackData ()
{
  MLock lazy (m_data_);

  UnplugAll ();

  for (int i = 0; i < m_data.size (); i++)
    delete m_data [i].selection;


  m_data.clear ();

  for (int i = 0; i < m_playback_tracks; i++)
    {
      m_data.push_back ();
      m_data.back () . selection = 0;
    }

  BuildSelection ();
}

void FXMixerSource::BuildSelection ()
{
  for (int i = 0; i < m_data.size (); i++)
    BuildSelection (i);
}

void FXMixerSource::BuildSelection (GSML::track_index track)
{
  if (m_data [track].selection)
    delete m_data [track].selection;

  m_data [track].selection = new TrackLayer::Selection ( (*m_layer)[track]->selection );

  //if (m_data [track].selection -> begin () == m_data [track].selection -> end ())
  //  { // Mark whole track
  //  }

  m_data [track].current_segment = m_data [track].selection -> begin ();
  m_data [track].end = m_data [track].selection -> end ();
}

void FXMixerSource::UnplugAll ()
{
  SignalAllUnpluged ();

  for (int i = 0; i < m_data.size (); i++)
    {
      m_data [i].plug_vector.clear ();
      m_data [i].m_plugs.clear ();
    }
}
