/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   
 */

#include <stdlib.h>
#include <functional>
#include "Factory.hh"
#ifndef GSMP_TESTING
#  include "App.hh"
#else
#  include "../tests/src/menutest.hh"
#endif
//#include "Activator.hh"

//using namespace GSMP;
using namespace GSMP::Menu;

Exception::Exception (const string& error)
  : GSMP::Exception (string ("Menu system :") + error) {}

MalformedLocation::MalformedLocation (const string& the_location)
  : Exception (string ("Malformed location string: '") + the_location + string ("'")) {}

DuplicateLocation::DuplicateLocation (const string& the_location)
  : Exception (string ("There is allready an Item at  '") + the_location + string ("'")) {}

MalformedProperty::MalformedProperty (const string& the_property)
  : Exception (string ("Malformed proprerty string: '") + the_property + string ("'")) {}

InvalidPropertyValue::InvalidPropertyValue (const string& the_property, const string& the_value)
  : Exception (string ("Property '") + the_property + string ("' has invalid value '") + the_value + string ("'")) {}

PropertyRequiered::PropertyRequiered (const string& expected_property, const string& for_type)
  : Exception (string ("Type '") + for_type + string ("' requires property '") + expected_property + string ("'")) {}

InvalidScope::InvalidScope (const string& the_scope)
  : Exception (string ("Invalid scope: '") + the_scope + string ("'")) {}

UnknownType::UnknownType (const string& the_type)
  : Exception (string ("Unknown item type: '") + the_type + string ("'")) {}

SetMismatch::SetMismatch (const string& the_set)
  : Exception (string ("Missmatching scopes in set'") + the_set + string ("'")) {}

IteratorException::IteratorException ()
  : Exception (string ("Invalid Iterator")) {}

UninitializedMenuData::UninitializedMenuData ()
  : Exception (string ("Uninitialized View/Project/Menu !")) {}

/* internal Exception classes */
class ParseError 
{
public:
  ParseError () 
  {
    DEBUG__(GSMP_MENU,low_level)
      cout << "ParseError exception thrown" << endl;
  };
};


/**************************************************************************************************/

SplitString :: SplitString (const string& s, char c, bool remove_escapes)
{
  string :: size_type position = 0;
  string :: size_type length = s.length ();
  string current_substring ("");
  bool ignore_next = false;

  while (position < length)
    { // I do not use the stl find/substring/replace string 
      // functions anymore, because they messed this code
      // up. If you can do better, feel free

      char x = s [position];

      if (!ignore_next && x == c)
	{
	  substrings.push_back (string(current_substring));
	  current_substring = "";
	}
      else 
	if (!ignore_next && x == '$')
	  {
	    ignore_next = true;
	    if (!remove_escapes)
	      current_substring += x;
	  }
	else
	  {
	    ignore_next = false;
	    current_substring += x;
	  }

      position ++;
    }

  substrings.push_back (current_substring);
}


Entity :: Entity (const string& init_entity) : my_string (init_entity)
{
  SplitString sub (init_entity, '&', true);
  if (sub.size () < 1 || sub.size () > 3)
    throw ParseError ();

  deque <string> :: const_iterator i = sub.begin ();

  name = *(i++);
  if (name == "")
    throw ParseError ();

  if (i != sub.end ())
    {
      group = *(i++);
      if (group == "")
	throw ParseError ();
    }
  else
    group = "zero";

  if (i != sub.end ())
    { // get priority
      const char* c_str = (*i) . c_str ();
      char* error_detect;
      sorting_priority = strtol (c_str, &error_detect, 10);
      if (*error_detect != '\0' || abs (sorting_priority) > 9999)
	throw ParseError ();
    }
  else
    sorting_priority = 0;

  DEBUG__(GSMP_MENU,detail)
    {
      cout << "New Entity Created:" << endl
	   << "name     : " << name << endl
	   << "group    : " << group << endl
	   << "priority : " << sorting_priority << endl;
    }
}


Entity :: Entity (const string& seperator_group, int ignored)
  : name ("<seperator>"), group (seperator_group)
{
  sorting_priority = 10000;
  DEBUG__(GSMP_MENU,detail)
    cout << "created seperator Entity for group " << seperator_group << endl;
}



bool Entity :: IsLess (const Entity& than) const
{
  if (group < than.group)
    return true;
  if (group > than.group)
    return false;

  if (sorting_priority < than.sorting_priority)
    return true;
  if (sorting_priority > than.sorting_priority)
    return false;

  return (name < than.name);
}

Location::Location (const string& path_string) : my_string (path_string)
{ 
  DEBUG__(GSMP_MENU,detail)
    cout << "Parsing Location string...." << endl;

  SplitString sub (path_string, '>', false);
  if (sub.size () < 2)
    throw MalformedLocation (path_string);

  try
    {
      for (deque <string> :: const_iterator i = sub.begin (); i != sub.end (); i++)
	entities.push_back (Entity (*i));
    }

  catch (ParseError e)
    {
      throw MalformedLocation (path_string);
    }

  DEBUG__(GSMP_MENU,detail)
    cout << "....Done." << endl;
}

deque <Entity> :: const_iterator Location :: Root ()
{
  return entities.begin ();
}

deque <Entity> :: const_iterator Location ::Leave ()
{
  return -- entities.end ();
}

/**************************************************************************************************/

const MenuData* GetMenuData (LeaveItem::Scope scope, const GSMP::View* view)
{
  switch (scope)
    {
    case LeaveItem::APP:
      return (view -> GetProject ()) -> GetApp ();
    case LeaveItem::PROJECT:
      return view -> GetProject ();
    case LeaveItem::VIEW:
      return view;
    }
}

MenuData* GetMenuData (LeaveItem::Scope scope, GSMP::View* view)
{
  switch (scope)
    {
    case LeaveItem::APP:
      return (view -> GetProject ()) -> GetApp ();
    case LeaveItem::PROJECT:
      return view -> GetProject ();
    case LeaveItem::VIEW:
      return view;
    }
}


/**************************************************************************************************/


Item :: Item (SubMenuItem* init_parent, const Entity& init_params)
  : my_params (init_params)
{
  my_parent = init_parent;
  DEBUG__(GSMP_MENU,low_level)
    cout << "new Item '" << my_params.name << "'" << endl 
	 << "parent is '" << ((init_parent == NULL) ? string ("<NULL>") : init_parent -> my_params.name) 
	 << "'" << endl;
}

string Item :: Location () const
{
  string result;
  if (my_parent)
    {
      if ((result = my_parent -> Location ()) != "")
	result = result + string (">") + my_params.name;
      else
	result = my_params.name;
    }
  else
    result = ""; // skip <root> name
  return result;
}

void Item :: Debug () const
{
  cout << "Item at '" << Location () << "'" << endl
       << "Name     : " << my_params.name << endl
       << "Group    : " << my_params.group << endl 
       << "Priority : " << my_params.sorting_priority << endl
       << "Which is a "; // continued by DebugTypeSpecific

  DebugTypeSpecific ();
}





LeaveItem :: LeaveItem (SubMenuItem* init_parent, const Entity& init_params, 
			const Scope& init_scope, const string& init_properties)
  : Item (init_parent, init_params)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Instanciating as LeaveItem" << endl; 
  my_scope = init_scope;
  ParseProperties (init_properties);
}

void LeaveItem :: DebugLeaveItem () const
{
  cout << "Scope    : " << ((my_scope == APP) ? string ("App") :
    (my_scope == PROJECT) ? string ("Project") : string ("View")) << endl
       << "ToolTip  : " << tooltip << endl
       << "Shortcut : " << shortcut << endl;
}


LeaveItem::Property :: Property (const string& init_property)
{
  SplitString sub (init_property, '=', true);
  if (sub.size () != 2)
    throw MalformedProperty (init_property);
  id = *(sub.begin ());
  value = *( -- (sub.end ())); 
}

LeaveItem::Property :: Property (const Property& copy)
  : id (copy.id), value (copy.value)
{ 
}





void LeaveItem :: ParseProperties (const string& property_part)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Parsing Properties" << endl;

  SplitString sub (property_part, '&', false);
  for (deque<string>::const_iterator i = sub.begin(); i != sub.end (); i++)
    {
      Property current (*i);
      my_properties [current.id] = current;
    }

  // check for tooltip and shortcut properties
  map <string, Property> :: const_iterator i = my_properties.find ("Tooltip");
  if (i != my_properties.end ())
    tooltip = i -> second.value;
  i = my_properties.find ("Shortcut");
  if (i != my_properties.end ())
    shortcut = i -> second.value;
}



SimpleItem :: SimpleItem (SubMenuItem* init_parent, const Entity& init_params, 
			  const Scope& init_scope, const string& init_properties)
  : LeaveItem (init_parent, init_params, init_scope, init_properties)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Specialisation to SimpleItem" << endl;
 
  map <string, Property> :: const_iterator i = my_properties.find ("State");
  if (i != my_properties.end ())
    {
      string s = i -> second.value;
      if (s == string ("on"))
	default_state = true;
      else if (s == string ("off"))
	default_state = false;
      else
	throw InvalidPropertyValue ("State", s);
    }
  else // default val
    default_state = true;

  blocked = false;
}

void SimpleItem :: DebugTypeSpecific () const
{
  cout << "SimpleItem" << endl;
  DebugLeaveItem ();
  cout << "Pixmap TODO" << endl
       << "Blocked = " << blocked << endl;
}

bool SimpleItem :: IsEnabled (const View* here) const
{
  string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Query MenuDataBase for " << s << endl;

  const MenuData* t_data = GetMenuData (my_scope, here);

  if (t_data -> initialized)
    {
      map <string, bool> :: const_iterator i =  t_data -> states.find (s);
      if (i == t_data -> states.end ())
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "Not in DB, inserting default_state" << endl;
	  t_data -> states [s] = default_state;
	  return (default_state && (!blocked));
	}
      else
	return (i -> second && (!blocked));
    }

  throw UninitializedMenuData ();
}

void SimpleItem :: BlockAll () const
{
  blocked = true;
  SignalEnabledChanged . emit (this);
}

void SimpleItem :: UnBlockAll () const
{
  blocked = false;
  SignalEnabledChanged . emit (this);
}

void SimpleItem :: SetState (View* there, bool state) const
{
 string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Accessing MenuDataBase for " << s << endl;

  MenuData* t_data = GetMenuData (my_scope, there);

  if (t_data -> initialized)
    {
      t_data -> states [s] = state;
      SignalEnabledChanged . emit (this);
      return;
    }

  throw UninitializedMenuData ();
}

void SimpleItem :: Activate (View* view) const
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "SimpleItem " << my_params.name << " activated !" << endl;

  if (!IsEnabled (view))
    {
      DEBUG__(GSMP_MENU, normal)
	cout << "Cannot activate an Disabled Item !" << endl;
      return;
    }

  App* t_app = view -> GetProject () -> GetApp ();
  Project* t_project = view -> GetProject ();

  Activator* t_activate;

  switch (my_scope)
    {
    case APP:
      t_activate = GSMX_utils::Plugin::Build <Activator, App*, Project*, View*> 
	(reactor_category, reactor_name, t_app, (Project*) 0, (View*) 0);
      break;
    case PROJECT:
      t_activate = GSMX_utils::Plugin::Build <Activator, App*, Project*, View*> 
	(reactor_category, reactor_name, t_app, t_project, (View*) 0);
      break;
    case VIEW:
      t_activate = GSMX_utils::Plugin::Build <Activator, App*, Project*, View*> 
	(reactor_category, reactor_name, t_app, t_project, view);
      break;
    }

  t_activate -> Execute ();
  delete t_activate;
}




CheckedItem :: CheckedItem (SubMenuItem* init_parent, const Entity& init_params, 
			  const Scope& init_scope, const string& init_properties)
  : LeaveItem (init_parent, init_params, init_scope, init_properties)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Specialisation to CheckedItem" << endl;
 
  map <string, Property> :: const_iterator i = my_properties.find ("State");
  if (i != my_properties.end ())
    {
      string s = i -> second.value;
      if (s == string ("on"))
	default_state = true;
      else if (s == string ("off"))
	default_state = false;
      else
	throw InvalidPropertyValue ("State", s);
    }
  else // default val
    default_state = true;

  i = my_properties.find ("Checked");
  if (i != my_properties.end ())
    {
      string s = i -> second.value;
      if (s == string ("checked"))
	default_checked = true;
      else if (s == string ("unchecked"))
	default_checked = false;
      else
	throw InvalidPropertyValue ("Checked", s);
    }
  else // default val
    default_checked = false;

  blocked = false;
}


void CheckedItem :: DebugTypeSpecific () const
{
  cout << "CheckedItem" << endl;
  DebugLeaveItem ();
  cout << "Pixmap TODO" << endl
       << "Default = " << default_checked << endl
       << "Blocked = " << blocked << endl;
}

bool CheckedItem :: IsEnabled (const View* here) const
{
  string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Query MenuDataBase for " << s << endl;

  const MenuData* t_data = GetMenuData (my_scope, here);

  if (t_data -> initialized)
    {
      map <string, bool> :: const_iterator i =  t_data -> states.find (s);
      if (i == t_data -> states.end ())
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "Not in DB, inserting default_state" << endl;
	  t_data -> states [s] = default_state;
	  return (default_state && (!blocked));
	}
      else
	return (i -> second && (!blocked));
    }

  throw UninitializedMenuData ();
}

void CheckedItem :: BlockAll () const
{
  blocked = true;
  SignalEnabledChanged . emit (this);
}

void CheckedItem :: UnBlockAll () const
{
  blocked = false;
  SignalEnabledChanged . emit (this);
}

void CheckedItem :: SetState (View* there, bool state) const
{
  string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Accessing MenuDataBase for " << s << endl;

  MenuData* t_data = GetMenuData (my_scope, there);

  if (t_data -> initialized)
    {
      t_data -> states [s] = state;
      SignalEnabledChanged . emit (this);
      return;
    }

  throw UninitializedMenuData ();
}


bool CheckedItem::IsChecked (const View* here) const
{
  string s = string (">Checked?") + Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Query MenuDataBase for " << s << endl;

  const MenuData* t_data = GetMenuData (my_scope, here);

  if (t_data -> initialized)
    {
      map <string, bool> :: const_iterator i =  t_data -> states.find (s);
      if (i == t_data -> states.end ())
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "Not in DB, inserting default_checked" << endl;
	  t_data -> states [s] = default_checked;
	  return default_checked;
	}
      else
	return i -> second;
    }

  throw UninitializedMenuData ();
};

void CheckedItem::SetCheck__ (View* here, bool state) const
{
  const MenuData* t_data = GetMenuData (my_scope, here);
  string s = string (">Checked?") + Location ();

  if (t_data -> initialized)
    t_data -> states [s] = state;
  else	
    throw UninitializedMenuData ();
}

bool CheckedItem::SetCheck (View* there, bool state) const
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "CheckedItem " << my_params.name << " checked shall be " << state << endl;

  if (!IsEnabled (there))
    {
      DEBUG__(GSMP_MENU, normal)
	cout << "Cannot activate an Disabled Item !" << endl;
      return IsChecked (there);
    }

  StatusActivator* parasite = Parasite (there);

  if (parasite)
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "Found Parasite" << endl;
      if (parasite -> IsParasite ())
	{
	  bool t_retval = parasite -> Execute (state);

	  if (!parasite -> IsParasite ())
	    {
	      DEBUG__(GSMP_MENU, low_level)
		cout << "Deleting Ex-parasite" << endl;
	      delete parasite;
	      RemoveParasite (there);
	    }

	  SetCheck__ (there, t_retval);
	  SignalCheckedChanged (this);
	  return t_retval;
	}
      else
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "..which has become invalid, deleting and rebuilding." << endl;
	  delete parasite;
	  RemoveParasite (there);
	}
    }

  App* t_app = there -> GetProject () -> GetApp ();
  Project* t_project = there -> GetProject ();

  StatusActivator* t_activate;


#ifdef GSMP_TESTING
  string t_name = reactor_name + string ("Siff");
#else
  string t_name = reactor_name;
#endif


  switch (my_scope)
    {
    case APP:
      t_activate = GSMX_utils::Plugin::Build <StatusActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, (Project*) 0, (View*) 0);
      break;
    case PROJECT:
      t_activate = GSMX_utils::Plugin::Build <StatusActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, t_project, (View*) 0);
      break;
    case VIEW:
      t_activate = GSMX_utils::Plugin::Build <StatusActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, t_project, there);
      break;
    }

  bool t_retval = t_activate -> Execute (state);
  if (t_activate -> IsParasite ())
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "..which has become a new parasite." << endl;
      InsertParasite (there, t_activate);
    }
  else
    delete t_activate;
 
  SetCheck__ (there, t_retval);
  SignalCheckedChanged (this);
  return t_retval;
};



StatusActivator* CheckedItem::Parasite (View* in_view) const
{
  string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Searching Parasite in " << s << endl;

  const MenuData* t_data = GetMenuData (my_scope, in_view);

  if (t_data -> initialized)
    {
      map <string, StatusActivator*> :: const_iterator i =  t_data -> parasites.find (s);
      if (i == t_data -> parasites.end ())
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "Not in DB" << endl;
	  t_data -> parasites [s] = (StatusActivator*) 0;
	  return (StatusActivator*) 0;
	}
      else
	return (i -> second);
    }
  throw UninitializedMenuData ();
}

void CheckedItem::InsertParasite (View* in_view, StatusActivator* parasite) const
{
  string s = Location ();

  DEBUG__(GSMP_MENU, low_level)
    cout << "Inserting Parasite for " << s << endl;

  MenuData* t_data = GetMenuData (my_scope, in_view);

  if (t_data -> initialized)
    {
      t_data -> parasites [s] = parasite;
      return;
    }

  throw UninitializedMenuData ();
}

void CheckedItem::RemoveParasite (View* from_view) const
{
  InsertParasite (from_view, (StatusActivator*) 0);
}




SelectionItem :: SelectionItem (MenuManager* i_mm, SubMenuItem* init_parent, const Entity& init_params, 
			  const Scope& init_scope, const string& init_properties)
  : LeaveItem (init_parent, init_params, init_scope, init_properties)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Specialisation to SelectionItem" << endl;
 
  map <string, Property> :: const_iterator i = my_properties.find ("Set");
  if (i == my_properties.end ())
    throw PropertyRequiered (string ("Set"), string ("Selection"));
  else
    m_set = i -> second.value;

  DEBUG__(GSMP_MENU, low_level)
    cout << "Set will be" << m_set << endl;

  bool is_default = false;

  i = my_properties.find ("Selected");
  if (i != my_properties.end ())
    {
      string s = i -> second.value;
      if (s == string ("yes"))
	is_default = true;
      else if (s != string ("no"))
	throw InvalidPropertyValue ("Selected", s);
    }

  DEBUG__(GSMP_MENU, low_level)
    cout << "Looking up Set " << m_set << endl;

  map <string, Set*> :: iterator j = i_mm -> all_sets.find (m_set);
  if (j == i_mm -> all_sets.end ())
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "Not found, creating " << m_set << endl;

      //i_mm -> all_sets [m_set] = new Set (my_scope);
      i_mm -> InsertSet (m_set, new Set (my_scope, (is_default) ? (this) : 0));
    }
  else
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "validating... " << m_set << endl;

      if (j -> second -> m_scope != my_scope)
	throw SetMismatch (m_set);
    
      if ((! (i_mm -> all_sets [m_set] -> default_selection)) && is_default)
	i_mm -> all_sets [m_set] -> default_selection = this;
    }
}


void SelectionItem :: DebugTypeSpecific () const
{
  cout << "SelectionItem" << endl;
  DebugLeaveItem ();
}


bool SelectionItem :: IsSelected (const View* here) const
{
  const MenuData* t_data = GetMenuData (my_scope, here);

  const Set* t_set = t_data -> selections [m_set];

  DEBUG__(GSMP_MENU, normal)
    if (!t_set)
      {
	cout << "Fatal: SelectionItem with Nonexistent Set " << m_set << endl;
	exit (-1); 
      }

  return (t_set -> current_selection == this);
}



bool SelectionItem :: Select (View * view) const
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Request for selection of " << my_params.name << endl;

  
  // TODO: Check State...

  App* t_app = view -> GetProject () -> GetApp ();
  Project* t_project = view -> GetProject ();

  PredicateActivator* t_activate;
  MenuData* t_data;

#ifdef GSMP_TESTING
  string t_name = reactor_name + string ("Sel");
#else
  string t_name = reactor_name;
#endif

  switch (my_scope)
    {
    case APP:
      t_activate = GSMX_utils::Plugin::Build <PredicateActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, (Project*) 0, (View*) 0);
      t_data = ((view -> GetProject ()) -> GetApp ());
      break;
    case PROJECT:
      t_activate = GSMX_utils::Plugin::Build <PredicateActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, t_project, (View*) 0);
      t_data = (view -> GetProject ());
      break;
    case VIEW:
      t_activate = GSMX_utils::Plugin::Build <PredicateActivator, App*, Project*, View*> 
	(reactor_category, t_name, t_app, t_project, view);
      t_data = view;
      break;
    }

  Set* t_set = t_data -> selections [m_set];

  DEBUG__(GSMP_MENU, normal)
    if (!t_set)
      {
	cout << "Fatal: SelectionItem with Nonexistent Set " << m_set << endl;
	exit (-1); 
      }

  bool t_return = t_activate -> Execute ();
    
  if (t_return)
    t_set -> ChangeCurrentSelection (this);

  delete t_activate;
  return t_return;
}




SeperatorItem :: SeperatorItem (SubMenuItem* init_parent, const string& groupname)
  : Item (init_parent, Entity (groupname, 0))
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "Specialisation to SeperatorItem" << endl; 
}


void SeperatorItem :: DebugTypeSpecific () const
{
  cout << "SeperatorItem" << endl;
}

void SeperatorItem :: Activate (View* view)
{
  DEBUG__(GSMP_MENU, normal)
    cout << "Can not Activate() an SeparatorItem..." << endl;
}

SubMenuItem :: SubMenuItem (SubMenuItem* init_parent, const Entity& init_params)
  : Item (init_parent, init_params)
{
  DEBUG__(GSMP_MENU,low_level)
    cout << "instantiated as SubMenuItem" << endl;
}

void SubMenuItem :: Insert (Item* this_item)
{ // need no conflict checking here, for it has (hopefully) been done before...
  DEBUG__(GSMP_MENU,low_level)
    cout << "Inserting '" << this_item -> my_params.name << "' into '" << my_params.name << "'" <<  endl;

  my_named_items [this_item -> my_params.name] = this_item;

  // insert into sorted deque
  deque <Item*> :: iterator insert_at = find_if (my_items.begin (), my_items.end (), IsAGreaterItem (this_item));
  my_items.insert (insert_at, this_item);

  // insert group seperator ?

  my_groups [this_item -> my_params.group] = my_groups [this_item -> my_params.group] + 1;
  if (my_groups [this_item -> my_params.group] == 1)
    Insert (new SeperatorItem (this, this_item -> my_params.group));  
}


void SubMenuItem :: DebugTypeSpecific () const
{
  cout << "SubMenuItem" << endl
       << "Content :" << endl << endl;

  for (deque <Item*> :: const_iterator i = my_items.begin (); i != my_items.end (); i++)
    {
      (*i) -> Debug ();
      cout << endl;
    }
  cout << "(End Content of " << my_params.name << ")" << endl;
}

/**************************************************************************************************/

Filter :: Filter (const string& location, const string& alias_title)
  : my_location (location), alias (alias_title)
{
  my_submenu = 0; 
}

bool Filter :: Matched () const
{
  return (my_submenu);
}


bool Filter :: Try (const LeaveItem& match)
{
  
  if (!match.my_parent)
    return false; // this should not happen with proper use of Menusystem.
  string s = match.my_parent -> Location ();

  DEBUG__(GSMP_MENU,low_level)
    cout << "Filter::Try () : Comparing " << my_location.m_string.size () << " chars from "
	 << my_location.m_string << " and " << s << endl; 
#if __GNUC__ < 3 
  if (s.compare (my_location.m_string, 0, my_location.m_string.size ()) == 0)
#else
  if (s.compare (0, my_location.m_string.size (), my_location.m_string) == 0)
#endif
    { //bingo, assign my_submenu
      DEBUG__(GSMP_MENU,low_level)
	cout << "Bingo !" << endl;

      SubMenuItem* current = match.my_parent; // Find Root
      while (current -> my_parent != 0)
	current = current -> my_parent;

      for (deque<string> :: const_iterator pos = my_location.begin (); pos != my_location.end (); pos++)
	{
	  map <string, Item*> :: const_iterator i = current -> my_named_items.find (*pos);
	  if (i == current -> my_named_items.end ())
	    {
	      DEBUG__(GSMP_MENU, normal)
		cout << "WARNING: Menu Filter Paradoxon for location " << my_location.m_string << " ???" << endl;  
	      return false; // not found....
	    }
	  Item* next_submenu = i -> second;
	  current = dynamic_cast <SubMenuItem*> (next_submenu);
	  if (!current)
	  {
	    DEBUG__(GSMP_MENU, normal)
	      cout << "WARNING: Menu Filter Paradoxon for location " << my_location.m_string << " ???" << endl;  
	    return false; // encountered leave item or something....
	  }
	}
      // Set my_submenu to corresponding item in global tree
      my_submenu = current; 

      return true;
    }
  else
    return false;
}

bool Filter::InSubTree (const LeaveItem& item)
{
  if (!Matched ())
    return Try (item);
  else
    {
      string s = item.my_parent -> Location ();
#if __GNUC__ < 3
      return (s.compare (my_location.m_string, 0, my_location.m_string.size ()) == 0);
#else
      return (s.compare (0, my_location.m_string.size (), my_location.m_string) == 0);
#endif
    }
}


Set::Set (LeaveItem::Scope i_scope, const SelectionItem* i_default)
{
  m_scope = i_scope;
  default_selection = i_default;
  current_selection = 0; 
}

Set::Set (const Set& i_set)
{
  m_scope = i_set . m_scope;
  default_selection = i_set . default_selection;
  current_selection = 0;
}



void Set::ChangeCurrentSelection(const SelectionItem *to_selection)
{
  if (to_selection != current_selection)
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "Changing selection in Set";

      const SelectionItem* t_old = current_selection;
      current_selection = to_selection;
      if (t_old)
	t_old -> SignalSelectionChanged (t_old);
      current_selection -> SignalSelectionChanged (current_selection);
    }
}


MenuData::MenuData (const MenuData& old_data)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Ctor";

  m_scope = old_data.m_scope;
  states = old_data.states;
  selections = old_data.selections;
  initialized = false;
}

MenuData::~MenuData ()
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Dtor";

  for (map <string, StatusActivator*>::iterator i = parasites.begin (); i != parasites.end (); i++)
    delete i -> second;

  SignalDeleted (this);
}


void MenuData::TestNewSet (string name, Set* set)
{
  if (set -> m_scope == m_scope)
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "MenuData: inserting new Set " << name << endl;

      selections [name] = new Set(*set);
    }
}

ItemPath::ItemPath (const SubMenuItem& from, const Item& to)
{
  //DEBUG__(GSMP_MENU, low_level)
  //  cout << "Itempath..." << endl;

  if (!to . my_parent)
    {
      //    DEBUG__(GSMP_MENU, low_level)
      //	cout << "...Failed" << endl;
      return;  
    }

  if (to . my_parent != &from)
    {
      //DEBUG__(GSMP_MENU, low_level)
      //	cout << "...UP" << endl;

      ItemPath t_recursion (from, *(to.my_parent));
      if (t_recursion.Matched ())
	subitem_no = t_recursion.subitem_no;
      else return;
    }

  int no = 0;
  deque <Item*>::const_iterator i =  to.my_parent -> my_items.begin ();
  while ((*i) != &to)
    {
      no++;
      ++i;
    }

  subitem_no.push_back (no);
}


MenuCollection::Entry :: Entry (const Filter& i_filter)
{
  m_submenu = i_filter.my_submenu;
  m_title = i_filter.alias;
}


MenuCollection::MenuCollection (/*View* i_view,*/ const MenuManager& i_menumanager, deque <Filter> i_filters, bool append_unknown)
  : m_menumanager (i_menumanager)
{
  DEBUG__(GSMP_MENU, detail)
    cout << "Creating new MenuCollection" << endl;

  append = append_unknown;
  //m_view = i_view;
  for (deque <Filter> :: const_iterator i = i_filters.begin (); i != i_filters.end (); i++)
    m_filters.push_back (*i);

  DEBUG__(GSMP_MENU, low_level)
    cout << "Iterating all LeaveItem s" << endl;

  MenuManager::const_iterator j = m_menumanager.end ();
  for (MenuManager::const_iterator i = m_menumanager.begin (); i != j; ++i)
    {
      bool has_matched = false;
      const LeaveItem* t_match = *i;

      DEBUG__(GSMP_MENU, low_level)
	cout << "Trying " << t_match -> my_params.name << " ... ";

      for (deque <Filter> :: iterator i = m_filters.begin (); i != m_filters.end (); i++)
	has_matched = has_matched || i -> InSubTree (*t_match);

      DEBUG__(GSMP_MENU, low_level)
	cout << ((has_matched) ? string ("yes") : string ("no")) << endl;

      if (append && (!has_matched))
	{
	  DEBUG__(GSMP_MENU, low_level)
	    cout << "Inserting generic Filter for " << t_match -> my_params.name << endl;

	  SimpleLocation t_location (t_match -> Location ());
	  string t_toplevel = *(t_location.begin ());

	  DEBUG__(GSMP_MENU, low_level)
	    cout << "filter & alias string : " << t_toplevel << endl;

	  m_filters.push_back (Filter (t_toplevel, t_toplevel));
	  bool t_check = m_filters.back () . Try (*t_match);
	  DEBUG__(GSMP_MENU, normal)
	    {
	      if (!t_check)
		cout << "MenuCollection Paradoxon: " << t_match -> my_params.name 
		     << " did not match its own generic Filter !?!" << endl;
	    }
	}

      // Conect LeaveItem signals
      if (append || has_matched)
	{
	  const SimpleItem* t_simple = dynamic_cast <const SimpleItem*> (t_match);
	  const SelectionItem* t_selection = dynamic_cast <const SelectionItem*> (t_match);
	  const CheckedItem* t_checked = dynamic_cast <const CheckedItem*> (t_match);


	  if (t_simple)
	    {
	      DEBUG__(GSMP_MENU, low_level)
		cout << "Connecting signals from SimpleItem" << endl;

	      t_simple -> SignalEnabledChanged.connect (SigC::slot (this, &MenuCollection::HandleSignalEC));
	    }

	  if (t_selection)
	    {
	      DEBUG__(GSMP_MENU, low_level)
		cout << "Connecting signals from SelectionItem" << endl;

	      t_selection -> SignalSelectionChanged.connect (SigC::slot (this, &MenuCollection::HandleSignalSC));
	    }

	  if (t_checked)
	    {
	      DEBUG__(GSMP_MENU, low_level)
		cout << "Connecting signals from CheckedItem" << endl;

	      t_checked -> SignalCheckedChanged.connect (SigC::slot (this, &MenuCollection::HandleSignalCI));
	      t_checked -> SignalEnabledChanged.connect (SigC::slot (this, &MenuCollection::HandleSignalCS));
	    }
	}

    }

  DEBUG__(GSMP_MENU, low_level)
    cout << "Creating Entrylist" << endl;

  for (deque <Filter> :: iterator i = m_filters.begin (); i != m_filters.end (); i++)
    if ((*i) . Matched ())
      m_entries.push_back (Entry ((*i)));
}


void MenuCollection :: HandleSignalEC (const SimpleItem* item)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuCollection: SimpleItem reports status change" << endl;

  int no = 0;

  for (const_iterator i = begin (); i != end (); i++)
    {
      ItemPath t_path (*((*i) . SubMenu ()), *item);
      if (t_path.Matched ())
	SimpleItemStateChanged (no, &t_path, *item);
      no++;
    }
} 

void MenuCollection :: HandleSignalCI (const CheckedItem* item)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuCollection: CheckedItem reports checked change" << endl;

  int no = 0;

  for (const_iterator i = begin (); i != end (); i++)
    {
      ItemPath t_path (*((*i) . SubMenu ()), *item);
      if (t_path.Matched ())
	SignalCheckedChanged (no, &t_path, *item);
      no++;
    }
} 

void MenuCollection :: HandleSignalCS (const CheckedItem* item)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuCollection: CheckedItem reports state change" << endl;

  int no = 0;

  for (const_iterator i = begin (); i != end (); i++)
    {
      ItemPath t_path (*((*i) . SubMenu ()), *item);
      if (t_path.Matched ())
	CheckedItemStateChanged (no, &t_path, *item);
      no++;
    }
} 


void MenuCollection :: HandleSignalSC (const SelectionItem* item)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuCollection: SelectionItem reports change" << endl;

  int no = 0;

  for (const_iterator i = begin (); i != end (); i++)
    {
      ItemPath t_path (*((*i) . SubMenu ()), *item);
      if (t_path.Matched ())
	SelectionChanged (no, &t_path, *item);
      no++;
    }
} 

MenuCollection::~MenuCollection ()
{
  DEBUG__(GSMP_MENU, detail)
    cout << "deleting MenuCollection" << endl;
}

void MenuCollection::Debug ()
{
  cout << "=== Start MenuCollection content ===" << endl << endl; 
  for (const_iterator i = begin (); i != end (); i++)
    cout << (*i) . Title () << "\t (" << (*i) . SubMenu () -> my_params.name << ")" << endl;
  cout << endl << "===  End MenuCollection content  ===" << endl;

}

void SimpleItemFinder :: HandleSignals (const SimpleItem* ignored)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "SimpleItemFinder: Item reports status change" << endl;
  SignalStateChanged ();
}

SimpleItemFinder :: SimpleItemFinder (const MenuManager& i_menumanager, const string& location)
  : m_menumanager (i_menumanager)
{
  DEBUG__(GSMP_MENU, detail)
    cout << "Instanciating SimpleItemFinder" << endl;

  MenuManager::const_iterator j = m_menumanager.end ();
  for (MenuManager::const_iterator i = m_menumanager.begin (); i != j; ++i)
    {
      const LeaveItem* t_item = *i;
      if ( (m_item = dynamic_cast <const SimpleItem*> (t_item)) )
	if (m_item -> Location () == location)
	  {
	    m_item -> SignalEnabledChanged.connect (SigC::slot (this, &SimpleItemFinder::HandleSignals));
	    return;
	  }
    }

  DEBUG__(GSMP_MENU, detail)
    cout << "Not found !" << endl;

  m_item = 0;
}



void CheckedItemFinder :: HandleSignalCh (const CheckedItem* ignored)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "CheckedItemFinder: Item reports checked change" << endl;
  SignalCheckedChanged ();
}

void CheckedItemFinder :: HandleSignalEn (const CheckedItem* ignored)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "CheckedItemFinder: Item reports status change" << endl;
  SignalStateChanged ();
}


CheckedItemFinder :: CheckedItemFinder (const MenuManager& i_menumanager, const string& location)
  : m_menumanager (i_menumanager)
{
  DEBUG__(GSMP_MENU, detail)
    cout << "Instanciating CheckedItemFinder" << endl;

  MenuManager::const_iterator j = m_menumanager.end ();
  for (MenuManager::const_iterator i = m_menumanager.begin (); i != j; ++i)
    {
      const LeaveItem* t_item = *i;
      if ( (m_item = dynamic_cast <const CheckedItem*> (t_item)) )
	if (m_item -> Location () == location)
	  {
	    m_item -> SignalEnabledChanged.connect (SigC::slot (this, &CheckedItemFinder::HandleSignalEn));
	    m_item -> SignalCheckedChanged.connect (SigC::slot (this, &CheckedItemFinder::HandleSignalCh));
	    return;
	  }
    }

  DEBUG__(GSMP_MENU, detail)
    cout << "Not found !" << endl;

  m_item = 0;
}




void SelectionItemFinder :: HandleSignals (const SelectionItem* ignored)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "SelectionItemFinder: Item reports status change" << endl;
  SignalSelectionChanged ();
}

SelectionItemFinder :: SelectionItemFinder (const MenuManager& i_menumanager, const string& location)
  : m_menumanager (i_menumanager)
{
  DEBUG__(GSMP_MENU, detail)
    cout << "Instanciating SelectionItemFinder" << endl;

  MenuManager::const_iterator j = m_menumanager.end ();
  for (MenuManager::const_iterator i = m_menumanager.begin (); i != j; ++i)
    {
      const LeaveItem* t_item = *i;
      if ( (m_item = dynamic_cast <const SelectionItem*> (t_item)) )
	if (m_item -> Location () == location)
	  {
	    m_item -> SignalSelectionChanged.connect (SigC::slot (this, &SelectionItemFinder::HandleSignals));
	    return;
	  }
    }

  DEBUG__(GSMP_MENU, detail)
    cout << "Not found !" << endl;

  m_item = 0;
}


/**************************************************************************************************/

/* some Item Factories */ 

class SimpleItemFactory : public GSMP::Menu::MenuManager::LeaveItemFactory
{
  friend class GSMP::Menu::MenuManager;

  virtual LeaveItem* Create (SubMenuItem* init_parent, const Entity& init_params, 
			     const LeaveItem::Scope& init_scope, const string& init_properties)
  {
    return new SimpleItem (init_parent, init_params, init_scope, init_properties);
  }

  virtual void Signal (MenuManager* through_this, LeaveItem* about_that, SubMenuItem* first_affected)
  {
    through_this -> SimpleItemAdded ((SimpleItem*)about_that, first_affected);
  }

  SimpleItemFactory (MenuManager* i_mm) : LeaveItemFactory (i_mm) {};
};

class CheckedItemFactory : public GSMP::Menu::MenuManager::LeaveItemFactory
{
  friend class GSMP::Menu::MenuManager;

  virtual LeaveItem* Create (SubMenuItem* init_parent, const Entity& init_params, 
			     const LeaveItem::Scope& init_scope, const string& init_properties)
  {
    return new CheckedItem (init_parent, init_params, init_scope, init_properties);
  }

  virtual void Signal (MenuManager* through_this, LeaveItem* about_that, SubMenuItem* first_affected)
  {
    through_this -> CheckedItemAdded ((CheckedItem*)about_that, first_affected);
  }

  CheckedItemFactory (MenuManager* i_mm) : LeaveItemFactory (i_mm) {};
};


class SelectionItemFactory : public GSMP::Menu::MenuManager::LeaveItemFactory
{
  friend class GSMP::Menu::MenuManager;

  virtual LeaveItem* Create (SubMenuItem* init_parent, const Entity& init_params, 
			     const LeaveItem::Scope& init_scope, const string& init_properties)
  {
    return new SelectionItem (m_mm, init_parent, init_params, init_scope, init_properties);
  }

  virtual void Signal (MenuManager* through_this, LeaveItem* about_that, SubMenuItem* first_affected)
  {
    through_this -> SelectionItemAdded ((SelectionItem*)about_that, first_affected);
  }

  SelectionItemFactory (MenuManager* i_mm) : LeaveItemFactory (i_mm) {};
};


/**************************************************************************************************/

MenuManager :: MenuManager ()
  : m_fmc ("Manager")
{
  DEBUG__(GSMP_MENU,low_level)
    cout << "Initializing MenuManager" << endl;
  root = new SubMenuItem (0, Entity ("<root>"));

  known_scopes ["App"] = LeaveItem :: APP;
  known_scopes ["Project"] = LeaveItem :: PROJECT;
  known_scopes ["View"] = LeaveItem :: VIEW;

  known_item_types ["Item"] = new SimpleItemFactory (this);
  known_item_types ["Selection"] = new SelectionItemFactory (this);
  known_item_types ["Checked"] = new CheckedItemFactory (this);

  m_fmc . Check4Corruption ();
}

MenuCollection* MenuManager :: CreateMenuCollection (/*View* init_view,*/ deque <Filter> i_filters, bool append_unknown)
{
  return new MenuCollection (/*init_view,*/ *this, i_filters, append_unknown);
}


void MenuManager :: Add (const string& activator, const string& reactor_category, const string& reactor_name)
{
  DEBUG__(GSMP_MENU,normal)
    cout << "Registering activator for [" << reactor_category << ":" << reactor_name << "]" << endl;

  SplitString sub (activator, ':', false); 
  deque <string> :: const_iterator i = sub.begin ();

  if (i == sub.end ())
    throw MalformedLocation ("<nil>");

  Location location (*i);

  /* now we check, wether the location is valid and create the missing branch
   * of SubMenuItems on the way from the root to the considered item position.
   * For we may encounter fatal errors in other parts of the activator string,
   * we do not append the new SubMenuItems to the tree right now, but store
   * them in <branch> */

  Item* branch;
  SubMenuItem* append_at = 0;      // where to append Branch to tree;
  SubMenuItem* tree_pos = root;    // used to iterate over tree

  DEBUG__(GSMP_MENU,detail)
    cout << "checking location string" << endl;

  // Location object has ensured that Root() is never Leave()
  // so we get no problems with algorithm below.
  deque<Entity> :: const_iterator loc_pos = location.Root (); 

  while (!append_at)
    { // Browse Existing part of tree.
      DEBUG__(GSMP_MENU,low_level)
	cout << " .. " << tree_pos -> my_params.name;
      
      map <string, Item*> :: iterator expected_sub_menu = tree_pos -> my_named_items . find ((*loc_pos) . name);
      if (expected_sub_menu == tree_pos -> my_named_items.end ())
	append_at = tree_pos;
      else
	{
	  Item* sub_menu = expected_sub_menu -> second;

	  // we have to check we realy found a submenu and no conflicting LeaveItem
	  tree_pos = dynamic_cast <SubMenuItem*> (sub_menu);
	  if (tree_pos == 0 || loc_pos == location.Leave ())
	    {
	      DEBUG__(GSMP_MENU, normal)
		cout << " <-> location conflict !!!" << endl;
	      throw DuplicateLocation (sub_menu -> Location ());
	    }

	  loc_pos ++;
	}
    }

  if (!append_at)
    { // make shure we can put out item here...
      DEBUG__(GSMP_MENU,normal)
	cout << " <-> location conflict !!!" << endl;
      throw DuplicateLocation (tree_pos -> Location ());
    }

  DEBUG__(GSMP_MENU,low_level)
    cout << " < branching here >" << endl;

  branch = append_at; 
  while (loc_pos != location.Leave ())
    branch = new SubMenuItem ((SubMenuItem*) branch, *loc_pos++);

  // TODO FREE RESOURCES ON EXCEPTION !! 

  DEBUG__(GSMP_MENU,low_level)
    cout << "Checking scope" << endl;

  i++;
  if (i == sub.end ())
    throw InvalidScope ("<nil>");

  map <string, LeaveItem::Scope> :: const_iterator scope = known_scopes.find (*i);
  if (scope == known_scopes.end ())
    throw InvalidScope (*i); 

  DEBUG__(GSMP_MENU,low_level)
    cout << "Checking type" << endl;

  i++;
  if (i == sub.end ())
    throw UnknownType ("<nil>");

  map <string, LeaveItemFactory*> :: const_iterator factory = known_item_types.find (*i);
  if (factory == known_item_types.end ())
    throw UnknownType (*i);

  DEBUG__(GSMP_MENU,low_level)
    cout << "Extracting property string" << endl;

  string properties ("");
  i++;
  if (i != sub.end ())
    properties = *i;

  /* instantiate LeaveItem */
  
  DEBUG__(GSMP_MENU,low_level)
    cout << "Invoking item factory" << endl;

  // TODO try block around  following calls to free resources on exceptions !!
  LeaveItem* new_item = (factory -> second) -> Create ((SubMenuItem *) branch, *loc_pos, scope -> second, properties);
  new_item -> reactor_name = reactor_name;
  new_item -> reactor_category = reactor_category;


  /* append branch and item to tree (means recursiv item insertion into Submenus)*/

  Item* insert = new_item;

  while (insert != append_at)
    {
      insert -> my_parent -> Insert (insert);
      insert = insert -> my_parent;
    }


  // notify others
  (factory -> second) -> Signal (this, new_item, append_at);
};



void MenuManager :: DebugMenuManager () const
{
  //m_fmc . Check4Corruption ();
  root -> Debug ();
}


MenuManager :: const_iterator :: const_iterator (SubMenuItem* i_root)
{
  m_root = i_root;
}

MenuManager :: const_iterator :: const_iterator (const MenuManager::const_iterator& copy)
{ 
  DEBUG__(GSMP_MENU, low_level)
    cout << "ctor MenuManager::const_iterator" << endl;
  positions = copy.positions;
  m_root = copy.m_root;
}


bool MenuManager :: const_iterator :: operator == (MenuManager::const_iterator cmp)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuManager :: const_iterator :: operator ==" << endl;

  if (positions.size () != cmp.positions.size ())
    return false;
  else
    {
      deque <deque <Item*> :: const_iterator> :: const_iterator j = cmp.positions.begin ();
      for (deque <deque <Item*> :: const_iterator> :: const_iterator i = positions.begin ();
	   i != positions.end (); i++)
	{
	  if ((*i) != (*j))
	    return false;
	  j++;
	}
    }
  return true;
}

bool MenuManager :: const_iterator :: operator != (MenuManager::const_iterator cmp)
{
  return !( operator == (cmp));
}


void MenuManager :: const_iterator :: operator ++ ()
{
 DEBUG__(GSMP_MENU, low_level)
    cout << "MenuManager::const_iterator ++" << endl;

 deque <Item*> :: const_iterator pos = positions.back ();

 if (positions.size () == 1)
   if (pos == m_root -> my_items.end ())
     {
       DEBUG__(GSMP_MENU, low_level)
	 cout << "already END" << endl;
       return;
     }

 while (true)
   {
     SubMenuItem* parent = (*pos) -> my_parent;
     pos++;

     if (pos == parent -> my_items.end ())
       {
	 if (positions.size () == 1)
	   {
	     positions.back () = pos;
	     DEBUG__(GSMP_MENU, low_level)
	       cout << " END" << endl;
	     return;
	   }

	 DEBUG__(GSMP_MENU, low_level)
	   cout << "Up ";
	 positions.pop_back ();
	 pos = positions.back ();
       }

     else if (!dynamic_cast <SeperatorItem*> (*pos)) // Skip SeperatorItem s
       {
	 DEBUG__(GSMP_MENU, low_level)
	   cout << "Right ";
	 positions.back () = pos;
	 SubMenuItem* current;
	 while ((current = dynamic_cast <SubMenuItem*> (*pos)))
	   {
	     DEBUG__(GSMP_MENU, low_level)
	       cout << "Down ";
	     pos = current -> my_items.begin ();
	     positions.push_back (pos);
	   }
	 DEBUG__(GSMP_MENU, low_level)
	       cout << endl;
	 return;
       }
   } 

}


const LeaveItem* MenuManager :: const_iterator :: operator * ()
{
  if (positions.size () == 1 && positions.front () == m_root -> my_items.end ())
    {
      DEBUG__(GSMP_MENU, normal)
	cout << "Error: MenuManager::const_iterator  == end ()" <<  endl;
      throw IteratorException ();
    }

  LeaveItem* r_item = dynamic_cast <LeaveItem*> (*(positions.back()));
  if (!r_item)
    {
      DEBUG__(GSMP_MENU, normal)
	cout << "Error: MenuManager::const_iterator is invalid" <<  endl;
      throw IteratorException ();
    }

  return r_item;
}


MenuManager :: const_iterator MenuManager :: begin () const
{ 
  const_iterator r_begin (root);
  SubMenuItem* current = root;
  deque <Item*> :: const_iterator pos;
  DEBUG__(GSMP_MENU, low_level)
    cout << "Constructing MenuManager::const_iterator begin ()" << endl;

  do
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << current -> my_params.name << " > ";
      pos = current -> my_items.begin ();
      r_begin.positions.push_back (pos);
      if (pos == current -> my_items.end ())
	current = 0;
      else
	current = dynamic_cast <SubMenuItem*> (*pos);
    }  while (current);

  DEBUG__(GSMP_MENU, low_level)
    cout << "OK" << endl;

  return r_begin;
}

MenuManager :: const_iterator MenuManager :: end () const
{ 
  const_iterator r_end (root);
  DEBUG__(GSMP_MENU, low_level)
    cout << "Constructing MenuManager::const_iterator end ()" << endl;
  r_end.positions.push_back (root -> my_items.end ());
  return r_end;
}

void MenuManager :: InitializeMenuData (App* new_app)
{
  m_fmc . Check4Corruption ();

  new_app -> m_scope = LeaveItem::APP;
  new_app -> initialized = true;
  new_app -> SignalDeleted . connect (SigC::slot (this, &MenuManager::MenuDataDeleted));
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Initialized, Scope = App" << endl;

  DEBUG__(GSMP_MENU, low_level)
   cout << "Connecting MenuData's Slots to personal Signals " << endl;

  SignalSetInserted.connect (slot (new_app, &MenuData::TestNewSet));

  m_fmc . Check4Corruption ();
}

void MenuManager :: InitializeMenuData (Project* new_project)
{
  m_fmc . Check4Corruption ();

  new_project -> m_scope = LeaveItem::PROJECT;
  new_project -> initialized = true;
  new_project -> SignalDeleted . connect (SigC::slot (this, &MenuManager::MenuDataDeleted));
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Initialized, Scope = Project" << endl;

  DEBUG__(GSMP_MENU, low_level)
   cout << "Connecting MenuData's Slots to personal Signals " << endl;

  SignalSetInserted.connect (slot (new_project, &MenuData::TestNewSet));

  m_fmc . Check4Corruption ();
}

void MenuManager :: InitializeMenuData (View* new_view)
{
  m_fmc . Check4Corruption ();

  new_view -> m_scope = LeaveItem::VIEW;
  new_view -> initialized = true;

  m_fmc . Check4Corruption ();

  new_view -> SignalDeleted . connect (SigC::slot (this, &MenuManager::MenuDataDeleted));
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Initialized, Scope = View" << endl
	 << "coping Sets" << endl;


  for (map <string, Set*>::iterator i = all_sets.begin (); i != all_sets.end (); i++)
    {
      Set* t_set = i -> second;
      MenuData* t_data;
      switch (t_set -> m_scope)
	{
	case LeaveItem::APP:
	  t_data = (new_view -> GetProject ()) -> GetApp ();
	  break;
	case LeaveItem::PROJECT:
	  t_data = (new_view -> GetProject ());
	  break;
	case LeaveItem::VIEW:
	  t_data = new_view;
	  break;
	}

      t_data -> selections [i -> first] = new Set(*t_set);
      if (t_data -> selections [i -> first] -> default_selection)
	t_data -> selections [i -> first] -> default_selection -> Activate (new_view);
    }

  DEBUG__(GSMP_MENU, low_level)
   cout << "Connecting MenuData's Slots to personal Signals " << endl;

  SignalSetInserted.connect (slot (new_view, &MenuData::TestNewSet));
  all_views.push_back (new_view);

  m_fmc . Check4Corruption ();
}



void MenuManager :: InitializeMenuDataCopy (View* new_view)
{
  m_fmc . Check4Corruption ();

  new_view -> initialized = true;
  new_view -> SignalDeleted . connect (SigC::slot (this, &MenuManager::MenuDataDeleted));
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData copy Initialized, Scope = View" << endl
	 << "Connecting MenuData's Slots to personal Signals " << endl;

  SignalSetInserted.connect (slot (new_view, &MenuData::TestNewSet));
  all_views.push_back (new_view);

  m_fmc . Check4Corruption ();
}





void MenuManager :: InsertSet (string set_name, Set* the_set)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuManager: Inserting new Set" << endl;
  all_sets [set_name] = the_set;
  SignalSetInserted (set_name, the_set);

  if (the_set -> default_selection)
    {
      DEBUG__(GSMP_MENU, low_level)
	cout << "MenuManager: Invoking DefaultSelection vor all Views" << endl;

      for (vector <View*>::iterator i = all_views.begin (); i != all_views.end (); i++)
	the_set -> default_selection -> Activate ((*i));
    }
}


void MenuManager :: MenuDataDeleted (MenuData* menu_data)
{
  DEBUG__(GSMP_MENU, low_level)
    cout << "MenuData Deleted, Scope = " << ((menu_data -> m_scope == LeaveItem::APP) ? string ("App") 
	 : ((menu_data -> m_scope == LeaveItem::PROJECT) ? string ("Project") : string ("View"))) << endl;

  // if VIEW, remove from all_views

  if (menu_data -> m_scope != LeaveItem::VIEW)
    return;

  for (vector <View*>::iterator i = all_views . begin (); i != all_views.end (); i++)
    if ((*i) == menu_data)
      {
	all_views .erase (i);
	return;
      }

  DEBUG__(GSMP_MENU, normal)
    cout << "BUG: View's MenuData not in Managers's all_views vector !?!?" << endl;
}
