/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   The GSMP Project class.
 */

#include "App.hh"
#include <sstream>
#include <algorithm>
#include "ParameterDialog.hh"

using namespace GSMP;
//using namespace GSMX_utils::Plugin;
using namespace GSML;

Project::Exception::Exception (const string& error)
  : GSMP::Exception (error) {}

Project::ViewNotFound::ViewNotFound ()
  : Exception (string ("Pointer reference to View not in Project Container")) {}


Project::MessageInfo::MessageInfo (int i_max_messages)
{
  max_messages = i_max_messages;
}

const string& Project::MessageInfo::GetMessage ()      // get newest message
{
  return m_messages . back ();
}

void Project::MessageInfo :: PutMessage (const string& message)
{
  DEBUG__(GSMP,normal)
    cout << message << endl;
  
  m_messages . push_back (message);
  if (m_messages . size () > max_messages)
    m_messages . pop_front ();
  SignalUpdated ();
}
	
Project::MessageInfo::const_iterator Project::MessageInfo::begin () // iterator to oldest message stored in queue
{
  return m_messages . begin ();
}

Project::MessageInfo::const_iterator Project::MessageInfo::end ()
{
  return m_messages . end ();
}


const string& Project::Name ()
{
  return m_name;
}

void Project::SetName (const string& name)
{
  DEBUG__(GSMP,normal)
    cout << "Project " << m_name << " renamed to " << name << endl;
  m_name = name;
  SignalNameChanged ();
}

WavePlotManager* Project::GetWavePlotManager ()
{
  return m_wave_plot_manager;
}


const WavePlotManager* Project::GetWavePlotManager () const
{
  return m_wave_plot_manager;
}


GSML::Mixer* Project::GetMixer ()
{
  return m_mixer;
}


Project :: Project (App* i_app, GSML::sample_rate i_sample_rate, int i_track_no, int i_max_datalayers, const string& i_name)
  : m_data (i_app -> m_file_index, i_max_datalayers, &(i_app -> pcm_type) ),
    m_fmc ("Project"),
    message_info (100),
    m_name (i_name),
    m_data_source (m_data . CurrentLayerPtr (), 0),
    m_fx_source (m_data . CurrentLayerPtr ()),
    m_selection_source (&m_data_source, m_data . CurrentLayerPtr () ),
    m_multiplexer_source (&m_selection_source, &m_fx_source),
    m_marker_source (&m_multiplexer_source, m_data . CurrentLayerPtr () ),
    m_loop_source (&m_marker_source),
    cancel_button (*(i_app -> GetManager () ), "Edit>Stop computation")
{
  m_app = i_app;
  prepared_layer = 0;
  m_fx = 0;

  m_fmc . Check4Corruption ();

  DEBUG__(GSMP,detail)
    cout << "Creating new Datalayer with " <<  i_track_no << " Tracks" <<  endl;

  // TODO: throw Exception if init_track_no < 1  or greater than [some utopic limit] !!!!!

  last_time = "";
  last_pos = 0;

  DEBUG__(GSMP,normal)
    cout << "Initializing Mixer" << endl;
  m_mixer = new Mixer ();
  m_mixer -> SetNode (m_app -> GetMixerNode () );
  m_mixer -> SetPlayback (true);
  m_mixer -> SetCapture (false);

  CurrentLayerPtr () -> markers . AddMarker ("cur_pos", 0, true);
  CurrentLayerPtr () -> markers . AddMarker ("io_pos", 0, true);

  DEBUG__(GSMP, low_level)
    cout << "Setting Mixer Parameters" << endl;

  MixerNode* t_mixer_node = m_app -> GetMixerNode ();

  m_pl_mixer_params = new UniversalMixerParams;
  m_ca_mixer_params = new UniversalMixerParams;

  m_pl_mixer_params -> SetChannels (t_mixer_node -> PlaybackChannels () );
  m_ca_mixer_params -> SetChannels (t_mixer_node -> CaptureChannels () );
  
  DataLayer* t_layer;
  t_layer = PrepareNewLayer ();
  t_layer -> CreateNEmptyTracks (i_track_no);
  FixateNewLayer ();

  m_fmc . Check4Corruption ();
  
  m_wave_plot_manager = new WavePlotManager (&m_data);

  //  AdjustMixerParams (CurrentLayerPtr ());

  message_info . PutMessage (i_name + string (": ") + GSML::itos (i_track_no) + string (" Tracks, ")
			     + GSML::itos (i_sample_rate) + string (" Hz") );

  progress_info () . SignalInfoPushed . connect (slot (this, &Project::SlotInfoPushed) );
  progress_info () . SignalInfoPopped . connect (slot (this, &Project::SlotInfoPopped) );

  m_fmc . Check4Corruption ();

  m_app -> m_manager -> InitializeMenuData (this);

  m_selection_source . SetSelectionForPlayback (false);
  m_selection_source . SetSelectionForCapture (false);

  m_loop_source . SetLooping (false);
  m_loop_source . SetStopAtEnd (false);

  view_border = m_views . end ();

  m_fmc . Check4Corruption ();
}

Project::~Project ()
{
  DEBUG__(GSMP,normal)
    {
      cout << "Killing Project " << m_name << endl;
      cout << "deleting " << m_views . size () << " views...." << endl;
    }

  for (view_iterator i = m_views . begin (); i != m_views.end (); ++ i)
    {
      (*i) -> SignalErased;
      delete (*i);
    }

  DEBUG__(GSMP,detail)
    cout << "deleting WavePlotManager... " << m_name << endl;
  delete m_wave_plot_manager;

  DEBUG__(GSMP,detail)
    cout << "deleting Mixer Params... " << m_name << endl;
  delete m_ca_mixer_params,
  delete m_pl_mixer_params;
  
  DEBUG__(GSMP,detail)
    cout << "deleting Mixer... " << m_name << endl;
  delete m_mixer;
}


int Project::Views ()
{
  cout << "             Project::Views " <<  m_views.size () << endl;
  return m_views.size ();
}
 
int Project::VisibleViews ()
{
  cout << "             Project::VisibleViews " << distance (m_views. begin (), view_border) << endl;
  return distance (m_views. begin (), view_border);
}

int Project::OrphanedViews ()
{
  cout << "             Project::OrphanedViews " << distance (view_border, m_views.end ()) << endl;
  return distance (view_border, m_views.end ());
}

Project::view_iterator Project::BeginViews ()
{
  return m_views . begin ();
}

Project::view_iterator Project::EndViews ()
{
  return m_views . end ();
}


Project::view_iterator Project::BeginVisibleViews ()
{
  return m_views . begin ();
}

Project::view_iterator Project::EndVisibleViews ()
{
  return view_border;
}

Project::view_iterator Project::BeginOrphanedViews ()
{
  return view_border;
}

Project::view_iterator Project::EndOrphanedViews ()
{
  return m_views . end ();
}


bool Project::EraseView (View* del_view)
{
  view_iterator t_view = FindView (del_view);
  return (EraseView (t_view) != t_view);
}

Project::view_iterator Project::EraseView (const view_iterator& del_view)
{
  if (! (*del_view) -> SignalEraseRequest () )
  {
    DEBUG__(GSMP,normal)
      cout << "View Erase cancelled" << endl;
    return del_view;
  }

  View* t_view = (*del_view);
  
  View* t_new_border = 0;

  if (del_view == view_border)
    view_border ++;
  if (view_border != m_views . end ())
    t_new_border = *view_border;

  view_iterator r_view = m_views . erase (del_view);

  view_border = (t_new_border) ? FindView (t_new_border) : m_views.end ();

  t_view -> SignalErased ();
  delete t_view;

  return r_view;
}


GSML::DataSource& Project::DataSource ()
{
  return m_data_source;
}


FXMixerSource& Project::FXFilter ()
{
  return m_fx_source;
}

MultiplexerFilterSource& Project::MultiplexerFilter ()
{
  return m_multiplexer_source;
}

SelectionFilterSource& Project::SelectionFilter ()
{
  return m_selection_source;
}

MarkerFilterSource& Project::MarkerFilter ()
{
  return m_marker_source;
}

LoopFilterSource& Project::LoopFilter ()
{
  return m_loop_source;
}

MixerSource* Project::GetToplevelSource ()
{
  return &m_loop_source;
}

void Project::DoFX (FXBase* fx)
{
  m_fx = fx;
  usleep (250); // HACK around stupid thread bug

  string error_msg;
  if (!m_fx -> Start (this, error_msg))
    {
      message_info.PutMessage (error_msg);
      delete m_fx;
    }
  else
    {
      slot_gui_refresh_fx = m_app -> GetKit () -> SignalGUIRefresh . connect (slot (this, &Project::SlotGUIRefreshFX) );
      cancel_button . Get () -> SetState (m_views . front (), true);
    }
}


SigC::Handle <ParameterDialog, SigC::Scopes::Extend> Project::FXConfigDialog (ParameterDialogAction* action, string title)
{
  string t_component =  m_app -> GuiComponentName (string ("ParameterDialog") );
  ParameterDialog* r_dialog = 0;

  try
    {
      r_dialog = GSMX_utils::Plugin::Build <ParameterDialog, ParameterDialogAction*, Project*, string>
	(string ("GUI"), t_component, action, this, title);
    }
  catch (GSMX_utils::Plugin::Exception e)
    {
      message_info.PutMessage ("Fatal: Could not create FXConfig Dialog !");
      delete action;
    }

  return manage (r_dialog);
}

SigC::Handle <EnhParameterDialog, SigC::Scopes::Extend> Project::FXConfigDialog (DialogAction* action, string title)
{
  string t_component =  m_app -> GuiComponentName (string ("EnhParameterDialog"));
  EnhParameterDialog* r_dialog = 0;

  try
    {
      r_dialog = GSMX_utils::Plugin::Build <EnhParameterDialog, DialogAction*, string>
	(string ("GUI"), t_component, action, title);
    }
  catch (GSMX_utils::Plugin::Exception e)
    {
      message_info.PutMessage ("Fatal: Could not create Enh FXConfig Dialog !");
      delete action;
    }

  return manage (r_dialog);
}


void Project::Cancel ()
{
  if (m_fx)
    {
      m_fx -> Cancel ();
      delete m_fx;
      m_fx = 0;
      slot_gui_refresh_fx . disconnect ();
      cancel_button . Get () -> SetState (m_views . front (), false);
    }
};


string Project::Time (sample_index pos)
{
  int i = pos / 441;
  int inaccurate = ((pos % 441) * 100) / 441;
  int dezi_secs = i % 100;
  i = (i - dezi_secs) / 100;
  int secs = i % 60;
  i = (i - secs) / 60;
  int mins = i;

  stringstream t_time;
  t_time << mins << ":";
  if (secs < 10)
    t_time << "0";
  t_time << secs << ".";
  if (dezi_secs < 10)
    t_time << "0";
  t_time << dezi_secs << ",";
  if (inaccurate < 10)
    t_time << "0";
  t_time << inaccurate;

  return t_time.str ();
}


void Project::StartPlaybackTracking (sample_index expected_latency, const string& stopitem)
{
  m_latency = expected_latency;
  m_stopitem = stopitem;
  slot_gui_refresh = m_app -> GetKit () -> SignalGUIRefresh . connect (slot (this, &Project::SlotGUIRefresh) );
  SlotGUIRefresh (); // make shure last_pos / last_time are set to realistic values
}

void Project::StopPlaybackTracking ()
{
  slot_gui_refresh . disconnect ();
}

GSML::sample_index Project::PlaybackPos ()
{
  return last_pos;
}

string Project::PlaybackTime ()
{
  return last_time;
}

void Project::AdjustMixerParams (GSML::DataLayer* for_layer)
{
  track_index t_tracks = for_layer->Tracks ();
  pcm_default hack;

  if (t_tracks != m_pl_mixer_params->Tracks () )
    {
      int t_old_tracks = m_pl_mixer_params->Tracks ();

      m_pl_mixer_params->SetTracks (t_tracks);

      for  (int i = t_old_tracks; i < t_tracks; ++i)
	for (int k = 0; k < m_pl_mixer_params->Channels (); ++k)
	  m_pl_mixer_params -> SetMatrixValue (i, k, (i == k) ? hack.dBfs () : (pcm_default::value_type) {0});
    }

  if (t_tracks != m_ca_mixer_params->Tracks () )
    {
      int t_old_tracks = m_ca_mixer_params->Tracks ();
     
      m_ca_mixer_params->SetTracks (t_tracks);
    
      for  (int i = t_old_tracks; i < t_tracks; ++i)
	for (int k = 0; k < m_ca_mixer_params->Channels (); ++k)
	  {
	    m_ca_mixer_params -> SetMatrixValue (i, k, (i == k) ? hack.dBfs () : (pcm_default::value_type) {0});
	    m_ca_mixer_params -> SetTrackMute (i, true);
	  }
    }
}


void Project::SetCurrentLayer (GSML::Data::iterator new_layer)
{
  AdjustMixerParams (*new_layer);

  try
    {
      m_data.SetCurrentLayer (new_layer);
    }
  catch (Data::Except e)
    {
      AdjustMixerParams (m_data . CurrentLayerPtr () );
      throw;
    }
}


DataLayer* Project::PrepareNewLayer ()
{
  prepared_layer = m_data.PrepareNewLayer ();
  return prepared_layer;
}

void Project::FixateNewLayer ()
{
  if (prepared_layer)
    AdjustMixerParams (prepared_layer);

  try
    {
      m_data.FixateNewLayer ();
    }
  catch (Data::Except e)
    {
      AdjustMixerParams (m_data.CurrentLayerPtr () );
      throw;
    }
}

void Project::CancelNewLayer ()
{
  m_data.CancelNewLayer ();
  prepared_layer = 0;
}



View* Project::CreateView (const string& view_impl_type = "")
{
  m_fmc . Check4Corruption ();
  // TODO: view_impl_type
  View* r_view = new View (this);

  m_fmc . Check4Corruption ();
  m_views.push_back (r_view);

  m_fmc . Check4Corruption ();
  // SignalViewCreated (r_view);

  return r_view;
}

View* Project::CopyView (View* copy_from)
{
  m_fmc . Check4Corruption ();
  View* r_view = new View (*copy_from);

  m_fmc . Check4Corruption ();
  m_views.push_back (r_view);

  m_fmc . Check4Corruption ();
  // SignalViewCreated (r_view);

  return r_view;
}


void Project::SetViewOphaned (View* view)
{
  View* t_new_border = 0;
  if (view_border != m_views . end ())
    t_new_border = *view_border;

  view_iterator t_view = FindView (view);
  m_views . erase (t_view);
  m_views . push_back (view);

  view_border = (t_new_border) ? FindView (t_new_border) : (-- (m_views.end ())); 
}

void Project::SetViewVisible (View* view)
{
  View* t_new_border = 0;

  view_iterator t_view = FindView (view);

  if (t_view == view_border)
    view_border ++;
  if (view_border != m_views . end ())
    t_new_border = *view_border;

  m_views . erase (t_view);
  m_views . push_front (view);

  view_border = (t_new_border) ? FindView (t_new_border) : m_views.end (); 
}

Project::view_iterator Project::FindView (View* view) // searches m_views for view
{
  view_iterator v = m_views . begin ();
  while (v != m_views . end () && *v != view)
    ++ v;
  
  if (v == m_views . end () )
    throw ViewNotFound ();

  return v;
}

void Project::SlotGUIRefresh ()
{
 if (m_mixer -> IsIORunning () )
   {
     sample_index t_cur_pos = (m_mixer -> IsPlayback () ) ? 
       GetToplevelSource () -> CurrentReadPos () : GetToplevelSource () -> CurrentWritePos ();

     t_cur_pos = limit_min (t_cur_pos - m_latency, 0);

     CurrentLayerPtr () -> markers . SetMarker ("io_pos", t_cur_pos);

     last_pos = t_cur_pos;
     last_time = Time (t_cur_pos);

     SignalTimeRefresh ();
   }
 else
   {
     DEBUG__ (GSMP, low_level)
       cout << "Mixer Stopped himself..." << endl;

     Menu::SimpleItemFinder t_item (*(m_app -> GetManager () ), m_stopitem);
     t_item . Get () -> Activate (m_views.front () );     
   }
}

void Project::SlotGUIRefreshFX ()
{
  if (!m_fx -> PollProgressData () )
    {
      //DEBUG__ (GSMP, low_level)
      //	cout << "FX Finished..." << endl;

      delete m_fx;
      //cout << "deleted m_fx" << endl;
      m_fx = 0;
      slot_gui_refresh_fx . disconnect ();
      //cout << "disconnected" << endl;
      cancel_button . Get () -> SetState (m_views . front (), false);
      //cout << "exiting" << endl;
    }
}

void Project::SlotInfoPushed (const GSML::Data::ProgressInfo::Info* info)
{
  message_info . PutMessage (info -> Text ());
}

void Project::SlotInfoPopped ()
{
  if (progress_info () . begin () != progress_info () . end())
    message_info . PutMessage ((*(progress_info () . begin () ) ) -> Text () );
  else
    message_info . PutMessage (string ("Done.") );
}
