/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2001 Valentin Ziegler and Ren Rebe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Window interface.
 */

#include "Window.hh"
#include "App.hh"

using namespace GSMP;

SimpleWindow :: SimpleWindow (App* i_app, View* i_view)
{
  DEBUG__(GSMP, detail)
    cout << "GSMP Window constructed" << endl;
  m_app = i_app;
  m_view = i_view;

  deque<App::Preferences::MenuFilter>::const_iterator i = m_app -> GetPreferences ().recommended_main_menu_l.begin ();
  deque<App::Preferences::MenuFilter>::const_iterator j = m_app -> GetPreferences ().recommended_main_menu_l.end ();

  deque <GSMP::Menu::Filter> t_filters;
  for (i; i != j; i++) 
    t_filters.push_back (Menu::Filter ((*i) . location, (*i) . title));
  m_main_menu_l = i_app -> m_manager -> CreateMenuCollection (t_filters, false);

  i = m_app -> GetPreferences ().recommended_main_menu_r.begin ();
  j = m_app -> GetPreferences ().recommended_main_menu_r.end ();

  deque <GSMP::Menu::Filter> t_filters2;
  for (i; i != j; i++)
    t_filters2.push_back (Menu::Filter ((*i) . location, (*i) . title));
  m_main_menu_r = i_app -> m_manager -> CreateMenuCollection (t_filters2, false);

  i_view -> SetWindow (this);

  view_impl_changed = i_view -> SignalViewImplExchanged.connect (this -> SignalViewImplChanged.slot ());
  view_erased = i_view -> SignalErased.connect (slot (this, &SimpleWindow::SlotViewErased));
  name_changed = m_view -> GetProject () -> SignalNameChanged.connect (slot (this, &SimpleWindow::SlotNameChanged));
  project_erased = m_view -> GetProject () -> SignalErased.connect (slot (this, &SimpleWindow::SlotProjectErased));
}

SimpleWindow::~SimpleWindow ()
{
  DEBUG__(GSMP, detail)
    cout << "SimpleWindow DTor" << endl;

  project_erased.disconnect ();
  view_erased.disconnect ();

  m_view -> SetWindow ((SimpleWindow *) 0);
  SignalDeleted ();

  delete m_main_menu_l;
  delete m_main_menu_r;
}

View* SimpleWindow :: GetView ()
{
  return m_view; 
}

const View* SimpleWindow :: GetView () const
{
  return m_view; 
}


void SimpleWindow::SlotNameChanged ()
{
  cout << "SimpleWindow::SlotNameChanged ..." << endl;
  
  GUI::Window* t_win = Win ();
  
  if (t_win)
    {
      t_win -> SetTitle (m_view -> GetProject () -> Name () );
    }
}

void SimpleWindow::SlotViewErased ()
{
  cout << "SimpleWindow::SlotViewErased ..." << endl;
  ViewErased_Impl ();
  cout << "done." << endl;
}

void SimpleWindow::SlotProjectErased ()
{
  cout << "SimpleWindow::SlotProjectErased ..." << endl;
  ProjectErased_Impl ();
  cout << "done." << endl;
}


MainWindow :: MainWindow (App* i_app, View* i_view)
  : SimpleWindow (i_app, i_view) 
{
  DEBUG__(GSMP, detail)
    cout << "  ... which is a MainWindow" << endl;

}; 


MainWindow :: ~MainWindow ()
{
  DEBUG__(GSMP, detail)
    cout << "MainWindow DTor" << endl;

  //  project_erased.disconnect ();
  //  view_erased.disconnect ();
};

void MainWindow :: SwitchToView (View* new_view)
{
  m_view -> SetWindow ((SimpleWindow *) 0);
  SwitchToView_impl (new_view);
}

void MainWindow :: SwitchToView_impl (View* new_view)
{
  DEBUG__(GSMP, detail)
    cout << "MainWindow: Switching View" << endl;

  m_view = new_view;
  m_view -> SetWindow (this);

  Win () -> SetTitle (new_view -> GetProject () -> Name ());

  view_impl_changed . disconnect ();
  view_impl_changed = new_view -> SignalViewImplExchanged.connect (this -> SignalViewImplChanged.slot ());

  view_erased . disconnect ();
  view_erased = new_view -> SignalErased.connect (slot (this, &SimpleWindow::SlotViewErased));

  project_erased . disconnect ();
  project_erased = m_view -> GetProject () -> SignalErased.connect (slot (this, &SimpleWindow::SlotProjectErased));

  name_changed . disconnect ();
  name_changed = m_view -> GetProject () -> SignalNameChanged.connect (slot (this, &SimpleWindow::SlotNameChanged));

  SignalViewImplChanged ();
}

void MainWindow :: ViewErased_Impl ()
{
  DEBUG__(GSMP, detail)
    cout << "MainWindow::ViewErased_Impl" << endl;

  Project* t_project = m_view -> GetProject (); // view still valid on ViewErased sig

  if (t_project -> BeginOrphanedViews () != t_project -> EndOrphanedViews ())
    {
       DEBUG__(GSMP, detail)
	 cout << "Switching to orphaned View, same Project..." << endl;

       SwitchToView_impl ( *(t_project -> BeginOrphanedViews ()) );
    }

  if (t_project -> BeginVisibleViews () != t_project -> EndVisibleViews ())
    {
       DEBUG__(GSMP, detail)
	 cout << "Switching to visible View, same Project..." << endl;

       SwitchToView_impl ( t_project -> CopyView ( *(t_project -> BeginVisibleViews ()) ) );
    }

  else
    {
       DEBUG__(GSMP, detail)
	 cout << "No more Views in crrent project (???)" << endl << "searching other projects via:" << endl;

       ProjectErased_Impl ();
    }
}

void MainWindow :: ProjectErased_Impl ()
{
  DEBUG__(GSMP, detail)
    cout << "MainWindow::ProjectErased_Impl" << endl
	 << "Searching for orphaned Views" << endl;

  for (App::project_iterator i = m_app -> BeginProjects (); i != m_app -> EndProjects (); i++)
    if ((*i) -> BeginOrphanedViews () != (*i) -> EndOrphanedViews ())
      {
	SwitchToView_impl ( *( (*i) -> BeginOrphanedViews ()) );
	return;
      }
      

  DEBUG__(GSMP, detail)
    cout << "None found, searching for visible Views" << endl;

  for (App::project_iterator i = m_app -> BeginProjects (); i != m_app -> EndProjects (); i++)
    if ((*i) -> BeginVisibleViews () != (*i) -> EndVisibleViews ())
      {
	SwitchToView_impl ( (*i) -> CopyView ( *( (*i) -> BeginVisibleViews () ) ) );
	return;
      }

  DEBUG__(GSMP, detail)
    cout << "None found, killing Window." << endl;

  Win () -> Kill ();
}


